Type: Package
Package: plumber
Title: An API Generator for R
Version: 1.3.3
Authors@R: c(
    person("Barret", "Schloerke", , "barret@posit.co", role = c("cre", "aut"),
           comment = c(ORCID = "0000-0001-9986-114X")),
    person("Jeff", "Allen", , "cran@trestletech.com", role = c("aut", "ccp")),
    person("Bruno", "Tremblay", , "cran@neoxone.com", role = "ctb"),
    person("Frans", "van Dunné", , "frans@ixpantia.com", role = "ctb"),
    person("Sebastiaan", "Vandewoude", , "sebastiaanvandewoude@gmail.com", role = "ctb"),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Gives the ability to automatically generate and serve an HTTP
    API from R functions using the annotations in the R documentation
    around your functions.
License: MIT + file LICENSE
URL: https://www.rplumber.io, https://github.com/rstudio/plumber
BugReports: https://github.com/rstudio/plumber/issues
Depends: R (>= 3.0.0)
Imports: crayon, httpuv (>= 1.5.5), jsonlite (>= 0.9.16), lifecycle (>=
        1.0.0), magrittr, mime, promises (>= 1.1.0), R6 (>= 2.0.0),
        rlang (>= 1.0.0), sodium, stringi (>= 0.3.0), swagger (>=
        3.33.0), webutils (>= 1.1)
Suggests: arrow, base64enc, coro, future, geojsonsf, htmlwidgets,
        later, ragg, rapidoc, readr, readxl, redoc, rmarkdown,
        rstudioapi, sf, spelling, svglite, testthat (>= 0.11.0), utils,
        visNetwork, withr, writexl, yaml
ByteCompile: TRUE
Config/Needs/check: Cairo, r-quantities/units
Config/Needs/website: tidyverse/tidytemplate
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.3
Collate: 'async.R' 'content-types.R' 'default-handlers.R' 'hookable.R'
        'shared-secret-filter.R' 'parser-cookie.R' 'parse-body.R'
        'parse-query.R' 'plumber.R' 'deprecated-R6.R' 'deprecated.R'
        'digital-ocean.R' 'find-port.R' 'globals.R' 'includes.R'
        'json.R' 'new-rstudio-project.R' 'openapi-spec.R'
        'openapi-types.R' 'options_plumber.R' 'paths.R' 'plumb-block.R'
        'plumb-globals.R' 'plumb.R' 'plumber-response.R'
        'plumber-static.R' 'plumber-step.R' 'pr.R' 'pr_set.R'
        'serializer.R' 'session-cookie.R' 'ui.R' 'utf8.R'
        'utils-pipe.R' 'utils.R' 'validate_api_spec.R' 'zzz.R'
NeedsCompilation: no
Packaged: 2026-01-27 21:50:00 UTC; barret
Author: Barret Schloerke [cre, aut] (ORCID:
    <https://orcid.org/0000-0001-9986-114X>),
  Jeff Allen [aut, ccp],
  Bruno Tremblay [ctb],
  Frans van Dunné [ctb],
  Sebastiaan Vandewoude [ctb],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Barret Schloerke <barret@posit.co>
Repository: CRAN
Date/Publication: 2026-01-28 06:10:08 UTC

# Generated by roxygen2: do not edit by hand

S3method(format,plumber_available_apis)
S3method(print,plumber_available_apis)
export("%>%")
export(Plumber)
export(PlumberEndpoint)
export(PlumberStatic)
export(addSerializer)
export(as_attachment)
export(available_apis)
export(do_configure_https)
export(do_deploy_api)
export(do_forward)
export(do_provision)
export(do_remove_api)
export(do_remove_forward)
export(endpoint_serializer)
export(forward)
export(getCharacterSet)
export(get_character_set)
export(get_option_or_env)
export(include_file)
export(include_html)
export(include_md)
export(include_rmd)
export(is_plumber)
export(options_plumber)
export(parser_arrow_ipc_stream)
export(parser_csv)
export(parser_excel)
export(parser_feather)
export(parser_form)
export(parser_geojson)
export(parser_json)
export(parser_multi)
export(parser_none)
export(parser_octet)
export(parser_parquet)
export(parser_rds)
export(parser_read_file)
export(parser_text)
export(parser_tsv)
export(parser_yaml)
export(plumb)
export(plumb_api)
export(plumber)
export(pr)
export(pr_cookie)
export(pr_delete)
export(pr_filter)
export(pr_get)
export(pr_handle)
export(pr_head)
export(pr_hook)
export(pr_hooks)
export(pr_mount)
export(pr_post)
export(pr_put)
export(pr_run)
export(pr_set_404)
export(pr_set_api_spec)
export(pr_set_debug)
export(pr_set_docs)
export(pr_set_docs_callback)
export(pr_set_error)
export(pr_set_parsers)
export(pr_set_serializer)
export(pr_static)
export(random_cookie_key)
export(register_docs)
export(register_parser)
export(register_serializer)
export(registered_docs)
export(registered_parsers)
export(registered_serializers)
export(serializer_agg_jpeg)
export(serializer_agg_png)
export(serializer_agg_tiff)
export(serializer_arrow_ipc_stream)
export(serializer_bmp)
export(serializer_cat)
export(serializer_content_type)
export(serializer_csv)
export(serializer_device)
export(serializer_excel)
export(serializer_feather)
export(serializer_format)
export(serializer_geojson)
export(serializer_headers)
export(serializer_html)
export(serializer_htmlwidget)
export(serializer_jpeg)
export(serializer_json)
export(serializer_octet)
export(serializer_parquet)
export(serializer_pdf)
export(serializer_png)
export(serializer_print)
export(serializer_rds)
export(serializer_svg)
export(serializer_svglite)
export(serializer_text)
export(serializer_tiff)
export(serializer_tsv)
export(serializer_unboxed_json)
export(serializer_write_file)
export(serializer_yaml)
export(sessionCookie)
export(session_cookie)
export(validate_api_spec)
import(R6)
import(promises)
import(stringi)
importFrom(crayon,silver)
importFrom(grDevices,dev.cur)
importFrom(grDevices,dev.set)
importFrom(jsonlite,parse_json)
importFrom(jsonlite,toJSON)
importFrom(lifecycle,deprecated)
importFrom(magrittr,"%>%")
importFrom(rlang,"%||%")
importFrom(rlang,missing_arg)
importFrom(stats,setNames)
importFrom(utils,file.edit)
importFrom(utils,installed.packages)
importFrom(utils,tail)
importFrom(webutils,parse_multipart)

# plumber <a href='https://www.rplumber.io/'><img src='man/figures/logo.svg' align="right" height="138.5" style="margin:10px;" /></a>

<!-- badges: start -->
[![R build status](https://github.com/rstudio/plumber/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/rstudio/plumber/actions)
[![CRAN version](https://www.r-pkg.org/badges/version/plumber)](https://www.r-pkg.org/pkg/plumber)
[![CRAN RStudio mirror downloads](https://cranlogs.r-pkg.org/badges/plumber?color=brightgreen)](https://www.r-pkg.org/pkg/plumber)
[![codecov](https://app.codecov.io/gh/rstudio/plumber/branch/main/graph/badge.svg)](https://app.codecov.io/gh/rstudio/plumber)
[![RStudio community](https://img.shields.io/badge/community-plumber-blue?style=social&logo=rstudio&logoColor=75AADB)](https://forum.posit.co/tag/plumber)
<!-- badges: end -->

Plumber allows you to create a web API by merely decorating your existing R
source code with `roxygen2`-like comments. Take a look at an example.

```r
# plumber.R

#* Echo back the input
#* @param msg The message to echo
#* @get /echo
function(msg="") {
  list(msg = paste0("The message is: '", msg, "'"))
}

#* Plot a histogram
#* @serializer png
#* @get /plot
function() {
  rand <- rnorm(100)
  hist(rand)
}

#* Return the sum of two numbers
#* @param a The first number to add
#* @param b The second number to add
#* @post /sum
function(a, b) {
  as.numeric(a) + as.numeric(b)
}
```

These comments allow `plumber` to make your R functions available as API
endpoints. You can use either `#*` as the prefix or `#'`, but we recommend the
former since `#'` will collide with `roxygen2`.

```r
library(plumber)
# 'plumber.R' is the location of the file shown above
pr("plumber.R") %>%
  pr_run(port=8000)
```

You can visit this URL using a browser or a terminal to run your R function and
get the results. For instance
`http://localhost:8000/plot` will show you a
histogram, and
`http://localhost:8000/echo?msg=hello`
will echo back the 'hello' message you provided.

Here we're using `curl` via a Mac/Linux terminal.

```
$ curl "http://localhost:8000/echo"
 {"msg":["The message is: ''"]}
$ curl "http://localhost:8000/echo?msg=hello"
 {"msg":["The message is: 'hello'"]}
```

As you might have guessed, the request's query string parameters are forwarded
to the R function as arguments (as character strings).

```
$ curl --data "a=4&b=3" "http://localhost:8000/sum"
 [7]
```

You can also send your data as JSON:

```
$ curl -H "Content-Type: application/json" --data '{"a":4, "b":5}' http://localhost:8000/sum
 [9]
```

## Installation

You can install the latest stable version from CRAN using the following command:

```r
install.packages("plumber")
```

If you want to try out the latest development version, you can install it from GitHub.

```r
pak::pkg_install("rstudio/plumber")
library(plumber)
```

## Cheat Sheet

<a href="https://github.com/rstudio/cheatsheets/blob/main/plumber.pdf"><img src="https://raw.githubusercontent.com/rstudio/cheatsheets/main/pngs/thumbnails/plumber-cheatsheet-thumbs.png" width="630" height="252" alt="plumber cheat sheet"/></a>

## Hosting

If you're just getting started with hosting cloud servers, the
[DigitalOcean](https://www.digitalocean.com) integration included in `plumber`
will be the best way to get started. You'll be able to get a server hosting your
custom API in just two R commands. To deploy to DigitalOcean, check out the
`plumber` companion package [`plumberDeploy`](https://github.com/meztez/plumberDeploy).

[Posit Connect](https://posit.co/products/enterprise/connect/) is a commercial
publishing platform that enables R developers to easily publish a variety of R
content types, including Plumber APIs. Additional documentation is available at
<https://www.rplumber.io/articles/hosting.html#rstudio-connect-1>.

A couple of other approaches to hosting plumber are also made available:

- PM2 - <https://www.rplumber.io/articles/hosting.html#pm2-1>
- Docker - <https://www.rplumber.io/articles/hosting.html#docker>

## Related Projects

- [OpenCPU](https://www.opencpu.org/) - A server designed for hosting R APIs
  with an eye towards scientific research.
- [jug](http://bart6114.github.io/jug/index.html) - *(development discontinued)*
  an R package similar to Plumber but uses a more programmatic approach to
  constructing the API.

## Static File Server

This example sets up two static file servers. One at the default path (`/public`), and another at an explicit path (`/static`). You should be able to access the two files in the `./files` directory at either of those paths. So try `http://localhost:8000/static/b.txt` or `http://localhost:8000/public/a.html`.

#* Get information about the currently available
#* @get /version
function(){
  desc <- read.dcf(system.file("DESCRIPTION", package="plumber"))
  resp <- list(
    version = unname(desc[1,"Version"]),
    built = unname(desc[1,"Built"])
  )

  if ("GithubSHA1" %in% colnames(desc)){
    resp["sha1"] <- unname(desc[1,"GithubSHA1"])
  }

  resp
}

#* Give GitHub Webhook a way to alert us about new pushes to the repo
#* https://developer.github.com/webhooks/
#* @post /update
function(req, res){
  secret <- readLines("./github-key.txt")[1]
  hm <- digest::hmac(secret, req$body, algo="sha1")
  hm <- paste0("sha1=", hm)
  if (!identical(hm, req$HTTP_X_HUB_SIGNATURE)){
    res$status <- 400
    res$body <- "invalid GitHub signature."
    return(res)
  }

  # DO...
  remotes::install_github("rstudio/plumber")

  TRUE
}

inventory <- read.csv("inventory.csv", stringsAsFactors = FALSE)

#* @apiTitle Auto Inventory Manager
#* @apiDescription Manage the inventory of an automobile
#*   store using an API.
#* @apiTag cars Functionality having to do with the management of
#*   car inventory.

#* List all cars in the inventory
#* @get /car/
#* @tag cars
listCars <- function(){
  inventory
}

#* Lookup a car by ID
#* @param id The ID of the car to get
#* @get /car/<id:int>
#* @response 404 No car with the given ID was found in the inventory.
#* @tag cars
getCar <- function(id, res){
  car <- inventory[inventory$id == id,]
  if (nrow(car) == 0){
    res$status <- 404
  }
  car
}

validateCar <- function(make, model, year){
  if (missing(make) || nchar(make) == 0){
    return("No make specified")
  }
  if (missing(model) || nchar(model) == 0){
    return("No make specified")
  }
  if (missing(year) || as.integer(year) == 0){
    return("No year specified")
  }
  NULL
}

#* Add a car to the inventory
#* @post /car/
#* @param make:character The make of the car
#* @param model:character The model of the car
#* @param edition:character Edition of the car
#* @param year:int Year the car was made
#* @param miles:int The number of miles the car has
#* @param price:numeric The price of the car in USD
#* @response 400 Invalid user input provided
#* @tag cars
addCar <- function(make, model, edition, year, miles, price, res){
  newId <- max(inventory$id) + 1

  valid <- validateCar(make, model, year)
  if (!is.null(valid)){
    res$status <- 400
    return(list(errors=paste0("Invalid car: ", valid)))
  }

  car <- list(
    id = newId,
    make = make,
    model = model,
    edition = edition,
    year = year,
    miles = miles,
    price = price
  )

  inventory <<- rbind(inventory, car)
  getCar(newId)
}

#* Update a car in the inventory
#* @param id:int The ID of the car to update
#* @param make:character The make of the car
#* @param model:character The model of the car
#* @param edition:character Edition of the car
#* @param year:int Year the car was made
#* @param miles:int The number of miles the car has
#* @param price:numeric The price of the car in USD
#* @put /car/<id:int>
#* @tag cars
updateCar <- function(id, make, model, edition, year, miles, price, res){

  valid <- validateCar(make, model, year)
  if (!is.null(valid)){
    res$status <- 400
    return(list(errors=paste0("Invalid car: ", valid)))
  }

  updated <- list(
    id = id,
    make = make,
    model = model,
    edition = edition,
    year = year,
    miles = miles,
    price = price
  )

  if (!(id %in% inventory$id)){
    stop("No such ID: ", id)
  }

  inventory[inventory$id == id, ] <<- updated
  getCar(id)
}

#* Delete a car from the inventory
#* @param id:int The ID of the car to delete
#* @delete /car/<id:int>
#* @tag cars
deleteCar <- function(id, res){
  if (!(id %in% inventory$id)){
    res$status <- 400
    return(list(errors=paste0("No such ID: ", id)))
  }
  inventory <<- inventory[inventory$id != id,]
}

#* Save a file with a particular file name. Ex: `time.txt`
#* @serializer text
#* @get /name
function() {
  as_attachment(Sys.time(), "time.txt")
}



#* Save a file as the route. Ex: `no_name`
#* @serializer text
#* @get /no_name
function() {
  as_attachment(Sys.time())
}

#* Display within browser. Possible as the mime type is `text/plain`
#* @serializer text
#* @get /inline
function() {
  Sys.time()
}


#* Write and return multiple files as an archive. Ex: `datasets.zip`
#* @serializer octet
#* @get /datasets
function() {

  # Create temporary directory structure
  tmp_dir <- tempfile()
  dir.create(tmp_dir, showWarnings = FALSE)
  on.exit(unlink(tmp_dir, recursive = TRUE), add = TRUE)

  # Save datasets to csv
  csv_files <- lapply(c("mtcars", "quakes", "airquality"), function(dataset) {
    csv_file <- file.path(tmp_dir, paste0(dataset, ".csv"))
    write.csv(get(dataset), csv_file)
    csv_file
  })

  # Create archive
  zip_file <- file.path(tmp_dir, "datasets.zip")
  zip(zip_file, csv_files, flags = "-jq9X")
  val <- readBin(zip_file, "raw", file.info(zip_file)$size)

  as_attachment(val, "datasets.zip")

}

# Instructions:
# 1. `plumb` API - `plumb_api("plumber", "14-future") %>% pr_run(port = 1234)`
# 2. In separate R session, source the test file - `source(system.file("plumber/14-future/test-future.R", package = "plumber"))`

local(withAutoprint({ # print when sourced

read_url <- function(...) {
  try(readLines(..., warn = FALSE))
}

# example
read_url("http://127.0.0.1:1234/divide?a=6&b=3") # 2
read_url("http://127.0.0.1:1234/divide-catch?a=6&b=3") # 2

# missing 'b' param
read_url("http://127.0.0.1:1234/divide?a=6") # fails
read_url("http://127.0.0.1:1234/divide-catch?a=6") # handles error; returns Inf

# missing 'a' param
read_url("http://127.0.0.1:1234/divide?b=3") # fails
read_url("http://127.0.0.1:1234/divide-catch?b=3") # fails

}))

# --------------------------
# --------------------------

local(withAutoprint({ # print when sourced

counter <- 1
log_file <- "_timings.log"
cat_file <- function(..., append = TRUE) {
  Sys.sleep(0.5) # for clean logging purposes only
  cat(..., "\n", file = log_file, append = append)
}
curl_route <- function(route) {
  Sys.sleep(0.5) # for clean logging purposes only
  local_counter <- counter
  counter <<- counter + 1
  system(
    paste0("curl 127.0.0.1:1234/", route, " >> ", log_file, " && echo ' - ", local_counter, "' >> ", log_file),
    wait = FALSE
  )
}


# Serially call each route
cat_file("--START route requests", append = FALSE)
curl_route("sync")
curl_route("future")
curl_route("sync")
curl_route("sync")
curl_route("sync")
curl_route("future")
curl_route("future") # will wait to execute if only two future processes allowed
curl_route("future")
curl_route("sync")
curl_route("sync")
curl_route("sync")
cat_file("--END route requests\n")


Sys.sleep(21) # wait for everything to finish

# display curl'ed output
cat(readLines(log_file), sep = "\n")


}))
# --------------------------

## Sample output using future::plan(future::multisession(workers = 2)) # only two workers
# --START route requests
# "/sync; 2019-10-07 13:11:06; pid:82424" - 1
# "/sync; 2019-10-07 13:11:07; pid:82424" - 3
# "/sync; 2019-10-07 13:11:07; pid:82424" - 4
# "/sync; 2019-10-07 13:11:08; pid:82424" - 5
# --END route requests
#
# "/sync; 2019-10-07 13:11:19; pid:82424" - 9
# "/sync; 2019-10-07 13:11:19; pid:82424" - 10
# "/sync; 2019-10-07 13:11:19; pid:82424" - 11
# "/future; 2019-10-07 13:11:16; pid:37135" - 2
# "/future; 2019-10-07 13:11:18; pid:37148" - 6
# "/future; 2019-10-07 13:11:27; pid:37272" - 7
# "/future; 2019-10-07 13:11:29; pid:37273" - 8


# --------------------------

## Sample output using future::plan("multisession") # a worker for each core
# --START route requests
# "/sync; 2019-10-07 13:16:22; pid:82424" - 1
# "/sync; 2019-10-07 13:16:23; pid:82424" - 3
# "/sync; 2019-10-07 13:16:24; pid:82424" - 4
# "/sync; 2019-10-07 13:16:25; pid:82424" - 5
# "/sync; 2019-10-07 13:16:27; pid:82424" - 9
# "/sync; 2019-10-07 13:16:27; pid:82424" - 10
# "/sync; 2019-10-07 13:16:28; pid:82424" - 11
# --END route requests
#
# "/future; 2019-10-07 13:16:33; pid:40613" - 2
# "/future; 2019-10-07 13:16:35; pid:40626" - 6
# "/future; 2019-10-07 13:16:36; pid:40630" - 7
# "/future; 2019-10-07 13:16:36; pid:40634" - 8

#* @get /counter
function(req){
  count <- 0
  if (!is.null(req$session$counter)){
    count <- as.numeric(req$session$counter)
  }
  req$session$counter <- count + 1
  return(paste0("This is visit #", count))
}

library(promises)
library(coro)
future::plan("multisession") # a worker for each core
# future::plan(future::multisession(workers = 2)) # only two workers


# Quick manual test:
# Within 10 seconds...
# 1. Visit /future
# 2. While /future is loading, visit /sync many times
# /future will not block /sync from being able to be loaded.


#' @serializer json list(auto_unbox = TRUE)
#' @get /sync
function() {
  # print route, time, and worker pid
  paste0("/sync; ", Sys.time(), "; pid:", Sys.getpid())
}

#' @contentType list(type = "text/html")
#' @serializer json list(auto_unbox = TRUE)
#' @get /future
function() {

  future_promise({
    # perform long running computations
    Sys.sleep(10)

    # print route, time, and worker pid
    paste0("/future; ", Sys.time(), "; pid:", Sys.getpid())
  })
}


# A function that will return a promise object
calc_using_promise <- function() {
  # Executes using `future`
  future_promise({
    # perform long computations
    Sys.sleep(1)

    # return which process id was used
    Sys.getpid()
  }) %...>% {
    exec_pid <- .
    main_pid <- Sys.getpid()
    # print route, time, and worker pid
    paste0("/coro; ", Sys.time(), "; exec pid:", exec_pid, "; main pid:", main_pid)

    list(exec_pid = exec_pid, main_pid = main_pid)
  }
}

#' For more info about `coro`, visit https://coro.r-lib.org
#' @contentType text
#' @serializer json list(auto_unbox = TRUE)
#' @get /coro
async(function() {

  # (a)wait for the promised value and use as a regular value
  # No need for a followup promise when using `await(p)`
  pids <- await(calc_using_promise())

  # Write code with the finalized async value as if it was synchronously found
  pids$exec_is_even <- pids$exec_pid %% 2 == 0

  # return info
  pids
})


# -----------------------------------


# Originally by @antoine-sachet from https://github.com/rstudio/plumber/issues/389
#' @get /divide
#' @serializer json list(auto_unbox = TRUE)
#' @param a number
#' @param b number
function(a = NA, b = NA) {
  future_promise({
    a <- as.numeric(a)
    b <- as.numeric(b)
    if (is.na(a)) stop("a is missing")
    if (is.na(b)) stop("b is missing")
    if (b == 0) stop("Cannot divide by 0")

    a / b
  })
}

#' @get /divide-catch
#' @serializer json list(auto_unbox = TRUE)
#' @param a number
#' @param b number
function(a = NA, b = NA) {
  future_promise({
    a <- as.numeric(a)
    b <- as.numeric(b)
    if (is.na(a)) stop("a is missing")
    if (is.na(b)) stop("b is missing")
    if (b == 0) stop("Cannot divide by 0")

    a / b
  }) %...!%
    # Handle `future` errors
    {
      error <- .
      # handle error here!
      if (error$message == "b is missing") {
        return(Inf)
      }

      # rethrow original error
      stop(error)
    }
}

# Visit /async
# While /async is loading, visit /sync many times

library(promises)

sleep_count <- 5
# add 5 seconds of sleep time
add_async_sleep <- function(p) {
  n <- 20
  for (i in 1:(sleep_count * n)) {
    p <- then(p, function(value) {
      Sys.sleep(1/n)
      "" # return value
    })
  }
  p
}

# use name_ as a placeholder for name when there are extra args
time <- function(name_, name = name_) {
  paste0(name, ": ", Sys.time())
}

new_promise <- function() {
  promise(function(resolve, reject){ resolve(NULL) })
}

#' @get /async
function() {
  new_promise() %>%
    add_async_sleep() %...>%
    time("async")
}

#' @get /sync
function() {
  time("sync")
}
#' @get /sync-slow
function() {
  Sys.sleep(sleep_count)
  time("sync-slow")
}

#' @get /async-bad
function() {
  new_promise() %>%
    add_async_sleep() %...>%
    stop("async-bad - expected error here") %>%
    add_async_sleep() %...>%
    time("async-bad")
}

#' @get /sync-bad
function() {
  stop("sync-bad - expected error here")
}

#Defining an array parameter
pr <- Plumber$new()

openapi_func <- function(spec) {
  spec$paths[["/sum"]]$get$summary <- "Sum numbers"
  spec$paths[["/sum"]]$get$parameters <- list(list(
    "description" = "numbers",
    "required" = TRUE,
    "in" = "query",
    "name" = "num",
    "schema" = list("type" = "array", "items" = list("type" = "integer"), "minItems" = 1),
    "style" = "form",
    "explode" = FALSE
  ))
  spec
}

handler <- function(num) {
  sum(as.integer(num))
}

pr$handle("GET", "/sum", handler, serializer = serializer_json())

pr$setApiSpec(api = openapi_func)

pr$getApiSpec()

# return Plumber object
pr

emails <- data.frame(from=character(0), time=character(0), subject=character(0), stringsAsFactors = FALSE)

#* @post /mail
function(from, subject){
  emails <<- rbind(emails, data.frame(from=from, time=date(), subject=htmltools::htmlEscape(subject), stringsAsFactors=FALSE))
  TRUE
}

#* @get /tail
function(){
  tail(emails[,-1], n=5)
}

#* Plumber allows for things like URI paths to be accessed via named R function arguments, but this is generally considered bad practice (see the examples below)
#* @serializer print
#* @post /bad-practice/<a>/<b>
function(a, b) {
  list(a = a, b = b)
}



#* Since URI paths, query params, and body arguments can have conflicting names, it's better practice to access arguments via the request object
#* If more information is needed from the body (such as filenames), inspect `req$body` for more information
#* @serializer print
#* @post /good-practice/<a>/<b>
function(req, res) {
  list(
    args = req$args,
    argsQuery = req$argsQuery,
    argsPath = req$argsPath,
    argsBody = req$argsBody,
    body = req$body
  )
}


# Test this api...

### In an R session...
### Run plumber API
# plumb_api("plumber", "17-arguments") %>% print() %>% pr_run(port = 1234)


### In a terminal...
### Curl API

## Fails (conflicting variable `a`)
# curl --data '' '127.0.0.1:1234/bad-practice/1/2?a=3'
# curl --data 'a=3' '127.0.0.1:1234/bad-practice/1/2'
# curl --data 'a=4' '127.0.0.1:1234/bad-practice/1/2?a=3'

## Works (but missing variable `d`)
# curl --data '' '127.0.0.1:1234/bad-practice/1/2?d=3'
#> $a
#> [1] "1"
#>
#> $b
#> [1] "2"

## Works (but missing variable `d`)
# curl --data 'd=3' '127.0.0.1:1234/bad-practice/1/2'
#> $a
#> [1] "1"
#>
#> $b
#> [1] "2"



## Safe endpoint setup
# curl --data 'a=5&b=6' '127.0.0.1:1234/good-practice/3/4?a=1&b=2&d=10'
#> $args
#> $args$req
#> <environment>
#>
#> $args$res
#> <PlumberResponse>
#>
#> $args$a
#> [1] "1"
#>
#> $args$b
#> [1] "2"
#>
#> $args$d
#> [1] "10"
#>
#> $args$a
#> [1] "3"
#>
#> $args$b
#> [1] "4"
#>
#> $args$a
#> [1] "5"
#>
#> $args$b
#> [1] "6"
#>
#>
#> $argsQuery
#> $argsQuery$a
#> [1] "1"
#>
#> $argsQuery$b
#> [1] "2"
#>
#> $argsQuery$d
#> [1] "10"
#>
#>
#> $argsPath
#> $argsPath$a
#> [1] "3"
#>
#> $argsPath$b
#> [1] "4"
#>
#>
#> $argsBody
#> $argsBody$a
#> [1] "5"
#>
#> $argsBody$b
#> [1] "6"
#>
#>
#> $body
#> $body$a
#> [1] "5"
#>
#> $body$b
#> [1] "6"

values <- 15

MAX_VALS <- 50

#* Append to our values
#* @post /append
function(val, res){
  v <- as.numeric(val)
  if (is.na(v)){
    res$status <- 400
    res$body <- "val parameter must be a number"
  }
  values <<- c(values, val)

  if (length(values) > MAX_VALS){
    values <<- tail(values, n=MAX_VALS)
  }

  list(result="success")
}

#* Get the last few values
#* @get /tail
function(n="10", res){
  n <- as.numeric(n)
  if (is.na(n) || n < 1 || n > MAX_VALS){
    res$status <- 400
    res$body <- "parameter 'n' must be a number between 1 and 100"
  }

  list(val=tail(values, n=n))
}

#* Get a graph of the values
#* @serializer png
#* @get /graph
function(){
  plot(values, type="b", ylim=c(1,100), main="Recent Values")
}

#* @get /counter
function(req, res) {
  count <- 0
  if (!is.null(req$cookies$visitcounter)) {
    count <- as.numeric(req$cookies$visitcounter)
  }
  # Most people won't need to concern themselves with the path argument.
  # I do because of some peculiarities in how I'm hosting the examples.
  res$setCookie("visitcounter", count+1, path="/")
  return(paste0("This is visit #", count))
}

#* Example using req$session. Requires adding "session_cookie()" support to your API in order
#* to work:
#* @get /sessionCounter
function(req){
  count <- 0
  if (!is.null(req$session$counter)){
    count <- as.numeric(req$session$counter)
  }
  req$session$counter <- count + 1
  return(paste0("This is visit #", count))
}

#* @assets static
list()


#* @plumber
function(pr) {
  pr %>%
    pr_cookie("secret", "cookieName")
}

library(plumber)

users <- data.frame(
  id=1:2,
  username=c("joe", "kim"),
  groups=c("users", "admin,users")
)

#* Filter that grabs the "username" querystring parameter.
#* You should, of course, use a real auth system, but
#* this shows the principles involved.
#* @filter auth-user
function(req, username=""){
  # Since username is a querystring param, we can just
  # expect it to be available as a parameter to the
  # filter (plumber magic).

  # This is a work-around for https://github.com/rstudio/plumber/issues/12
  # and shouldn't be necessary long-term
  req$user <- NULL

  if (username == ""){
    # No username provided
  } else if (username %in% users$username){
    # username is valid

    req$user <- users[users$username == username,]

  } else {
    # username was provided, but invalid
    stop("No such username: ", username)
  }

  # Continue on
  forward()
}

#* Now require that all users must be authenticated.
#* @filter require-auth
function(req, res){
  if (is.null(req$user)){
    # User isn't logged in

    res$status <- 401 # Unauthorized
    list(error="You must login to access this resource.")
  } else {
    # user is logged in. Move on...
    forward()
  }
}

#* @get /me
function(req){
  # Safe to assume we have a user, since we've been
  # through all the filters and would have gotten an
  # error earlier if we weren't.
  list(user=req$user)
}

#* Get info about the service. We preempt the
#* require-auth filter because we want authenticated and
#* unauthenticated users alike to be able to access this
#* endpoint.
#* @preempt require-auth
#* @get /about
function(){
  list(descript="This is a demo service that uses authentication!")
}

#* @serializer contentType list(type="application/pdf")
#* @get /pdf
function(){
  tmp <- tempfile()
  pdf(tmp)
  plot(1:10, type="b")
  text(4, 8, "PDF from plumber!")
  dev.off()

  readBin(tmp, "raw", n=file.info(tmp)$size)
}

#* @serializer contentType list(type="text/plain")
#* @get /text
function(){
  "just plain text here..."
}

#* @serializer contentType list(type="text/html")
#* @get /html
function(){
  "<html><h1>HTML!</h1>HTML here!</html>"
}

#* Download a binary file.
#* @serializer contentType list(type="application/octet-stream")
#* @get /download-binary
function(res){
  # TODO: Stream the data into the response rather than loading it all in memory
  # first.

  # Create a temporary example RDS file
  x <- list(a=123, b="hi!")
  tmp <- tempfile(fileext=".rds")
  saveRDS(x, tmp)

  # This header is a convention that instructs browsers to present the response
  # as a download named "mydata.Rds" rather than trying to render it inline.
  res$setHeader("Content-Disposition", "attachment; filename=mydata.Rds")

  # Read in the raw contents of the binary file
  bin <- readBin(tmp, "raw", n=file.info(tmp)$size)

  # Delete the temp file
  file.remove(tmp)

  # Return the binary contents
  bin
}

#' Store session data in encrypted cookies.
#'
#' \code{plumber} uses the crypto R package \code{sodium}, to encrypt/decrypt
#' \code{req$session} information for each server request.
#'
#' The cookie's secret encryption \code{key} value must be consistent to maintain
#' \code{req$session} information between server restarts.
#'
#' @section Storing secure keys:
#' While it is very quick to get started with user session cookies using
#' \code{plumber}, please exercise precaution when storing secure key information.
#' If a malicious person were to gain access to the secret \code{key}, they would
#' be able to eavesdrop on all \code{req$session} information and/or tamper with
#' \code{req$session} information being processed.
#'
#' Please: \itemize{
#' \item Do NOT store keys in source control.
#' \item Do NOT store keys on disk with permissions that allow it to be accessed by everyone.
#' \item Do NOT store keys in databases which can be queried by everyone.
#' }
#'
#' Instead, please: \itemize{
#' \item Use a key management system, such as
#' \href{https://github.com/r-lib/keyring}{'keyring'} (preferred)
#' \item Store the secret in a file on disk with appropriately secure permissions,
#'   such as "user read only" (\code{Sys.chmod("myfile.txt", mode = "0600")}),
#'   to prevent others from reading it.
#' } Examples of both of these solutions are done in the Examples section.
#'
#' @param key The secret key to use. This must be consistent across all R sessions
#'   where you want to save/restore encrypted cookies. It should be produced using
#'   \code{\link{random_cookie_key}}. Please see the "Storing secure keys" section for more details
#'   complex character string to bolster security.
#' @param name The name of the cookie in the user's browser.
#' @param path The URI path that the cookie will be available in future requests.
#'    Defaults to the request URI. Set to \code{"/"} to make cookie available to
#'    all requests at the host.
#' @param expiration A number representing the number of seconds into the future
#'   before the cookie expires or a \code{POSIXt} date object of when the cookie expires.
#'   Defaults to the end of the user's browser session.
#' @param http Boolean that adds the \code{HttpOnly} cookie flag that tells the browser
#'   to save the cookie and to NOT send it to client-side scripts. This mitigates \href{https://developer.mozilla.org/en-US/docs/Glossary/Cross-site_scripting}{cross-site scripting}.
#'   Defaults to \code{TRUE}.
#' @param secure Boolean that adds the \code{Secure} cookie flag.  This should be set
#'   when the route is eventually delivered over \href{https://en.wikipedia.org/wiki/HTTPS}{HTTPS}.
#' @param same_site A character specifying the SameSite policy to attach to the cookie.
#'   If specified, one of the following values should be given: "Strict", "Lax", or "None".
#'   If "None" is specified, then the \code{secure} flag MUST also be set for the modern browsers to
#'   accept the cookie. An error will be returned if \code{same_site = "None"} and \code{secure = FALSE}.
#'   If not specified or a non-character is given, no SameSite policy is attached to the cookie.
#' @export
#' @seealso \itemize{
#' \item \href{https://github.com/r-lib/sodium}{'sodium'}: R bindings to 'libsodium'
#' \item \href{https://doc.libsodium.org/}{'libsodium'}: A Modern and Easy-to-Use Crypto Library
#' \item \href{https://github.com/r-lib/keyring}{'keyring'}: Access the system credential store from R
#' \item \href{https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie#Directives}{Set-Cookie flags}: Descriptions of different flags for \code{Set-Cookie}
#' \item \href{https://developer.mozilla.org/en-US/docs/Glossary/Cross-site_scripting}{Cross-site scripting}: A security exploit which allows an attacker to inject into a website malicious client-side code
#' }
#' @examples
#' \dontrun{
#'
#' ## Set secret key using `keyring` (preferred method)
#' keyring::key_set_with_value("plumber_api", plumber::random_cookie_key())
#'
#'
#' # Load a plumber API
#' plumb_api("plumber", "01-append") %>%
#'   # Add cookie support via `keyring`
#'   pr_cookie(
#'     keyring::key_get("plumber_api")
#'   ) %>%
#'   pr_run()
#'
#'
#' #### -------------------------------- ###
#'
#'
#' ## Save key to a local file
#' pswd_file <- "normal_file.txt"
#' cat(plumber::random_cookie_key(), file = pswd_file)
#' # Make file read-only
#' Sys.chmod(pswd_file, mode = "0600")
#'
#'
#' # Load a plumber API
#' plumb_api("plumber", "01-append") %>%
#'   # Add cookie support and retrieve secret key from file
#'   pr_cookie(
#'     readLines(pswd_file, warn = FALSE)
#'   ) %>%
#'   pr_run()
#' }
session_cookie <- function(
  key,
  name = "plumber",
  expiration = FALSE,
  http = TRUE,
  secure = FALSE,
  same_site = FALSE,
  path = NULL
) {

  if (missing(key)) {
    stop("You must define an encryption key. Please see `?session_cookie` for more details")
  }
  key <- asCookieKey(key)

  # force the args to evaluate
  list(expiration, http, secure, same_site, path)

  # sanity check the same_site and secure arguments
  if (is.character(same_site)) {
    same_site <- match.arg(same_site, c("Strict", "Lax", "None"))
  } else {
    same_site <- FALSE
  }
  if (identical(same_site, "None")) {
    if (!secure) {
      stop("You must set `secure = TRUE` when `same_site = \"None\"`. See: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie")
    }
  }

  # Return a list that can be added to registerHooks()
  list(
    preroute = function(req, res, data) {

      cookies <- req$cookies
      if (is.null(cookies)){
        # The cookie-parser filter has probably not run yet. Parse the cookies ourselves
        # TODO: would be more performant not to run this cookie parsing twice.
        cookies <- parseCookies(req$HTTP_COOKIE)
      }
      session <- cookies[[name]]
      tryCatch({
        req$session <- decodeCookie(session, key)
      }, error = function(e) {
        # silently fail
        NULL # kept to not re-throw warning
      })

    },
    postroute = function(value, req, res, data) {
      session <- req$session
      # save session in a cookie
      if (!is.null(session)) {
        res$setCookie(
          name,
          encodeCookie(session, key),
          path = path,
          expiration = expiration,
          http = http,
          secure = secure,
          same_site = same_site
        )
      } else {
        # session is null
        if (!is.null(req$cookies[[name]])) {
          # no session to save, but had session to parse
          # remove cookie session cookie
          res$removeCookie(
            name,
            path = path,
            expiration = expiration,
            http = http,
            secure = secure,
            same_site = same_site
          )
        }
      }

      value
    }
  )
}

#' Random cookie key generator
#'
#' Uses a cryptographically secure pseudorandom number generator from [sodium::helpers()] to generate a 64 digit hexadecimal string.  \href{https://github.com/r-lib/sodium}{'sodium'} wraps around \href{https://doc.libsodium.org/}{'libsodium'}.
#'
#' Please see \code{\link{session_cookie}} for more information on how to save the generated key.
#'
#' @return A 64 digit hexadecimal string to be used as a key for cookie encryption.
#' @export
#' @seealso \code{\link{session_cookie}}
random_cookie_key <- function() {
  sodium::bin2hex(
    sodium::random(32)
  )
}


asCookieKey <- function(key) {
  if (is.null(key) || identical(key, "")) {
    warning(
      "\n",
      "\n\t!! Cookie secret 'key' is `NULL`. Cookies will not be encrypted.      !!",
      "\n\t!! Support for unencrypted cookies is deprecated and will be removed. !!",
      "\n\t!! Please see `?session_cookie` for details.                           !!",
      "\n"
    )
    return(NULL)
  }

  if (!is.character(key)) {
    stop(
      "Illegal cookie secret 'key' detected.",
      "\nPlease see `?session_cookie` for details."
    )
  }

  # trim away white space
  key <- stri_trim_both(key)

  if (
    nchar(key) != 64 ||
    grepl("[^0-9a-fA-F]", key)
  ) {
    warning(
      "\n",
      "\n\t!! Legacy cookie secret 'key' detected!                                         !!",
      "\n\t!! Support for legacy cookie secret 'key' is deprecated and will be removed.    !!",
      "\n\t!! Please follow the instructions in `?session_cookie` for creating a new secret key. !!",
      "\n"
    )

    # turn key into 64 digit hex str by hashing it
    key <-
      key %>%
      charToRaw() %>%
      sodium::sha256() %>%
      sodium::bin2hex()
  }

  sodium::hex2bin(key)
}


encodeCookie <- function(x, key) {
  if (is.null(x)) {
    return("")
  }
  xRaw <-
    x %>%
    toJSON() %>%
    charToRaw()

  if (is.null(key)) {
    encodedCookie <- base64enc::base64encode(xRaw)
    return(encodedCookie)
  }

  # key provided

  # random nonce for each request
  nonce <- sodium::random(24)
  # toJSON -> as raw -> encrypt
  encryptedX <- sodium::data_encrypt(xRaw, key, nonce)

  encodedCookie <- combine_val_and_nonce(encryptedX, nonce)
  return(encodedCookie)
}


# any bad result returns NULL
# any invalid input returns NULL
decodeCookie <- function(encodedValue, key) {
  if (is.null(encodedValue) || identical(encodedValue, "")) {
    # nothing to decode
    return(NULL)
  }

  # if no key provided
  if (is.null(key)) {
    value <-
      encodedValue %>%
      base64enc::base64decode() %>%
      rawToChar() %>%
      safeFromJSON()
    return(value)
  }

  # key provided

  # split parts, then decrypt -> as char -> fromJSON
  valueParts <- split_val_and_nonce(encodedValue)
  xChar <-
    sodium::data_decrypt(valueParts$value, key, valueParts$nonce) %>%
    rawToChar()

  tryCatch({
    x <- safeFromJSON(xChar)
  }, error = function(e) {
    # print warning
    warning("Cookie information could not be converted from JSON.  Please make sure session objects being stored can be converted to and from JSON cleanly.")
    # re throw error
    stop(e)
  })
  return(x)
}


# combine/split with "_" as it is not in the base64 vocab
combine_val_and_nonce <- function(val, nonce) {
  paste0(base64enc::base64encode(val), "_", base64enc::base64encode(nonce))
}
split_val_and_nonce <- function(x) {
  vals <- strsplit(x, "_")[[1]]
  if (length(vals) != 2) {
    stop("Could not separate secure cookie parts")
  }
  list(
    value = base64enc::base64decode(vals[[1]]),
    nonce = base64enc::base64decode(vals[[2]])
  )
}

context("Cookies")

skip_if_no_cookie_support <- function() {
  skip_if_not_installed("sodium")
  skip_if_not_installed("base64enc")
}

test_that("cookies are parsed", {

  cookies <- parseCookies("spaced=cookie%2C%20here; another=2")
  expect_equal(names(cookies), c("spaced", "another"))
  expect_equal(cookies$spaced, "cookie, here")
  expect_equal(cookies$another, "2")

  cookies <- parseCookies("a=zxcv=asdf; missingVal=; b=qwer=ttyui")
  expect_equal(names(cookies), c("a", "missingVal", "b"))
  expect_equal(cookies$a, "zxcv=asdf")
  expect_equal(cookies$missingVal, "")
  expect_equal(cookies$b, "qwer=ttyui")

})

test_that("missing cookies are an empty list", {
  cookies <- parseCookies("")

  expect_equal(cookies, list())
})

cookieReq <- function(cookieStr) {
  req <- new.env()
  req$HTTP_COOKIE <- cookieStr
  cookieFilter(req)
  req
}

test_that("the cookies list is set", {
  req <- cookieReq("abc=123")
  expect_equal(req$cookies$abc, "123")
})

test_that("missing cookie values are empty string", {
  req <- cookieReq("abc=")
  expect_equal(req$cookies$abc, "")
})

test_that("cookies can convert to string", {
  testthat::skip_on_cran()

  expect_equal(cookieToStr("abc", 123), "abc=123")
  expect_equal(cookieToStr("complex", "string with spaces"), "complex=string%20with%20spaces")
  expect_equal(cookieToStr("complex2", "forbidden:,%/"), "complex2=forbidden%3A%2C%25%2F")
  expect_equal(cookieToStr("abc", 123, path="/somepath"), "abc=123; Path=/somepath")
  expect_equal(cookieToStr("abc", 123, http=TRUE, secure=TRUE), "abc=123; HttpOnly; Secure")
  expect_equal(cookieToStr("abc", 123, http=TRUE, secure=TRUE, same_site="None"), "abc=123; HttpOnly; Secure; SameSite=None")

  now <- force(Sys.time())
  cookieToStr_ <- function(expiration, ...) {
    cookieToStr("abc", 123, expiration = expiration, ..., now = now)
  }
  cookie_match <- function(expirationStr, expiresSec) {
    # difftime is exclusive, so the Max-Age may be off by one on positive time diffs.
    #   Using a regex from 0 to 9 incase a slow machine is encountered
    # match from 0 to 9 seconds
    paste0("abc=123; Expires= ", expirationStr, "; Max-Age= ", expiresSec)
  }
  expect_cookie <- function(expiresSec) {
    expires <- now + expiresSec
    expyStr <- format(expires, format="%a, %e %b %Y %T", tz="GMT", usetz=TRUE)

    # When given as a number of seconds
    expect_equal(cookieToStr_(expiresSec), cookie_match(expyStr, expiresSec), label = "Raw seconds expiration cookie")

    # When given as a POSIXct
    expect_equal(cookieToStr_(expires), cookie_match(expyStr, expiresSec), label = "POSIXct expiration cookie")
  }

  # Test date in the future
  expect_cookie(9)

  # Works with a negative number of seconds
  expect_cookie(-8)
})

test_that("remove cookie string works", {
  expect_equal(
    removeCookieStr("asdf"),
    "asdf=; Expires=Thu, 01 Jan 1970 00:00:00 GMT"
  )
  expect_equal(
    removeCookieStr("asdf", path = "/"),
    "asdf=; Path=/; Expires=Thu, 01 Jan 1970 00:00:00 GMT"
  )
  expect_equal(
    removeCookieStr("asdf", http = TRUE),
    "asdf=; HttpOnly; Expires=Thu, 01 Jan 1970 00:00:00 GMT"
  )
  expect_equal(
    removeCookieStr("asdf", secure = TRUE),
    "asdf=; Secure; Expires=Thu, 01 Jan 1970 00:00:00 GMT"
  )
  expect_equal(
    removeCookieStr("asdf", path = "/", http = TRUE, secure = TRUE),
    "asdf=; Path=/; HttpOnly; Secure; Expires=Thu, 01 Jan 1970 00:00:00 GMT"
  )
  expect_equal(
    removeCookieStr("asdf", path = "/", http = TRUE, secure = TRUE, same_site = "None"),
    "asdf=; Path=/; HttpOnly; Secure; SameSite=None; Expires=Thu, 01 Jan 1970 00:00:00 GMT"
  )
})


test_that("asCookieKey conforms entropy", {
  skip_if_no_cookie_support()


  secretFromStr <- function(val, count) {
    rep(val, count) %>%
      paste0(collapse = "")
  }

  expect_cookie_key <- function(key) {
    expect_type(key, "raw")
    expect_length(key, 32)
  }
  expect_invalid_cookie <- function(secret) {
    expect_error({
      asCookieKey(secret)
    }, "Illegal cookie")
  }
  expect_legacy_cookie <- function(secret) {
    expect_warning({
      cookie <- asCookieKey(secret)
    }, "Legacy cookie secret")
    ret <- expect_cookie_key(cookie)
    invisible(ret)
  }

  expect_warning({
    expect_null(asCookieKey(NULL))
  }, "Cookies will not be encrypted")

  # not char
  expect_invalid_cookie(42)
  expect_invalid_cookie(sodium::random(31))
  expect_invalid_cookie(sodium::random(100))

  # legacy cookie
  # convert non hexadecimal to hexadecimal
  expect_legacy_cookie(secretFromStr("a", 63))
  expect_legacy_cookie(secretFromStr("a", 65))
  expect_legacy_cookie(secretFromStr("/", 64))

  # Used as 64 digit hex bin
  ## lower case a
  char <- secretFromStr("a", 64)
  key <- asCookieKey(char)
  expect_cookie_key(key)
  expect_equal(key, sodium::hex2bin(char))

  ## upper case a
  char <- secretFromStr("A", 64)
  key <- asCookieKey(char)
  expect_cookie_key(key)
  expect_equal(key, sodium::hex2bin(char))

  ## hex char input
  randomRaw <- sodium::random(32) %>% sodium::bin2hex()
  key <- asCookieKey(randomRaw)
  expect_cookie_key(key)
  expect_equal(sodium::bin2hex(key), randomRaw)
})


test_that("cookie encryption works", {
  skip_if_no_cookie_support()

  # check that you can't encode a NULL value
  expect_equal(encodeCookie(NULL, NULL), "")
  expect_equal(encodeCookie(NULL, asCookieKey(random_cookie_key())), "")

  xVals <- list(
    list(),
    "",
    list(a = 4, b = 3),
    rep("1234567890", 100) %>% paste0(collapse = "")
  )
  keys <- list(
    NULL, # no key
    asCookieKey(random_cookie_key()), # random key
    asCookieKey(random_cookie_key()) # different random key
  )

  for (key in keys) {
    for (x in xVals) {
      encrypted <- encodeCookie(x, key)
      encryptedStr <- cookieToStr("cookieVal", encrypted)

      encryptedParsed <- parseCookies(encryptedStr)
      maybeX <- decodeCookie(encryptedParsed$cookieVal, key)
      expect_equal(x, maybeX)
    }
  }

})

test_that("cookie encyption fails smoothly", {
  skip_if_no_cookie_support()

  x <- list(x = 4, y = 5)

  # garbage in
  garbage <- x %>%
    serialize(NULL) %>%
    sodium::hash() %>% # make "garbage"
    base64enc::base64encode()

  # garbage in, no key
  expect_error({
    decodeCookie(garbage, NULL)
  }) # error from jsonlite::parse_json()
  # garbage in, key
  expect_error({
    decodeCookie(garbage, asCookieKey(random_cookie_key()))
  }, "Could not separate")

  infoList <- list(
    # different cookies
    list(
      a = asCookieKey(random_cookie_key()),
      b = asCookieKey(random_cookie_key()),
      error = "Failed to decrypt"
    ),
    # not encrypted, try to decrypt
    list(
      a = NULL,
      b = asCookieKey(random_cookie_key()),
      error = "Could not separate"
    ),
    # encrypted, no decryption
    list(
      a = asCookieKey(random_cookie_key()),
      b = NULL
      # error from jsonlite::parse_json()
    )
  )

  for (info in infoList) {
    keyA <- info$a
    keyB <- info$b
    err <- info$error

    expect_error({
      encrypted <- encodeCookie(x, keyA)
      encryptedStr <- cookieToStr("cookieVal", encrypted)

      encryptedParsed <- parseCookies(encryptedStr)
      maybeX <- decodeCookie(encryptedParsed$cookieVal, keyB)
    }, err)
  }

})


test_that("large cookie size makes warning", {
  skip_if_no_cookie_support()

  largeObj <- rbind(iris, iris)
  encrypted <- encodeCookie(largeObj, NULL)
  expect_warning({
    cookieToStr("cookieVal", encrypted)
  }, "Cookie being saved is too large")
})

context("Sessions")

skip_if_no_cookie_support <- function() {
  skip_if_not_installed("sodium")
  skip_if_not_installed("base64enc")
}


make_req_cookie <- function(verb, path, cookie) {
  req <- new.env()
  req$REQUEST_METHOD <- toupper(verb)
  req$PATH_INFO <- path
  req$rook.input <- list(read_lines = function() { "" },
                         rewind = function() {},
                         read = function() { charToRaw("") })
  if (!missing(cookie)){
    req$HTTP_COOKIE <- cookie
  }
  req
}

test_that("session_cookie throws missing key", {
  expect_error(
    session_cookie(),
    "You must define an encryption key"
  )
})

test_that("cookies are set", {
  skip_if_no_cookie_support()

  r <- pr()
  expr <- expression(function(req, res){ req$session <- list(abc = 1234); TRUE })

  r$handle("GET", "/", expr)

  key <- random_cookie_key()
  sc <- session_cookie(
    key,
    name = "plcook"
  )

  r$registerHooks(sc)

  res <- PlumberResponse$new()
  r$serve(make_req_cookie("GET", "/"), res)

  cook <- res$headers[["Set-Cookie"]]
  expect_match(cook, "^plcook")
  cook <- parseCookies(cook)$plcook
  expect_equal(decodeCookie(cook, asCookieKey(key)), list(abc = 1234))
})

test_that("cookies are unset", {
  skip_if_no_cookie_support()

  r <- pr()
  exprRemoveSession <- expression(function(req, res){ req$session <- NULL; TRUE })

  r$handle("GET", "/", exprRemoveSession)

  key <- random_cookie_key()
  sc <- session_cookie(
    key,
    name = "plcook"
  )

  r$registerHooks(sc)

  res <- PlumberResponse$new()
  r$serve(
    make_req_cookie(
      "GET", "/",
      # start with a session cookie
      paste0("plcook=", encodeCookie(list(abc = 1234), asCookieKey(key)))
    ),
    res
  )

  cook <- res$headers[["Set-Cookie"]]
  expect_match(cook, "^plcook=;")
  expect_true(grepl("Thu, 01 Jan 1970", cook, fixed = TRUE))
})

test_that("cookies are read", {
  skip_if_no_cookie_support()

  r <- pr()

  expr <- expression(function(req, res){ req$session$abc })

  r$handle("GET", "/", expr)

  key <- random_cookie_key()
  sc <- session_cookie(
    key,
    name = "plcook"
  )
  r$registerHooks(sc)


  # Create the request with an encrypted cookie
  res <- PlumberResponse$new()
  r$serve(
    make_req_cookie(
      "GET", "/",
      # start with a session cookie
      paste0("plcook=", encodeCookie(list(abc = 1234), asCookieKey(key)))
    ),
    res
  )

  expect_equal(res$body, jsonlite::toJSON(1234))
})

test_that("invalid cookies/JSON are handled", {
  skip_if_no_cookie_support()

  r <- pr()

  expr <- expression(function(req, res){ ifelse(is.null(req$session), "no session", req$session) })

  r$handle("GET", "/", expr)

  key <- random_cookie_key()
  sc <- session_cookie(
    key,
    name = "plcook"
  )
  r$registerHooks(sc)

  res <- PlumberResponse$new()

  badKey <- random_cookie_key()
  x <- list(abc = 1234)
  encodedX <- encodeCookie(x, asCookieKey(badKey))
  expect_silent({
    r$serve(
      make_req_cookie(
        "GET", "/",
        paste0('plcook=', encodedX)
      ),
      res
    )
  })
  expect_equal(res$body, jsonlite::toJSON("no session"))
})

test_that("cookie attributes are set", {
  skip_if_no_cookie_support()

  r <- pr()
  expr <- expression(function(req, res){ req$session <- list(abc = 1234); TRUE })

  r$handle("GET", "/", expr)

  key <- random_cookie_key()
  sc <- session_cookie(
    key,
    name = "plcook",
    expiration = 10,
    http = TRUE,
    secure = TRUE,
    same_site = "None"
  )

  r$registerHooks(sc)

  res <- PlumberResponse$new()
  r$serve(make_req_cookie("GET", "/"), res)

  cook <- res$headers[["Set-Cookie"]]
  expect_match(cook, "^plcook")
  expect_match(cook, "Expires=[^;]+(?:;|$)")
  expect_match(cook, "Max-Age=\\s*\\d+(?:;|$)")
  expect_match(cook, "HttpOnly(?:;|$)")
  expect_match(cook, "Secure(?:;|$)")
  expect_match(cook, "SameSite=None(?:;|$)")
  cook <- parseCookies(cook)$plcook
  expect_equal(decodeCookie(cook, asCookieKey(key)), list(abc = 1234))
})

#' DigitalOcean Plumber server
#'
#' These methods are now defunct.
#' Please use the [`plumberDeploy`](https://github.com/meztez/plumberDeploy) R package.
#'
#' @keywords internal
#' @rdname digitalocean
#' @export
do_provision <- function(...) {
  plumberDeploy_helper("do_provision", list(...))
}

#' @export
#' @rdname digitalocean
do_configure_https <- function(...) {
  plumberDeploy_helper("do_configure_https", list(...))
}

#' @export
#' @rdname digitalocean
do_deploy_api <- function(...) {
  plumberDeploy_helper("do_deploy_api", list(...))
}

#' @export
#' @rdname digitalocean
do_forward <- function(...) {
  plumberDeploy_helper("do_forward", list(...))
}

#' @export
#' @rdname digitalocean
do_remove_api <- function(...) {
  plumberDeploy_helper("do_remove_api", list(...))
}

#' @export
#' @rdname digitalocean
do_remove_forward <- function(...) {
  plumberDeploy_helper("do_remove_forward", list(...))
}


plumberDeploy_helper <- function(fn_name, args, new_fn_name = fn_name) {
  cur_fn <- paste0(fn_name, "()")
  new_fn <- paste0("plumberDeploy", "::", new_fn_name, "()")

  # check if plumberDeploy is called
  if (!plumberDeploy_is_available()) {
    # not found
    # throw error
    lifecycle::deprecate_stop("1.0.0", cur_fn, new_fn)
  }

  # plumberDeploy is found
  # 1. Throw warning
  # 2. Call method
  lifecycle::deprecate_warn("1.0.0", cur_fn, new_fn)

  fn <- utils::getFromNamespace(fn_name, "plumberDeploy")
  do.call(fn, args)
}

plumberDeploy_is_available <- function() {
  is_available("plumberDeploy")
}

context("Deprecated")

test_that("addEndpoint continues to work", {
  pr <- Plumber$new()
  expect_warning(pr$addEndpoint("GET", "/", function(){ 123 }))
  expect_error(expect_warning(pr$addEndpoint("GET", "/", function(){ 123 }, comments="break")))

  val <- pr$route(make_req("GET", "/"), PlumberResponse$new())
  expect_equal(val, 123)
})

test_that("addFilter continues to work", {
  pr <- Plumber$new()
  expect_warning(pr$addFilter("f1", function(req){ req$filtered <- TRUE }))
  pr$handle("GET", "/", function(req){ req$filtered })

  val <- pr$route(make_req("GET", "/"), PlumberResponse$new())
  expect_true(val)
})

test_that("addGlobalProcessor continues to work", {
  pr <- Plumber$new()
  expect_warning(pr$addGlobalProcessor(session_cookie("secret", "cookieName")))
})

test_that("addAssets continues to work", {
  pr <- Plumber$new()
  expect_warning(pr$addAssets(test_path("./files/static"), "/public"))
  res <- PlumberResponse$new()
  val <- pr$route(make_req("GET", "/public/test.txt"), res)
  expect_true(inherits(val, "PlumberResponse"))
})


test_that("getCharacterSet continues to work", {
  expect_equal(
    lifecycle::expect_deprecated(getCharacterSet(contentType = "foo")),
    "UTF-8"
  )
})


test_that("sessionCookie continues to work", {
  key <- random_cookie_key()
  cookie_hooks_old <- lifecycle::expect_deprecated(sessionCookie(key))
  cookie_hooks_new <- expect_silent(session_cookie(key))

  expect_equal(names(cookie_hooks_old), names(cookie_hooks_new))
})

test_that("hookable throws deprecated warning", {
  expect_warning(hookable$new(), "Hookable")
})
test_that("plumber throws deprecated warning", {
  expect_warning(plumber$new(), "Plumber")
})


test_that("Digital Ocean functions throw errors", {
  skip_on_cran()

  # Do not test if plumberDeploy is installed, as real functions will executed
  skip_if(plumberDeploy_is_available())

  expect_error(do_provision(), class = "lifecycle_error_deprecated")
  expect_error(do_configure_https(), class = "lifecycle_error_deprecated")
  expect_error(do_deploy_api(), class = "lifecycle_error_deprecated")
  expect_error(do_forward(), class = "lifecycle_error_deprecated")
  expect_error(do_remove_api(), class = "lifecycle_error_deprecated")
  expect_error(do_remove_forward(), class = "lifecycle_error_deprecated")
})

queryStringFilter <- function(req){
  handled <- req$.internal$queryStringHandled
  if (is.null(handled) || handled != TRUE) {
    qs <- req$QUERY_STRING
    args <- parseQS(qs)
    req$args <- c(req$args, args)
    req$argsQuery <- args
    req$.internal$queryStringHandled <- TRUE
  }
  forward()
}

#' @noRd
parseQS <- function(qs){

  if (is.null(qs) || length(qs) == 0L || qs == "") {
    return(list())
  }

  # Looked into using webutils::parse_query()
  # Currently not pursuing `webutils::parse_query` as it does not handle Encoding issues handled below
  # (Combining keys are also not handled by `webutils::parse_query`)

  qs <- stri_replace_first_regex(qs, "^[?]", "")
  qs <- chartr("+", " ", qs)

  args <- stri_split_fixed(qs, "&", omit_empty = TRUE)[[1L]]
  kv <- lapply(args, function(x) {
    # returns utf8 strings
    httpuv::decodeURIComponent(stri_split_fixed(x, "=", omit_empty = TRUE)[[1]])
  })
  kv <- kv[vapply(kv, length, numeric(1)) == 2] # Ignore incompletes

  if (length(kv) == 0) {
    # return a blank list of args if there is nothing to parse
    return(list())
  }

  keys <- vapply(kv, `[`, character(1), 1)
  kenc <- unique(Encoding(keys))
  if (any(kenc != "unknown")) {
    # https://github.com/rstudio/plumber/pull/314#discussion_r239992879
    non_ascii <- setdiff(kenc, "unknown")
    warning(
      "Query string parameter received in non-ASCII encoding. Received: ",
      paste0(non_ascii, collapse = ", ")
    )
  }

  vals <- lapply(kv, `[`, 2)
  names(vals) <- keys

  # If duplicates, combine
  combine_keys(vals, type = "query")
}

createPathRegex <- function(pathDef, funcParams = NULL){
  # Create a regex from the defined path, substituting variables where appropriate
  match <- stri_match_all(
    pathDef,
    # capture any plumber type (<arg:TYPE>).
    # plumberToApiType(types) will yell if it is unknown
    # and can not be guessed from endpoint function args)
    # <arg> will be given the TYPE `defaultApiType`
    regex = "/<(\\.?[a-zA-Z][\\w_\\.]*)(?::([^>]*))?>"
  )[[1]]
  names <- match[,2]
  # No path params
  if (length(names) <= 1 && is.na(names)) {
    return(
      list(
        names = character(),
        types = NULL,
        regex = paste0("^", pathDef, "$"),
        converters = NULL,
        areArrays = NULL
      )
    )
  }

  plumberTypes <- stri_replace_all(match[,3], "$1", regex = "^\\[([^\\]]*)\\]$")
  if (length(funcParams) > 0) {
    # Override with detection of function args if type not found in map
    idx <- !(plumberTypes %in% names(plumberToApiTypeMap))
    plumberTypes[idx] <- sapply(funcParams, `[[`, "type")[names[idx]]
  }
  apiTypes <- plumberToApiType(plumberTypes, inPath = TRUE)

  areArrays <- stri_detect_regex(match[,3], "^\\[[^\\]]*\\]$")
  if (length(funcParams) > 0) {
    # Override with detection of function args when false or na
    idx <- (is.na(areArrays) | !areArrays)
    areArrays[idx] <- sapply(funcParams, `[[`, "isArray")[names[idx]]
  }
  areArrays[is.na(areArrays)] <- defaultIsArray

  pathRegex <- pathDef
  regexps <- typesToRegexps(apiTypes, areArrays)
  for (regex in regexps) {
    pathRegex <- stri_replace_first_regex(
      pathRegex,
      pattern = "/(?:<\\.?[a-zA-Z][\\w_\\.:\\[\\]]*>)(/?)",
      replacement = paste0("/(", regex, ")$1")
    )
  }

  list(
    names = names,
    types = apiTypes,
    regex = paste0("^", pathRegex, "$"),
    converters = typesToConverters(apiTypes, areArrays),
    areArrays = areArrays
  )
}


typesToRegexps <- function(apiTypes, areArrays = FALSE) {
  # return vector of regex strings
  mapply(
    function(x, y) {x[[y]]},
    apiTypesInfo[apiTypes],
    ifelse(areArrays, "regexArray", "regex"),
    USE.NAMES = FALSE
  )
}


typesToConverters <- function(apiTypes, areArrays = FALSE) {
  # return list of functions
  mapply(
    function(x, y) {x[[y]]},
    apiTypesInfo[apiTypes],
    ifelse(areArrays, "converterArray", "converter"),
    USE.NAMES = FALSE
  )
}


# Extract the params from a given path
# @param def is the output from createPathRegex
extractPathParams <- function(def, path){
  vals <- as.list(stri_match(path, regex = def$regex)[,-1])
  names(vals) <- def$names

  if (!is.null(def$converters)){
    # Run each value through its converter
    for (i in 1:length(vals)){
      vals[[i]] <- def$converters[[i]](vals[[i]])
    }
  }

  vals
}


#' combine args that share the same name
#' @noRd
#' @importFrom stats setNames
combine_keys <- function(obj, type) {

  keys <- names(obj)
  unique_keys <- unique(keys)

  # If a query string as the same amount of unique keys as keys,
  # then return it as it
  # (`"multi"` type objects MUST be processed, regardless if the unique key count is the same)
  if (
    length(unique_keys) == length(keys) &&
    identical(type, "query")
  ) {
    return(obj)
  }

  vals <- unname(obj)

  cleanup_item <- switch(
    type,
    "query" =
      function(x) {
        unname(unlist(x))
      },
    "multi" =
      function(x) {
        if (length(x) == 1) {
          part <- x[[1]]
          filename <- part$filename
          parsed <- part$parsed

          if (!is.null(filename)) {
            # list(
            #   "myfile.json" = list(
            #     a = 1, b = 2
            #   )
            # )
            return(
              setNames(
                list(parsed),
                filename
              )
            )
          }
          # list(
          #   a = 1, b = 2
          # )
          return(parsed)
        }

        # length is > 1

        has_a_filename <- FALSE
        filenames <- lapply(x, function(part) {
          filename <- part$filename
          if (is.null(filename)) return("")
          has_a_filename <<- TRUE
          filename
        })

        parsed_items <- lapply(unname(x), `[[`, "parsed")

        if (!has_a_filename) {
          # return as is
          return(parsed_items)
        }

        return(setNames(parsed_items, filenames))
      },
    stop("unknown type: ", type)
  )

  # equivalent code output, `split` is much faster with larger objects
  # Testing on personal machine had a breakpoint around 150 letters as query parameters
  ## n <- 150
  ## k <- sample(letters, n, replace = TRUE)
  ## v <- as.list(sample(1L, n, replace = TRUE))
  ## microbenchmark::microbenchmark(
  ##   split = {
  ##     lapply(split(v, k), function(x) unname(unlist(x)))
  ##   },
  ##   not_split = {
  ##     lapply(unique(k), function(x) {
  ##       unname(unlist(v[k == x]))
  ##     })
  ##   }
  ## )
  vals <-
    if (length(unique_keys) > 150) {
      lapply(split(vals, keys), function(items) {
        cleanup_item(items)
      })
    } else {
      # n < 150
      lapply(unique_keys, function(key) {
        cleanup_item(vals[keys == key])
      })
    }
  names(vals) <- unique_keys

  vals
}

context("combine multipart values")

test_that("multi part keys with same name are combined", {
  # no lists
  a <- list(
    A = list(name = "A", parsed = 1),
    A = list(name = "A", parsed = 2),
    B = list(name = "B", parsed = 3),
    A = list(name = "A", parsed = 4),
    B = list(name = "B", parsed = 5)
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(1,2,4), B = list(3,5)))

  # unnamed lists
  a <- list(
    A = list(name = "A", parsed = 1),
    A = list(name = "A", parsed = list(11, 12)),
    B = list(name = "B", parsed = 3),
    A = list(name = "A", parsed = 4),
    B = list(name = "B", parsed = 5)
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(1,list(11,12),4), B = list(3,5)))

  # partial named lists
  a <- list(
    A = list(name = "A", parsed = 1),
    A = list(name = "A", parsed = list(X = 11, 12)),
    B = list(name = "B", parsed = 3),
    A = list(name = "A", parsed = list(Y = 4)),
    B = list(name = "B", parsed = 5)
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(1,list(X = 11, 12), list(Y = 4)), B = list(3,5)))

  # named lists
  a <- list(
    A = list(name = "A", parsed = 1),
    A = list(name = "A", parsed = list(X = 11, Y = 12)),
    B = list(name = "B", parsed = 3),
    A = list(name = "A", parsed = list(Z = 4)),
    B = list(name = "B", parsed = 5)
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(1,list(X = 11, Y = 12), list(Z = 4)), B = list(3,5)))
})


test_that("multi part keys with all same name", {
  a <- list(
    A = list(name = "A", parsed = 1),
    A = list(name = "A", parsed = 2),
    A = list(name = "A", parsed = 3),
    A = list(name = "A", parsed = 4),
    A = list(name = "A", parsed = 5)
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(1,2,3,4,5)))

  a <- list(
    A = list(name = "A", parsed = 1),
    A = list(name = "A", parsed = list(2)),
    A = list(name = "A", parsed = 3),
    A = list(name = "A", parsed = 4),
    A = list(name = "A", parsed = 5)
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(1,list(2),3,4,5)))

  a <- list(
    A = list(name = "A", parsed = 1),
    A = list(name = "A", parsed = list(2)),
    A = list(name = "A", parsed = 3),
    A = list(name = "A", parsed = list(4, 5))
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(1,list(2),3,list(4,5))))
})


test_that("raw values are not combined", {
  x <- charToRaw("testval")
  y <- charToRaw("other testval")

  a <- list(
    A = list(name = "A", parsed = x),
    B = list(name = "B", parsed = 2),
    A = list(name = "A", parsed = y),
    A = list(name = "A", parsed = 4)
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(x, y, 4), B = 2))

  a <- list(
    A = list(name = "A", parsed = x, filename = "x"),
    B = list(name = "B", parsed = 2),
    A = list(name = "A", parsed = y),
    A = list(name = "A", parsed = 4)
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(x = x, y, 4), B = 2))

  a <- list(
    A = list(name = "A", parsed = x, filename = "x"),
    B = list(name = "B", parsed = 2, filename = "two"),
    A = list(name = "A", parsed = y),
    A = list(name = "A", parsed = 4)
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(x = x, y, 4), B = list(two = 2)))

  a <- list(
    A = list(name = "A", parsed = x),
    B = list(name = "B", parsed = 2, filename = "two"),
    A = list(name = "A", parsed = y, filename = "y"),
    A = list(name = "A", parsed = 4)
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(x, y = y, 4), B = list(two = 2)))

  a <- list(
    A = list(name = "A", parsed = x, filename = "x"),
    B = list(name = "B", parsed = 2, filename = "two"),
    A = list(name = "A", parsed = y, filename = "y"),
    A = list(name = "A", parsed = 4)
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(x = x, y = y, 4), B = list(two = 2)))

  a <- list(
    A = list(name = "A", parsed = x, filename = "x"),
    B = list(name = "B", parsed = 2, filename = "two"),
    A = list(name = "A", parsed = y, filename = "y"),
    A = list(name = "A", parsed = 4, filename = "four")
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(x = x, y = y, four = 4), B = list(two = 2)))

  a <- list(
    A = list(name = "A", parsed = list(x), filename = "x"),
    B = list(name = "B", parsed = 2, filename = "two"),
    A = list(name = "A", parsed = y, filename = "y"),
    A = list(name = "A", parsed = 4, filename = "four")
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(x = list(x), y = y, four = 4), B = list(two = 2)))

  a <- list(
    A = list(name = "A", parsed = list(x), filename = "same"),
    B = list(name = "B", parsed = 2, filename = "same"),
    A = list(name = "A", parsed = y, filename = "same"),
    A = list(name = "A", parsed = 4, filename = "same")
  )
  expect_equal(combine_keys(a, "multi"), list(A = list(same = list(x), same = y, same = 4), B = list(same = 2)))
})

context("query strings")

test_that("query strings are properly parsed", {
  expect_equal(parseQS("?a=1"), list(a="1"))
  expect_equal(parseQS("b=2"), list(b="2"))
  expect_equal(parseQS("a=1&b=2&c=url%20encoded"), list(a="1", b="2", c="url encoded"))
})

test_that("path parameters do not convert + to space", {
  r <- pr(test_path("files/path-params.R"))
  expect_equal(r$route(make_req("GET", "/car/a+b"), PlumberResponse$new()), "a+b")
})

test_that("special characters in query strings are handled properly", {
  expect_equal(parseQS("?a=1+.#"), list(a="1 .#"))
  expect_equal(parseQS("?a=a%20b"), list(a="a b"))
  expect_equal(parseQS('?a=%2C%2B%2F%3F%25%26'), list(a=",+/?%&"))
})

test_that("null an empty strings return empty list", {
  expect_equal(parseQS(NULL), list())
  expect_equal(parseQS(""), list())
})

test_that("incomplete query strings are ignored", {
  expect_equivalent(parseQS("a="), list()) # It's technically a named list()
  expect_equal(parseQS("a=1&b=&c&d=1"), list(a="1", d="1"))
})

test_that("query strings with duplicates are made into vectors", {
  expect_equal(parseQS("a=1&a=2&a=3&a=4"), list(a=c("1", "2", "3", "4")))
})

test_that("parseQS() will mark UTF-8 explicitly", {
  expect_warning(
    {
      out <- parseQS("%E5%8F%82%E6%95%B01=%E4%B8%AD%E6%96%87")
    },
    "received in non-ASCII encoding"
  )
  expect_equal(Encoding(names(out)), "UTF-8")
  expect_identical(
    charToRaw(names(out)),
    as.raw(c(0xe5, 0x8f, 0x82, 0xe6, 0x95, 0xb0, 0x31))
  )
  expect_identical(
    charToRaw(out[[1L]]),
    as.raw(c(0xe4, 0xb8, 0xad, 0xe6, 0x96, 0x87))
  )
  expect_equal(Encoding(out[[1L]]), "UTF-8")
})


test_that("different lengths of query string return same shape", {
  for (n in c(5, 50, 500, 1000)) {
    keys <- sample(letters, n, replace = TRUE)
    vals <- as.list(sample(letters, n, replace = TRUE))

    expect_equal(
      parseQS(
        paste0("?", paste0(keys, "=", vals, collapse = "&"))
      ),
      stats::setNames(
        lapply(unique(keys), function(key) {
          unname(unlist(vals[keys == key]))
        }),
        unique(keys)
      )
    )

  }
})

context("Paths")

test_that("paths are properly converted", {
  varRegex <- "([^/]+)"
  p <- createPathRegex("/car/")
  expect_equal(p$names, character())
  expect_equal(p$regex, "^/car/$")

  p <- createPathRegex("/car/<id>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", varRegex, "$"))

  p <- createPathRegex("/car/<id>/sell")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", varRegex, "/sell$"))

  p <- createPathRegex("/car/<id>/sell/<price>")
  expect_equal(p$names, c("id", "price"))
  expect_equal(p$regex, paste0("^/car/", varRegex, "/sell/", varRegex, "$"))
})

test_that("variables are typed", {
  p <- createPathRegex("/car/<id:int>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", "(-?\\d+)", "$"))

  p <- createPathRegex("/car/<id:[int]>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", "((?:(?:-?\\d+),?)+)", "$"))

  p <- createPathRegex("/car/<id:double>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", "(-?\\d*\\.?\\d+)", "$"))
  p <- createPathRegex("/car/<id:[double]>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", "((?:(?:-?\\d*\\.?\\d+),?)+)", "$"))

  p <- createPathRegex("/car/<id:numeric>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", "(-?\\d*\\.?\\d+)", "$"))
  p <- createPathRegex("/car/<id:[numeric]>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", "((?:(?:-?\\d*\\.?\\d+),?)+)", "$"))

  p <- createPathRegex("/car/<id:bool>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", "([01tfTF]|true|false|TRUE|FALSE)", "$"))
  p <- createPathRegex("/car/<id:[bool]>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", "((?:(?:[01tfTF]|true|false|TRUE|FALSE),?)+)", "$"))

  p <- createPathRegex("/car/<id:logical>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", "([01tfTF]|true|false|TRUE|FALSE)", "$"))
  p <- createPathRegex("/car/<id:[logical]>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", "((?:(?:[01tfTF]|true|false|TRUE|FALSE),?)+)", "$"))

  p <- createPathRegex("/car/<id:chr>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", "([^/]+)", "$"))
  p <- createPathRegex("/car/<id:[chr]>")
  expect_equal(p$names, "id")
  expect_equal(p$regex, paste0("^/car/", "((?:(?:[^/]+),?)+)", "$"))
  expect_equal(p$areArrays, TRUE)
  expect_equal(p$converters[[1]]("BOB,LUKE,GUY"), c("BOB", "LUKE", "GUY"))

  #Check that warnings happen on typo or unsupported type
  expect_warning(createPathRegex("/car/<id:motor>"),
                 "Unrecognized type")
  expect_warning(createPathRegex("/car/<id:[motor>"),
                 "Unrecognized type")
  expect_warning(createPathRegex("/car/<id:[df*>"),
                 "Unrecognized type")
  expect_warning(createPathRegex("/car/<id:df]*>"),
                 "Unrecognized type")
  expect_warning(createPathRegex("/car/<id:df]>"),
                 "Unrecognized type")

})

test_that("path regex's are created properly", {
  expect_equivalent(extractPathParams(createPathRegex("/car/"), "/car/"),  list())
  expect_equal(extractPathParams(createPathRegex("/car/<id>"), "/car/15"), list(id="15") )
  expect_equal(extractPathParams(createPathRegex("/car/<id>/sell"), "/car/12/sell"), list(id="12") )
  expect_equal(extractPathParams(createPathRegex("/car/<id>/sell/<price>"), "/car/15/sell/$15,000"), list(id="15", price="$15,000"))
})

test_that("integration of path parsing works", {
  r <- pr(test_path("files/path-params.R"))

  expect_equal(r$route(make_req("GET", "/car/13"), PlumberResponse$new()), "13")
  expect_equal(r$route(make_req("GET", "/car/int/13"), PlumberResponse$new()), 13)
  expect_equal(r$route(make_req("GET", "/car/int/-13"), PlumberResponse$new()), -13)
  expect_equal(r$route(make_req("GET", "/car/15/sell/$15,000"), PlumberResponse$new()), list(id="15", price="$15,000"))
  expect_equal(r$route(make_req("POST", "/car/13"), PlumberResponse$new()), "13")
  expect_equal(r$route(make_req("GET", "/car/15/buy/$15,000"), PlumberResponse$new()),
               list(id=15, price="$15,000"))
  expect_equal(r$route(make_req("GET", "/car/15/buy/$15,000.99"), PlumberResponse$new()),
               list(id=15, price="$15,000.99"))
  expect_equal(r$route(make_req("GET", "/car/ratio/1.5"), PlumberResponse$new()), 1.5)
  expect_equal(r$route(make_req("GET", "/car/ratio/-1.5"), PlumberResponse$new()), -1.5)
  expect_equal(r$route(make_req("GET", "/car/ratio/-.5"), PlumberResponse$new()), -.5)
  expect_equal(r$route(make_req("GET", "/car/ratio/.5"), PlumberResponse$new()), .5)
  expect_equal(r$route(make_req("GET", "/car/ratio/a"), PlumberResponse$new()),
               list(error = "404 - Resource Not Found"))
  expect_equal(r$route(make_req("GET", "/car/ratio/"), PlumberResponse$new()),
               list(error = "404 - Resource Not Found"))
  expect_equal(r$route(make_req("GET", "/car/ratio/."), PlumberResponse$new()),
               list(error = "404 - Resource Not Found"))
  expect_equal(r$route(make_req("GET", "/car/sold/true"), PlumberResponse$new()), TRUE)
  expect_match(r$call(make_req("POST", "/car/sold/true"))$body,
               "405 - Method Not Allowed")
})

test_that("multiple variations in path works nicely with function args detection", {

  # Check when detection is not provided
  pathDef <- "/<var0:str>/<var1:chr*>/<var2:[int]>/<var3>/<var4:*>/<var5:[*>/<var6:[]*>/<var7:[]>/<var8*>"
  expect_warning(regex <- createPathRegex(pathDef), "Unrecognized type")
  expect_equal(regex$types, c("string", "string", "integer", "string", "string", "string", "string", "string"))
  expect_equal(regex$areArrays, c(FALSE, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE))

  # Check when no args in endpoint function
  dummy <- function() {}
  funcParams <- getArgsMetadata(dummy)
  expect_warning(regex <- createPathRegex(pathDef, funcParams), "Unrecognized type")
  expect_equal(regex$types, c("string", "string", "integer", "string", "string", "string", "string", "string"))
  expect_equal(regex$areArrays, c(FALSE, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE))

  # Mix and match
  dummy <- function(var0 = 420.69,
                    var1,
                    var2 = c(1L, 2L),
                    var3 = rnorm,
                    var4 = NULL,
                    var5 = c(TRUE, FALSE),
                    var6 = list(name = c("luke", "bob"), lastname = c("skywalker", "ross")),
                    var7 = new.env(parent = .GlobalEnv),
                    var8 = list(a = 2, b = mean, c = new.env(parent = .GlobalEnv))) {}
  funcParams <- getArgsMetadata(dummy)
  expect_warning(regex <- createPathRegex(pathDef, funcParams), "Unsupported path parameter type")
  expect_equal(regex$types, c("string", "string", "integer", "string", "string", "boolean", "string", "string"))
  expect_equal(regex$areArrays, c(FALSE, FALSE, TRUE, FALSE, FALSE, TRUE, TRUE, TRUE))

  # Throw sand at it
  pathDef <- "/<>/<:chr*>/<:chr>/<henry:[IV]>"
  regex <- createPathRegex(pathDef, funcParams)
  expect_equivalent(regex$types, "string")
  expect_equal(regex$names, "henry")
  # Since type IV is converted to string, areArrays can be TRUE
  expect_equal(regex$areArrays, TRUE)

})

# testthat:::on_cran
is_interactive <- rlang::is_interactive()
not_cran <- isTRUE(as.logical(Sys.getenv("NOT_CRAN", "false")))
on_cran <- !is_interactive && !not_cran

if (!on_cran) {
  if (requireNamespace("spelling", quietly = TRUE)) {
    spelling::spell_check_test(
      vignettes = TRUE,
      error = TRUE,
      skip_on_cran = TRUE
    )
  }
}

#' Process a Plumber API
#'
#' @details API routers are the core request handler in plumber. A router is responsible for
#' taking an incoming request, submitting it through the appropriate filters and
#' eventually to a corresponding endpoint, if one is found.
#'
#' See the [Programmatic Usage](https://www.rplumber.io/articles/programmatic-usage.html) article for additional
#' details on the methods available on this object.
#' @param file The file to parse as the plumber router definition.
#' @param dir The directory containing the `plumber.R` file to parse as the
#'   plumber router definition. Alternatively, if an `entrypoint.R` file is
#'   found, it will take precedence and be responsible for returning a runnable
#'   router.
#' @export
plumb <- function(file = NULL, dir = ".") {

  if (!is.null(file) && !identical(dir, ".")) {
    # both were explicitly set.
    # assume it is a file in that dir and continue like normal
    file <- file.path(
      # removing trailing slash in dir
      normalize_dir_path(dir),
      file
    )
  }

  if (is.null(file)) {
    if (identical(dir, "")) {
      # dir and file are both empty. Error
      stop("You must specify either a file or directory parameter")
    }

    dir <- normalize_dir_path(dir)

    # if the entrypoint file exists...
    entrypoint <- list.files(dir, "^entrypoint\\.r$", ignore.case = TRUE)
    if (length(entrypoint) >= 1) {
      if (length(entrypoint) > 1) {
        entrypoint <- entrypoint[1]
        warning("Found multiple files named 'entrypoint.R'. Using: '", entrypoint, "'")
      }

      # set working directory to dir before sourcing
      old_wd <- setwd(dir)
      on.exit(setwd(old_wd), add = TRUE)

      # Expect that entrypoint will provide us with the router
      #   Do not 'poison' the global env. Using a local environment that points to global env.
      #   sourceUTF8 returns the (visible) value object. No need to call source()$value()
      pr <- sourceUTF8(entrypoint, new.env(parent = globalenv()))

      if (!is_plumber(pr)) {
        stop("'", entrypoint, "' must return a runnable Plumber router.")
      }

      # return plumber object
      return(pr)
    }

    # Find plumber.R in the directory case-insensitive
    file <- list.files(dir, "^plumber\\.r$", ignore.case = TRUE, full.names = TRUE)
    if (length(file) == 0) {
      stop("No plumber.R file found in the specified directory: ", dir)
    }
    if (length(file) > 1) {
      file <- file[1]
      warning("Found multiple files named 'plumber.R' in directory: '", dir, "'.\nUsing: '", file, "'")
    }
    # continue as if a file has been provided...
  }

  if (!file.exists(file)) {
    # Couldn't find the Plumber file nor an entrypoint
    stop("File does not exist: ", file)
  }

  # Plumber file found
  Plumber$new(file)
}




#' Process a Package's Plumber API
#'
#' So that packages can ship multiple plumber routers, users should store their Plumber APIs
#' in the `inst` subfolder `plumber` (`./inst/plumber/API_1/plumber.R`).
#'
#' To view all available Plumber APIs across all packages, please call `available_apis()`.
#' A `package` value may be provided to only display a particular package's Plumber APIs.
#'
#' @param package Package to inspect
#' @param name Name of the package folder to [plumb()].
#' @param edit Whether or not to open the API source code for viewing / editing
#' @describeIn plumb_api [plumb()]s a package's Plumber API. Returns a [`Plumber`] router object
#' @return A [`Plumber`] object. If either `package` or `name` is null, the appropriate [available_apis()] will be returned.
#' @importFrom utils file.edit
#' @export
plumb_api <- function(package = NULL, name = NULL, edit = FALSE) {

  if (is.null(package)) {
    return(available_apis(package = NULL))
  }
  if (is.null(name)) {
    return(available_apis(package = package))
  }

  stopifnot(length(package) == 1)
  stopifnot(length(name) == 1)
  stopifnot(is.character(package))
  stopifnot(is.character(name))

  apis <- available_apis(package = package)
  apis_sub <- (apis$package == package) & (apis$name == name)
  if (sum(apis_sub) == 0) {
    stop("Could not find Plumber API for package '", package, "'  with name '", name, "'")
  }

  if (isTRUE(edit)) {
    api_dir <- apis[apis_sub, "source_directory"]

    if (file.exists(file.path(api_dir, "entrypoint.R"))) {
      file_loc <- file.path(api_dir, "entrypoint.R")
    } else {
      file_loc <- file.path(api_dir, "plumber.R")
    }

    # Check for RStudio running
    if (requireNamespace("rstudioapi", quietly = TRUE)) {
      if (rstudioapi::isAvailable()) {
        rstudioapi::navigateToFile(file_loc)
      } else {
        file.edit(file_loc)
      }
    } else {
      file.edit(file_loc)
    }
    warning(
      file_loc, " has been opened in the editor.",
      " Any changes saved to this file are permanent until the ", package, " package is reinstalled.",
      " If you would like to make persistent changes, consider copying the contents of this file to a new file.",
      call. = FALSE
    )
  }

  plumb(
    dir = system.file(
      file.path("plumber", name),
      package = package
    )
  )
}


#' @describeIn plumb_api Displays all available package Plumber APIs. Returns a `data.frame` of `package`, `name`, and `source_directory` information.
#' @export
available_apis <- function(package = NULL) {
  info <-
    if (is.null(package)) {
      all_available_apis()
    } else {
      available_apis_for_package(package)
    }
  if (!is.null(info$error)) {
    stop(info$error, call. = FALSE)
  }
  apis <- info$apis
  return(apis)
}


#' @return will return a list of `error` and `apis`.
#'   `apis` is a \code{data.frame} containing
#'    "package": name of package; string
#'    "name": API directory. (can be passed in as `plumb_api(PKG, NAME)`; string
#' @noRd
available_apis_for_package <- function(package) {

  an_error <- function(...) {
    list(
      apis = NULL,
      error = paste0(...)
    )
  }

  if (!nzchar(system.file(package = package))) {
    return(an_error(
      "No package found with name: \"", package, "\""
    ))
  }

  apis_dir <- system.file("plumber", package = package)
  if (!nzchar(apis_dir)) {
    return(an_error(
      "No Plumber APIs found for package: \"", package, "\""
    ))
  }

  api_folders <- list.dirs(apis_dir, full.names = TRUE, recursive = FALSE)
  names(api_folders) <- basename(api_folders)

  api_info <- lapply(api_folders, function(api_dir) {
    api_files <- dir(api_dir)
    if (!any(c("plumber.r", "entrypoint.r") %in% tolower(api_files))) {
      # could not find any plumber files. Quitting
      return(NULL)
    }

    # did find a plumb file. Can return the dir
    data.frame(
      package = package,
      name = basename(api_dir),
      source_directory = api_dir,
      stringsAsFactors = FALSE,
      row.names = FALSE
    )
  })

  has_no_api <- vapply(api_info, is.null, logical(1))
  if (all(has_no_api)) {
    return(an_error(
      "No Plumber APIs found for package: \"", package, "\""
    ))
  }

  api_info <- api_info[!has_no_api]

  apis <- do.call(rbind, api_info)
  class(apis) <- c("plumber_available_apis", class(apis))
  rownames(apis) <- NULL

  list(
    apis = apis,
    error = NULL
  )
}

#' @return will return a list of `error` and `apis` which is a \code{data.frame} containing "package", and "name"
#'
#' @importFrom utils installed.packages
#' @noRd
all_available_apis <- function() {
  ret <- list()
  all_pkgs <- installed.packages()[,"Package"]

  for (pkg in all_pkgs) {
    info <- available_apis_for_package(pkg)
    if (!is.null(info$apis)) {
      ret[[length(ret) + 1]] <- info$apis
    }
  }

  # do not check for size 0, as plumber contains apis.

  apis <- do.call(rbind, ret)

  list(
    apis = apis, # will maintain class
    error = NULL
  )
}


#' @export
format.plumber_available_apis <- function(x, ...) {
  apis <- x
  split_apis <- split(apis, apis$package)

  pkg_apis <- vapply(
    split_apis,
    function(apis_sub) {
      paste0(
        "* ", apis_sub$package[1], "\n",
        paste0("  - ", apis_sub$name, collapse = "\n")
      )
    },
    character(1)
  )

  paste0(
    "Available Plumber APIs:\n",
    paste0(pkg_apis, collapse = "\n")
  )
}


#' @export
print.plumber_available_apis <- function(x, ...) {
  cat(format(x, ...), "\n", sep = "")
}

for_each_plumber_api <- function(fn, ...) {
  package <- "plumber"
  lapply(
    available_apis(package)$name,
    function(name) {
      if (name == "14-future") {
        if (.Platform$OS.type == "windows" && getRversion() >= "4.0.0") {
          # Future is not cleaned up properly and produces detritus temp files
          # Debug PR: https://github.com/rstudio/plumber/pull/870
          # Explain: https://github.com/HenrikBengtsson/future/blame/7a15808391b604903659bf67f9aa809f4c9a54c4/cran-comments.md
          # Work around: https://github.com/HenrikBengtsson/future/blob/12cf573dc60cad82a22f03ff6e144201ba1bd42c/tests/nested_futures%2Cmc.cores.R#L5-L15
          # R devel post: https://stat.ethz.ch/pipermail/r-devel/2021-June/080830.html
          message("Skipping test on 14-future example on Windows R >= 4.0")
          return()
        }
        if (!require("future", character.only = TRUE, quietly = TRUE)) {
          return()
        }
      }

      pr <-
        if (name %in% c("06-sessions", "12-entrypoint")) {
          expect_warning({
            plumb_api(package, name)
          }, "Legacy cookie secret")
        } else {
          plumb_api(package, name)
        }
      expect_true(is_plumber(pr), paste0("plumb_api(\"", package, "\", \"", name, "\")"))


      fn(pr, ...)
    }
  )
}

context("plumb() package APIs")


expect_apis <- function(apis) {
  expect_s3_class(apis, "plumber_available_apis")
  expect_s3_class(apis, "data.frame")
  expect_true(all(c("package", "name") %in% names(apis)))
}
test_that("available_apis() works with no package", {
  skip_on_cran()

  apis <- available_apis()
  expect_apis(apis)
})
test_that("available_apis() works with a package", {
  apis <- available_apis("plumber")
  expect_apis(apis)
})
test_that("available_apis() print method works", {
  apis_output <- capture.output({
    available_apis("plumber")
  })

  expected_apis_output <- c(
    "Available Plumber APIs:",
    "* plumber",
    paste0("  - ", dir(system.file("plumber", package = "plumber")))
  )

  expect_equal(
    apis_output,
    expected_apis_output
  )
})
test_that("available_apis() print method works with two packages", {
  top <- available_apis("plumber")
  bottom <- top
  top$package <- "top"
  bottom$package <- "bottom"

  apis_output <- capture.output({
    rbind(top, bottom)
  })

  plumber_apis <- paste0("  - ", dir(system.file("plumber", package = "plumber")))

  # printed in alpha order
  expected_apis_output <- c(
    "Available Plumber APIs:",
    "* bottom",
    plumber_apis,
    "* top",
    plumber_apis
  )

  expect_equal(
    apis_output,
    expected_apis_output
  )
})

test_that("missing args are handled", {
  expect_equal(plumb_api("plumber", NULL), available_apis("plumber"))

  skip_on_cran()
  all_apis <- available_apis()
  expect_equal(plumb_api(NULL, "01-append"), all_apis)
  expect_equal(plumb_api(NULL, NULL), all_apis)
})

test_that("errors are thrown", {


  expect_error(plumb_api(c("plumber", "plumber"), "01-append"))
  expect_error(plumb_api("plumber", c("01-append", "01-append")))

  expect_error(plumb_api(TRUE, "01-append"))
  expect_error(plumb_api("plumber", TRUE))

  expect_error(plumb_api("plumber", "not an api"))

  expect_error(available_apis("not a package"), "No package found with name")
  expect_error(available_apis("crayon"), "No Plumber APIs found for package")
})

test_that("edit opens correct file", {
  # Redefine editor so that file.edit doesn't try to open a file
  with_options(
    list(
      editor = function(name, file, title) {
        cat(file, " test file attempted to open\n", sep = "")
      }
    ),
    {
      apis <- available_apis()

      selected_api <- apis$package == "plumber" & apis$name == "01-append"

      expect_warning(
        expect_output(
          plumb_api("plumber", "01-append", edit = TRUE),
          "plumber.R test file attempted to open",
          fixed = TRUE
        ),
        "plumber.R has been opened in the editor"
      )

      selected_api <- apis$package == "plumber" & apis$name == "12-entrypoint"

      expect_warning(
        expect_output(
          plumb_api("plumber", "12-entrypoint", edit = TRUE),
          "entrypoint.R test file attempted to open",
          fixed = TRUE
        ),
        "entrypoint.R has been opened in the editor"
      )
    }
  )
})

test_that("edit throws a warning", {
  with_options(
    list(editor = function(name, file, title) NULL),
    expect_warning(plumb_api("plumber", "01-append", edit = TRUE))
  )
})

context("plumb() plumber APIs")
test_that("all example plumber apis plumb", {
  # plumb each api and validate they return a plumber object
  for_each_plumber_api(identity)
})

#* @serializer png
#* @get /png
function() {
  plot(1:10)
}

#* @serializer jpeg
#* @get /jpeg
function() {
  plot(1:10)
}

#* @serializer svg
#* @get /svg
function() {
  plot(1:10)
}

#' @serializer png list(width = 150, height=150)
#' @get /littlepng
function(){
  plot(1:10)
}

#' @serializer jpeg list(width = 150, height=150)
#' @get /littlejpeg
function(){
  plot(1:10)
}

#* @serializer svg list(width=4, height=4)
#* @get /littlesvg
function() {
  warning("Should not test. Image size does not decrease with dimension decrease")
  plot(1:10)
}

#' Static file router
#'
#' Creates a router that is backed by a directory of files on disk.
#' @include plumber.R
#' @export
PlumberStatic <- R6Class(
  "PlumberStatic",
  inherit = Plumber,
  public = list(
    #' @description Create a new `PlumberStatic` router
    #' @param direc a path to an asset directory.
    #' @param options options to be evaluated in the `PlumberStatic` router environment
    #' @return A new `PlumberStatic` router
    initialize = function(direc, options) {
      super$initialize(filters=NULL)

      private$dir <- direc

      if (missing(direc)) {
        stop("Cannot add asset directory when no directory was specified")
      }

      # Relative paths
      if (substr(direc, 1, 2) == "./") {
        direc <- substr(direc, 3, nchar(direc))
      }

      if (missing(options)) {
        options <- list()
      }

      # Evaluate to convert to list
      if (is.function(options)) {
        options <- options()
      } else if (is.expression(options)) {
        options <- eval(options, private$envir)
      }

      badRequest <- function(res) {
        res$body <- "<h1>Bad Request</h1>"
        res$status <- 400
        res
      }

      expr <- function(req, res) {
        # Adapted from shiny:::staticHandler
        if (!req$REQUEST_METHOD %in% c('GET', 'HEAD')) {
          return(badRequest(res))
        }

        path <- req$PATH_INFO

        if (is.null(path)) {
          return(badRequest(res))
        }

        if (path == '/') {
          path <- '/index.html'
        }

        path <- httpuv::decodeURIComponent(path)
        abs.path <- resolve_path(direc, path)
        if (is.null(abs.path)) {
          # TODO: Should this be inherited from a parent router?
          val <- private$notFoundHandler(req=req, res=res)
          return(val)
        }

        ext <- tools::file_ext(abs.path)
        info <- file.info(abs.path)
        contentType <- getContentType(ext)
        # Similar to: https://github.com/rstudio/httpuv/blob/220319d/src/webapplication.cpp#L617-L623
        res$setHeader("Content-Type", contentType)
        res$setHeader("Content-Length", info$size)
        res$setHeader("Last-Modified", http_date_string(info$mtime))
        res$body <-
          if (req$REQUEST_METHOD == 'GET') {
            readBin(abs.path, 'raw', n = info$size)
          } else {
            # HEAD request
            NULL
          }
        res$status <- 200
        res
      }

      filter <- PlumberFilter$new(paste("static-asset", direc, sep="|"), expr, private$envir)
      private$addFilterInternal(filter)
    },
    #' @description Print representation of `PlumberStatic()` router.
    #' @param prefix a character string. Prefix to append to representation.
    #' @param topLevel a logical value. When method executed on top level
    #' router, set to `TRUE`.
    #' @param ... additional arguments for recursive calls
    #' @return A terminal friendly representation of a `PlumberStatic()` router.
    print = function(prefix="", topLevel=TRUE, ...) {
      cat(prefix)
      if (!topLevel) {
        cat("\u2502 ")
      }
      cat(crayon::silver("# Plumber static router serving from directory: ", private$dir, "\n", sep = ""))
    }
  ), private=list(
    dir = NULL
  )
)

#* @stuff
#* @more- stuff
#*
#* comments
function(req, res, forward){
    4
}

#* @get /test
#* @post /test
#* @more stuff
#* hey
foo <- function(a, b, ..., req, res, forward){
    5
}

#* Not an API
#* @param things
foo <- function(req, res, forward){
    6
}

#* @post /
#* one line
function(req, res, forward){x <- 9; x + 1}

#* Spaces in between
#* @delete /

function(req, res, forward){
    x <- 10
    x + 2
}

#*@get nowhitespace
function(req, res, forward){
    14
}

#* @get /count
{
  if (!exists("count")) {
    count <- 0
  }
  # will not init if variable exists,
  # testing against sourcing an endpoint twice in the same envir
  count <- count + 1
  function() {
    count
  }
}

#* @get /static_count
function() {
  static_count
}

# will not init if variable exists,
# testing against sourcing a script twice in the same envir
if (!exists("static_count")) {
  static_count <- 0
}
static_count <- static_count + 1

#* @use /some
function(){

}

#* @get /get
function() {

}


#* @put /put
function(req, res, forward){
  7
}

#* @post /post
function() {

}

#* @delete /delete
function() {

}

#* @post /test
#* @get /test
function(){

}

#* @head /head
function() {

}

#* @options /options
function(){

}

#* @patch /patch
function(){

}

library(promises)
library(coro)
future::plan("multisession") # a worker for each core
# future::plan(future::multisession(workers = 2)) # only two workers


# Quick manual test:
# Within 10 seconds...
# 1. Visit /future
# 2. While /future is loading, visit /sync many times
# /future will not block /sync from being able to be loaded.


#' @serializer json list(auto_unbox = TRUE)
#' @get /sync
function() {
  # print route, time, and worker pid
  paste0("/sync; ", Sys.time(), "; pid:", Sys.getpid())
}

#' @contentType list(type = "text/html")
#' @serializer json list(auto_unbox = TRUE)
#' @get /future
function() {

  future_promise({
    # perform long running computations
    Sys.sleep(10)

    # print route, time, and worker pid
    paste0("/future; ", Sys.time(), "; pid:", Sys.getpid())
  })
}


# A function that will return a promise object
calc_using_promise <- function() {
  # Executes using `future`
  future_promise({
    # perform long computations
    Sys.sleep(1)

    # return which process id was used
    Sys.getpid()
  }) %...>% {
    exec_pid <- .
    main_pid <- Sys.getpid()
    # print route, time, and worker pid
    paste0("/coro; ", Sys.time(), "; exec pid:", exec_pid, "; main pid:", main_pid)

    list(exec_pid = exec_pid, main_pid = main_pid)
  }
}

#' For more info about `coro`, visit https://coro.r-lib.org
#' @contentType text
#' @serializer json list(auto_unbox = TRUE)
#' @get /coro
async(function() {

  # (a)wait for the promised value and use as a regular value
  # No neeed for a followup promise when using `await(p)`
  pids <- await(calc_using_promise())

  # Write code with the finalized async value as if it was synchronously found
  pids$exec_is_even <- pids$exec_pid %% 2 == 0

  # return info
  pids
})


# -----------------------------------


# Originally by @antoine-sachet from https://github.com/rstudio/plumber/issues/389
#' @get /divide
#' @serializer json list(auto_unbox = TRUE)
#' @param a number
#' @param b number
function(a = NA, b = NA) {
  future_promise({
    a <- as.numeric(a)
    b <- as.numeric(b)
    if (is.na(a)) stop("a is missing")
    if (is.na(b)) stop("b is missing")
    if (b == 0) stop("Cannot divide by 0")

    a / b
  })
}

#' @get /divide-catch
#' @serializer json list(auto_unbox = TRUE)
#' @param a number
#' @param b number
function(a = NA, b = NA) {
  future_promise({
    a <- as.numeric(a)
    b <- as.numeric(b)
    if (is.na(a)) stop("a is missing")
    if (is.na(b)) stop("b is missing")
    if (b == 0) stop("Cannot divide by 0")

    a / b
  }) %...!%
    # Handle `future` errors
    {
      error <- .
      # handle error here!
      if (error$message == "b is missing") {
        return(Inf)
      }

      # rethrow original error
      stop(error)
    }
}

#* @get /
function(req, res){
  include_file("include/test.txt", res)
}

#* @get /html
function(req, res){
  include_html("include/test-html.html", res)
}

#* @get /rmd
function(req, res){
  include_rmd("include/test.Rmd", res)
}

#* @get /md
function(req, res){
  include_md("include/test.md", res)
}

#' @stuff
#' @more- stuff
#'
#' comments
function(req, res, forward){
  4
}

#' @get /test
#' @post /test
#' @more stuff
#' hey
foo <- function(a, b, ..., req, res, forward){
  5
}

#' Not an API
#' @param things
foo <- function(req, res, forward){
  6
}

#' @post /
#' one line
function(req, res, forward){x <- 9; x + 1}

#' Spaces in between
#' @delete /

function(req, res, forward){
  x <- 10
  x + 2
}

#'@get nowhitespace
function(req, res, forward){
  14
}

default404Handler <- function(req, res) {
  res$status <- 404
  res$serializer <- serializer_unboxed_json()
  list(error="404 - Resource Not Found")
}

defaultErrorHandler <- function(){
  function(req, res, err){
    print(err)

    li <- list()

    # always serialize error with unboxed json
    res$serializer <- serializer_unboxed_json()

    if (res$status == 200L){
      # The default is a 200. If that's still set, then we should probably override with a 500.
      # It's possible, however, than a handler set a 40x and then wants to use this function to
      # render an error, though.
      res$status <- 500
      li$error <- "500 - Internal server error"
    } else {
      li$error <- "Internal error"
    }


    # Don't overly leak data unless they opt-in
    if (is.function(req$pr$getDebug) && isTRUE(req$pr$getDebug())) {
      li["message"] <- as.character(err)
    }

    li
  }
}

# this function works under the assumption that the regular route was not found.
# Here we are only trying to find if there are other routes that can be
#' @noRd
allowed_verbs <- function(pr, path_to_find) {

  verbs_allowed <- c()

  # look at all possible endpoints
  for (endpoint_group in pr$endpoints) {
    for (endpoint in endpoint_group) {
      if (endpoint$matchesPath(path_to_find)) {
        verbs_allowed <- c(verbs_allowed, endpoint$verbs)
      }
    }
  }

  # look at all possible mounts
  for (i in seq_along(pr$mounts)) {
    mount <- pr$mounts[[i]]
    mount_path <- sub("/$", "", names(pr$mounts)[i]) # trim trailing `/`

    # if the front of the urls don't match, move on to next mount
    if (!identical(
      substr(path_to_find, 1, nchar(mount_path)),
      mount_path
    )) {
      next
    }

    # remove path and recurse
    mount_path_to_find <- substr(path_to_find, nchar(mount_path) + 1, nchar(path_to_find))
    m_verbs <- allowed_verbs(mount, mount_path_to_find)

    # add any verbs found
    if (length(m_verbs) > 0) {
      verbs_allowed <- c(verbs_allowed, m_verbs)
    }
  }

  # return verbs
  sort(unique(verbs_allowed))
}

#' @noRd
is_405 <- function(pr, path_to_find, verb_to_find) {
  verbs_allowed <- allowed_verbs(pr, path_to_find)

  # nothing found, not 405
  if (length(verbs_allowed) == 0) return(FALSE)

  # is verb excluded?
  !(verb_to_find %in% verbs_allowed)
}
router_has_route <- function(pr, path_to_find, verb_to_find) {
  verbs_allowed <- allowed_verbs(pr, path_to_find)

  # nothing found, not a route
  if (length(verbs_allowed) == 0) return(FALSE)

  # is verb found?
  verb_to_find %in% verbs_allowed
}

context("default handlers")

test_that("404 handler sets 404", {
  res <- PlumberResponse$new()
  val <- default404Handler(list(), res)
  expect_equal(res$status, 404)
  expect_match(val$error, "404")
  expect_match(val$error, "Not Found")
})

test_that("405 handling is ok, get the right verbs", {
  pr <- pr()
  sub <- pr()
  sub$handle("GET", "/test", force)
  pr$mount("/barret", sub)

  expect_equal(allowed_verbs(pr, path_to_find = "/barret/test"), "GET")
  expect_null(allowed_verbs(pr, path_to_find = "/subroute/not_found"))
  expect_null(allowed_verbs(pr, path_to_find = "/barret/"))
  expect_null(allowed_verbs(pr, path_to_find = "/barret/wrong"))

  expect_false(is_405(pr, path_to_find = "/barret/test", "GET"))
  expect_true(is_405(pr, path_to_find = "/barret/test", "POST"))
  expect_false(is_405(pr, path_to_find = "/subroute/not_found"))

  pr <- pr()
  sub <- pr()
  sub$handle("GET", "/test", force)
  pr$mount("/", sub)

  expect_equal(allowed_verbs(pr, path_to_find = "/test"), "GET")
  expect_null(allowed_verbs(pr, path_to_find = "/subroute/not_found"))
  expect_null(allowed_verbs(pr, path_to_find = "/barret/"))
  expect_null(allowed_verbs(pr, path_to_find = "/barret/wrong"))

})

test_that("default error handler returns an object with an error property", {
  res <- PlumberResponse$new()
  capture.output(val <- defaultErrorHandler()(list(), res, "I'm an error!"))
  expect_match(val$error, "500")
  expect_match(val$error, "Internal server error")
  expect_equal(res$status, 500)
})
test_that("error handler doesn't clobber non-200 status", {
  res <- PlumberResponse$new()
  res$status <- 403
  capture.output(val <- defaultErrorHandler()(list(), res, "I'm an error!"))
  expect_match(val$error, "Internal error")
  expect_equal(res$status, 403)
})

test_that("error handler only includes message in debug mode.", {
  res <- PlumberResponse$new()
  capture.output(val <- defaultErrorHandler()(list(), res, "I'm an error!"))
  expect_null(val$message)

  res <- PlumberResponse$new()
  capture.output({
    val <- defaultErrorHandler()(
      req = list(pr = list(getDebug = function() { TRUE })),
      res,
      "I'm an error!"
    )
  })
  expect_equal(val$message, "I'm an error!")
})

#' @include plumber.R
NULL

#' Deprecated R6 functions
#'
#' @keywords internal
#' @describeIn deprecated_r6 See [Hookable()]
hookable <- R6Class(
  "hookable",
  inherit = Hookable,
  public = list(
    #' @description Initialize a new `hookable`. Throws deprecated warning prompting user to use [`Hookable`]
    initialize = function() {
      # use `.Deprecated` until `lifecycle` gets R6 support
      .Deprecated(msg = paste0(
        "`hookable` is deprecated as of plumber 1.0.0.\n",
        "Please use `Hookable` instead."
      ))
      # lifecycle::deprecate_warn("1.0.0", "hookable$new()", "Hookable$new()")


      ## no initialize method
      # super$initialize()
    }
  )
)


#' Deprecated R6 functions
#'
#' @keywords internal
#' @export
#' @describeIn deprecated_r6 See [Plumber()]
plumber <- R6Class(
  "plumber",
  inherit = Plumber,
  public = list(
    #' @description Initialize a new `plumber`. Throws deprecated warning prompting user to use [`Plumber`]
    #' @param ... params passed in to `Plumber$new()`
    initialize = function(...) {
      # use `.Deprecated` until `lifecycle` gets R6 support
      .Deprecated(msg = paste0(
        "`plumber` is deprecated as of plumber 1.0.0.\n",
        "Please use `Plumber` instead."
      ))
      # lifecycle::deprecate_warn("1.0.0", "hookable$new()", "Hookable$new()")

      super$initialize(...)
    }
  )
)

#' Set the default serializer of the router
#'
#' By default, Plumber serializes responses to JSON. This function updates the
#' default serializer to the function supplied via \code{serializer}
#'
#' @template param_pr
#' @param serializer A serializer function
#'
#' @return The Plumber router with the new default serializer
#'
#' @export
pr_set_serializer <- function(pr, serializer) {
  validate_pr(pr)
  pr$setSerializer(serializer)
  pr
}

#' Set the default endpoint parsers for the router
#'
#' By default, Plumber will parse JSON, text, query strings, octet streams, and multipart bodies. This function updates the
#' default parsers for any endpoint that does not define their own parsers.
#'
#' Note: The default set of parsers will be completely replaced if any value is supplied. Be sure to include all of your parsers that you would like to include.
#' Use `registered_parsers()` to get a list of available parser names.
#'
#' @template param_pr
#' @template pr_setParsers__parsers
#'
#' @return The Plumber router with the new default [PlumberEndpoint] parsers
#'
#' @export
pr_set_parsers <- function(pr, parsers) {
  validate_pr(pr)
  pr$setParsers(parsers)
  pr
}

#' Set the handler that is called when the incoming request can't be served
#'
#' This function allows a custom error message to be returned when a request
#' cannot be served by an existing endpoint or filter.
#'
#' @template param_pr
#' @param fun A handler function
#'
#' @return The Plumber router with a modified 404 handler
#'
#' @examples
#' \dontrun{
#' handler_404 <- function(req, res) {
#'   res$status <- 404
#'   res$body <- "Oops"
#' }
#'
#' pr() %>%
#'   pr_get("/hi", function() "Hello") %>%
#'   pr_set_404(handler_404) %>%
#'   pr_run()
#' }
#'
#' @export
pr_set_404 <- function(pr, fun) {
  validate_pr(pr)
  pr$set404Handler(fun)
  pr
}

#' Set the error handler that is invoked if any filter or endpoint generates an
#' error
#'
#' @template param_pr
#' @param fun An error handler function. This should accept `req`, `res`, and the error value
#'
#' @return The Plumber router with a modified error handler
#'
#' @examples
#' \dontrun{
#' handler_error <- function(req, res, err){
#'   res$status <- 500
#'   list(error = "Custom Error Message")
#' }
#'
#' pr() %>%
#'   pr_get("/error", function() log("a")) %>%
#'   pr_set_error(handler_error) %>%
#'   pr_run()
#' }
#' @export
pr_set_error <- function(pr, fun) {
  validate_pr(pr)
  pr$setErrorHandler(fun)
  pr
}




#' Set debug value to include error messages of routes cause an error
#'
#' To hide any error messages in production, set the debug value to `FALSE`.
#' The `debug` value is enabled by default for [interactive()] sessions.
#'
#' @template param_pr
#' @param debug `TRUE` provides more insight into your API errors.
#' @return The Plumber router with the new debug setting.
#' @export
#' @examples
#' \dontrun{
#' # Will contain the original error message
#' pr() %>%
#'   pr_set_debug(TRUE) %>%
#'   pr_get("/boom", function() stop("boom")) %>%
#'   pr_run()
#'
#' # Will NOT contain an error message
#' pr() %>%
#'   pr_set_debug(FALSE) %>%
#'   pr_get("/boom", function() stop("boom")) %>%
#'   pr_run()
#' }
pr_set_debug <- function(pr, debug = interactive()) {
  validate_pr(pr)
  pr$setDebug(debug = debug)
  pr
}


#' Set the API visual documentation
#'
#' `docs` should be either a logical or a character value matching a registered visual documentation.
#' Multiple handles will be added to [`Plumber`] object. OpenAPI json
#' file will be served on paths `/openapi.json`. Documentation
#' will be served on paths `/__docs__/index.html` and `/__docs__/`.
#'
#' @template param_pr

#' @param docs a character value or a logical value.
#' If using [options_plumber()], the value must be set before initializing your Plumber router.
#' @param ... Arguments for the visual documentation. See each visual documentation package for further details.
#' @return The Plumber router with the new docs settings.
#' @export
#' @examples
#' \dontrun{
#' ## View API using Swagger UI
#' # Official Website: https://swagger.io/tools/swagger-ui/
#' # install.packages("swagger")
#' if (require(swagger)) {
#'   pr() %>%
#'     pr_set_docs("swagger") %>%
#'     pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) %>%
#'     pr_run()
#' }
#'
#' ## View API using Redoc
#' # Official Website: https://github.com/Redocly/redoc
#' if (require(redoc)) {
#'   pr() %>%
#'     pr_set_docs("redoc") %>%
#'     pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) %>%
#'     pr_run()
#' }
#'
#' ## View API using RapiDoc
#' # Official Website: https://github.com/mrin9/RapiDoc
#' if (require(rapidoc)) {
#'   pr() %>%
#'     pr_set_docs("rapidoc") %>%
#'     pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) %>%
#'     pr_run()
#' }
#'
#' ## Disable the OpenAPI Spec UI
#' pr() %>%
#'   pr_set_docs(FALSE) %>%
#'   pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) %>%
#'   pr_run()
#' }
pr_set_docs <- function(
  pr,
  docs = getOption("plumber.docs", TRUE),
  ...
) {
  validate_pr(pr)
  pr$setDocs(docs = docs, ...)
  pr
}


#' Set the `callback` to tell where the API visual documentation is located
#'
#' When set, it will be called with a character string corresponding
#' to the API visual documentation url. This allows RStudio to locate visual documentation.
#'
#' If using [options_plumber()], the value must be set before initializing your Plumber router.
#'
#' @template param_pr
#' @param callback a callback function for taking action on the docs url.
#' @return The Plumber router with the new docs callback setting.
#' @export
#' @examples
#' \dontrun{
#' pr() %>%
#'   pr_set_docs_callback(function(url) { message("API location: ", url) }) %>%
#'   pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) %>%
#'   pr_run()
#' }
pr_set_docs_callback <- function(
  pr,
  callback = getOption('plumber.docs.callback', NULL)
) {
  validate_pr(pr)
  pr$setDocsCallback(callback = callback)
  pr
}


#' Set the OpenAPI Specification
#'
#' Allows to modify OpenAPI Specification autogenerated by `plumber`.
#'
#' Note, the returned value will be sent through [serializer_unboxed_json()] which will turn all length 1 vectors into atomic values.
#' To force a vector to serialize to an array of size 1, be sure to call [as.list()] on your value. `list()` objects are always serialized to an array value.
#'
#' @template param_pr
#' @template pr_setApiSpec__api
#' @return The Plumber router with the new OpenAPI Specification object or function.
#' @export
#' @examples
#' \dontrun{
#' # Set the API Spec to a function to use the auto-generated OAS object
#' pr() %>%
#'   pr_set_api_spec(function(spec) {
#'     spec$info$title <- Sys.time()
#'     spec
#'   }) %>%
#'   pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) %>%
#'   pr_run()
#'
#' # Set the API Spec using an object
#' pr() %>%
#'   pr_set_api_spec(my_custom_object) %>%
#'   pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) %>%
#'   pr_run()
#' }
pr_set_api_spec <- function(
  pr,
  api
) {
  validate_pr(pr)
  pr$setApiSpec(api = api)
  pr
}

context("Promise")
library(promises) # reexports


# Block until all pending later tasks have executed
wait_for_async <- function() {
  skip_if_not_installed("later")
  while (!later::loop_empty()) {
    later::run_now()
    Sys.sleep(0.00001)
  }
}


get_result <- function(result) {
  if (!promises::is.promising(result)) {
    return(result)
  }


  # return async value... synchronously
  ret <- NULL
  err <- NULL
  set_result <- function(value) {
    ret <<- value
  }
  result %...>% set_result() %...!% (function(error_value) {
    err <<- error_value
  })
  wait_for_async()
  if (!is.null(err)) {
    stop(err)
  }
  ret
}

expect_not_promise <- function(x) {
  expect_false(promises::is.promising(x))
  invisible(x)
}
expect_promise <- function(x) {
  expect_true(promises::is.promising(x))
  invisible(x)
}

expect_route_sync <- function(x) {
  expect_equal(x$body$name, "sync")
  invisible(x)
}
expect_route_async <- function(x) {
  expect_equal(x$body$name, "async")
  invisible(x)
}

async_router <- function() {
  "files/async.R" %>%
    test_path() %>%
    pr()
}

serve_route <- function(pr, route) {
  pr %>%
    pr_set_serializer(serializer_identity()) %>%
    {
      pr <- .
      pr$call(make_req("GET", route))
    }
}


test_that("sync works", {
  async_router() %>%
    serve_route("/sync") %>%
    expect_not_promise() %>%
    get_result() %>%
    expect_route_sync()
})

test_that("async works", {
  async_router() %>%
    serve_route("/async") %>%
    expect_promise() %>%
    get_result() %>%
    expect_route_async()
})


context("Promise - hooks")

hooks <- c(
  # PlumberEndpoint only
  "preexec", "postexec", "aroundexec",
  # Plumber router only
  "preserialize", "postserialize",
  "preroute", "postroute"
)

test_that("async hooks create async execution", {
  # exhaustive test of all public hooks

  # make an exhaustive matrix of T/F values of which hooks are async
  hooks_are_async <- do.call(
    expand.grid,
    lapply(stats::setNames(hooks, hooks), function(...) c(FALSE, TRUE))
  )

  # remove the all FALSE row
  hooks_are_async <- hooks_are_async[-1, ]
  # make sure there is at least one async hook
  expect_true(all(apply(hooks_are_async, 1, sum) > 0))

  # for each row in `hooks_are_async`
  apply(hooks_are_async, 1, function(hooks_are_async_row) {
    async_hook_count <- 0
    async_hook <- function(...) {
      args <- list(...)
      p <- promise_resolve(args$value)
      # add extra promises
      for (i in 1:10) {
        p <- then(p, function(val) {
          val
        })
      }
      # increment the counter
      p <- then(p, function(val) {
        async_hook_count <<- async_hook_count + 1
        val
      })
      p
    }
    pr <- async_router()
    # for each hook, register it if it should be async
    for (hook in hooks) {
      hook_is_async <- hooks_are_async_row[[hook]]
      if (hook_is_async) {
        switch(hook,
          # PlumberEndpoint hooks
          "preexec" = ,
          "postexec" = {
            expect_equal(pr$endpoints[[1]][[2]]$path, "/sync")
            pr$endpoints[[1]][[2]]$registerHook(hook, async_hook)
          },
          "aroundexec" = {
            expect_equal(pr$endpoints[[1]][[2]]$path, "/sync")
            pr$endpoints[[1]][[2]]$registerHook(hook, function(..., .next) {
              p <- promise_resolve(TRUE)
              # add extra promises
              for (i in 1:10) {
                p <- then(p, function(val) {
                  val
                })
              }
              p <- then(p, function(val) {
                .next(...)
              })
              # increment the counter
              p <- then(p, function(val) {
                async_hook_count <<- async_hook_count + 1
                val
              })
              p
            })
          },
          # default case
          {
            # Plumber hooks
            pr$registerHook(hook, async_hook)
          }
        )
      }
    }
    # run the sync route with some async hooks
    pr %>%
      serve_route("/sync") %>%
      expect_promise() %>%
      get_result() %>%
      expect_route_sync()
    # make sure all the hooks were hit once
    expect_equal(async_hook_count, sum(hooks_are_async_row))
  })

})




context("Promise - multiple hooks can change the value")

test_that("async hooks change value being passed through",  {

  pr <- async_router()

  lapply(1:10, function(i) {
    pr$registerHook("preroute", function(...) {
      # no value arg
      promise_resolve(TRUE) # make execution in a promise
    })
  })
  lapply(seq(1, 20 - 1, by = 1), function(i) {
    pr$registerHook("postroute", function(value, ...) {
      expect_equal(value, i)
      value + 1
    })
  })
  # finishes at 20

  lapply(seq(20, 40 - 2, by = 2), function(i) {
    pr$registerHook("preserialize", function(value, ...) {
      expect_equal(value, i)
      value + 2
    })
  })
  # finishes at 40

  lapply(seq(40, 70 - 3, by = 3), function(i) {
    pr$registerHook("postserialize", function(value, ...) {
      # in return object format
      expect_equal(value$body, i)
      value$body <- value$body + 3
      value
    })
  })
  # finishes at 70

  pr %>%
    serve_route("/number-one") %>%
    expect_promise() %>%
    get_result() %>% {
      x <- .
      expect_equal(x$body, 70)
    }
})


context("Promise - errors are handled")

expect_route_error <- function(response, txt) {
  if (!is.list(response$body)) {
    response$body <- jsonlite::parse_json(response$body, simplifyVector = TRUE)
  }
  expect_equal(response$body$error, "500 - Internal server error")
  expect_true(grepl(txt, response$body$message))
}

test_that("sync error is caught", {
  bad_expression <- "sync-bad - expected error here"
  expect_output(
    {
      async_router() %>%
        pr_set_debug(TRUE) %>%
        serve_route("/sync-bad") %>%
        expect_not_promise() %>%
        get_result() %>%
        expect_route_error(bad_expression)
    },
    bad_expression
  )
})


test_that("async error is caught", {
  bad_expression <- "async-bad - expected error here"
  expect_output(
    {
      async_router() %>%
        pr_set_debug(TRUE) %>%
        serve_route("/async-bad") %>%
        expect_promise() %>%
        get_result() %>%
        expect_route_error(bad_expression)
    },
    bad_expression
  )
})




test_that("hook errors are caught", {

  for (pr_is_async in c(TRUE, FALSE)) {
    for (error_is_async in c(TRUE, FALSE)) {
      for (hook in hooks) {
        local({
          bad_expression <- paste0("boom ", hook, " sync")
          pr <- async_router()

          if (pr_is_async) {
            pr$registerHook("preroute", function(...) {
              # no value arg
              promise_resolve(TRUE) # make execution in a promise
            })
          }

          bad_hook <-
            if (error_is_async) {
              function(...) {
                # no value arg
                p <- promise_resolve(TRUE)
                p <- then(p, function(value) {
                  stop(bad_expression, call. = FALSE)
                })
                p
              }
            } else {
              function(...) {
                # no value arg
                stop(bad_expression, call. = FALSE)
              }
            }
          switch(hook,
            # PlumberEndpoint hooks
            "preexec" = ,
            "postexec" = {
              expect_equal(pr$endpoints[[1]][[2]]$path, "/sync")
              pr$endpoints[[1]][[2]]$registerHook(hook, bad_hook)
            },
            "aroundexec" = {
              expect_equal(pr$endpoints[[1]][[2]]$path, "/sync")
              pr$endpoints[[1]][[2]]$registerHook(hook, bad_hook)
            },
            # default
            pr$registerHook(hook, bad_hook)
          )

          expect_output(
            {
              pr %>%
                pr_set_debug(TRUE) %>%
                serve_route("/sync") %>%
                {
                  if (pr_is_async || error_is_async) {
                    expect_promise(.)
                  } else {
                    expect_not_promise(.)
                  }
                } %>%
                get_result() %>%
                expect_route_error(bad_expression)
            },
            bad_expression
          ) # expect_output
        }) # local
      } # hook
    } # error_is_sync
  } # pr_is_sync

})



test_that("accessing two images created using promises does not create an error", {

  root <- async_router() %>% pr_set_debug(TRUE)

  # access route 1
  p1 <-
    # Open p1 on p1
    root$call(make_req("GET", "/promise_plot1")) %>%
    expect_promise()
  # access route 2
  p2 <-
    # Open p2 on p2
    root$call(make_req("GET", "/promise_plot2")) %>%
    expect_promise()

  # if the graphics device was not maintained for the promises, two promises could break how graphics are recorded
  ## Bad
  ## * Open p1 device
  ## * Open p2 device
  ## * Draw p1 in p2 device
  ## * Draw p2 in p2 device
  ## * Close cur device (p2)
  ## * Close cur device (p1) (which is empty)
  ##
  ## Good
  ## * Open p1 device in p1
  ## * Open p2 device in p2
  ## * Draw p1 in p1 device in p1
  ## * Draw p2 in p2 device in p2
  ## * Close p1 device in p1
  ## * Close p2 device in p2

  # These actual steps may be interwoven with each other (, but commented to where they might occur)
  expect_silent({
    # Draw p1 in p1 device
    # Close p1 device in p1
    p1_val <- get_result(p1)
    # Draw p2 in p2 device
    # Close p2 device in p2
    p2_val <- get_result(p2)
  })

  expect_equal(p1_val$status, 200L)
  expect_equal(p1_val$headers$`Content-Type`, "image/png")
  expect_true(is.raw(p1_val$body))
  expect_gt(length(p1_val$body), 1000)

  expect_equal(p2_val$status, 200L)
  expect_equal(p2_val$headers$`Content-Type`, "image/png")
  expect_true(is.raw(p2_val$body))
  expect_gt(length(p2_val$body), 1000)

  # make sure they are not the same image
  expect_false(
    identical(p1_val$body, p2_val$body)
  )

})

with_interrupt <- function(expr) {
  # Causes pr_run() to immediately exit
  later::later(httpuv::interrupt)
  force(expr)
}

test_that("quiet=TRUE suppresses startup messages", {
  with_interrupt({
    expect_message(pr() %>% pr_run(quiet = TRUE), NA)
  })
  with_interrupt({
    expect_message(pr()$run(quiet = TRUE), NA)
  })
})

test_that("`docs` does not not permanetly set pr information", {
  doc_name <- "swagger"
  root <- pr() %>% pr_set_docs(doc_name)
  # no docs. do not find printed message
  with_interrupt({
    expect_failure(
      expect_message({
        root %>% pr_run(docs = FALSE)
      }, doc_name)
    )
  })
  # no docs. do not find printed message
  with_interrupt({
    expect_failure(
      expect_message({
        root$run(docs = FALSE)
      }, doc_name)
    )
  })
  # docs enabled by default. Find printed message
  with_interrupt({
    expect_message({
      root %>% pr_run(quiet = FALSE)
    }, doc_name)
  })
})

test_that("`swaggerCallback` does not not permanetly set pr information", {
  call_count <- 0
  m <- function(url) {
    call_count <<- call_count + 1
    TRUE
  }

  # Initialize call count
  m("init")
  expect_equal(call_count, 1)

  root <- pr() %>% pr_set_docs_callback(m)

  # m not used
  with_interrupt({
    expect_equal(call_count, 1)
    root %>% pr_run(swaggerCallback = NULL)
    expect_equal(call_count, 1)
  })

  # m not used
  with_interrupt({
    expect_equal(call_count, 1)
    root$run(swaggerCallback = NULL)
    expect_equal(call_count, 1)
  })

  # m is used
  with_interrupt({
    expect_equal(call_count, 1)
    root %>% pr_run(quiet = FALSE)
    expect_equal(call_count, 2)
  })
})

test_that("`swaggerCallback` can be set by option after the pr is created", {
  call_count <- 0
  m <- function(url) {
    call_count <<- call_count + 1
    TRUE
  }

  # must initialize before options are set for this test
  root <- pr()

  with_options(
    list(
      plumber.swagger.url = get_option_or_env("plumber.swagger.url"),
      plumber.docs.callback = get_option_or_env("plumber.docs.callback")
    ),
    {
      # set option after init
      options_plumber(docs.callback = m)
      with_interrupt({
        expect_equal(call_count, 0)
        pr_run(root)
      })
    }
  )
  # m is used
  expect_equal(call_count, 1)
})

plumber::register_serializer("fake", function(){
  function(val, req, res, errorHandler){
    tryCatch({
      json <- jsonlite::toJSON(val)

      res$setHeader("Content-Type", "application/json")
      res$body <- paste0("FAKE", json)

      return(res$toResponse())
    }, error=function(e){
      errorHandler(req, res, e)
    })
  }
})

pr("./plumber.R")

#' @import R6
#' @import stringi
#' @importFrom rlang %||%
NULL

# used to identify annotation flags.
verbs <- c("GET", "PUT", "POST", "DELETE", "HEAD", "OPTIONS", "PATCH")
enumerateVerbs <- function(v) {
  if (identical(v, "use")) {
    return(verbs)
  }
  toupper(v)
}




#' @include parse-query.R
#' @include parse-body.R
#' @include parser-cookie.R
#' @include shared-secret-filter.R
defaultPlumberFilters <- list(
  queryString = queryStringFilter,
  body = bodyFilter,
  cookieParser = cookieFilter,
  sharedSecret = sharedSecretFilter
)




#' Package Plumber Router
#'
#' Routers are the core request handler in \pkg{plumber}. A router is responsible for
#' taking an incoming request, submitting it through the appropriate filters and
#' eventually to a corresponding endpoint, if one is found.
#'
#' See the [Programmatic Usage](https://www.rplumber.io/articles/programmatic-usage.html) article for additional
#' details on the methods available on this object.
#' @seealso
#'  [pr()],
#'  [pr_run()],
#'  [pr_get()], [pr_post()],
#'  [pr_mount()],
#'  [pr_hook()], [pr_hooks()], [pr_cookie()],
#'  [pr_filter()],
#'  [pr_set_api_spec()], [pr_set_docs()],
#'  [pr_set_serializer()], [pr_set_parsers()],
#'  [pr_set_404()], [pr_set_error()],
#'  [pr_set_debug()],
#'  [pr_set_docs_callback()]
#' @include hookable.R
#' @export
Plumber <- R6Class(
  "Plumber",
  inherit = Hookable,
  public = list(

    #' @description Create a new `Plumber` router
    #'
    #' See also [plumb()], [pr()]
    #' @param filters a list of Plumber filters
    #' @param file path to file to plumb
    #' @param envir an environment to be used as the enclosure for the routers execution
    #' @return A new `Plumber` router
    initialize = function(file = NULL, filters = defaultPlumberFilters, envir) {

      if (!is.null(file)){
        if (!file.exists(file)){
          stop("File does not exist: ", file)
        } else {
          inf <- file.info(file)
          if (inf$isdir){
            stop("Expecting a file but found a directory: '", file, "'.")
          }
        }
      }

      if (missing(envir)){
        private$envir <- new.env(parent=.GlobalEnv)
      } else {
        private$envir <- envir
      }

      if (is.null(filters)){
        filters <- list()
      }

      # Initialize
      self$setSerializer(serializer_json())
      # Default parsers to maintain legacy features
      self$setParsers(c("json", "form", "text", "octet", "multi"))
      self$setErrorHandler(defaultErrorHandler())
      self$set404Handler(default404Handler)
      self$setDocs(TRUE)
      private$docs_info$has_not_been_set <- TRUE # set to know if `$setDocs()` has been called before `$run()`
      private$docs_callback <- rlang::missing_arg()
      private$debug <- NULL
      self$setApiSpec(NULL)

      # Add in the initial filters
      for (fn in names(filters)){
        fil <- PlumberFilter$new(fn, filters[[fn]], private$envir, private$default_serializer, NULL, NULL)
        private$filts <- c(private$filts, fil)
      }

      if (!is.null(file)) {
        # plumb() the file in the working directory
        # The directory is also set when running the plumber object
        private$filename <- file
        old_wd <- setwd(dirname(file))
        on.exit({setwd(old_wd)}, add = TRUE)
        file <- basename(file)

        private$lines <- readUTF8(file)
        private$parsed <- parseUTF8(file)
        private$disable_run <- TRUE
        on.exit({
          private$disable_run <- FALSE
        }, add = TRUE)

        for (i in seq_len(length(private$parsed))) {
          e <- private$parsed[i]

          srcref <- attr(e, "srcref")[[1]]

          evaluateBlock(srcref, private$lines, e, private$envir, private$addEndpointInternal,
                        private$addFilterInternal, self)
        }

        private$globalSettings <- plumbGlobals(private$lines, private$envir)
      }

    },
    #' @description Start a server using `Plumber` object.
    #'
    #' See also: [pr_run()]
    #' @param host a string that is a valid IPv4 or IPv6 address that is owned by
    #' this server, which the application will listen on. "0.0.0.0" represents
    #' all IPv4 addresses and "::/0" represents all IPv6 addresses.
    #' @param port a number or integer that indicates the server port that should
    #' be listened on. Note that on most Unix-like systems including Linux and
    #' Mac OS X, port numbers smaller than 1025 require root privileges.
    #'
    #' This value does not need to be explicitly assigned. To explicitly set it, see [options_plumber()].
    #' @param debug If `TRUE`, it will provide more insight into your API errors. Using this value will only last for the duration of the run. If a `$setDebug()` has not been called, `debug` will default to `interactive()` at `$run()` time. See `$setDebug()` for more details.
    #' @param swagger Deprecated. Please use `docs` instead. See `$setDocs(docs)` or `$setApiSpec()` for more customization.
    #' @param swaggerCallback An optional single-argument function that is
    #'   called back with the URL to an OpenAPI user interface when one becomes
    #'   ready. If missing, defaults to information previously set with `$setDocsCallback()`.
    #'   This value will only be used while running the router.
    #' @param docs Visual documentation value to use while running the API.
    #'   This value will only be used while running the router.
    #'   If missing, defaults to information previously set with `setDocs()`.
    #'   For more customization, see `$setDocs()` or [pr_set_docs()] for examples.
    #' @param quiet If `TRUE`, don't print routine startup messages.
    #' @param ... Should be empty.
    #' @importFrom lifecycle deprecated
    #' @importFrom rlang missing_arg
    run = function(
      host = '127.0.0.1',
      port = getOption('plumber.port', NULL),
      swagger = deprecated(),
      debug = missing_arg(),
      swaggerCallback = missing_arg(),
      ...,
      # any new args should go below `...`
      docs = missing_arg(),
      quiet = FALSE
    ) {

      if (isTRUE(private$disable_run)) {
        stop("Plumber router `$run()` method should not be called while `plumb()`ing a file")
      }

      ellipsis::check_dots_empty()

      # Legacy support for RStudio pro products.
      # Checks must be kept for >= 2 yrs after plumber v1.0.0 release date
      if (lifecycle::is_present(swagger)) {
        if (!rlang::is_missing(docs)) {
          lifecycle::deprecate_warn("1.0.0", "Plumber$run(swagger = )", "Plumber$run(docs = )", details = "`docs` will take preference (ignoring `swagger`)")
          # (`docs` is resolved after `swagger` checks)
        } else {
          if (is.function(swagger)) {
            # between v0.4.6 and v1.0.0
            lifecycle::deprecate_warn("1.0.0", "Plumber$run(swagger = )", "Plumber$setApiSpec(api = )")
            # set the new api function and force turn on the docs
            old_api_spec_handler <- private$api_spec_handler
            self$setApiSpec(swagger)
            on.exit({
              private$api_spec_handler <- old_api_spec_handler
            }, add = TRUE)
            docs <- TRUE
          } else {
            if (isTRUE(private$docs_info$has_not_been_set)) {
              # <= v0.4.6
              lifecycle::deprecate_warn("1.0.0", "Plumber$run(swagger = )", "Plumber$run(docs = )")
              docs <- swagger
            } else {
              # $setDocs() has been called (other than during initialization).
              # `docs` is not provided
              # Believe that prior `$setDocs()` behavior is the correct behavior
              # Warn about updating the run method
              lifecycle::deprecate_warn("1.0.0", "Plumber$run(swagger = )", "Plumber$run(docs = )", details = "The Plumber docs have already been set. Ignoring `swagger` parameter.")
            }
          }
        }
      }

      port <- findPort(port)

      # Delay setting max size option. It could be set in `plumber.R`, which is after initialization
      private$maxSize <- getOption('plumber.maxRequestSize', 0) #0  Unlimited

      # Delay the setting of swaggerCallback as long as possible.
      # An option could be set in `plumber.R`, which is after initialization
      # Order: Run method parameter, internally set value, option, fallback option, NULL
      swaggerCallback <-
        rlang::maybe_missing(swaggerCallback,
          rlang::maybe_missing(private$docs_callback,
            getOption('plumber.docs.callback', getOption('plumber.swagger.url', NULL))
          )
        )

      # Delay the setting of debug as long as possible.
      # The router could be made in an interactive setting and used in background process.
      # Do not determine if interactive until run time
      prev_debug <- private$debug
      on.exit({
        private$debug <- prev_debug
      }, add = TRUE)
      # Fix the debug value while running.
      self$setDebug(
        # Order: Run method param, internally set value, is interactive()
        # `$getDebug()` is dynamic given `setDebug()` has never been called.
        rlang::maybe_missing(debug, self$getDebug())
      )

      docs_info <-
        if (!rlang::is_missing(docs)) {
          # Manually provided. Need to upgrade the parameter
          upgrade_docs_parameter(docs)
        } else {
          private$docs_info
        }

      if (!isTRUE(quiet)) {
        message("Running plumber API at ", urlHost(host = host, port = port, changeHostLocation = FALSE))
      }

      # Set and restore the wd to make it appear that the proc is running local to the file's definition.
      if (!is.null(private$filename)) {
        old_wd <- setwd(dirname(private$filename))
        on.exit({setwd(old_wd)}, add = TRUE)
      }

      if (isTRUE(docs_info$enabled)) {
        mount_docs(
          pr = self,
          host = host,
          port = port,
          docs_info = docs_info,
          callback = swaggerCallback,
          quiet = quiet
        )
        on.exit(unmount_docs(self, docs_info), add = TRUE)
      }

      on.exit(private$runHooks("exit"), add = TRUE)

      httpuv::runServer(host, port, self)
    },
    #' @description Mount a Plumber router
    #'
    #' Plumber routers can be “nested” by mounting one into another
    #' using the `mount()` method. This allows you to compartmentalize your API
    #' by paths which is a great technique for decomposing large APIs into smaller files.
    #'
    #' See also: [pr_mount()]
    #' @param path a character string. Where to mount router.
    #' @param router a Plumber router. Router to be mounted.
    #' @examples
    #' \dontrun{
    #' root <- pr()
    #'
    #' users <- Plumber$new("users.R")
    #' root$mount("/users", users)
    #'
    #' products <- Plumber$new("products.R")
    #' root$mount("/products", products)
    #' }
    mount = function(path, router) {
      # Ensure that the path has both a leading and trailing slash.
      if (!grepl("^/", path)) {
        path <- paste0("/", path)
      }
      if (!grepl("/$", path)) {
        path <- paste0(path, "/")
      }

      private$mnts[[path]] <- router
    },
    #' @description Unmount a Plumber router
    #' @param path a character string. Where to unmount router.
    unmount = function(path) {
      # Ensure that the path has both a leading and trailing slash.
      if (!grepl("^/", path)) {
        path <- paste0("/", path)
      }
      if (!grepl("/$", path)) {
        path <- paste0(path, "/")
      }
      private$mnts[[path]] <- NULL
    },
    #' @param stage a character string. Point in the lifecycle of a request.
    #' @param handler a hook function.
    #' @description Register a hook
    #'
    #' Plumber routers support the notion of "hooks" that can be registered
    #' to execute some code at a particular point in the lifecycle of a request.
    #' Plumber routers currently support four hooks:
    #' \enumerate{
    #'  \item `preroute(data, req, res)`
    #'  \item `postroute(data, req, res, value)`
    #'  \item `preserialize(data, req, res, value)`
    #'  \item `postserialize(data, req, res, value)`
    #' }
    #' In all of the above you have access to a disposable environment in the `data`
    #' parameter that is created as a temporary data store for each request. Hooks
    #' can store temporary data in these hooks that can be reused by other hooks
    #' processing this same request.
    #'
    #' One feature when defining hooks in Plumber routers is the ability to modify
    #' the returned value. The convention for such hooks is: any function that accepts
    #' a parameter named `value` is expected to return the new value. This could
    #' be an unmodified version of the value that was passed in, or it could be a
    #' mutated value. But in either case, if your hook accepts a parameter
    #' named `value`, whatever your hook returns will be used as the new value
    #' for the response.
    #'
    #' You can add hooks using the `registerHook` method, or you can add multiple
    #' hooks at once using the `registerHooks` method which takes a name list in
    #' which the names are the names of the hooks, and the values are the
    #' handlers themselves.
    #'
    #' See also: [pr_hook()], [pr_hooks()]
    #' @examples
    #' \dontrun{
    #' pr <- pr()
    #' pr$registerHook("preroute", function(req){
    #'   cat("Routing a request for", req$PATH_INFO, "...\n")
    #' })
    #' pr$registerHooks(list(
    #'   preserialize=function(req, value){
    #'     print("About to serialize this value:")
    #'     print(value)
    #'
    #'     # Must return the value since we took one in. Here we're not choosing
    #'     # to mutate it, but we could.
    #'     value
    #'   },
    #'   postserialize=function(res){
    #'     print("We serialized the value as:")
    #'     print(res$body)
    #'   }
    #' ))
    #'
    #' pr$handle("GET", "/", function(){ 123 })
    #' }
    registerHook = function(stage=c("preroute", "postroute",
                                    "preserialize", "postserialize", "exit"), handler){
      stage <- match.arg(stage)
      super$registerHook(stage, handler)
    },
    #' @description Define endpoints
    #'
    #' The “handler” functions that you define in these handle calls
    #' are identical to the code you would have defined in your plumber.R file
    #' if you were using annotations to define your API. The handle() method
    #' takes additional arguments that allow you to control nuanced behavior
    #' of the endpoint like which filter it might preempt or which serializer
    #' it should use.
    #'
    #' See also: [pr_handle()], [pr_get()], [pr_post()], [pr_put()], [pr_delete()]
    #' @param methods a character string. http method.
    #' @param path a character string. Api endpoints
    #' @param handler a handler function.
    #' @param preempt a preempt function.
    #' @param serializer a serializer function.
    #' @param parsers a named list of parsers.
    #' @param endpoint a `PlumberEndpoint` object.
    #' @param ... additional arguments for [PlumberEndpoint] `new` method (namely `lines`, `params`, `comments`, `responses` and `tags`. Excludes `envir`).
    #' @examples
    #' \dontrun{
    #' pr <- pr()
    #' pr$handle("GET", "/", function(){
    #'   "<html><h1>Programmatic Plumber!</h1></html>"
    #' }, serializer=plumber::serializer_html())
    #' }
    handle = function(methods, path, handler, preempt, serializer, parsers, endpoint, ...) {
      epdef <- !missing(methods) || !missing(path) || !missing(handler) || !missing(serializer) || !missing(parsers)
      if (!missing(endpoint) && epdef) {
        stop("You must provide either the components for an endpoint (handler and serializer) OR provide the endpoint yourself. You cannot do both.")
      }

      if (epdef) {
        if (missing(serializer)) {
          serializer <- private$default_serializer
        }
        if (missing(parsers)) {
          parsers <- private$parsers
        }
        forbid <- c("verbs", "expr", "envir")
        forbid_check <- forbid %in% names(list(...))
        if (any(forbid_check)) {
          stop(paste0("`", forbid[forbid_check], "`", collapse = ", "), " can not be supplied to `pr$handle()` method.")
        }

        endpoint <- PlumberEndpoint$new(verbs = methods,
                                        path = path,
                                        expr = handler,
                                        envir = private$envir,
                                        serializer = serializer,
                                        parsers = parsers, ...)
      }
      private$addEndpointInternal(endpoint, preempt)
    },
    #' @description Remove endpoints
    #' @param methods a character string. http method.
    #' @param path a character string. Api endpoints
    #' @param preempt a preempt function.
    removeHandle = function(methods, path, preempt = NULL){
      private$removeEndpointInternal(methods, path, preempt)
    },
    #' @description Print representation of plumber router.
    #' @param prefix a character string. Prefix to append to representation.
    #' @param topLevel a logical value. When method executed on top level
    #' router, set to `TRUE`.
    #' @param ... additional arguments for recursive calls
    #' @return A terminal friendly representation of a plumber router.
    print = function(prefix="", topLevel=TRUE, ...){
      endCount <- as.character(sum(unlist(lapply(self$endpoints, length))))

      # Reference on box characters: https://en.wikipedia.org/wiki/Box-drawing_character

      cat(prefix)
      if (!topLevel){
        cat("\u2502 ") # "| "
      }

      # Avoid printing recursion (mount on mount on mount on ...)
      if (!isTRUE(topLevel)) {
        if (isTRUE(self$flags$is_printing)) {
          cat(
            crayon::bgYellow(
              crayon::black(
                "# Circular Plumber router definition detected")),
            "\n", sep=""
          )
          return()
        }
        # set flags to avoid inf recursion
        on.exit({ self$flags$is_printing <- NULL }, add = TRUE)
        self$flags$is_printing <- TRUE
      }

      cat(crayon::silver("# Plumber router with ", endCount, " endpoint", ifelse(endCount == 1, "", "s"),", ",
                         as.character(length(private$filts)), " filter", ifelse(length(private$filts) == 1, "", "s"),", and ",
                         as.character(length(self$mounts)), " sub-router", ifelse(length(self$mounts) == 1, "", "s"),".\n", sep=""))

      if(topLevel){
        cat(prefix, crayon::silver("# Use `pr_run()` on this object to start the API.\n"), sep="")
      }

      # Filters
      # TODO: scrub internal filters?
      for (f in private$filts){
        cat(prefix, "\u251c\u2500\u2500", crayon::green("[", f$name, "]", sep=""), "\n", sep="") # "+--"
      }

      printEndpoints <- function(prefix, name, nodes, isLast){
        if (is.list(nodes)){
          verbs <- paste(sapply(nodes, function(n){ n$verbs }), collapse=", ")
        } else {
          verbs <- nodes$verbs
        }
        cat(prefix)
        if (isLast){
          cat("\u2514") # "|_"
        } else {
          cat("\u251c")  # "+"
        }
        cat(crayon::blue("\u2500\u2500/", name, " (", verbs, ")\n", sep=""), sep="") # "+--"
      }

      printNode <- function(node, name="", prefix="", isRoot=FALSE, isLast = FALSE){

        childPref <- paste0(prefix, "\u2502  ")
        if (isRoot){
          childPref <- prefix
        }

        if (inherits(node, "PlumberEndpoint")) {
          # base case
          printEndpoints(prefix, name, node, isLast)

        } else if (is_plumber(node)){
          # base case
          cat(prefix, "\u251c\u2500\u2500/", name, "\n", sep="") # "+--"
          # It's a router, let it print itself
          print(node, prefix=childPref, topLevel=FALSE)

        } else if (is.list(node)) {

          has_no_name <-
            if (is.null(names(node))) {
              TRUE
            } else {
              # there are other endpoints, so get only nodes with name ""
              # which path does not end with / and path is not root
              node_path <- function(node) {
                path <- node$path %||% ""
                if (!is.character(path)) path <-""
                path
              }
              names(node) == "" & !grepl(".+/$", vapply(node, node_path, character(1)))
            }

          # mounted routers at root location will also have a missing name.
          # only look for endpoints
          are_endpoints <- has_no_name & vapply(node, inherits, logical(1), "PlumberEndpoint")
          # print all endpoints in a single line with verbs attached together
          if (any(are_endpoints)) {
            printEndpoints(prefix, name, node[are_endpoints], isLast)
          }

          # recurse
          if (any(!are_endpoints)) {
            node <- node[!are_endpoints]
            if (!isRoot){
              cat(prefix, "\u251c\u2500\u2500/", name, "\n", sep="") # "+--"
            }
            for (i in seq_along(node)) {
              name <- names(node)[i]
              printNode(node[[i]], name, childPref, isLast = (i == length(node)))
            }
          }

        } else {
          cat("??")
        }
      }
      printNode(self$routes, "", prefix, TRUE)

      invisible(self) # actually needs to be invisible
    },
    #' @description Serve a request
    #' @param req request object
    #' @param res response object
    serve = function(req, res) {
      hookEnv <- new.env()

      prerouteStep <- function(...) {
        private$runHooks("preroute", list(data = hookEnv, req = req, res = res))
      }
      routeStep <- function(...) {
        self$route(req, res)
      }
      postrouteStep <- function(value, ...) {
        private$runHooks("postroute", list(data = hookEnv, req = req, res = res, value = value))
      }

      serializeSteps <- function(value, ...) {
        if ("PlumberResponse" %in% class(value)) {
          return(res$toResponse())
        }

        ser <- res$serializer
        if (typeof(ser) != "closure") {
          stop("Serializers must be closures: '", ser, "'")
        }

        preserializeStep <- function(value, ...) {
          private$runHooks("preserialize", list(data = hookEnv, req = req, res = res, value = value))
        }
        serializeStep <- function(value, ...) {
          ser(value, req, res, private$errorHandler)
        }
        postserializeStep <- function(value, ...) {
          private$runHooks("postserialize", list(data = hookEnv, req = req, res = res, value = value))
        }

        runSteps(
          value,
          stop,
          list(
            preserializeStep,
            serializeStep,
            postserializeStep
          )
        )
      }

      errorHandlerStep <- function(error, ...) {
        # must set the body and return as this is after the serialize step
        res$body <- private$errorHandler(req, res, error)
        return(res$toResponse())
      }

      runSteps(
        NULL,
        errorHandlerStep,
        list(
          prerouteStep,
          routeStep,
          postrouteStep,
          serializeSteps
        )
      )

      #
      # conclude <- function(v) {
      #   v <- private$runHooks("postroute", list(data=hookEnv, req=req, res=res, value=v))
      #
      #   if ("PlumberResponse" %in% class(v)){
      #     # They returned the response directly, don't serialize.
      #     res$toResponse()
      #   } else {
      #     ser <- res$serializer
      #
      #     if (typeof(ser) != "closure") {
      #       stop("Serializers must be closures: '", ser, "'")
      #     }
      #
      #     v <- private$runHooks("preserialize", list(data=hookEnv, req=req, res=res, value=v))
      #     out <- ser(v, req, res, private$errorHandler)
      #     out <- private$runHooks("postserialize", list(data=hookEnv, req=req, res=res, value=out))
      #     out
      #   }
      # }
      #
      # if (hasPromises() && promises::is.promise(val)){
      #   # The endpoint returned a promise, we should wait on it
      #   then(val, conclude, function(error){
      #     # The original error handler would not have run because the endpoint didn't
      #     # synchronously produce any errors. We have to run our error handling logic now.
      #     # TODO: Dry this up with the error handler in route()
      #     v <- private$errorHandler(req, res, error)
      #     conclude(v)
      #   })
      # } else {
      #   conclude(val)
      # }
    },
    #' @description Route a request
    #' @param req request object
    #' @param res response object
    route = function(req, res) {
      getHandle <- function(filt) {
        handlers <- private$ends[[filt]]
        if (!is.null(handlers)) {
          for (h in handlers) {
            if (h$canServe(req)) {
              return(h)
            }
          }
        }
        return(NULL)
      }

      path <- req$PATH_INFO
      makeHandleStep <- function(name) {
        function(...) {
          resetForward()
          h <- getHandle(name)
          if (is.null(h)) {
            return(forward())
          }
          if (!is.null(h$serializer)) {
            res$serializer <- h$serializer
          }
          parsers <-
            if (!is.null(h$parsers)) {
              h$parsers
            } else {
              private$default_parsers
            }
          req$argsPath <- h$getPathParams(path)
          # `req_body_parser()` will also set `req$body` with the untouched body value
          req$body <- req_body_parser(req, parsers)
          req$argsBody <- req_body_args(req)

          req$args <- c(
            # (does not contain req or res)
            # will contain all args added in filters
            # `req$argsQuery` is available, but already absorbed into `req$args`
            req$args,
            # path is more important than body
            req$argsPath,
            # body is added last
            req$argsBody
          )

          return(
            h$exec(req, res)
          )
        }
      }

      steps <- list(
        # first step
        makeHandleStep("__first__")
      )

      # Start running through filters until we find a matching endpoint.
      # returns 2 functions which need to be flattened overall
      filterSteps <- unlist(recursive = FALSE, lapply(private$filts, function(fi) {
        # Check for endpoints preempting in this filter.
        handleStep <- makeHandleStep(fi$name)

        # Execute this filter
        # Do not stop if the filter returned a non-forward object
        # If a non-forward object is returned, serialize it according to the filter
        filterStep <- function(...) {

          filterExecStep <- function(...) {
            resetForward()
            fi$exec(req, res)
          }
          postFilterStep <- function(fres, ...) {
            if (hasForwarded()) {
              # return like normal
              return(fres)
            }
            # forward() wasn't called, presumably meaning the request was
            # handled inside of this filter.
            if (!is.null(fi$serializer)) {
              res$serializer <- fi$serializer
            }
            return(fres)
          }

          runSteps(
            NULL,
            stop,
            list(
              filterExecStep,
              postFilterStep
            )
          )
        }

        list(
          handleStep,
          filterStep
        )
      }))
      steps <- append(steps, filterSteps)

      # If we still haven't found a match, check the un-preempt'd endpoints.
      steps <- append(steps, list(makeHandleStep("__no-preempt__")))

      # We aren't going to serve this endpoint; see if any mounted routers will
      mountSteps <- lapply(names(private$mnts), function(mountPath) {
        # (make step function)
        function(...) {
          resetForward()
          # TODO: support globbing?

          if (nchar(path) >= nchar(mountPath) && substr(path, 0, nchar(mountPath)) == mountPath) {
            # This is a prefix match or exact match. Let this router handle.

            # First trim the prefix off of the PATH_INFO element
            req$PATH_INFO <- substr(req$PATH_INFO, nchar(mountPath), nchar(req$PATH_INFO))
            return(private$mnts[[mountPath]]$route(req, res))
          } else {
            return(forward())
          }
        }
      })
      steps <- append(steps, mountSteps)

      # No endpoint could handle this request. 404
      notFoundStep <- function(...) {

        if (isTRUE(getOption("plumber.trailingSlash", FALSE))) {
          # Redirect to the slash route, if it exists
          path <- req$PATH_INFO
          # If the path does not end in a slash,
          if (!grepl("/$", path)) {
            new_path <- paste0(path, "/")
            # and a route with a slash exists...
            if (router_has_route(req$pr, new_path, req$REQUEST_METHOD)) {

              # Temp redirect with same REQUEST_METHOD
              # Add on the query string manually. They do not auto transfer
              # The POST body will be reissued by caller
              new_location <- paste0(new_path, req$QUERY_STRING)
              res$status <- 307
              res$setHeader(
                name = "Location",
                value = new_location
              )
              res$serializer <- serializer_unboxed_json()
              return(
                list(message = "307 - Redirecting with trailing slash")
              )
            }
          }
        }

        # No trailing-slash route exists...
        # Try allowed verbs

        if (isTRUE(getOption("plumber.methodNotAllowed", TRUE))) {
          # Notify about allowed verbs
          if (is_405(req$pr, req$PATH_INFO, req$REQUEST_METHOD)) {
            res$status <- 405L
            # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Allow
            res$setHeader("Allow", paste(req$verbsAllowed, collapse = ", "))
            res$serializer <- serializer_unboxed_json()
            return(list(error = "405 - Method Not Allowed"))
          }
        }

        # Notify that there is no route found
        private$notFoundHandler(req = req, res = res)
      }
      steps <- append(steps, list(notFoundStep))

      errorHandlerStep <- function(error, ...) {
        private$errorHandler(req, res, error)
      }

      withCurrentExecDomain(req, res, { # used to allow `hasForwarded` to work
        withWarn1({
          runStepsIfForwarding(NULL, errorHandlerStep, steps)
        })
      })
    },
    #' @description \pkg{httpuv} interface call function. (Required for \pkg{httpuv})
    #' @param req request object
    call = function(req) {
      # Set the arguments to an empty list
      req$pr <- self
      req$.internal <- new.env()

      res <- PlumberResponse$new(private$default_serializer)
      req$args <- list()

      # maybe return a promise object
      self$serve(req, res)
    },
    #' @description httpuv interface onHeaders function. (Required for \pkg{httpuv})
    #' @param req request object
    onHeaders = function(req) {
      maxSize <- private$maxSize
      if (isTRUE(maxSize <= 0))
        return(NULL)

      reqSize <- 0
      # https://github.com/rstudio/shiny/blob/a022a2b4/R/middleware.R#L298-L301
      if (length(req$CONTENT_LENGTH) > 0)
        reqSize <- as.numeric(req$CONTENT_LENGTH)
      else if (length(req$HTTP_TRANSFER_ENCODING) > 0)
        reqSize <- Inf

      if (isTRUE(reqSize > maxSize)) {
        return(list(status = 413L,
                    headers = list('Content-Type' = 'text/plain'),
                    body = 'Maximum upload size exceeded'))
      }
      else {
        return(NULL)
      }
    },
    #' @description httpuv interface onWSOpen function. (Required for \pkg{httpuv})
    #' @param ws WebSocket object
    onWSOpen = function(ws){
      warning("WebSockets not supported.")
    },
    #' @description Sets the default serializer of the router.
    #'
    #' See also: [pr_set_serializer()]
    #' @param serializer a serializer function
    #' @examples
    #' \dontrun{
    #' pr <- pr()
    #' pr$setSerializer(serializer_unboxed_json())
    #' }
    setSerializer = function(serializer) {
      private$default_serializer <- serializer
    },
    #' @description Sets the default parsers of the router. Initialized to `c("json", "form", "text", "octet", "multi")`
    #' @template pr_setParsers__parsers
    setParsers = function(parsers) {
      private$default_parsers <- make_parser(parsers)
    },
    #' @description Sets the handler that gets called if an
    #' incoming request can’t be served by any filter, endpoint, or sub-router.
    #'
    #' See also: [pr_set_404()]
    #' @param fun a handler function.
    #' @examples
    #' \dontrun{
    #' pr <- pr()
    #' pr$set404Handler(function(req, res) {cat(req$PATH_INFO)})
    #' }
    set404Handler = function(fun){
      private$notFoundHandler <- fun
    },
    #' @description Sets the error handler which gets invoked if any filter or
    #' endpoint generates an error.
    #'
    #' See also: [pr_set_404()]
    #' @param fun a handler function.
    #' @examples
    #' \dontrun{
    #' pr <- pr()
    #' pr$setErrorHandler(function(req, res, err) {
    #'   message("Found error: ")
    #'   str(err)
    #' })
    #' }
    setErrorHandler = function(fun){
      private$errorHandler <- fun
    },
    #' @description Set visual documentation to use for API
    #'
    #' See also: [pr_set_docs()], [register_docs()], [registered_docs()]
    #' @param docs a character value or a logical value. See [pr_set_docs()] for examples.
    #'  If using [options_plumber()], the value must be set before initializing your Plumber router.
    #' @param ... Arguments for the visual documentation. See each visual documentation package for further details.
    setDocs = function(
      docs = getOption("plumber.docs", TRUE),
      ...
    ) {
      private$docs_info <- upgrade_docs_parameter(docs, ...)
    },
    #' @description Set a callback to notify where the API's visual documentation is located.
    #'
    #' When set, it will be called with a character string corresponding
    #' to the API docs url. This allows RStudio to locate visual documentation.
    #'
    #' If using [options_plumber()], the value must be set before initializing your Plumber router.
    #'
    #' See also: [pr_set_docs_callback()]
    #' @param callback a callback function for taking action on the docs url. (Also accepts `NULL` values to disable the `callback`.)
    setDocsCallback = function(
      callback = getOption('plumber.docs.callback', NULL)
    ) {
      # Use callback when defined
      if (!length(callback) || !is.function(callback)) {
        callback <- function(...) { NULL }
      }
      if (length(formals(callback)) == 0) {
        stop("`callback` must accept at least 1 argument. (`api_url`)")
      }
      private$docs_callback <- callback
    },
    #' @description Set debug value to include error messages.
    #'
    #' See also: `$getDebug()` and [pr_set_debug()]
    #' @param debug `TRUE` provides more insight into your API errors.
    setDebug = function(debug = interactive()) {
      stopifnot(length(debug) == 1)
      private$debug <- isTRUE(debug)
    },
    #' @description Retrieve the `debug` value. If it has never been set, the result of `interactive()` will be used.
    #'
    #' See also: `$getDebug()` and [pr_set_debug()]
    getDebug = function() {
      private$debug %||% default_debug()
    },
    #' @description Add a filter to plumber router
    #'
    #' See also: [pr_filter()]
    #' @param name a character string. Name of filter
    #' @param expr an expr that resolve to a filter function or a filter function
    #' @param serializer a serializer function
    filter = function(name, expr, serializer){
      filter <- PlumberFilter$new(name, expr, private$envir, serializer)
      private$addFilterInternal(filter)
    },
    #' @description
    #' Allows to modify router autogenerated OpenAPI Specification
    #'
    #' Note, the returned value will be sent through [serializer_unboxed_json()] which will turn all length 1 vectors into atomic values.
    #' To force a vector to serialize to an array of size 1, be sure to call [as.list()] on your value. `list()` objects are always serialized to an array value.
    #'
    #' See also: [pr_set_api_spec()]
    #' @template pr_setApiSpec__api
    setApiSpec = function(api = NULL) {
      if (is.character(api) && length(api) == 1 && file.exists(api)) {
        if (tools::file_ext(api) %in% c("yaml", "yml")) {
          if (!requireNamespace("yaml", quietly = TRUE)) {
            stop("yaml must be installed to read yaml format")
          }
          api <- yaml::read_yaml(api, eval.expr = FALSE)
        } else {
          api <- jsonlite::read_json(api, simplifyVector = TRUE)
        }
      }
      api_fun <-
        if (is.null(api)) {
          identity
        } else {
          if (!is.function(api)) {
            # function to return the api object
            function(x) {
              api
            }
          } else {
            api
          }
        }
      private$api_spec_handler <- api_fun
    },
    #' @description Retrieve OpenAPI file
    getApiSpec = function() { #FIXME: test

      routerSpec <- private$routerSpecificationInternal(self)

      # Extend the previously parsed settings with the endpoints
      def <- utils::modifyList(private$globalSettings, list(paths = routerSpec))

      # Lay those over the default globals so we ensure that the required fields
      # (like API version) are satisfied.
      ret <- utils::modifyList(defaultGlobals, def)

      ret <- private$api_spec_handler(ret)

      # remove NA or NULL values, which OpenAPI parsers do not like
      ret <- removeNaOrNulls(ret)

      ret
    },

    # list of key/value pairs that should be temporarily set. Ex: is_printing = 1
    #' @field flags For internal use only
    flags = list(),


    ### Legacy/Deprecated
    #' @description addEndpoint has been deprecated in v0.4.0 and will be removed in a coming release. Please use `handle()` instead.
    #' @param verbs verbs
    #' @param path path
    #' @param expr expr
    #' @param serializer serializer
    #' @param processors processors
    #' @param preempt preempt
    #' @param params params
    #' @param comments comments
    addEndpoint = function(verbs, path, expr, serializer, processors, preempt=NULL, params=NULL, comments){
      warning("addEndpoint has been deprecated in v0.4.0 and will be removed in a coming release. Please use `handle()` instead.")
      if (!missing(processors) || !missing(params) || !missing(comments)){
        stop("The processors, params, and comments parameters are no longer supported.")
      }

      self$handle(verbs, path, expr, preempt, serializer)
    },
    #' @description addAssets has been deprecated in v0.4.0 and will be removed in a coming release. Please use `mount` and `PlumberStatic$new()` instead.
    #' @param dir dir
    #' @param path path
    #' @param options options
    addAssets = function(dir, path="/public", options=list()){
      warning("addAssets has been deprecated in v0.4.0 and will be removed in a coming release. Please use `mount` and `PlumberStatic$new()` instead.")
      if (substr(path, 1,1) != "/"){
        path <- paste0("/", path)
      }

      stat <- PlumberStatic$new(dir, options)
      self$mount(path, stat)
    },
    #' @description `$addFilter()` has been deprecated in v0.4.0 and will be removed in a coming release. Please use `$filter()` instead.
    #' @param name name
    #' @param expr expr
    #' @param serializer serializer
    #' @param processors processors
    addFilter = function(name, expr, serializer, processors){
      warning("addFilter has been deprecated in v0.4.0 and will be removed in a coming release. Please use `filter` instead.")
      if (!missing(processors)){
        stop("The processors parameter is no longer supported.")
      }

      filter <- PlumberFilter$new(name, expr, private$envir, serializer)
      private$addFilterInternal(filter)
    },
    #' @description `$addGlobalProcessor()` has been deprecated in v0.4.0 and will be removed in a coming release. Please use `$registerHook`(s) instead.
    #' @param proc proc
    addGlobalProcessor = function(proc){
      warning("addGlobalProcessor has been deprecated in v0.4.0 and will be removed in a coming release. Please use `registerHook`(s) instead.")
      self$registerHooks(proc)
    },
    #' @description Deprecated. Retrieve OpenAPI file
    openAPIFile = function() {
      warning("`$openAPIFile()` has been deprecated in v1.0.0 and will be removed in a coming release. Please use `$getApiSpec()`.")
      self$getApiSpec()
    },
    #' @description Deprecated. Retrieve OpenAPI file
    swaggerFile = function() {
      warning("`$swaggerFile()` has been deprecated in v1.0.0 and will be removed in a coming release. Please use `$getApiSpec()`.")
      self$getApiSpec()
    }
  ), active = list(
    #' @field endpoints Plumber router endpoints read-only
    endpoints = function(){
      private$ends
    },
    #' @field filters Plumber router filters read-only
    filters = function(){
      private$filts
    },
    #' @field mounts Plumber router mounts read-only
    mounts = function(){
      private$mnts
    },
    #' @field environment Plumber router environment read-only
    environment = function() {
      private$envir
    },
    #' @field routes Plumber router routes read-only
    routes = function(){
      paths <- list()

      addPath <- function(node, children, endpoint){
        if (length(children) == 0){
          if (is.null(node)){
            return(endpoint)
          } else {
            # Concat to existing.
            return(c(node, endpoint))
          }

        }
        if (is.null(node)){
          node <- list()
        }

        # Check for existing endpoints at current children node that share the same name
        matching_name_nodes <- node[names(node) == children[1]]
        existing_endpoints <- vapply(matching_name_nodes, inherits, logical(1), "PlumberEndpoint")

        # This is for situation where an endpoint is on `/A` and you
        # also have route with an endpoint on `A/B`. Resulting nested list
        # already has an endpoint on the children node and you need a deeper nested
        # list for the current children node. Combine them.
        if (any(existing_endpoints) && length(children) > 1) {
          node <- c(
            # Nodes with preexisting endpoints sharing the same name
            matching_name_nodes[existing_endpoints],
            # New nested list to combine with, passing the nodes that are not endpoints
            addPath(matching_name_nodes[!existing_endpoints], children, endpoint)
          )
        } else {
          # Keep building the nested list until you hit an endpoint
          node[[children[1]]] <- addPath(node[[children[1]]], children[-1], endpoint)
        }

        node
      }

      lapply(self$endpoints, function(ends){
        lapply(ends, function(e){
          # Trim leading slash
          path <- sub("^/", "", e$path)

          levels <- strsplit(path, "/", fixed=TRUE)[[1]]
          # If there is a trailing `/`, add a blank level for an extra print line
          if (grepl("/$", path)) {levels <- c(levels, "")}
          paths <<- addPath(paths, levels, e)
        })
      })

      # Sub-routers
      if (length(self$mounts) > 0){
        for(i in 1:length(self$mounts)){
          # Trim leading slash
          path <- sub("^/", "", names(self$mounts)[i])

          levels <- strsplit(path, "/", fixed=TRUE)[[1]]

          m <- self$mounts[[i]]
          paths <- addPath(paths, levels, m)
        }
      }

      lexisort <- function(paths) {
        if (is.list(paths)) {
          paths <- lapply(paths, lexisort)
          if (!is.null(names(paths))) {
            paths <- paths[order(names(paths))]
          }
        }
        paths
      }

      lexisort(paths)
    }
  ),
  private = list(

    default_serializer = NULL, # The default serializer for the router
    default_parsers = NULL, # The default parsers for the router

    ends = list(), # List of endpoints indexed by their pre-empted filter.
    filts = NULL, # Array of filters
    mnts = list(),

    envir = NULL, # The environment in which all API execution will be conducted
    filename = NULL, # The file which was plumbed
    lines = NULL, # The lines constituting the API
    parsed = NULL, # The parsed representation of the API
    globalSettings = list(info=list()), # Global settings for this API. Primarily used for OpenAPI Specification.
    disable_run = NULL, # Disable run method during parsing of the Plumber file

    errorHandler = NULL,
    notFoundHandler = NULL,
    maxSize = 0, # Max request size in bytes. (0 is a no-op)

    api_spec_handler = NULL,
    docs_info = NULL,
    docs_callback = NULL,
    debug = NULL,

    addFilterInternal = function(filter){
      # Create a new filter and add it to the router
      private$filts <- c(private$filts, filter)

      self
    },
    addEndpointInternal = function(ep, preempt){
      noPreempt <- missing(preempt) || is.null(preempt)

      filterNames <- "__first__"
      for (f in private$filts){
        filterNames <- c(filterNames, f$name)
      }
      if (!noPreempt && ! preempt %in% filterNames){
        if (!is.null(ep$lines)){
          stopOnLine(ep$lines[1], private$fileLines[ep$lines[1]], paste0("The given @preempt filter does not exist in this Plumber router: '", preempt, "'"))
        } else {
          stop(paste0("The given preempt filter does not exist in this Plumber router: '", preempt, "'"))
        }
      }

      if (noPreempt){
        preempt <- "__no-preempt__"
      }

      private$ends[[preempt]] <- c(private$ends[[preempt]], ep)
    },
    removeEndpointInternal = function(methods, path, preempt){
      noPreempt <- is.null(preempt)

      if (noPreempt){
        preempt <- "__no-preempt__"
      }
      toRemove <- vapply(
        private$ends[[preempt]],
        function(ep) {
          isTRUE(all(ep$verbs %in% methods)) && isTRUE(ep$path == path)
        },
        logical(1))

      private$ends[[preempt]][toRemove] <- NULL
    },

    routerSpecificationInternal = function(router, parentPath = "") {
      remove_trailing_slash <- function(x) {
        sub("[/]$", "", x)
      }
      remove_leading_slash <- function(x) {
        sub("^[/]", "", x)
      }
      join_paths <- function(x, y) {
        x <- remove_trailing_slash(x)
        y <- remove_leading_slash(y)
        paste(x, y, sep = "/")
      }

      # make sure to use the full path
      endpointList <- list()

      for (endpoint in router$endpoints) {
        for (endpointEntry in endpoint) {
          endpointSpec <- endpointSpecification(
            endpointEntry,
            join_paths(parentPath, endpointEntry$path)
          )
          endpointList <- utils::modifyList(endpointList, endpointSpec)
        }
      }

      # recursively gather mounted enpoint entries
      if (length(router$mounts) > 0) {
        for (mountPath in names(router$mounts)) {
          mountEndpoints <- private$routerSpecificationInternal(
            router$mounts[[mountPath]],
            join_paths(parentPath, mountPath)
          )
          endpointList <- utils::modifyList(endpointList, mountEndpoints)
        }
      }

      # returning a single list of OpenAPI Paths Objects
      endpointList
    }
  )
)



upgrade_docs_parameter <- function(docs, ...) {
  stopifnot(length(docs) == 1)
  stopifnot(is.logical(docs) || is.character(docs))
  if (isTRUE(docs)) {
    docs <- "swagger"
  }
  if (is.character(docs) && is_docs_available(docs)) {
    enabled <- TRUE
  } else {
    enabled <- FALSE
    docs <- "__not_enabled__"
  }

  list(
    enabled = enabled,
    docs = docs,
    args = list(...),
    has_not_been_set = FALSE
  )
}



default_debug <- function() {
  interactive()
}


urlHost <- function(scheme = "http", host, port, path = "", changeHostLocation = FALSE) {
  if (isTRUE(changeHostLocation)) {
    # upgrade callback location to be localhost and not catch-all addresses
    # shiny: https://github.com/rstudio/shiny/blob/95173f6/R/server.R#L781-L786
    if (identical(host, "0.0.0.0")) {
      # RStudio IDE does NOT like 0.0.0.0 locations.
      # Must use 127.0.0.1 instead.
      host <- "127.0.0.1"
    } else if (identical(host, "::")) {
      # upgrade ipv6 catch-all to ipv6 "localhost"
      host <- "::1"
    }
  }

  # if ipv6 address, surround in brackets
  if (grepl(":[^/]", host)) {
    host <- paste0("[", host, "]")
  }

  if (is.null(scheme) || !nzchar(scheme)) {
    scheme <- "http"
  }

  paste0(scheme, "://", host, ":", port, path)
}

context("OpenAPI")

test_that("plumberToApiType works", {
  expect_equal(plumberToApiType("bool"), "boolean")
  expect_equal(plumberToApiType("logical"), "boolean")

  expect_equal(plumberToApiType("double"), "number")
  expect_equal(plumberToApiType("numeric"), "number")

  expect_equal(plumberToApiType("int"), "integer")

  expect_equal(plumberToApiType("character"), "string")

  expect_equal(plumberToApiType("df"), "object")
  expect_equal(plumberToApiType("list"), "object")
  expect_equal(plumberToApiType("data.frame"), "object")

  expect_warning({
    expect_equal(plumberToApiType("flargdarg"),  defaultApiType)
  }, "Unrecognized type:")
})

test_that("response attributes are parsed", {
  lines <- c(
    "#' @get /",
    "#' @response 201 This is response 201",
    "#' @response 202 Here's second",
    "#' @response 203 Here's third",
    "#' @response default And default")
  b <- plumbBlock(length(lines), lines)
  expect_length(b$responses, 4)
  expect_equal(b$responses$`201`, list(description="This is response 201"))
  expect_equal(b$responses$`202`, list(description="Here's second"))
  expect_equal(b$responses$`203`, list(description="Here's third"))
  expect_equal(b$responses$default, list(description="And default"))

  b <- plumbBlock(1, "")
  expect_null(b$responses)
})

test_that("params are parsed", {
  lines <- c(
    "#' @get /",
    "#' @param test Test docs",
    "#' @param required:character* Required param",
    "#' @param another:int Another docs",
    "#' @param multi:[int]* Required array param")
  b <- plumbBlock(length(lines), lines)
  expect_length(b$params, 4)
  expect_equal(b$params$another, list(desc="Another docs", type="integer", required=FALSE, isArray = FALSE))
  expect_equal(b$params$test, list(desc="Test docs", type=defaultApiType, required=FALSE, isArray = FALSE))
  expect_equal(b$params$required, list(desc="Required param", type="string", required=TRUE, isArray = FALSE))
  expect_equal(b$params$multi, list(desc="Required array param", type="integer", required=TRUE, isArray = TRUE))

  b <- plumbBlock(1, "")
  expect_null(b$params)
})

# TODO
#test_that("endpointSpecification works", {
#})

test_that("getApiSpec works with mounted routers", {
  # parameter in path
  pr1 <- pr()
  pr1$handle("GET", "/nested/:path/here", function(){})
  pr1$handle("POST", "/nested/:path/here", function(){})

    # static file handler
  stat <- PlumberStatic$new(".")

  # multiple entries
  pr2 <- pr()
  pr2$handle("GET", "/something", function(){})
  pr2$handle("POST", "/something", function(){})
  pr2$handle("GET", "/", function(){})

  # test with a filter
  pr3 <- pr()
  pr3$filter("filter1", function(){})
  pr3$handle("POST", "/else", function(){}, "filter1")
  pr3$handle("PUT", "/else", function(){})
  pr3$handle("GET", "/", function(){})

  # nested mount
  pr4 <- pr()
  pr4$handle("GET", "/completely", function(){})

  # trailing slash in route
  pr5 <- pr()
  pr5$handle("GET", "/trailing_slash/", function(){})

  # ├──/nested
  # │  ├──/:path
  # │  │  └──/here (GET, POST)
  # ├──/static
  # ├──/sub2
  # │  ├──/something (GET, POST)
  # │  ├──/ (GET)
  # │  ├──/sub3
  # │  │  ├──/else (POST, PUT)
  # │  │  └──/ (GET)
  # ├──/sub4
  # │  ├──/completely (GET)
  # │  ├──/
  # │  │  └──/trailing_slash (GET)
  pr1$mount("/static", stat)
  pr2$mount("/sub3", pr3)
  pr1$mount("/sub2", pr2)
  pr1$mount("/sub4", pr4)
  pr4$mount("/", pr5)

  paths <- names(pr1$getApiSpec()$paths)
  expect_length(paths, 7)
  expect_equal(paths, c("/nested/:path/here", "/sub2/something",
    "/sub2/", "/sub2/sub3/else", "/sub2/sub3/", "/sub4/completely",
    "/sub4/trailing_slash/"
  ))
})

test_that("responsesSpecification works", {
  # Empty
  r <- responsesSpecification(NULL)
  expect_equal(r, defaultResponse)

  # Response constructor actually defaults to NA, so that's an important case, too
  r <- responsesSpecification(NA)
  expect_equal(r, defaultResponse)

  # Responses with no default
  customResps <- list("400" = list())
  endpts <- plumber::PlumberEndpoint$new(
    path = "/a",
    responses = customResps,
    verbs = "GET",
    expr = function() {},
    envir = new.env(parent = globalenv()))
  r <- responsesSpecification(endpts)
  expect_length(r, 4)
  expect_equal(r$default, defaultResponse$default)
  expect_equal(r$`400`, customResps$`400`)
})

test_that("parametersSpecification works", {
  ep <- list(id=list(desc="Description", type="integer", required=FALSE),
             id2=list(desc="Description2", required=FALSE), # No redundant type specification
             make=list(desc="Make description", type="string", required=FALSE),
             prices=list(desc="Historic sell prices", type="numeric", required = FALSE, isArray = TRUE),
             claims=list(desc="Insurance claims", type="object", required = FALSE))
  pp <- data.frame(name=c("id", "id2", "owners"), type=c("int", "int", "chr"), isArray = c(FALSE, FALSE, TRUE), stringsAsFactors = FALSE)

  params <- parametersSpecification(ep, pp)
  expect_equal(params$parameters[[1]],
               list(name="id",
                    description="Description",
                    `in`="path",
                    required=TRUE, # Made required b/c path arg
                    schema = list(
                      type="integer",
                      format="int64",
                      default=NULL)))
  expect_equal(params$parameters[[2]],
               list(name="id2",
                    description="Description2",
                    `in`="path",
                    required=TRUE, # Made required b/c path arg
                    schema = list(
                      type="integer",
                      format="int64",
                      default=NULL)))
  expect_equal(params$parameters[[3]],
               list(name="make",
                    description="Make description",
                    `in`="query",
                    required=FALSE,
                    schema = list(
                      type="string",
                      format=NULL,
                      default=NULL)))
  expect_equal(params$parameters[[4]],
               list(name="prices",
                    description="Historic sell prices",
                    `in`="query",
                    required=FALSE,
                    schema = list(
                      type="array",
                      items= list(
                        type="number",
                        format="double"),
                      default = NULL),
                    style="form",
                    explode=TRUE))
  expect_equal(params$parameters[[5]],
               list(name="owners",
                    description=NULL,
                    `in`="path",
                    required=TRUE,
                    schema = list(
                      type="array",
                      items= list(
                        type="string",
                        format=NULL),
                      default = NULL),
                    style="simple",
                    explode=FALSE))
  expect_equal(params$requestBody,
               list(content = list(
                 `application/json` = list(
                   schema = list(
                     type = "object",
                     properties = list(
                       claims = list(
                         type = "object",
                         format = NULL,
                         example = NULL,
                         description = "Insurance claims")))))))

  # If id were not a path param it should not be promoted to required
  params <- parametersSpecification(ep, NULL)
  idParam <- params$parameters[[which(vapply(params$parameters, `[[`, character(1), "name") == "id")]]
  expect_equal(idParam$required, FALSE)
  expect_equal(idParam$schema$type, "integer")

  for (param in params$parameters) {
    if (param$schema$type != "array") {
      expect_equal(length(param), 5)
    } else {
      expect_equal(length(param), 7)
    }
  }

  # Check if we can pass a single path parameter without a @param line match
  params <- parametersSpecification(NULL, pp[3,])
  expect_equal(params$parameters[[1]],
               list(name="owners",
                    description=NULL,
                    `in`="path",
                    required=TRUE,
                    schema = list(
                      type="array",
                      items= list(
                        type="string",
                        format=NULL),
                      default=NULL),
                    style="simple",
                    explode=FALSE))

  params <- parametersSpecification(NULL, NULL)
  expect_equal(sum(sapply(params, length)), 0)
})

test_that("api kitchen sink", {

  skip_on_cran()
  skip_on_bioc()
  skip_on_os(setdiff(c("windows", "mac", "linux", "solaris"), c("mac", "linux")))

  ## install brew - https://brew.sh/
  # /usr/bin/ruby -e "$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/master/install)"
  ## install yarn
  # brew install yarn
  ## install yarn
  # yarn add swagger-ui

  # yarn install

  skip_if_not(nzchar(Sys.which("node")), "node not installed")
  skip_if_not(nzchar(Sys.which("npm")), "`npm` system dep not installed")

  for_each_plumber_api(validate_api_spec, verbose = FALSE)

  # TODO test more situations
})

test_that("multiple variations in function extract correct metadata", {
  dummy <- function(var0 = 420.69,
                    var1,
                    var2 = c(1L, 2L),
                    var3 = rnorm,
                    var4 = NULL,
                    var5 = FALSE,
                    var6 = list(name = c("luke", "bob"), lastname = c("skywalker", "ross")),
                    var7 = new.env(parent = .GlobalEnv),
                    var8 = list(a = 2, b = mean, c = new.env(parent = .GlobalEnv))) {}
  funcParams <- getArgsMetadata(dummy)
  expect_identical(sapply(funcParams, `[[`, "required"),
                   c(var0 = FALSE, var1 = TRUE, var2 = FALSE, var3 = FALSE, var4 = FALSE,
                     var5 = FALSE, var6 = FALSE, var7 = FALSE, var8 = FALSE))
  expect_identical(lapply(funcParams, `[[`, "default"),
                   list(var0 = 420.69, var1 = NA, var2 = 1L:2L, var3 = NA, var4 = NA, var5 = FALSE,
                        var6 = list(name = c("luke", "bob"), lastname = c("skywalker", "ross")), var7 = NA, var8 = NA))
  expect_identical(lapply(funcParams, `[[`, "example"),
                   list(var0 = 420.69, var1 = NA, var2 = 1L:2L, var3 = NA, var4 = NA, var5 = FALSE,
                        var6 = list(name = c("luke", "bob"), lastname = c("skywalker", "ross")), var7 = NA, var8 = NA))
  expect_identical(lapply(funcParams, `[[`, "isArray"),
                   list(var0 = defaultIsArray, var1 = defaultIsArray, var2 = TRUE,
                        var3 = defaultIsArray, var4 = defaultIsArray,
                        var5 = defaultIsArray, var6 = TRUE,
                        var7 = defaultIsArray, var8 = defaultIsArray))
  expect_identical(lapply(funcParams, `[[`, "type"),
                   list(var0 = "number", var1 = defaultApiType, var2 = "integer", var3 = defaultApiType, var4 = defaultApiType,
                        var5 = "boolean", var6 = "object", var7 = defaultApiType, var8 = defaultApiType))

})

test_that("priorize works as expected", {
  expect_identical("abc", priorizeProperty(structure("zzz", default = TRUE), NULL, "abc"))
  expect_identical(NULL, priorizeProperty(NULL, NULL, NULL))
  expect_identical(structure("zzz", default = TRUE), priorizeProperty(structure("zzz", default = TRUE), NULL, NA))
  expect_identical(NULL, priorizeProperty())
})

test_that("custom spec works", {
  pr <- pr()
  pr$handle("POST", "/func1", function(){})
  pr$handle("GET", "/func2", function(){})
  pr$handle("GET", "/func3", function(){})
  customSpec <- function(spec) {
    custom <- list(info = list(description = "My Custom Spec", title = "This is only a test"))
    return(utils::modifyList(spec, custom))
  }
  pr$setApiSpec(customSpec)
  spec <- pr$getApiSpec()
  expect_equal(spec$info$description, "My Custom Spec")
  expect_equal(spec$info$title, "This is only a test")
  expect_equal(class(spec$openapi), "character")
})

test_that("no params plumber router still produces spec when there is a func params", {
  pr <- pr()
  handler <- function(num) { sum(as.integer(num)) }
  pr$handle("GET", "/sum", handler, serializer = serializer_json())
  spec <- pr$getApiSpec()
  expect_equal(spec$paths$`/sum`$get$parameters[[1]]$name, "num")
})

test_that("Response content type set with serializer", {
  a <- pr()
  pr_get(a, "/json", function() {"OK"}, serializer = serializer_json())
  pr_get(a, "/csv", function() {"OK"}, serializer = serializer_csv())
  spec <- a$getApiSpec()
  expect_equal(spec$paths$`/json`$get$responses$`200`$content, list("application/json" = list(schema = list(type = "object"))))
  expect_equal(spec$paths$`/csv`$get$responses$`200`$content, list("text/csv" = list(schema = list(type = "string"))))
})

test_that("Api spec can be set using a file path", {
  pr <- pr() %>% pr_set_api_spec(test_path("files/openapi.yaml"))
  expect_equal(pr$getApiSpec()$paths$`/health-check`$get$summary, " Determine if the API is running and listening as expected")
})

test_that("RemoveNAorNULLs preserve nested examples", {
  a <- list(schema = list(example = list(a = 5, b = list(NULL))), examples = list(a = 5, value = list(NULL)))
  expect_identical(removeNaOrNulls(a), a)

  a <- list(schema = list(value = list(NULL), b = list(NULL)))
  expect_identical(removeNaOrNulls(a), list(schema = list(value = list(), b = list())))
})

context("test-query-with-multibytes")

test_that("Support multi-bytes queries", {
  r <- pr(test_path("files/query-with-multibytes.R"))
  res <- PlumberResponse$new()

  req <- make_req("GET", "/msg", "?param1=%E4%B8%AD%E6%96%87&param2=%E4%BD%A0%E5%A5%BD")
  out <- r$serve(req, res)$body
  expect_equal(Encoding(out), "UTF-8")
  expect_identical(charToRaw(out), charToRaw(jsonlite::toJSON("\u4e2d\u6587-\u4f60\u597d")))

  req <- make_req("POST", "/msg", "?param1=%E4%B8%AD%E6%96%87&param2=%E4%BD%A0%E5%A5%BD")
  out <- r$serve(req, res)$body
  expect_equal(Encoding(out), "UTF-8")
  expect_identical(charToRaw(out), charToRaw(jsonlite::toJSON("\u4e2d\u6587-\u4f60\u597d")))
})

context("Router modifier (@plumber tag)")

test_that("router modifier works", {
  pr <- pr(test_path("files/router-modifier.R"))
  expect_equal(class(pr$routes[[1]])[1], "PlumberEndpoint")
  expect_equal(names(pr$routes), "avatartare")
})

test_that("$run() causes error", {
  expect_error(
    pr(test_path("files/router-modifier-run.R")),
    "method should not be called while")
})


test_that("new routers can not be returned", {
  expect_error(
    pr(test_path("files/router-modifier-new.R")),
    "not the same as the one provided")
})

context("Plumber")

exec_endpoint <- function(pr, ep_pos, subset = 1) {
  # This is a poor setup of `req` and `res`. But it works for testing purposes
  pr$endpoints[[subset]][[ep_pos]]$exec(make_req(), PlumberResponse$new())
}

test_that("Endpoints are properly identified", {
  r <- pr(test_path("files/endpoints.R"))
  expect_equal(length(r$endpoints), 1)
  expect_equal(length(r$endpoints[[1]]), 5)
  expect_equal(exec_endpoint(r, 1), 5)
  expect_equal(exec_endpoint(r, 2), 5)
  expect_equal(exec_endpoint(r, 3), 10)
  expect_equal(exec_endpoint(r, 4), 12)
  expect_equal(exec_endpoint(r, 5), 14)
})

test_that("Empty file argument is OK", {
  r <- pr()
  expect_equal(length(r$endpoints), 0)
})

test_that("Empty file is OK", {
  f <- tempfile()
  writeLines(character(), f)
  on.exit(unlink(f), add = TRUE)
  r <- pr(f)
  expect_equal(length(r$endpoints), 0)
})

test_that("The file is sourced in the envir", {
  r <- pr(test_path("files/in-env.R"))
  expect_equal(length(r$endpoints), 1)
  expect_equal(length(r$endpoints[[1]]), 2)
  expect_equal(exec_endpoint(r, 1), 15)
})

test_that("Verbs translate correctly", {
  r <- pr(test_path("files/verbs.R"))
  expect_equal(length(r$endpoints), 1)
  expect_equal(length(r$endpoints[[1]]), 10)
  expect_equal(r$endpoints[[1]][[1]]$verbs, c("GET", "PUT", "POST", "DELETE", "HEAD", "OPTIONS", "PATCH"))
  expect_equal(r$endpoints[[1]][[2]]$verbs, "GET")
  expect_equal(r$endpoints[[1]][[3]]$verbs, "PUT")
  expect_equal(r$endpoints[[1]][[4]]$verbs, "POST")
  expect_equal(r$endpoints[[1]][[5]]$verbs, "DELETE")
  expect_equal(r$endpoints[[1]][[6]]$verbs, "POST")
  expect_equal(r$endpoints[[1]][[7]]$verbs, "GET")
  expect_equal(r$endpoints[[1]][[8]]$verbs, "HEAD")
  expect_equal(r$endpoints[[1]][[9]]$verbs, "OPTIONS")
  expect_equal(r$endpoints[[1]][[10]]$verbs, "PATCH")
})

test_that("Invalid file fails gracefully", {
  expect_error(pr("asdfsadf"), regexp="File does not exist.*asdfsadf")
})

test_that("plumb accepts a file", {
  r <- plumb(test_path("files/endpoints.R"))
  expect_length(r$endpoints[[1]], 5)
})

test_that("plumb gives a good error when passing in a dir instead of a file", {

  # brittle test. Fails on r-devel-windows-x86_64-gcc10-UCRT
  skip_on_cran()

  if (isWindows()) {
    # File paths are hard to work with and are inconsistent
    skip_on_os("windows")
  }

  expect_error(plumb(test_path("files/")), "Expecting a file but found a directory: 'files/'")
})

test_that("plumb accepts a directory with a `plumber.R` file", {
  # works without trailing slash
  r <- plumb(dir = test_path('files'))
  expect_equal(length(r$endpoints), 1)
  expect_equal(length(r$endpoints[[1]]), 5)

  # works with trailing slash
  r <- plumb(dir = paste0(test_path('files'), "/"))
  expect_equal(length(r$endpoints), 1)
  expect_equal(length(r$endpoints[[1]]), 5)

  # errors when no plumber.R found
  expect_error(plumb(dir = test_path("files/static")), regexp="No plumber.R file found in the specified directory: ")

  # errors when neither dir is empty and file is not given
  expect_error(plumb(dir=""), regexp="You must specify either a file or directory*")

  # reads from working dir if no args
  expect_error(plumb(), regexp="No plumber.R file found in the specified directory: .")

  # errors when both dir and file are given
  expect_silent(plumb(file = "endpoints.R", dir = test_path("files")))

})

test_that("plumb() a dir leverages `entrypoint.R`", {

  with_tmp_serializers({
    expect_false(
      "fake" %in% registered_serializers(),
      "This just that your Plumber environment is dirty. Restart your R session."
    )

    r <- plumb(dir = test_path("files/entrypoint/"))
    expect_equal(length(r$endpoints), 1)
    expect_equal(length(r$endpoints[[1]]), 1)

    # A global serializer was added by entrypoint.R before parsing
    expect_true(
      "fake" %in% registered_serializers(),
      "This just that your Plumber environment is dirty. Restart your R session."
    )
  })

  expect_false(
    "fake" %in% registered_serializers(),
    "This just that your Plumber environment is dirty. Restart your R session."
  )
})

test_that("bad `entrypoint.R`s throw", {
  expect_error(plumb(dir = test_path("files/entrypoint-bad/")), "runnable Plumber router")
})

test_that("plumb() a dir works with `entrypoint.R` and without `plumber.R`", {
  r <- plumb(dir = test_path("files/no-plumber/"))
  expect_equal(length(r$endpoints), 1)
  expect_equal(length(r$endpoints[[1]]), 1)
})

test_that("Empty endpoints error", {
  expect_error(pr(test_path("files/endpoints-empty.R")), regexp="No path specified")
})

test_that("The old roxygen-style comments work", {
  r <- pr(test_path("files/endpoints-old.R"))
  expect_equal(length(r$endpoints), 1)
  expect_equal(length(r$endpoints[[1]]), 5)
  expect_equal(exec_endpoint(r, 1), 5)
  expect_equal(exec_endpoint(r, 2), 5)
  expect_equal(exec_endpoint(r, 3), 10)
  expect_equal(exec_endpoint(r, 4), 12)
  expect_equal(exec_endpoint(r, 5), 14)
})

test_that("routes can be constructed correctly", {
  pr <- pr()
  pr$handle("GET", "/nested/path/here", function(){})
  pr$handle("POST", "/nested/path/here", function(){})

  pr2 <- pr()
  pr2$handle("POST", "/something", function(){})
  pr2$handle("GET", "/", function(){})
  pr$mount("/mysubpath", pr2)

  stat <- PlumberStatic$new(".")
  pr$mount("/static", stat)

  pr3 <- pr()
  pr_get(pr3, "a/b", function(){})
  pr_get(pr3, "a", function(){})
  expect_length(pr3$routes$a, 2)

  pr4 <- pr()
  pr_get(pr4, "a", function(){})
  pr_post(pr4, "a/b/c/f", function(){})
  pr_get(pr4, "a/b/c/f", function(){})
  pr_get(pr4, "a/b/c/f/g/h/j/k", function(){})
  pr_get(pr4, "v/b/c/f", function(){})
  pr_get(pr4, "v/b/c/b", function(){})
  pr_get(pr4, "v/b/c/a", function(){})
  pr_get(pr4, "t", function(){})
  pr_post(pr4, "u/b/c/f", function(){})
  pr_get(pr4, "i/b/c/f/g/h/j/k", function(){})
  expect_equal(names(pr4$routes), c("a", "a", "i", "t", "u", "v"))
  expect_equal(names(pr4$routes$v$b$c), c("a", "b", "f"))

  expect_length(pr$routes, 3)
  expect_s3_class(pr$routes[["static"]], "PlumberStatic")
  expect_s3_class(pr$routes[["mysubpath"]], "Plumber")

  # 2 endpoints at the same location (different verbs)
  expect_length(pr$routes$nested$path$here, 2)
})

test_that("mounts can be read correctly", {
  pr <- pr()
  pr$handle("GET", "/nested/path/here", function(){})
  pr$handle("POST", "/nested/path/here", function(){})

  pr2 <- pr()
  pr2$handle("POST", "/something", function(){})
  pr2$handle("GET", "/", function(){})
  pr$mount("/mysubpath", pr2)

  stat <- PlumberStatic$new(".")
  pr$mount("/static", stat)
  pr$mount("missing-slashes", stat)
  pr$mount("/both-slashes/", stat)
  pr$mount("trailing-slash/", stat)
  pr$mount("/extra-slash//", stat)

  expect_length(pr$routes, 7)
  expect_s3_class(pr$mounts[["/static/"]], "PlumberStatic")
  expect_s3_class(pr$mounts[["/missing-slashes/"]], "PlumberStatic")
  expect_s3_class(pr$mounts[["/both-slashes/"]], "PlumberStatic")
  expect_s3_class(pr$mounts[["/trailing-slash/"]], "PlumberStatic")
  expect_s3_class(pr$mounts[["/extra-slash//"]], "PlumberStatic")
  expect_s3_class(pr$mounts[["/mysubpath/"]], "Plumber")
})



test_that("mounts work", {
  pr <- pr()
  sub <- pr()
  sub$handle("GET", "/", function(){ 1 })
  sub$handle("GET", "/nested/path", function(a){ a })

  pr$mount("/subpath", sub)

  res <- PlumberResponse$new()
  pr$route(make_req("GET", "/nested/path"), res)
  expect_equal(res$status, 404)

  val <- pr$route(make_req("GET", "/subpath/nested/path", qs="?a=123"), PlumberResponse$new())
  expect_equal(val, "123")

  val <- pr$route(make_req("GET", "/subpath/nested/path", body='{"a":123}'), PlumberResponse$new())
  expect_equal(val, 123)

  val <- pr$route(make_req("GET", "/subpath/"), PlumberResponse$new())
  expect_equal(val, 1)
})

test_that("mounting at root path works", {
  pr <- pr()
  sub <- pr()
  sub$handle("GET", "/", function(){ 1 })
  sub$handle("GET", "/nested/path", function(){ 2 })

  pr$mount("/", sub)

  val <- pr$route(make_req("GET", "/nested/path"), PlumberResponse$new())
  expect_equal(val, 2)

  val <- pr$route(make_req("GET", "/"), PlumberResponse$new())
  expect_equal(val, 1)
})

test_that("conflicting mounts behave consistently", {
  pr <- pr()

  sub <- pr()
  sub$handle("GET", "/", function(){ 1 })
  pr$mount("/subpath", sub)

  val <- pr$route(make_req("GET", "/subpath/"), PlumberResponse$new())
  expect_equal(val, 1)

  pr$handle("GET", "/subpath/", function(){ 2 })

  val <- pr$route(make_req("GET", "/subpath/"), PlumberResponse$new())
  expect_equal(val, 2)
})

test_that("hooks can be registered", {
  pr <- pr()
  events <- NULL
  pr$handle("GET", "/", function(){ events <<- c(events, "exec") })
  pr$registerHook("preroute", function(){ events <<- c(events, "preroute") })
  pr$registerHook("postroute", function(){ events <<- c(events, "postroute") })
  pr$registerHook("preserialize", function(){ events <<- c(events, "preserialize") })
  pr$registerHook("postserialize", function(){ events <<- c(events, "postserialize") })

  pr$call(make_req("GET", "/"))
  expect_equal(events, c("preroute", "exec", "postroute", "preserialize", "postserialize"))
})

test_that("preroute hook gets the right data", {
  pr <- pr()
  pr$handle("GET", "/", function(){ })
  rqst <- make_req("GET", "/")

  pr$registerHook("preroute", function(data, req, res){
    expect_s3_class(res, "PlumberResponse")
    expect_equal(rqst, req)
    expect_true(is.environment(data))
  })
  pr$call(rqst)
})

test_that("postroute hook gets the right data and can modify", {
  pr <- pr()
  pr$handle("GET", "/abc", function(){ 123 })

  pr$registerHook("postroute", function(data, req, res, value){
    expect_s3_class(res, "PlumberResponse")
    expect_equal(req$PATH_INFO, "/abc")
    expect_true(is.environment(data))
    expect_equal(value, 123)
    "new val"
  })
  res <- pr$call(make_req("GET", "/abc"))
  expect_equal(as.character(res$body), '["new val"]')
})

test_that("preserialize hook gets the right data and can modify", {
  pr <- pr()
  pr$handle("GET", "/abc", function(){ 123 })

  pr$registerHook("preserialize", function(data, req, res, value){
    expect_s3_class(res, "PlumberResponse")
    expect_equal(req$PATH_INFO, "/abc")
    expect_true(is.environment(data))
    expect_equal(value, 123)
    "new val"
  })
  res <- pr$call(make_req("GET", "/abc"))
  expect_equal(as.character(res$body), '["new val"]')
})

test_that("postserialize hook gets the right data and can modify", {
  pr <- pr()
  pr$handle("GET", "/abc", function(){ 123 })

  pr$registerHook("postserialize", function(data, req, res, value){
    expect_s3_class(res, "PlumberResponse")
    expect_equal(req$PATH_INFO, "/abc")
    expect_true(is.environment(data))
    expect_equal(as.character(value$body), "[123]")
    value$body <- "new val"
    value
  })
  res <- pr$call(make_req("GET", "/abc"))
  expect_equal(as.character(res$body), 'new val')
})

test_that("invalid hooks err", {
  pr <- pr()
  expect_error(pr$registerHook("flargdarg"))
})

test_that("handle invokes correctly", {
  with_options(
    list(plumber.trailingSlash = NULL),
    {
      pr <- pr()
      pr$handle("GET", "/trailslash", function(){ "getter" })
      pr$handle("POST", "/trailslashp/", function(){ "poster" })

      expect_equal(pr$call(make_req("GET", "/trailslash"))$body, jsonlite::toJSON("getter"))
      res <- pr$call(make_req("GET", "/trailslash/")) # With trailing slash
      expect_equal(res$status, 404)
      res <- pr$call(make_req("POST", "/trailslash")) # Wrong verb
      expect_equal(res$status, 405)

      expect_equal(pr$call(make_req("POST", "/trailslashp/"))$body, jsonlite::toJSON("poster"))
      res <- pr$call(make_req("POST", "/trailslashp")) # w/o trailing slash
      expect_equal(res$status, 404)
      res <- pr$call(make_req("GET", "/trailslashp/")) # Wrong verb
      expect_equal(res$status, 405)
    }
  )

})

test_that("trailing slashes are redirected", {

  pr <- pr() %>%
    pr_get("/get/", function(a) a) %>%
    pr_post("/post/", function(a) a) %>%
    pr_mount(
      "/mnt",
      pr() %>%
        pr_get("/", function(a) a)
    )

  with_options(list(plumber.trailingSlash = FALSE), {
    res <- pr$call(make_req("GET", "/get", "?a=1"))
    expect_equal(res$status, 404)

    res <- pr$call(make_req("POST", "/post", "?a=1"))
    expect_equal(res$status, 404)

    res <- pr$call(make_req("GET", "/mnt", "?a=1"))
    expect_equal(res$status, 404)
  })

  with_options(list(plumber.trailingSlash = TRUE), {
    res <- pr$call(make_req("GET", "/get", "?a=1"))
    expect_equal(res$status, 307)
    expect_equal(res$headers$Location, "/get/?a=1")

    res <- pr$call(make_req("POST", "/post", "?a=1"))
    expect_equal(res$status, 307)
    expect_equal(res$headers$Location, "/post/?a=1")

    res <- pr$call(make_req("GET", "/mnt", "?a=1"))
    expect_equal(res$status, 307)
    expect_equal(res$headers$Location, "/mnt/?a=1")
  })
})


test_that("No 405 on same path, different verb", {

  pr <- pr()
  pr$handle("GET", "/apathow", function(){ "getter" })
  pr$handle("POST", "/apathow", function(){ "poster" })

  expect_equal(pr$route(make_req("GET", "/apathow"), PlumberResponse$new()), "getter")
  expect_equal(pr$route(make_req("POST", "/apathow"), PlumberResponse$new()), "poster")

})

test_that("handle with an endpoint works", {
  pr <- pr()
  ep <- PlumberEndpoint$new("GET", "/", function(){ "manual endpoint" }, pr$environment, serializer_json())
  pr$handle(endpoint=ep)

  val <- pr$route(make_req("GET", "/"), PlumberResponse$new())
  expect_equal(val, "manual endpoint")
})

test_that("handle with and enpoint and endpoint def fails", {
  pr <- pr()
  ep <- PlumberEndpoint$new("GET", "/", function(){ "manual endpoint" }, pr$environment, serializer_json())
  expect_error(pr$handle("GET", "/", endpoint=ep))
})

test_that("full handle call works", {
  pr <- pr()
  pr$filter("f1", function(req){ req$filtered <- TRUE; forward() })

  pr$handle("GET", "/preempt", function(req){
    expect_null(req$filtered)
    "preempted"
  }, "f1", serializer_unboxed_json())

  pr$handle("GET", "/dontpreempt", function(req){
    expect_true(req$filtered)
    "unpreempted"
  }, serializer=serializer_unboxed_json())

  res <- PlumberResponse$new()
  val <- pr$route(make_req("GET", "/preempt"), res)
  expect_equal(val, "preempted") # no JSON box
  res <- PlumberResponse$new()
  val <- pr$route(make_req("GET", "/dontpreempt"), res)
  expect_equal(val, "unpreempted") # no JSON box
})

test_that("Expressions and functions both work on handle", {
  pr <- pr()
  pr$handle("GET", "/function", function(req){ req[["PATH_INFO"]] })
  pr$handle("GET", "/expression", expression(function(req){ req[["PATH_INFO"]] }))

  val <- pr$route(make_req("GET", "/function"), PlumberResponse$new())
  expect_equal(val, "/function")
  val <- pr$route(make_req("GET", "/expression"), PlumberResponse$new())
  expect_equal(val, "/expression")
})

test_that("Expressions and functions both work on filter", {
  pr <- pr()
  pr$filter("ff", function(req){ req$filteredF <- TRUE; forward() })
  pr$filter("fe", expression(function(req){ req$filteredE <- TRUE; forward() }))
  pr$handle("GET", "/", function(req){
    req$filteredE && req$filteredF
  })

  val <- pr$route(make_req("GET", "/"), PlumberResponse$new())
  expect_true(val)

  pr$handle("GET", "/expr", expression(function(req){
    req$filteredE && req$filteredF
  }))

  val <- pr$route(make_req("GET", "/expr"), PlumberResponse$new())
  expect_true(val)
})

test_that("filters and endpoint expressions evaluated in the appropriate (possibly injected) environment", {
  # Create an environment that contains a variable named `y`.
  env <- new.env(parent=.GlobalEnv)
  env$y <- 10

  # We provide expressions so that they get closurified in the right environment
  # and will be able to find `y`.
  # This would all fail without an injected environment that contains `y`.
  pr <- pr(envir=env)
  pr$filter("ff", expression(function(req){ req$ys <- y^2; forward() }))
  pr$handle("GET", "/", expression(function(req){ paste(y, req$ys) }))

  # Send a request through and we should see an assign to our env.
  val <- pr$route(make_req("GET", "/"), PlumberResponse$new())
  expect_equal(val, "10 100")
})

test_that("filters and endpoints executed in the appropriate environment", {
  # We've already seen that, if expressions, they're going to be evaluated in the
  # appropriate environment, but we can also confirm that once they've been evaluated,
  # they're then executed in the appropriate environment.

  # This almost certainly doesn't matter unless a function is inspecting the call stack,
  # but for the sake of consistency we not only ensure that any given expressions are
  # evaluated in the appropriate environment, but also that they are then called in the
  # given environment, as well.

  env <- new.env(parent=.GlobalEnv)

  pr <- pr(envir=env)
  pr$filter("ff", expression(function(req){ req$filterEnv <- parent.frame(); forward() }))
  pr$handle("GET", "/", expression(function(req){
    expect_identical(req$filterEnv, parent.frame())
    parent.frame()
  }))

  # Send a request through and we should see an assign to our env.
  val <- pr$route(make_req("GET", "/"), PlumberResponse$new())
  expect_identical(env, val)
})


test_that("host is updated properly for printing", {

  expect_identical(
    urlHost(host = "1:1:1", port = 1234),
    "http://[1:1:1]:1234"
  )
  expect_identical(
    urlHost(host = "::", port = 1234, changeHostLocation = FALSE),
    "http://[::]:1234"
  )
  expect_identical(
    urlHost(host = "::", port = 1234, changeHostLocation = TRUE),
    "http://[::1]:1234"
  )
  expect_identical(
    urlHost(host = "1.2.3.4", port = 1234),
    "http://1.2.3.4:1234"
  )
  expect_identical(
    urlHost(host = "0.0.0.0", port = 1234, changeHostLocation = FALSE),
    "http://0.0.0.0:1234"
  )
  expect_identical(
    urlHost(host = "0.0.0.0", port = 1234, changeHostLocation = TRUE),
    "http://127.0.0.1:1234"
  )
  expect_identical(
    urlHost(scheme = "http", host = "0.0.0.0", port = 1234, path = "/v1", changeHostLocation = TRUE),
    "http://127.0.0.1:1234/v1"
  )
})

test_that("unmount works", {
  pr <- pr()
  sub <- pr()
  sub$handle("GET", "/", function(){ 1 })
  sub$handle("GET", "/nested/path", function(){ 2 })
  pr$mount("/mount", sub)
  pr$mount("/mount2", sub)
  expect_equal(names(pr$mounts), c("/mount/", "/mount2/"))
  expect_invisible(pr$unmount("/henry"))
  expect_invisible(pr$unmount("/mount2/"))
  expect_equal(names(pr$mounts), "/mount/")
})

test_that("removeHandle works", {
  pr <- pr()
  pr$handle("GET", "/path1", function(){ 1 })
  pr$handle("GET", "/path2", function(){ 2 })
  expect_equal(length(pr$endpoints[[1]]), 2L)
  expect_invisible(pr$removeHandle("GET", "/path1"))
  expect_equal(length(pr$endpoints[[1]]), 1L)
  expect_equal(pr$endpoints[[1]][[1]]$path, "/path2")
})


test_that("routes that don't start with a slash are prepended with a slash", {
  pr <- pr()
  pr$handle("GET", "nested/path/here", function(){})

  expect_equal(length(pr$endpoints[[1]]), 1L)
  expect_equal(pr$endpoints[[1]][[1]]$path, "/nested/path/here")
})

test_that("handle method rejects forbidden arguments", {
  pr <- pr()
  expect_error(pr$handle("GET", "nested/path/here", function(){}, envir = new.env()),
               "can not be supplied to", )
  expect_error(pr$handle("GET", "nested/path/here", function(){}, verbs = "GET"),
               "can not be supplied to")
  expect_error(pr$handle("GET", "nested/path/here", function(){}, expr = function(){}),
               "can not be supplied to")
})

context("Routing")

test_that("Routing to errors and 404s works", {
  r <- pr(test_path("files/router.R"))
  errors <- 0
  notFounds <- 0

  errRes <- list(a=1)
  notFoundRes <- list(b=2)

  r$setErrorHandler(function(req, res, err){ errors <<- errors + 1; errRes })
  r$set404Handler(function(req, res){ notFounds <<- notFounds + 1; notFoundRes })

  res <- PlumberResponse$new()

  expect_equal(r$route(make_req("GET", "/"), res), "first")
  expect_equal(r$route(make_req("GET", "/abc"), res), "abc get")
  expect_equal(r$route(make_req("GET", "/dog"), res), "dog get")
  expect_equal(r$route(make_req("POST", "/dog"), res), "dog use")
  expect_equal(r$route(make_req("GET", "/path1"), res), "dual path")
  expect_equal(r$route(make_req("GET", "/path2"), res), "dual path")

  expect_equal(errors, 0)
  expect_equal(notFounds, 0)

  nf <- r$route(make_req("GET", "/something-crazy"), res)
  expect_equal(res$serializer, serializer_json())
  expect_equal(nf, notFoundRes)
  expect_equal(notFounds, 1)

  er <- r$route(make_req("GET", "/error"), res)
  expect_equal(res$serializer, serializer_json())
  expect_equal(er, errRes)
  expect_equal(errors, 1)
})

context("Warnings")

test_that("Routing to errors and 404s works", {
  r <- pr(test_path("files/warn.R"))

  res <- plumber:::PlumberResponse$new("json")

  expect_equal(options("warn")[[1]], 0)
  expect_warning(r$route(make_req("GET", "/warning"), res), "this is a warning")
  expect_equal(res$status, 1)
  expect_equal(options("warn")[[1]], 0)
})

context("Plumber Environment")

test_that(
  "Variables are populated in local environment, even if created after functions that use them",
  {
    r <- pr(test_path("files/source_values.R"))

    errors <- 0
    notFounds <- 0
    errRes <- list(a=1)
    notFoundRes <- list(b=2)
    r$setErrorHandler(function(req, res, err){ errors <<- errors + 1; errRes })
    r$set404Handler(function(req, res){ notFounds <<- notFounds + 1; notFoundRes })

    res <- PlumberResponse$new()

    expected <- "value_a"
    expect_equal(r$route(make_req("GET", "/count"), res), 1)
    expect_equal(r$route(make_req("GET", "/static_count"), res), 1)

    expect_equal(errors, 0)
    expect_equal(notFounds, 0)
  }
)

context("Endpoint around exec hook")

test_that("aroundexec function wrap like onions", {

  r <- PlumberEndpoint$new(
    'verb', '/path',
    expr = function(n = 100) {
      cat("exec\n")
      n
    },
    envir = new.env(parent = globalenv())
  )

  i <- 0
  expected_sum <- 100
  register_hook <- function(stage) {
    i <<- i + 1
    ii <- force(i)
    switch(stage,
      "preexec" = {
        r$registerHook(stage, function(req, res) {
          cat("preexec ", ii, "\n", sep = "")
          NULL
        })
      },
      "postexec" = {
        expected_sum <<- expected_sum + ii
        r$registerHook(stage, function(value, req, res, data) {
          cat("postexec ", ii, "\n", sep = "")
          value + ii
        })
      },
      "aroundexec" = {
        expected_sum <<- expected_sum + ii
        r$registerHook(stage, function(..., .next) {
          cat("aroundexec pre ", ii, "\n", sep = "")
          value <- .next(...)
          cat("aroundexec post ", ii, "\n", sep = "")
          value + ii
        })
      },
      stop("not a valid hook stage!")
    )
  }
  register_hook("preexec")
  register_hook("aroundexec")
  register_hook("postexec")

  register_hook("preexec")
  register_hook("postexec")
  register_hook("aroundexec")

  register_hook("aroundexec")
  register_hook("preexec")
  register_hook("postexec")

  register_hook("aroundexec")
  register_hook("postexec")
  register_hook("preexec")

  register_hook("postexec")
  register_hook("aroundexec")
  register_hook("preexec")

  register_hook("postexec")
  register_hook("preexec")
  register_hook("aroundexec")


  expect_output(
    result <- r$exec(req = list(), res = -1),
    paste(
      sep = "\n",
      "preexec 1",
      "preexec 4",
      "preexec 8",
      "preexec 12",
      "preexec 15",
      "preexec 17",
      "aroundexec pre 18",
      "aroundexec pre 14",
      "aroundexec pre 10",
      "aroundexec pre 7",
      "aroundexec pre 6",
      "aroundexec pre 2",
      "exec",
      "aroundexec post 2",
      "aroundexec post 6",
      "aroundexec post 7",
      "aroundexec post 10",
      "aroundexec post 14",
      "aroundexec post 18",
      "postexec 3",
      "postexec 5",
      "postexec 9",
      "postexec 11",
      "postexec 13",
      "postexec 16"
    )

  )
  expect_equal(result, expected_sum)
})





test_that("serializers can register all pre,post,aroundexec stages", {

  root <- with_tmp_serializers({
    plumb(test_path("files/endpoint-serializer.R"))
  })

  expect_output(
    result <- root$call(make_req("GET", "/test_serializer")),
    paste(
      "preexec",
      "around pre",
      "exec",
      "around post",
      "postexec",
      sep = "\n"
    )
  )
  expect_equal(result$body, "4")
})


test_that("not producing an image produces an error", {
  skip_on_os("windows")

  root <- with_tmp_serializers({
    plumb(test_path("files/endpoint-serializer.R")) %>% pr_set_debug(TRUE)
  })

  expect_output(
    result <- root$call(make_req("GET", "/no_plot", pr = root)),
    "device output file is missing"
  )
  expect_equal(result$status, 500)
  if (!is.list(result$body)) {
    result$body <- jsonlite::parse_json(result$body, simplifyVector = TRUE)
  }
  expect_true(grepl("device output file is missing", result$body$message))
})

context("legacy")

test_that("postBody is still available", {
  r <- pr(test_path("files/legacy.R"))
  res <- PlumberResponse$new()

  test_body <- 'testing body\nline 2'

  expect_identical(
    r$route(make_req("POST", "/postBody", body = test_body), res),
    test_body
  )
})

test_that("postBody is not enabled by default", {

  # Windows does not like the null pointer / non-utf8 body
  skip_on_os("windows") # Flavor: r-oldrel-windows-ix86+x86_64
  skip_on_cran() # Flavor: r-devel-linux-x86_64-debian-clang

  bin_file <- test_path("files/multipart-form.bin")
  bin_body <- readBin(bin_file, "raw", file.info(bin_file)$size)

  req <- make_req(
    "POST", "/postBody",
    body = bin_body,
    HTTP_CONTENT_TYPE = "multipart/form-data; boundary=----WebKitFormBoundaryMYdShB9nBc32BUhQ"
  )

  r <- pr(test_path("files/legacy.R"))
  # when parsing a multipart file, null pointers don't behave well in strings
  expect_output({
    r$route(req, PlumberResponse$new())
  }, "embedded nul in string")

  req <- make_req(
    "POST", "/body",
    body = bin_body,
    HTTP_CONTENT_TYPE = "multipart/form-data; boundary=----WebKitFormBoundaryMYdShB9nBc32BUhQ"
  )
  # when parsing a multipart file, null pointers don't behave well in strings
  # So if `req$postBody` is not touched, there is no error.
  expect_silent({
    r$route(req, PlumberResponse$new())
  })

})

context("Includes")

test_that("Includes work", {
  r <- pr(test_path("files/includes.R"))

  # When running, we setwd to the file's dir. Simulate that here.
  cwd <- getwd()
  on.exit( { setwd(cwd) } )
  setwd(test_path("files"))

  res <- PlumberResponse$new()
  val <- r$route(make_req("GET", "/"), res)
  expect_equal(val$body, "test.txt content")
  expect_equal(val$headers$`Content-Type`, "text/plain")

  res <- PlumberResponse$new()
  val <- r$route(make_req("GET", "/html"), res)
  expect_match(val$body, ".*<html.*</html>\\s*$")
  expect_equal(val$headers$`Content-Type`, "text/html; charset=UTF-8")

  # Skip these tests on some CRAN instances
  if (rmarkdown::pandoc_available()){
    res <- PlumberResponse$new()
    val <- r$route(make_req("GET", "/md"), res)
    expect_match(val$body, "<html.*<h2>R Output</h2>.*</html>\\s*$")
    expect_equal(val$headers$`Content-Type`, "text/html; charset=UTF-8")

    res <- PlumberResponse$new()
    val <- r$route(make_req("GET", "/rmd"), res)
    expect_match(val$body, "<html.*<img src=\"data:image/png;base64.*</html>\\s*$")
    expect_equal(val$headers$`Content-Type`, "text/html; charset=UTF-8")
  }
})

context("Verb enumeration")

test_that("enumerate returns all on 'use'", {
  expect_equal(enumerateVerbs("use"), c("GET", "PUT", "POST", "DELETE", "HEAD", "OPTIONS", "PATCH"))
})

test_that("regular verbs return themselves", {
  expect_equal(enumerateVerbs("get"), "GET")
  expect_equal(enumerateVerbs("post"), "POST")
})

context("multibytes source file")

with_locale <- function(locale, locale_val, expr) {
  prev_locale_val <- Sys.getlocale(locale)
  Sys.setlocale(locale, locale_val)
  on.exit({
    Sys.setlocale(locale, prev_locale_val)
  })
  force(expr)
}

expect_utf8 <- function(x) {
  expect_equal(Encoding(x), "UTF-8")
}

test_that("support files with multibytes", {

  # LC_ALL was not used as Mac had warnings when trying to restore LC_ALL
  # ?Sys.setlocale
  # `R sets ‘"LC_CTYPE"’ and ‘"LC_COLLATE"’,
  #  which allow the use of a different character set and alphabetic
  #  comparisons in that character set`
  with_locale("LC_CTYPE", "C", {
    with_locale("LC_COLLATE", "C", {

      r <- pr(test_path("files/multibytes.R"))
      req <- make_req("GET", "/echo")
      out <- r$call(req)$body

      # ?Quotes
      # `Unicode escapes can be used to enter Unicode characters not in the current locale's
      # charset (when the string will be stored internally in UTF-8).`
      expected <- jsonlite::toJSON("\u4e2d\u6587\u6d88\u606f")

      expect_identical(Sys.getlocale("LC_CTYPE"), "C")
      expect_identical(Sys.getlocale("LC_COLLATE"), "C")
      expect_utf8(expected)
      expect_utf8(out)
      # using charToRaw as identical coerces strings to different locales
      # ?identical
      # `Character strings are regarded as identical if they are in
      # different marked encodings but would agree when translated to
      # UTF-8.`
      expect_identical(charToRaw(out), charToRaw(expected))

    })
  })
})

context("filters")

test_that("Filters work", {
  r <- pr(test_path("files/filters.R"))
  expect_equal(length(r$filters), 4+2) #4 for post, query string, cookie, and shared secret filters

  expect_equal(r$filters[[5]]$name, "something")
  expect_equal(r$filters[[6]]$name, "nospace")
})

test_that("Filters can update req$args", {
  r <- pr(test_path("files/filters.R"))
  expect_equal(r$call(make_req("GET", "/"))$body, jsonlite::toJSON(23))
})

test_that("Redundant filters fail", {
  expect_error(pr(test_path("files/filter-redundant.R")), regexp="Multiple @filters")
})

test_that("Empty filters fail", {
  expect_error(pr(test_path("files/filter-empty.R")), regexp="No @filter name specified")
})

test_that("Filter and path fails", {
  expect_error(pr(test_path("files/filterpath.R")), regexp="can only be")
})

test_that("Filter and assets fails", {
  expect_error(pr(test_path("files/filterasset.R")), regexp="can only be")
})

test_that("Terminal filters indeed terminate", {
  r <- pr(test_path("files/terminal-filter.R"))
  expect_equal(r$call(make_req("GET", "/"))$body, jsonlite::toJSON(1))
})

test_that("complete addFilter works", {
  r <- pr()

  serializer <- "ser"

  name <- "fullFilter"
  expr <- expression(function(req, res){res$setHeader("expr", TRUE)})

  baseFilters <- length(r$filters)
  r$filter(name, expr, serializer)
  expect_equal(length(r$filters), baseFilters+1)

  fil <- r$filters[[baseFilters+1]]
  expect_equal(fil$name, "fullFilter")
  expect_equal(fil$lines, NA)
  expect_equal(fil$serializer, serializer)

  res <- PlumberResponse$new()
  req <- list()
  fil$exec(req=req, res=res)

  h <- res$headers
  expect_true(h$expr)
})

# No processors or serializer
test_that("sparse addFilter works", {
  r <- pr()

  name <- "sparseFilter"
  expr <- expression(function(req, res){res$setHeader("expr", TRUE)})

  baseFilters <- length(r$filters)
  r$filter(name, expr)
  expect_equal(length(r$filters), baseFilters+1)

  fil <- r$filters[[baseFilters+1]]
  expect_equal(fil$name, "sparseFilter")
  expect_equal(fil$lines, NA)

  res <- PlumberResponse$new()
  req <- list()
  fil$exec(req=req, res=res)

  h <- res$headers
  expect_true(h$expr)
})

test_that("sparse addFilter with a function works", {
  r <- pr()

  name <- "sparseFilter"
  expr <- function(req, res){res$setHeader("expr", TRUE)}

  baseFilters <- length(r$filters)
  r$filter(name, expr)
  expect_equal(length(r$filters), baseFilters+1)

  fil <- r$filters[[baseFilters+1]]
  expect_equal(fil$name, "sparseFilter")
  expect_equal(fil$lines, NA)

  res <- PlumberResponse$new()
  req <- list()
  fil$exec(req=req, res=res)

  h <- res$headers
  expect_true(h$expr)
})

context("Static")

# Windows-note. Convert all text to char and trim to avoid differences involving `\r\n` and `\n`

pr <- PlumberStatic$new(test_path("files/static"))

test_that("the response is returned", {
  res <- PlumberResponse$new()
  val <- pr$route(make_req("GET", "/test.txt"), res)
  expect_true(inherits(val, "PlumberResponse"))
})

test_that("static txt file is served", {
  res <- PlumberResponse$new()
  pr$route(make_req("GET", "/test.txt"), res)
  expect_equal(res$headers$`Content-Type`, "text/plain")
  expect_equal(trimws(rawToChar(res$body)), "I am a text file.")
})

test_that("static txt file with encoded URI is served", {

  # Some file systems cannot handle these characters.
  testthat::skip_on_cran()

  res <- PlumberResponse$new()
  f <- test_path("files/static/测试.txt")
  file.create(f)
  on.exit(unlink(f), add = TRUE)
  writeChar("here be dragons", f)
  pr$route(make_req("GET", "/测试.txt"), res)
  unlink(test_path("files/static/测试.txt"))
  expect_equal(res$headers$`Content-Type`, "text/plain")
  expect_equal(trimws(rawToChar(res$body)), "here be dragons")
})

test_that("static html file is served", {
  res <- PlumberResponse$new()
  pr$route(make_req("GET", "/index.html"), res)
  expect_equal(res$headers$`Content-Type`, "text/html; charset=UTF-8")
  expect_equal(trimws(rawToChar(res$body)), "<html>I am HTML</html>")
})

test_that("root requests are routed to index.html", {
  res <- PlumberResponse$new()
  pr$route(make_req("GET", "/"), res)
  expect_equal(res$headers$`Content-Type`, "text/html; charset=UTF-8")
  expect_equal(trimws(rawToChar(res$body)), "<html>I am HTML</html>")
})

test_that("HEAD requests are served correctly", {
  for (path in c("/", "/index.html", "/test.txt", "/test.txt.zip")) {
    resg <- PlumberResponse$new()
    resh <- PlumberResponse$new()
    pr$route(make_req("GET", path), resg)
    pr$route(make_req("HEAD", path), resh)
    expect_equal(resg$headers, resh$headers)
    expect_type(resg$headers$`Last-Modified`, "character")
  }
})

test_that("static binary file is served", {
  res <- PlumberResponse$new()
  pr$route(make_req("GET", "/test.txt.zip"), res)
  expect_equal(res$headers$`Content-Type`, "application/zip")
  bod <- res$body
  zipf <- file(test_path("files/static/test.txt.zip"), "rb")
  bin <- readBin(zipf, "raw", n=1000)
  close(zipf)
  expect_equal(bin, bod)
})

test_that("404s are handled", {
  res <- PlumberResponse$new()
  pr$route(make_req("GET", "/i-dont-exist"), res)
  expect_equal(res$status, 404)
})

test_that("PUTs error", {
  res <- PlumberResponse$new()
  pr$route(make_req("PUT", "/"), res)
  expect_equal(res$status, 400)
})

test_that("files are parsed properly", {
  p <- pr(test_path("files/static.R"))
  expect_length(p$mounts, 2)

  res <- PlumberResponse$new()
  req <- make_req("GET", "/static/test.txt")
  p$route(req=req, res=res)
  expect_equal(nchar(trimws(rawToChar(res$body))), 17)
  expect_equal(res$status, 200)
  expect_equal(res$headers$`Content-Type`, "text/plain")

  res <- PlumberResponse$new()
  req <- make_req("GET", "/public/test.txt")
  p$route(req=req, res=res)
  expect_equal(nchar(trimws(rawToChar(res$body))), 17)
  expect_equal(res$status, 200)
  expect_equal(res$headers$`Content-Type`, "text/plain")
})

test_that("no directory throws error", {
  expect_error(pr(test_path("files/static-nodir.R")), "No directory specified")
})

test_that("expressions work as options", {
  pr <- pr()
  stat <- PlumberStatic$new(test_path("files/static"), {list()})
  pr$mount("/public", stat)

  res <- PlumberResponse$new()
  pr$route(make_req("GET", "/public/test.txt"), res)
  expect_equal(res$headers$`Content-Type`, "text/plain")
  expect_equal(trimws(rawToChar(res$body)), "I am a text file.")
})

context("device serializer")

test_that("graphics device promise domains must have a device", {
  expect_error(
    # send in a value that is returned for a null device
    createGraphicsDevicePromiseDomain(which = c(`null device` = 1L)),
    "was called without opening a device"
  )
})

test_that("you should not call `dev_set()` with a null device", {

  expect_warning(
    # send in a value that is returned for a null device
    dev_set(c(`null device` = 1L)),
    "null device"
  )
})






# Only test on CI
skip_on_cran()

expect_device_output <- function(name, content_type, capability_type = name) {

  if (!is.null(capability_type)) {
    if (!capabilities(capability_type)) {
      testthat::skip(paste0("Graphics device type not supported: ", name))
    }
  }

  ep <- NULL
  pr <- pr()

  evaluateBlock(
    srcref = 3, # which evaluates to line 2
    file = c("#' @get /test", paste0("#' @serializer ", name, " list()")),
    expr = function() {
      plot(1:10)
    },
    envir = new.env(parent=globalenv()),
    addEndpoint = function(a, b, ...) { pr$handle(endpoint = a) },
    addFilter = as.null,
    pr = pr
  )

  ret <- pr$call(make_req("GET", "/test"))

  expect_equal(ret$status, 200)
  expect_equal(ret$headers$`Content-Type`, content_type)
  expect_true(is.raw(ret$body))
  expect_gt(length(ret$body), 1000)
}



test_that("jpeg produces an image", {
  expect_device_output("jpeg", "image/jpeg")
})
test_that("png produces an image", {
  expect_device_output("png", "image/png")
})
test_that("svg produces an image", {
  expect_device_output("svg", "image/svg+xml", "cairo")
})
test_that("bmp produces an image", {
  expect_device_output("bmp", "image/bmp", "cairo")
})
test_that("tiff produces an image", {
  expect_device_output("tiff", "image/tiff")
})
test_that("pdf produces an image", {
  expect_device_output("pdf", "application/pdf", NULL)
})




context("plumb() device serializer")

test_device <- local({
    r <- pr(test_path("files/device.R"))

  function(name, content_type, capability_type = name, test_little = TRUE) {
    if (!capabilities(capability_type)) {
      testthat::skip("Graphics type not supported: ", name)
    }

    resp <- r$serve(make_req("GET", paste0("/", name)), PlumberResponse$new())
    expect_equal(resp$status, 200)
    expect_equal(resp$headers$`Content-Type`, content_type)
    fullsize <- length(resp$body)
    expect_gt(fullsize, 1000) # This changes based on R ver/OS, may not be useful.

    if (!isTRUE(test_little)) {
      # do not test the smaller device route
      return()
    }
    resp <- r$serve(make_req("GET", paste0("/little", name)), PlumberResponse$new())
    expect_equal(resp$status, 200)
    expect_equal(resp$headers$`Content-Type`, content_type)
    expect_gt(length(resp$body), 100) # This changes based on R ver/OS, may not be useful.
    expect_lt(length(resp$body), fullsize) # Should be smaller than the full one
  }
})

test_that("png are properly rendered", {
  test_device("png", "image/png")
})
test_that("jpeg are properly rendered", {
  test_device("jpeg", "image/jpeg")
})
test_that("svg are properly rendered", {
  test_device("svg", "image/svg+xml", capability_type = "cairo", test_little = FALSE)
})

context("Injection")

test_that("Injected arguments on req$args get passed on.", {
  r <- pr(test_path("files/filter-inject.R"))

  expect_equal(r$call(make_req("GET", "/"))$body, jsonlite::toJSON(13))
})

context("shared secret")

test_that("requests with shared secrets pass, w/o fail", {
  options(`plumber.sharedSecret`="abcdefg")

  pr <- pr()
  pr$handle("GET", "/", function(){ 123 })
  req <- make_req("GET", "/", pr = pr)

  # No shared secret
  res <- PlumberResponse$new()
  output <- pr$route(req, res)
  expect_equal(res$status, 400)
  expect_equal(output, list(error = "400 - Bad request"))

  # When debugging, we get additional details in the error.
  pr$setDebug(TRUE)
  res <- PlumberResponse$new()
  output <- pr$route(req, res)
  expect_equal(res$status, 400)
  expect_equal(output, list(
    error = "400 - Bad request",
    message = "Shared secret mismatch"))
  pr$setDebug(FALSE)

  # Set shared secret
  assign("HTTP_PLUMBER_SHARED_SECRET", "abcdefg", envir=req)
  res <- PlumberResponse$new()
  pr$route(req, res)
  expect_equal(res$status, 200)

  options(`plumber.sharedSecret`=NULL)
})

test_that("prints correctly", {
  testthat::skip_on_cran()
  testthat::skip_on_os("windows") # has issues comparing text values


  pr1 <- pr()
  pr1$handle("GET", "/nested/path/here", function(){})
  pr1$handle("POST", "/nested/path/here", function(){})

  pr2 <- pr()
  pr2$handle("POST", "/something", function(){})
  pr2$handle("GET", "/", function(){})
  pr1$mount("/mysubpath", pr2)

  stat <- PlumberStatic$new(".")
  pr1$mount("/static", stat)

  printed <- capture.output(print(pr1))

  expected_output <- c(
    "# Plumber router with 2 endpoints, 4 filters, and 2 sub-routers.",
    "# Use `pr_run()` on this object to start the API.",
    "├──[queryString]",
    "├──[body]",
    "├──[cookieParser]",
    "├──[sharedSecret]",
    "├──/mysubpath",
    "│  │ # Plumber router with 2 endpoints, 4 filters, and 0 sub-routers.",
    "│  ├──[queryString]",
    "│  ├──[body]",
    "│  ├──[cookieParser]",
    "│  ├──[sharedSecret]",
    "│  ├──/ (GET)",
    "│  └──/something (POST)",
    "├──/nested",
    "│  ├──/path",
    "│  │  └──/here (GET, POST)",
    "├──/static",
    "│  │ # Plumber static router serving from directory: ."
  )

  expect_equal(printed, expected_output)

  expected_output2 <- c(
    "# Plumber router with 1 endpoint, 4 filters, and 0 sub-routers.",
    "# Use `pr_run()` on this object to start the API.",
    "├──[queryString]",
    "├──[body]",
    "├──[cookieParser]",
    "├──[sharedSecret]",
    "├──/A",
    "│  ├──/B",
    "│  │  └──/ (GET)"
  )
  printed2 <- capture.output(print(pr_get(pr(), "/A/B/", identity)))
  expect_equal(printed2, expected_output2)
})

test_that("prints correctly", {
  testthat::skip_on_cran()
  testthat::skip_on_os("windows") # has issues comparing text values

  pr1 <- pr()
  sub <- pr()
  sub$handle("GET", "/", force)
  sub$handle("POST", "/something", force)
  sub$handle("GET", "/nested/path", force)
  sub$handle("POST", "/", force)
  sub$handle("POST", "/nested/path", force)

  pr1$mount("/", sub)

  printed <- capture.output(print(pr1))

  expected_output <- c(
    "# Plumber router with 0 endpoints, 4 filters, and 1 sub-router.",
    "# Use `pr_run()` on this object to start the API.",
    "├──[queryString]",
    "├──[body]",
    "├──[cookieParser]",
    "├──[sharedSecret]",
    "├──/",
    "│  │ # Plumber router with 5 endpoints, 4 filters, and 0 sub-routers.",
    "│  ├──[queryString]",
    "│  ├──[body]",
    "│  ├──[cookieParser]",
    "│  ├──[sharedSecret]",
    "│  ├──/ (GET, POST)",
    "│  ├──/nested",
    "│  │  └──/path (GET, POST)",
    "│  └──/something (POST)"
  )

  expect_equal(printed, expected_output)
})


test_that("prints recursion is detected", {
  a <- pr()
  b <- pr()

  a$mount("B", b)
  b$mount("A", a)

  printed <- capture.output(print(a))
  printed2 <- capture.output(print(a))
  printed3 <- capture.output(print(b))

  expect_match(printed, "Circular Plumber router definition detected", all = FALSE)
  expect_match(printed3, "Circular Plumber router definition detected", all = FALSE)
  expect_equal(printed, printed2)

})

#* @get /car/<id>
function(id){
  id
}

#* @get /car/int/<id:int>
function(id){
  id
}

#* @get /car/<id>/sell/<price>
function(id, price){
  list(id=id, price=price)
}

#* @get /car/ratio/<ratio:double>
function(ratio){
  ratio
}

#* @get /car/sold/<sold:bool>
function(sold){
  sold
}

#* @post /car/<id>
function(req){
  req$args$id
}

#* @get /car/<id:int>/buy/<price>
function(id, price){
  list(id=id, price=price)
}

# FROM Shiny
# @author Shiny package authors
knownContentTypes <- c(
  html = "text/html; charset=UTF-8",
  htm = "text/html; charset=UTF-8",
  js = "text/javascript",
  css = "text/css",
  png = "image/png",
  jpg = "image/jpeg",
  jpeg = "image/jpeg",
  gif = "image/gif",
  svg = "image/svg+xml",
  txt = "text/plain",
  pdf = "application/pdf",
  ps = "application/postscript",
  xml = "application/xml",
  m3u = "audio/x-mpegurl",
  m4a = "audio/mp4a-latm",
  m4b = "audio/mp4a-latm",
  m4p = "audio/mp4a-latm",
  mp3 = "audio/mpeg",
  wav = "audio/x-wav",
  m4u = "video/vnd.mpegurl",
  m4v = "video/x-m4v",
  mp4 = "video/mp4",
  mpeg = "video/mpeg",
  mpg = "video/mpeg",
  avi = "video/x-msvideo",
  mov = "video/quicktime",
  ogg = "application/ogg",
  swf = "application/x-shockwave-flash",
  doc = "application/msword",
  xls = "application/vnd.ms-excel",
  ppt = "application/vnd.ms-powerpoint",
  xlsx = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
  xltx = "application/vnd.openxmlformats-officedocument.spreadsheetml.template",
  potx = "application/vnd.openxmlformats-officedocument.presentationml.template",
  ppsx = "application/vnd.openxmlformats-officedocument.presentationml.slideshow",
  pptx = "application/vnd.openxmlformats-officedocument.presentationml.presentation",
  sldx = "application/vnd.openxmlformats-officedocument.presentationml.slide",
  docx = "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
  dotx = "application/vnd.openxmlformats-officedocument.wordprocessingml.template",
  xlam = "application/vnd.ms-excel.addin.macroEnabled.12",
  xlsb = "application/vnd.ms-excel.sheet.binary.macroEnabled.12",
  feather = "application/vnd.apache.arrow.file",
  parquet = "application/vnd.apache.parquet",
  rds = "application/rds",
  tsv = "application/tab-separated-values",
  csv = "application/csv",
  json = "application/json",
  yml = "application/yaml",
  yaml = "application/yaml"
)

getContentType <- function(ext, defaultType = 'application/octet-stream') {
  ext <- tolower(ext)

  ret <-
    knownContentTypes[ext] %|%
    mime::mimemap[ext] %|%
    defaultType

  ret[[1]]
}

cleanup_content_type <- function(type) {
  if (length(type) == 0) return(type)

  type <- tolower(type)

  # remove trailing content type information
  # "text/yaml; charset=UTF-8"
  # to
  # "text/yaml"
  if (stri_detect_fixed(type, ";")) {
    type <- stri_split_fixed(type, ";")[[1]][1]
  }

  type
}

get_fileext <- function(type) {
  type <- cleanup_content_type(type)

  all_content_types <- c(knownContentTypes, mime::mimemap)

  type_to_ext <- setNames(names(all_content_types), all_content_types)

  ret <- type_to_ext[type] %|% NULL
  ret[[1]]
}

#' Request character set
#' @param content_type Request Content-Type header
#' @return Default to `UTF-8`. Otherwise return `charset` defined in request header.
#' @export
get_character_set <- function(content_type = NULL) {
  if (is.null(content_type)) return("UTF-8")
  stri_match_first_regex(paste(content_type,"; charset=UTF-8"), "charset=([^;\\s]*)")[,2]
}

context("Content Types")

test_that("contentType serializes properly", {
  l <- list(a=1, b=2, c="hi")
  val <- serializer_content_type("somethinghere")(l, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "somethinghere")
  expect_equal(val$body, l)
})

test_that("empty contentType errors", {
  expect_error(serializer_content_type())
})

test_that("contentType works in files", {

  r <- pr(test_path("files/content-type.R"))
  val <- r$call(make_req("GET", "/"))
  expect_equal(val$headers$`Content-Type`, "text/plain")
})

test_that("Parses charset properly", {
  charset <- get_character_set("Content-Type: text/html; charset=latin1")
  expect_equal(charset, "latin1")
  charset <- get_character_set("Content-Type: text/html; charset=greek8")
  expect_equal(charset, "greek8")
})

test_that("Defaults charset when not there", {
  charset <- get_character_set("Content-Type: text/html")
  expect_equal(charset, "UTF-8")
  charset <- get_character_set(NULL)
  expect_equal(charset, "UTF-8")
})


test_that("File extensions can be found from the content type", {
  expect_equal(get_fileext("application/json"), "json")
  expect_equal(get_fileext("not a match"), NULL)
})

# Ports that are considered unsafe by Chrome
# http://superuser.com/questions/188058/which-ports-are-considered-unsafe-on-chrome
# https://github.com/rstudio/shiny/issues/1784
unsafePorts <- function() {
  asNamespace("httpuv")[["unsafe_ports"]]
}


randomPort <- function(..., n = 10) {
  tryCatch({
    port <- httpuv::randomPort(..., n = n)
  }, httpuv_unavailable_port = function(e) {
    stop(
      "Unable to start a Plumber server. ",
      paste0("We were unable to find a free port in ", n, " tries.")
    )
  })
  return(port)
}

portIsAvailable <- function(port) {
  tryCatch(
    {
      randomPort(min = port, max = port)
      TRUE
    },
    error = function(e) {
      FALSE
    }
  )
}

#' Find a port either using the assigned port or randomly search 10 times for an
#' available port. If a port was manually assigned, just return it and assume it
#' will work.
#' @noRd
findPort <- function(port = NULL) {
  if (is.null(port)) {
    # Try to use the most recently used _random_ port
    if (
      (!is.null(.globals$last_random_port)) &&
        portIsAvailable(.globals$last_random_port)
    ) {
      return(.globals$last_random_port)
    }

    # Find an available port
    port <- randomPort()

    # Save the random port for future use
    .globals$last_random_port <- port
    return(.globals$last_random_port)
  }

  port_og <- port

  if (rlang::is_character(port)) {
    port <- suppressWarnings(as.integer(port))
  }

  if (!rlang::is_integerish(port, n = 1, finite = TRUE) || port != port_og) {
    stop("Port must be an integer value, not '", port_og, "'.")
  }

  port <- as.integer(port)

  # Ports must be in [1024-49151]
  # https://www.iana.org/assignments/service-names-port-numbers/service-names-port-numbers.xhtml
  if (port < 1024 || port > 49151) {
    stop("Port must be an integer in the range 1024 to 49151 (inclusive).")
  }

  if (port == 0 || port %in% unsafePorts()) {
    stop("Port ", port, " is an unsafe port. Please choose another port.")
  }

  port
}

context("CSV serializer")

test_that("CSV serializes properly", {
  skip_if_not_installed("readr")

  d <- data.frame(a=1, b=2, c="hi")
  val <- serializer_csv()(d, data.frame(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "text/csv; charset=UTF-8")
  expect_equal(val$body, readr::format_csv(d))

  d <- data.frame(a=1, b=2, c="hi", na=NA)
  val <- serializer_csv()(d, data.frame(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "text/csv; charset=UTF-8")
  expect_equal(val$body, readr::format_csv(d, na = "NA"))

  d <- data.frame(a=1, b=2, c="hi", na=NA)
  val <- serializer_csv(na = 'string')(d, data.frame(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "text/csv; charset=UTF-8")
  expect_equal(val$body, readr::format_csv(d, na = 'string'))
})

test_that("Errors call error handler", {
  skip_if_not_installed("readr")

  errors <- 0
  errHandler <- function(req, res, err){
    errors <<- errors + 1
  }

  expect_equal(errors, 0)
  serializer_csv()(parse(text="hi"), data.frame(), PlumberResponse$new("csv"), errorHandler = errHandler)
  expect_equal(errors, 1)
})

test_that("GeoJSON serializes properly", {
  skip_if_not_installed("geojsonsf")
  skip_if_not_installed("sf")

  # Objects taken from ?st_sf() examples.
  sfc <- sf::st_sfc(sf::st_point(1:2), sf::st_point(3:4))
  sf <- sf::st_sf(a = 3:4, sfc)

  # Test sfc
  val <- serializer_geojson()(sfc, data.frame(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "application/geo+json")
  expect_equal(val$body, geojsonsf::sfc_geojson(sfc))

  # Test sf
  val <- serializer_geojson()(sf, data.frame(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "application/geo+json")
  expect_equal(val$body, geojsonsf::sf_geojson(sf))

})

test_that("Errors call error handler", {
  skip_if_not_installed("geojsonsf")
  skip_if_not_installed("sf")

  errors <- 0
  errHandler <- function(req, res, err){
    errors <<- errors + 1
  }

  expect_equal(errors, 0)
  serializer_geojson()(parse(text="h$534i} {!"), data.frame(), PlumberResponse$new(), errorHandler = errHandler)
  expect_equal(errors, 1)
})

context("find port")

test_that("ports can be randomly found", {
  foundPorts <- unlist(lapply(1:50, function(i) {
    p <- randomPort()
    # Use findPort to verify the port's value
    findPort(p)
  }))

  # It's possible we got a collision or two, but shouldn't have many.
  expect_gt(length(unique(foundPorts)), 45)
})

test_that("global port used if available", {
  .globals$last_random_port <- 12345
  expect_equal(findPort(), 12345)
  rm("last_random_port", envir = .globals)
})

test_that("integer type is returned", {
  expect_type(findPort(), "integer")
  expect_equal(findPort("8000"), 8000L)
  expect_equal(findPort(8000.00000), 8000L)
})

test_that("throws if provided non-integerish port", {
  expect_error(findPort("blue"))
  expect_error(findPort(8000.0001))
  expect_error(findPort(8000:8002))
})

test_that("throws for invalid ports", {
  expect_error(findPort(800)) # in 0-1024
  expect_error(findPort(123456)) # out of range
  expect_error(findPort(0)) # unsafe
  expect_error(findPort(6666)) # unsafe
})

test_that("finds a good port and persists it", {
  testthat::skip_on_cran()

  p <- findPort()

  # Persisted
  expect_equal(.globals$last_random_port, p)

  # Check that we can actually start a server
  srv <- httpuv::startServer("127.0.0.1", p, list())

  # Cleanup
  rm("last_random_port", envir = .globals)
  httpuv::stopServer(srv)
})

test_that("we don't pin to globals$port if it's occupied", {
  testthat::skip_on_cran()

  ex_p <- 12345
  srv <- httpuv::startServer("127.0.0.1", ex_p, list())
  .globals$last_random_port <- ex_p

  p <- findPort()

  # It should shuffle to find another port.
  expect_true(p != ex_p)

  rm("last_random_port", envir = .globals)
  httpuv::stopServer(srv)
})

context("excel serializer")

test_that("excel serializes properly", {
  skip_if_not_installed("writexl")

  d <- data.frame(a=1, b=2, c="hi")
  val <- serializer_excel()(d, data.frame(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet")

  # the remaining relies on the fact that xlsx files start as zip files
  # https://en.wikipedia.org/wiki/List_of_file_signatures
  expect_equal(val$body[1:4], as.raw(c(0x50, 0x4b, 0x03, 0x04)))
  tf <- tempfile()
  on.exit(unlink(tf), add = TRUE)
  writeBin(val$body, tf)
  zipcontents <- expect_silent(utils::unzip(tf, list = TRUE))
  expect_s3_class(zipcontents, "data.frame")
  expect_true("xl/workbook.xml" %in% zipcontents$Name)

})

test_that("Errors call error handler", {
  skip_if_not_installed("writexl")

  errors <- 0
  errHandler <- function(req, res, err){
    errors <<- errors + 1
  }

  expect_equal(errors, 0)
  serializer_excel()(parse(text="hi"), data.frame(), PlumberResponse$new("csv"), errorHandler = errHandler)
  expect_equal(errors, 1)
})

context("YAML serializer")

test_that("YAML serializes properly", {
  skip_if_not_installed("yaml")

  l <- list(a=1, b=2, c="hi")
  val <- serializer_yaml()(l, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "text/x-yaml; charset=UTF-8")
  expect_equal(val$body, yaml::as.yaml(l))

  l <- list(a=1, b=2, c="hi", na=NA)
  val <- serializer_yaml()(l, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "text/x-yaml; charset=UTF-8")
  expect_equal(val$body, yaml::as.yaml(l))

  l <- list(a=1, b=2, c="hi", na=NA)
  val <- serializer_yaml(indent = 4)(l, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "text/x-yaml; charset=UTF-8")
  expect_equal(val$body, yaml::as.yaml(l, indent = 4))
})

test_that("Errors call error handler", {
  skip_if_not_installed("yaml")

  errors <- 0
  errHandler <- function(req, res, err){
    errors <<- errors + 1
  }

  expect_equal(errors, 0)
  serializer_yaml()(parse(text="hi"), list(), PlumberResponse$new("yaml"), errorHandler = errHandler)
  expect_equal(errors, 1)
})

context("HTML serializer")

test_that("HTML serializes properly", {
  v <- "<html><h1>Hi!</h1></html>"
  val <- serializer_html()(v, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "text/html; charset=UTF-8")
  expect_equal(val$body, v)
})

test_that("Errors call error handler", {
  errors <- 0
  errHandler <- function(req, res, err){
    errors <<- errors + 1
  }

  expect_equal(errors, 0)
  serializer_html()(parse(stop("I crash")), list(), PlumberResponse$new("json"), errorHandler = errHandler)
  expect_equal(errors, 1)
})

context("htmlwidgets serializer")

# Render a simple HTML widget using the visNetwork package
renderWidget <- function(){
  skip_if_not_installed("visNetwork")

  nodes <- data.frame(id = 1:6, title = paste("node", 1:6),
                      shape = c("dot", "square"),
                      size = 10:15, color = c("blue", "red"))
  edges <- data.frame(from = 1:5, to = c(5, 4, 6, 3, 3))
  visNetwork::visNetwork(nodes, edges) %>%
    visNetwork::visOptions(highlightNearest = TRUE, nodesIdSelection = TRUE)

}

test_that("htmlwidgets serialize properly", {
  # Solaris doesn't have htmlwidgets available for some reason.
  skip_on_cran()
  # Too many moving parts on an inconsistent os
  skip_on_os("windows")

  w <- renderWidget()
  val <- serializer_htmlwidget()(w, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "text/html; charset=UTF-8")
  # Check that content is encoded
  expect_match(val$body, "url\\(['\"]?data:image\\/png;base64")
})

test_that("Errors call error handler", {
  errors <- 0
  errHandler <- function(req, res, err){
    errors <<- errors + 1
  }

  expect_equal(errors, 0)
  suppressWarnings(
    serializer_htmlwidget()(parse(text="hi"), list(), PlumberResponse$new("htmlwidget"), errorHandler = errHandler)
  )
  expect_equal(errors, 1)
})

#' HTTP Date String
#'
#' Given a POSIXct object, return a date string in the format required for a
#' HTTP Date header. For example: "Wed, 21 Oct 2015 07:28:00 GMT"
#'
#' @noRd
http_date_string <- function(time) {
  lt <- as.POSIXlt(time, tz = "GMT")
  weekdays <- c("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat")
  months <- c("Jan", "Feb", "Mar", "Apr", "May", "Jun",
              "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")
  weekday <- weekdays[lt$wday + 1]
  month <- months[lt$mon + 1]
  fmt <- paste0(weekday, ", %d ", month, " %Y %H:%M:%S GMT")
  strftime(time, fmt, tz = "GMT")
}

is_available <- function(package, version = NULL) {
  installed <- nzchar(system.file(package = package))
  if (is.null(version)) {
    return(installed)
  }
  installed && isTRUE(utils::packageVersion(package) >= version)
}

`%||%` <- function(x, y) {
  if (is.null(x)) {
    y
  } else {
    x
  }
}

`%|%` <- function(x, y) {
  if (length(x) > 1) {
    stopifnot(length(y) == 1)
    x[is.na(x)] <- y
    return(x)
  }

  if (is.na(x)) {
    y
  } else {
    x
  }
}

once <- function(f) {
  called <- FALSE

  function() {
    if (!called) {
      called <<- TRUE
      f()
      invisible(TRUE)
    } else {
      invisible(FALSE)
    }
  }
}

#' Validate OpenAPI Spec
#'
#' Validate an OpenAPI Spec using [`@redocly/cli`](https://redocly.com/docs/cli/commands/lint).
#'
#' If any warning or error is presented, an error will be thrown.
#'
#' This function is `r lifecycle::badge("experimental")` and may be altered, changed, or removed in the future.
#'
#' @param pr A Plumber API
#' @param ... Ignored
#' @param ruleset Character that determines the ruleset to use for validation. Can be one of "minimal", "recommended",
#' or "recommended-strict". Defaults to "minimal". See [`@redocly/cli`
#' options](https://redocly.com/docs/cli/commands/lint#options) for more details.
#' @param verbose Logical that determines if a "is valid" statement is displayed. Defaults to `TRUE`
#' @export
#' @examples
#' \dontrun{
#' pr <- plumb_api("plumber", "01-append")
#' validate_api_spec(pr)
#' }
validate_api_spec <- function(pr, ..., ruleset = c("minimal", "recommended", "recommended-strict"), verbose = TRUE) {

  ruleset <- match.arg(ruleset, several.ok = FALSE)
  if (!nzchar(Sys.which("node"))) {
    stop("`node` command not found. Please install Node.js")
  }
  if (!nzchar(Sys.which("npx"))) {
    stop("`npx` not installed. Please install Node.js w/ `npx` command available.")
  }
  
  spec <- jsonlite::toJSON(pr$getApiSpec(), auto_unbox = TRUE, pretty = TRUE)

  tmpfile <- tempfile(fileext = ".json")
  on.exit({
    unlink(tmpfile)
  }, add = TRUE)
  cat(spec, file = tmpfile)

  output <- system2(
    "npx",
    c(
      "--yes", # auto install `@redocly/cli`
      "-p", "@redocly/cli",
      "redocly",
      "lint",
      "--extends", ruleset,
      "--skip-rule", "no-empty-servers", # We don't know the end servers by default
      "--skip-rule", "security-defined", # We don't know the security by default
      "--skip-rule", "operation-summary", # operation summary values are optional. Not required
      "--skip-rule", "operation-operationId-url-safe", # By default, it wants to have all operationId values be URL safe. This does not work with path parameters that want to use `{``}`.
      "--skip-rule", "no-path-trailing-slash", # This is OK
      tmpfile
    ),
    stdout = TRUE,
    stderr = TRUE
  )
  
  output <- paste0(output, collapse = "\n")

  has_warn_or_error <- grepl("\n[1] ", output, fixed = TRUE)
  if (has_warn_or_error) {
    cat("Plumber Spec: \n", as.character(spec), "\n\nOutput:\n", output, sep = "")
    stop("Plumber OpenAPI Spec is not valid")
  }

  invisible(TRUE)
}

## plumber::plumb("using_plumber_tag.R")

library(plumber)

#* @get /counter
function(req){
  count <- 0
  if (!is.null(req$session$counter)){
    count <- as.numeric(req$session$counter)
  }
  req$session$counter <- count + 1
  return(paste0("This is visit #", count))
}


#' @plumber
function(pr) {
  pr %>%
    pr_cookie(
      key = "pleasechangeme",
      name = "cookieName"
    )
}

#* @stuff
#* @more- stuff
#*
#* comments
function(req, res, forward){
  4
}

#* @get /test
#* @post /test
#* @more stuff
#' @param req roxygen2 params
#* hey
foo <- function(a, b, ..., req, res, forward){
  5
}

#* Not an API
#* @param things
foo <- function(req, res, forward){
  6
}

#* @post /
#* one line
function(req, res, forward){x <- 9; x + 1}

#* Spaces in between
#* @delete /

function(req, res, forward){
  x <- 10
  x + 2
}

#*@get nowhitespace
function(req, res, forward){
  14
}

#' Parse the given argument and extend the given fields list with new data.
#' @param fields A list that contains at least an `info` sub-list.
#' @param argument The line (including the plumber comment prefix) to append.
#'   If this line represents what was once multiple lines, intermediate comment
#'   prefixes should have been removed.
#' @param envir An environment where to evaluate parsed expressions
#' @noRd
plumbOneGlobal <- function(fields, argument, envir = parent.frame()){
  if (nchar(argument) == 0){
    return(fields)
  }

  parsedLine <- regmatches(argument, regexec(
    argRegex, argument, ignore.case=TRUE))[[1]]

  if (length(parsedLine) != 4){
    return(fields)
  }

  name <- parsedLine[3]
  def <- parsedLine[4]
  def <- gsub("^\\s*|\\s*$", "", def)

  switch(name,
         apiTitle={
           fields$info$title <- def
         },
         apiDescription={
           fields$info$description <- def
         },
         apiTOS={
           fields$info$termsOfService <- def
         },
         apiContact={
           if (grepl("^list\\(", def)) {
             def <- eval(parse(text = def), envir)
           }
           fields$info$contact <- def
         },
         apiLicense={
           if (grepl("^list\\(", def)) {
             def <- eval(parse(text = def), envir)
           }
           fields$info$license <- def
         },
         apiVersion={
           fields$info$version <- def
         },
         apiTag={
           tagMat <- stri_match(def, regex="^\\s*(\"[^\"]+\"|'[^']+'|\\S+)\\s+(\\S.*)\\s*$")
           name <- stri_trim_both(tagMat[1,2], pattern = "[[\\P{Wspace}]-[\"']]")
           description <- tagMat[1,3]
           if(!is.null(fields$tags) && name %in% unlist(lapply(fields$tags, "[[", "name"))) {
             stop("Error: '", argument, "' - ","Duplicate tag definition specified.")
           }
           fields$tags <- c(fields$tags, list(list(name=name, description=description)))
         })
  fields
}

argRegex <- "^#['\\*]\\s*(@(api\\w+)\\s+)?(.*)$"

#' Parse out the global API settings of a given set of lines and return a
#' OpenAPI-compliant list describing the global API.
#' @noRd
plumbGlobals <- function(lines, envir = parent.frame()){
  # Build up the entire argument here; needed since a single directive
  # might wrap multiple lines
  fullArg <- ""

  # Build up the fields that we want to return as globals
  fields <- list(info=list())

  # Parse the global docs
  for (line in lines){
    parsedLine <- regmatches(line, regexec(
      argRegex, line, ignore.case=TRUE))[[1]]
    if (length(parsedLine) == 4){
      if (nchar(parsedLine[3]) == 0){
        # Not a new argument, continue existing one
        fullArg <- paste(fullArg, parsedLine[4])
      } else {
        # New argument, parse the buffer and start a new one
        fields <- plumbOneGlobal(fields, fullArg, envir)
        fullArg <- line
      }
    } else {
      # This isn't a line we can underestand. Parse what we have in the
      # buffer and then reset
      fields <- plumbOneGlobal(fields, fullArg, envir)
      fullArg <- ""
    }
  }

  # Clear out the buffer
  fields <- plumbOneGlobal(fields, fullArg, envir)

  fields
}

#' The default set of OpenAPI Specification (OAS) globals. Some of these properties
#' are subject to being overridden by @api* annotations.
#' @noRd
defaultGlobals <- list(
  openapi = "3.0.3",
  info = list(description = "API Description", title = "API Title", version = "1.0.0"),
  paths = list()
)

context("global settings")

test_that("plumbOneGlobal parses with various formats", {
  fields <- list(info=list())

  # No leading space
  g <- plumbOneGlobal(fields, "#'@apiTitle Title")
  expect_equal(g$info$title, "Title")

  # Plumber-style
  g <- plumbOneGlobal(fields, "#* @apiTitle Title")
  expect_equal(g$info$title, "Title")

  #Extra space
  g <- plumbOneGlobal(fields, "#*    @apiTitle     Title   ")
  expect_equal(g$info$title, "Title")
})

test_that("plumbGlobals works", {
  # Test all fields
  lines <- c("#' @apiTitle title",
             "#' @apiDescription description",
             "#' @apiTOS tos",
             "#' @apiContact contact",
             "#' @apiLicense license",
             "#' @apiVersion version",
             "#' @apiTag t d",
             "#' @apiTag tag description",
             "#' @apiTag tag2 description2",
             "#' @apiTag tag3 description in part",
             "#' @apiTag 'tag4 space' spaces",
             "#' @apiTag \"tag5 space\" spaces")

  fields <- plumbGlobals(lines)

  expect_equal(fields, list(
    info=list(
      title="title",
      description="description",
      termsOfService="tos",
      contact="contact",
      license="license",
      version="version"
    ),
    tags=list(list(name="t", description="d"),
              list(name="tag", description="description"),
              list(name="tag2", description="description2"),
              list(name="tag3", description="description in part"),
              list(name="tag4 space", description="spaces"),
              list(name="tag5 space", description="spaces"))
  ))

  # Test contact and licence object
  lines <- c('#* @apiContact list(name = "API Support", url = "http://www.example.com/support", email = "support@example.com")',
             '#* @apiLicense list(name = "Apache 2.0", url = "https://www.apache.org/licenses/LICENSE-2.0.html")')

  fields <- plumbGlobals(lines)

  expect_equal(fields, list(
    info=list(
      contact=list(name = "API Support", url = "http://www.example.com/support", email = "support@example.com"),
      license=list(name = "Apache 2.0", url = "https://www.apache.org/licenses/LICENSE-2.0.html")
    )
  ))
})

test_that("Globals can't contain duplicate tags", {
  lines <- c("#* @apiTag test description1",
             "#* @apiTag test description2")
  expect_error(plumbGlobals(lines), "Duplicate tag definition specified.")
})

# This function is invoked when creating a new Plumber API project in the
# RStudio IDE. The function will be called when the user invokes the
# New Project wizard using the project template defined in the file at:
#
#   inst/rstudio/templates/project/new-rstudio-project.dcf

# The new project template mechanism is documented at:
# https://rstudio.github.io/rstudio-extensions/rstudio_project_templates.html

newRStudioProject <- function(path, ...) {

  # ensure path exists
  dir.create(path, recursive = TRUE, showWarnings = FALSE)

  # copy 'resources' folder to path
  resources <- system.file("rstudio", "templates", "project", "resources",
                           package = "plumber", mustWork = TRUE)

  files <- list.files(resources, recursive = TRUE, include.dirs = FALSE)
  source <- file.path(resources, files)
  target <- file.path(path, files)
  file.copy(source, target)
}

with_tmp_serializers <- function(expr) {
  ## Do not _actually_ register test-only serializers
  cur_serializers <- .globals$serializers
  on.exit({
    .globals$serializers <- cur_serializers
  }, add = TRUE)

  force(expr)
}

#' @include globals.R

# Mount OpenAPI and Docs
#' @noRd
mount_docs <- function(pr, host, port, docs_info, callback, quiet = FALSE) {

  # return early if not enabled
  if (!isTRUE(docs_info$enabled)) {
    return()
  }

  # Build api url
  api_url <- getOption(
    "plumber.apiURL",
    urlHost(
      scheme = getOption("plumber.apiScheme", "http"),
      host   = getOption("plumber.apiHost", host),
      port   = getOption("plumber.apiPort", port),
      path   = getOption("plumber.apiPath", ""),
      changeHostLocation = TRUE
    )
  )

  # Mount OpenAPI spec paths openapi.json
  mount_openapi(pr, api_url)

  # Mount Docs
  if (length(registered_docs()) == 0) {
    if (!isTRUE(quiet)) {
      message("No visual documentation options registered. See help(register_docs).")
    }
    return()
  }

  if (is_docs_available(docs_info$docs)) {
    docs_mount <- .globals$docs[[docs_info$docs]]$mount
    docs_url <- do.call(docs_mount, c(list(pr, api_url), docs_info$args))
    if (!isTRUE(quiet)) {
      message("Running ", docs_info$docs, " Docs at ", docs_url, sep = "")
    }
  } else {
    return()
  }

  # Use callback
  if (is.function(callback)) {
    callback(docs_url)
  }

  invisible()

}

# Check is Docs is available
#' @noRd
is_docs_available <- function(docs) {
  if (isTRUE(docs %in% registered_docs())) {
    return(TRUE)
  } else {
    message("Unknown docs \"", docs,"\". Maybe try library(", docs,").")
    return(FALSE)
  }
}

# Unmount OpenAPI and Docs
#' @noRd
unmount_docs <- function(pr, docs_info) {

  # return early if not enabled
  if (!isTRUE(docs_info$enabled)) {
    return()
  }

  # Unount OpenAPI spec paths openapi.json
  unmount_openapi(pr)

  # Mount Docs
  docs_unmount <- .globals$docs[[docs_info$docs]]$unmount
  if (length(docs_unmount) && is.function(docs_unmount)) {
    docs_unmount(pr = pr)
  }
}

#' Mount OpenAPI Specification to a plumber router
#' @noRd
mount_openapi <- function(pr, api_url) {

  spec <- pr$getApiSpec()

  # Create a function that's hardcoded to return the OpenAPI specification -- regardless of env.
  openapi_fun <- function(req) {
    # use the HTTP_REFERER so RSC can find the Docs location to ask
    ## (can't directly ask for 127.0.0.1)
    if (is.null(getOption("plumber.apiURL")) &&
        is.null(getOption("plumber.apiHost"))) {
      if (is.null(req$HTTP_REFERER)) {
        # Prevent leaking host and port if option is not set
        api_url <- character(1)
      }
      else {
        # Use HTTP_REFERER as fallback
        api_url <- req$HTTP_REFERER
        api_url <- sub("(\\?.*)?$", "", api_url)
        api_url <- sub("index\\.html$", "", api_url)
        api_url <- sub("__docs__/$", "", api_url)
      }
    }

    utils::modifyList(list(servers = list(list(url = api_url))), spec)

  }
  # http://spec.openapis.org/oas/v3.0.3#document-structure
  # "It is RECOMMENDED that the root OpenAPI document be named: openapi.json"
  for (ep in pr$endpoints[["__no-preempt__"]]) {
    if (ep$path == "/openapi.json") {
      message("Overwritting existing `/openapi.json` route. Use `$setApiSpec()` to define your OpenAPI Spec")
      break
    }
  }
  pr$handle("GET", "/openapi.json", openapi_fun, serializer = serializer_unboxed_json())

  invisible()
}

#' Mount OpenAPI Specification to a plumber router
#' @noRd
unmount_openapi <- function(pr) {

  pr$removeHandle("GET", "/openapi.json")
  invisible()

}

#' Add visual documentation for plumber to use
#'
#' [register_docs()] is used by other packages like `swagger`, `rapidoc`, and `redoc`.
#' When you load these packages, it calls [register_docs()] to provide a user
#' interface that can interpret your plumber OpenAPI Specifications.
#'
#' @param name Name of the visual documentation
#' @param index A function that returns the HTML content of the landing page of the documentation.
#'   Parameters (besides `req` and `res`) will be supplied as if it is a regular `GET` route.
#'   Default parameter values may be used when setting the documentation `index` function.
#'   See the example below.
#' @param static A function that returns the path to the static assets (images, javascript, css, fonts) the Docs will use.
#'
#' @export
#' @examples
#' \dontrun{
#' # Example from the `swagger` R package
#' register_docs(
#'   name = "swagger",
#'   index = function(version = "3", ...) {
#'     swagger::swagger_spec(
#'       api_path = paste0(
#'         "window.location.origin + ",
#'         "window.location.pathname.replace(",
#'           "/\\(__docs__\\\\/|__docs__\\\\/index.html\\)$/, \"\"",
#'         ") + ",
#'         "\"openapi.json\""
#'       ),
#'       version = version
#'     )
#'   },
#'   static = function(version = "3", ...) {
#'     swagger::swagger_path(version)
#'   }
#' )
#'
#' # When setting the docs, `index` and `static` function arguments can be supplied
#' # * via `pr_set_docs()`
#' # * or through URL query string variables
#' pr() %>%
#'   # Set default argument `version = 3` for the swagger `index` and `static` functions
#'   pr_set_docs("swagger", version = 3) %>%
#'   pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) %>%
#'   pr_run()
#' }
#' @rdname register_docs
register_docs <- function(name, index, static = NULL) {

  stopifnot(is.character(name) && length(name) == 1L)
  stopifnot(grepl("^[a-zA-Z0-9_]+$", name))
  stopifnot(is.function(index))
  if (!is.null(static)) stopifnot(is.function(static))

  docs_root <- paste0("/__docs__/")
  docs_paths <- c("/index.html", "/")
  mount_docs_func <- function(pr, api_url, ...) {
    # Save initial extra argument values
    args_index <- list(...)

    docs_index <- function(...) {
      # Override with arguments provided live with URI (i.e. index.html?version=2)
      args <- utils::modifyList(args_index, list(...))
      # Remove default arguments req and res
      args <- args[!(names(args) %in% c("req", "res"))]
      do.call(index, args)
    }

    docs_router <- Plumber$new()
    for (path in docs_paths) {
      docs_router$handle("GET", path, docs_index, serializer = serializer_html())
    }
    if (!is.null(static)) {
      docs_router$mount("/", PlumberStatic$new(static(...)))
    }

    if (!is.null(pr$mounts[[docs_root]])) {
      message("Overwritting existing `", docs_root, "` mount")
      message("")
    }

    pr$mount(docs_root, docs_router)

    # add legacy swagger redirects (RStudio Connect)
    redirect_info <- swagger_redirects()
    for (path in names(redirect_info)) {
      if (router_has_route(pr, path, "GET")) {
        message("Overwriting existing GET endpoint: ", path, ". Disable by setting `options_plumber(legacyRedirects = FALSE)`")
      }
      if (router_has_route(pr, redirect_info[[path]]$route, "GET")) {
        message("Overwriting existing GET endpoint: ", redirect_info[[path]]$route, ". Disable by setting `options_plumber(legacyRedirects = FALSE)`")
      }
      pr_get(pr, path, redirect_info[[path]]$handler)
    }

    docs_url <- paste0(api_url, docs_root)
    return(docs_url)
  }
  unmount_docs_func <- function(pr) {
    pr$unmount(docs_root)

    # remove legacy swagger redirects
    redirect_info <- swagger_redirects()
    for (path in names(redirect_info)) {
      pr$removeHandle("GET", path)
    }

    invisible()
  }

  if (is.null(.globals$docs[[name]])) {
    .globals$docs[[name]] <- list()
  }
  .globals$docs[[name]]$mount <- mount_docs_func
  .globals$docs[[name]]$unmount <- unmount_docs_func

  invisible(name)
}
#' @export
#' @rdname register_docs
registered_docs <- function() {
  sort(names(.globals$docs))
}


swagger_redirects <- function() {
  if (!isTRUE(getOption("plumber.legacyRedirects", TRUE))) {
    return(list())
  }

  to_route <- function(route) {
    list(
      route = route,
      handler = function(req, res) {
        res$status <- 301 # redirect permanently
        res$setHeader("Location", route)
        res$body <- "redirecting..."
        res
      }
    )
  }
  list(
    "/__swagger__/" = to_route("../__docs__/"),
    "/__swagger__/index.html"  = to_route("../__docs__/index.html")
  )
}


register_swagger_docs_onLoad <- function() {
  tryCatch({
    do.call(register_docs, swagger::plumber_docs())
  }, error = function(e) {
    message("Could not register `swagger` docs. ", e)
    NULL
  })
}

context("Options")

test_that("Options set and get", {
  with_options(list(plumber.port = NULL), {
    options_plumber(port = FALSE)
    expect_false(getOption("plumber.port"))
    options_plumber(port = NULL)
    expect_null(getOption("plumber.port"))
  })
})

test_that("all options used are `options_plumber()` parameters", {
  skip_on_cran() # only test in CI / locally

  ## `./R` will not exist when shiny is installed
  r_folder <- "../../R"
  if (
    # not local structure
    !dir.exists(r_folder) ||
    # must contain many files, not just the three files typically found in installed folder
    length(dir(r_folder)) < 10
  ) {
    skip("Not testing locally. Skipping")
  }

  matches <- character()
  for (r_file in dir(r_folder, full.names = T)) {
    file_content <- paste0(readLines(r_file, warn = F), collapse = "")
    match <- stringi::stri_match_all_regex(file_content, "getOption\\([^,\\)]+,?\\)?")[[1]][,1]
    match <- gsub("\\s", "", match)
    if (length(match) > 0 && !is.na(match)) {
      matches <- c(matches, match)
    }
  }
  options_used <- unique(sort(gsub("getOption|\\(|\"|,|'|\\)", "", matches)))
  plumber_options_used <- grep("^plumber", options_used, value = TRUE)
  deprecated_options <-  c("plumber.swagger.url")
  plumber_options_used <- plumber_options_used[!(plumber_options_used %in% deprecated_options)]
  ### code to match formals
  formals_to_match <-
    sort(setdiff(
      names(formals(options_plumber)),
      "..."
    ))
  options_plumber_formals <- paste0("plumber.", formals_to_match)

  expect_equal(
    plumber_options_used,
    options_plumber_formals
  )
})


test_that("Legacy swagger redirect can be disabled", {
  with_options(
    list(
      plumber.legacyRedirets = getOption("plumber.legacyRedirects")
    ), {
      options_plumber(legacyRedirects = TRUE)
      redirects <- swagger_redirects()
      expect_gt(length(redirects), 0)

      options_plumber(legacyRedirects = FALSE)
      redirects <- swagger_redirects()
      expect_equal(length(redirects), 0)
    }
  )
})

test_that("docs.callback sync plumber.swagger.url", {
  with_options(
    list(
      plumber.swagger.url = getOption("plumber.swagger.url"),
      plumber.docs.callback = getOption("plumber.docs.callback")
    ), {
      options("plumber.swagger.url" = function(api_url) {cat(api_url)})
      opt <- options_plumber(docs.callback = NULL)
      expect_null(getOption("plumber.swagger.url"))
      expect_null(opt$plumber.docs.callback)
    }
  )
})

context("Swagger redirects")

test_that("swagger_redirects returns correct redirect structure with default apiPath", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = NULL
    ),
    {
      redirects <- swagger_redirects()

      # Check that we have the expected redirect paths
      expect_true("/__swagger__/" %in% names(redirects))
      expect_true("/__swagger__/index.html" %in% names(redirects))

      # Verify redirect structure uses relative paths
      expect_equal(redirects[["/__swagger__/"]]$route, "../__docs__/")
      expect_equal(
        redirects[["/__swagger__/index.html"]]$route,
        "../__docs__/index.html"
      )

      # Verify handler is a function
      expect_true(is.function(redirects[["/__swagger__/"]]$handler))
      expect_true(is.function(redirects[["/__swagger__/index.html"]]$handler))
    }
  )
})

test_that("swagger_redirects respects custom apiPath", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = "/api/v1"
    ),
    {
      redirects <- swagger_redirects()

      # Redirect keys should include the apiPath prefix
      expect_true("/api/v1/__swagger__/" %in% names(redirects))
      expect_true("/api/v1/__swagger__/index.html" %in% names(redirects))

      # Targets use relative paths
      expect_equal(redirects[["/api/v1/__swagger__/"]]$route, "../__docs__/")
      expect_equal(
        redirects[["/api/v1/__swagger__/index.html"]]$route,
        "../__docs__/index.html"
      )
    }
  )
})

test_that("swagger_redirects returns empty list when disabled", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(plumber.legacyRedirects = FALSE),
    {
      redirects <- swagger_redirects()
      expect_length(redirects, 0)
      expect_equal(redirects, list())
    }
  )
})

test_that("swagger redirect handler returns 301 response", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = NULL
    ),
    {
      redirects <- swagger_redirects()
      handler <- redirects[["/__swagger__/"]]$handler

      # Create mock request and response
      req <- make_req("GET", "/__swagger__/")
      res <- PlumberResponse$new()

      # Execute handler
      result <- handler(req, res)

      # Verify redirect response uses relative path
      expect_equal(res$status, 301)
      expect_equal(res$headers$Location, "../__docs__/")
      expect_equal(res$body, "redirecting...")
    }
  )
})

test_that("swagger redirect handler works with apiPath", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = "/api/v2"
    ),
    {
      redirects <- swagger_redirects()
      handler <- redirects[["/api/v2/__swagger__/index.html"]]$handler

      # Create mock request and response
      req <- make_req("GET", "/api/v2/__swagger__/index.html")
      res <- PlumberResponse$new()

      # Execute handler
      result <- handler(req, res)

      # Verify redirect response uses relative path
      expect_equal(res$status, 301)
      expect_equal(res$headers$Location, "../__docs__/index.html")
      expect_equal(res$body, "redirecting...")
    }
  )
})

test_that("swagger redirects are mounted on plumber router", {
  skip_if_not_installed("withr")
  skip_if_not_installed("swagger")

  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = ""
    ),
    {
      pr <- pr()
      pr$handle("GET", "/test", function() "test")

      docs_info <- list(
        enabled = TRUE,
        docs = "swagger",
        args = list()
      )

      # Mount docs (which includes redirects)
      mount_docs(
        pr = pr,
        host = "127.0.0.1",
        port = 8000,
        docs_info = docs_info,
        callback = NULL,
        quiet = TRUE
      )

      # Check that redirect endpoints are registered
      endpoints <- pr$endpoints[["__no-preempt__"]]
      endpoint_paths <- sapply(endpoints, function(e) e$path)

      expect_true("/__swagger__/" %in% endpoint_paths)
      expect_true("/__swagger__/index.html" %in% endpoint_paths)
    }
  )
})

test_that("swagger redirects work on mounted router with apiPath", {
  skip_if_not_installed("withr")
  skip_if_not_installed("swagger")

  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = "/myapp"
    ),
    {
      pr <- pr()
      pr$handle("GET", "/endpoint", function() "data")

      docs_info <- list(
        enabled = TRUE,
        docs = "swagger",
        args = list()
      )

      mount_docs(
        pr = pr,
        host = "127.0.0.1",
        port = 8000,
        docs_info = docs_info,
        callback = NULL,
        quiet = TRUE
      )

      # Find the redirect endpoint (should be prefixed with apiPath)
      endpoints <- pr$endpoints[["__no-preempt__"]]
      swagger_endpoint <- NULL
      for (ep in endpoints) {
        if (ep$path == "/myapp/__swagger__/") {
          swagger_endpoint <- ep
          break
        }
      }

      expect_false(is.null(swagger_endpoint))

      # Test the redirect by routing a request
      req <- make_req("GET", "/myapp/__swagger__/", pr = pr)
      res <- PlumberResponse$new()

      result <- pr$route(req, res)

      # Verify redirect uses relative path
      expect_equal(res$status, 301)
      expect_equal(res$headers$Location, "../__docs__/")
    }
  )
})

test_that("swagger redirects can be unmounted", {
  skip_if_not_installed("withr")
  skip_if_not_installed("swagger")

  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = ""
    ),
    {
      pr <- pr()
      pr$handle("GET", "/test", function() "test")

      docs_info <- list(
        enabled = TRUE,
        docs = "swagger",
        args = list()
      )

      # Mount docs
      mount_docs(
        pr = pr,
        host = "127.0.0.1",
        port = 8000,
        docs_info = docs_info,
        callback = NULL,
        quiet = TRUE
      )

      # Verify redirects are mounted
      endpoints_before <- pr$endpoints[["__no-preempt__"]]
      paths_before <- sapply(endpoints_before, function(e) e$path)
      expect_true("/__swagger__/" %in% paths_before)

      # Unmount docs
      unmount_docs(pr, docs_info)

      # Verify redirects are unmounted
      endpoints_after <- pr$endpoints[["__no-preempt__"]]
      if (length(endpoints_after) > 0) {
        paths_after <- sapply(endpoints_after, function(e) e$path)
        expect_false("/__swagger__/" %in% paths_after)
        expect_false("/__swagger__/index.html" %in% paths_after)
      } else {
        # No endpoints left is also valid
        expect_length(endpoints_after, 0)
      }
    }
  )
})

test_that("swagger redirect overwrite warning appears when route exists", {
  skip_if_not_installed("withr")
  skip_if_not_installed("swagger")

  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = ""
    ),
    {
      pr <- pr()
      # Pre-register a route at the same path as the redirect
      pr$handle("GET", "/__swagger__/", function() "existing")

      docs_info <- list(
        enabled = TRUE,
        docs = "swagger",
        args = list()
      )

      # Mounting docs should show overwrite warning
      expect_message(
        mount_docs(
          pr = pr,
          host = "127.0.0.1",
          port = 8000,
          docs_info = docs_info,
          callback = NULL,
          quiet = TRUE
        ),
        "Overwriting existing GET endpoint: /__swagger__/"
      )
    }
  )
})

test_that("swagger redirects work with nested apiPath", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = "/api/v1/myapp"
    ),
    {
      redirects <- swagger_redirects()

      # Verify redirect keys include full nested path
      expect_true("/api/v1/myapp/__swagger__/" %in% names(redirects))
      expect_true("/api/v1/myapp/__swagger__/index.html" %in% names(redirects))

      # Targets use relative paths
      expect_equal(
        redirects[["/api/v1/myapp/__swagger__/"]]$route,
        "../__docs__/"
      )
      expect_equal(
        redirects[["/api/v1/myapp/__swagger__/index.html"]]$route,
        "../__docs__/index.html"
      )
    }
  )
})

test_that("swagger redirects respect environment variable for apiPath", {
  skip_if_not_installed("withr")

  withr::with_envvar(
    list(PLUMBER_APIPATH = "/env/api"),
    {
      withr::with_options(
        list(
          plumber.legacyRedirects = TRUE,
          plumber.apiPath = NULL  # Clear option to use env var
        ),
        {
          redirects <- swagger_redirects()

          # Should use environment variable in redirect keys
          expect_true("/env/api/__swagger__/" %in% names(redirects))
          # Targets use relative paths
          expect_equal(
            redirects[["/env/api/__swagger__/"]]$route,
            "../__docs__/"
          )
        }
      )
    }
  )
})

test_that("swagger redirects handle empty apiPath correctly", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = ""
    ),
    {
      redirects <- swagger_redirects()

      # Empty string should behave same as NULL, uses relative paths
      expect_equal(redirects[["/__swagger__/"]]$route, "../__docs__/")
      expect_equal(
        redirects[["/__swagger__/index.html"]]$route,
        "../__docs__/index.html"
      )
    }
  )
})

test_that("multiple swagger redirects on same router work independently", {
  skip_if_not_installed("withr")
  skip_if_not_installed("swagger")

  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = "/v1"
    ),
    {
      pr <- pr()
      pr$handle("GET", "/test", function() "test")

      docs_info <- list(
        enabled = TRUE,
        docs = "swagger",
        args = list()
      )

      mount_docs(pr, "127.0.0.1", 8000, docs_info, NULL, TRUE)

      # Test both redirect endpoints (with apiPath prefix)
      req1 <- make_req("GET", "/v1/__swagger__/", pr = pr)
      res1 <- PlumberResponse$new()
      pr$route(req1, res1)

      req2 <- make_req("GET", "/v1/__swagger__/index.html", pr = pr)
      res2 <- PlumberResponse$new()
      pr$route(req2, res2)

      # Both should redirect correctly using relative paths
      expect_equal(res1$status, 301)
      expect_equal(res1$headers$Location, "../__docs__/")

      expect_equal(res2$status, 301)
      expect_equal(res2$headers$Location, "../__docs__/index.html")
    }
  )
})

context("UI with apiPath")

test_that("mount_openapi respects plumber.apiPath option", {
  # Test with default (empty) apiPath
  pr <- pr()
  pr$handle("GET", "/test", function() "test")

  api_url <- "http://localhost:8000"
  mount_openapi(pr, api_url)

  # Should mount at /openapi.json by default
  openapi_endpoint <- pr$endpoints[["__no-preempt__"]]
  openapi_paths <- sapply(openapi_endpoint, function(e) e$path)
  expect_true("/openapi.json" %in% openapi_paths)
})

test_that("mount_openapi uses plumber.apiPath when set", {
  skip_if_not_installed("withr")

  # Test with apiPath set
  withr::with_options(
    list(plumber.apiPath = "/api/v1"),
    {
      pr <- pr()
      pr$handle("GET", "/test", function() "test")

      api_url <- "http://localhost:8000/api/v1"
      mount_openapi(pr, api_url)

      # Should mount at /api/v1/openapi.json
      openapi_endpoint <- pr$endpoints[["__no-preempt__"]]
      openapi_paths <- sapply(openapi_endpoint, function(e) e$path)
      expect_true("/api/v1/openapi.json" %in% openapi_paths)
      expect_false("/openapi.json" %in% openapi_paths)
    }
  )
})

test_that("mount_openapi works with nested apiPath", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(plumber.apiPath = "/api/v2/myapp"),
    {
      pr <- pr()
      pr$handle("GET", "/endpoint", function() "data")

      api_url <- "http://localhost:8000/api/v2/myapp"
      mount_openapi(pr, api_url)

      # Should mount at /api/v2/myapp/openapi.json
      openapi_endpoint <- pr$endpoints[["__no-preempt__"]]
      openapi_paths <- sapply(openapi_endpoint, function(e) e$path)
      expect_true("/api/v2/myapp/openapi.json" %in% openapi_paths)
    }
  )
})

test_that("unmount_openapi respects plumber.apiPath option", {
  # Test with default (empty) apiPath
  pr <- pr()
  pr$handle("GET", "/test", function() "test")
  mount_openapi(pr, "http://localhost:8000")

  # Verify it's mounted
  openapi_endpoint <- pr$endpoints[["__no-preempt__"]]
  openapi_paths_before <- sapply(openapi_endpoint, function(e) e$path)
  expect_true("/openapi.json" %in% openapi_paths_before)

  # Unmount
  unmount_openapi(pr)

  # Verify it's unmounted
  openapi_endpoint_after <- pr$endpoints[["__no-preempt__"]]
  if (length(openapi_endpoint_after) > 0) {
    openapi_paths_after <- sapply(openapi_endpoint_after, function(e) e$path)
    expect_false("/openapi.json" %in% openapi_paths_after)
  } else {
    # No endpoints left, which is also valid
    expect_length(openapi_endpoint_after, 0)
  }
})

test_that("unmount_openapi uses plumber.apiPath when set", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(plumber.apiPath = "/api/v1"),
    {
      pr <- pr()
      pr$handle("GET", "/test", function() "test")
      mount_openapi(pr, "http://localhost:8000/api/v1")

      # Verify it's mounted at the correct path
      openapi_endpoint <- pr$endpoints[["__no-preempt__"]]
      openapi_paths_before <- sapply(openapi_endpoint, function(e) e$path)
      expect_true("/api/v1/openapi.json" %in% openapi_paths_before)

      # Unmount
      unmount_openapi(pr)

      # Verify it's unmounted
      openapi_endpoint_after <- pr$endpoints[["__no-preempt__"]]
      if (length(openapi_endpoint_after) > 0) {
        openapi_paths_after <- sapply(openapi_endpoint_after, function(e) e$path)
        expect_false("/api/v1/openapi.json" %in% openapi_paths_after)
      }
    }
  )
})

test_that("register_docs mounts docs at correct path with apiPath", {
  # Test with default (empty) apiPath
  pr <- pr()
  pr$handle("GET", "/test", function() "test")

  # register_docs is called when swagger package loads, but we can test the mount behavior
  # by checking the mount path
  api_url <- "http://localhost:8000"

  # The register_docs function should be available (registered by swagger or similar)
  if (length(registered_docs()) > 0) {
    doc_name <- registered_docs()[1]
    mount_func <- .globals$docs[[doc_name]]$mount

    # Call the mount function
    docs_url <- mount_func(pr, api_url)

    # Should mount at /__docs__/
    expect_true(!is.null(pr$mounts[["/__docs__/"]]))
    expect_equal(docs_url, "http://localhost:8000/__docs__/")
  } else {
    skip("No registered docs available for testing")
  }
})

test_that("register_docs respects plumber.apiPath option", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(plumber.apiPath = "/api/v1"),
    {
      pr <- pr()
      pr$handle("GET", "/test", function() "test")

      api_url <- "http://localhost:8000/api/v1"

      if (length(registered_docs()) > 0) {
        doc_name <- registered_docs()[1]
        mount_func <- .globals$docs[[doc_name]]$mount

        # Call the mount function
        docs_url <- mount_func(pr, api_url)

        # Should mount at /api/v1/__docs__/
        expect_true(!is.null(pr$mounts[["/api/v1/__docs__/"]]))
        expect_equal(docs_url, "http://localhost:8000/api/v1/__docs__/")
        expect_null(pr$mounts[["/__docs__/"]])  # Should NOT be at default location
      } else {
        skip("No registered docs available for testing")
      }
    }
  )
})

test_that("swagger_redirects respects plumber.apiPath option", {
  skip_if_not_installed("withr")

  # Ensure legacyRedirects is enabled for this test
  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = NULL
    ),
    {
      # Test with default (empty) apiPath
      redirects <- swagger_redirects()
      expect_true("/__swagger__/" %in% names(redirects))
      expect_equal(redirects[["/__swagger__/"]]$route, "../__docs__/")
    }
  )

  # Test with apiPath set
  withr::with_options(
    list(
      plumber.legacyRedirects = TRUE,
      plumber.apiPath = "/api/v1"
    ),
    {
      redirects <- swagger_redirects()
      # With apiPath, redirect keys include the prefix
      expect_true("/api/v1/__swagger__/" %in% names(redirects))
      # Targets use relative paths
      expect_equal(redirects[["/api/v1/__swagger__/"]]$route, "../__docs__/")
      expect_equal(redirects[["/api/v1/__swagger__/index.html"]]$route, "../__docs__/index.html")
    }
  )
})

test_that("mount_docs integrates openapi and docs paths correctly", {
  skip_if_not_installed("swagger")
  skip_if_not_installed("withr")

  # Test with apiPath set
  withr::with_options(
    list(plumber.apiPath = "/api/v1"),
    {
      pr <- pr()
      pr$handle("GET", "/endpoint", function() "data")

      docs_info <- list(
        enabled = TRUE,
        docs = "swagger",
        args = list()
      )

      # Mount docs (which also mounts openapi)
      mount_docs(
        pr = pr,
        host = "127.0.0.1",
        port = 8000,
        docs_info = docs_info,
        callback = NULL,
        quiet = TRUE
      )

      # Check that openapi.json is mounted at the correct path
      openapi_endpoint <- pr$endpoints[["__no-preempt__"]]
      openapi_paths <- sapply(openapi_endpoint, function(e) e$path)
      expect_true("/api/v1/openapi.json" %in% openapi_paths)

      # Check that docs are mounted at the correct path
      expect_true(!is.null(pr$mounts[["/api/v1/__docs__/"]]))
      expect_null(pr$mounts[["/__docs__/"]])  # Should NOT be at default location
    }
  )
})

test_that("multiple sequential mount/unmount operations work correctly", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(plumber.apiPath = "/v1"),
    {
      pr <- pr()
      pr$handle("GET", "/test", function() "test")

      # Mount
      mount_openapi(pr, "http://localhost:8000/v1")
      openapi_endpoint <- pr$endpoints[["__no-preempt__"]]
      openapi_paths <- sapply(openapi_endpoint, function(e) e$path)
      expect_true("/v1/openapi.json" %in% openapi_paths)

      # Unmount
      unmount_openapi(pr)
      openapi_endpoint_after <- pr$endpoints[["__no-preempt__"]]
      if (length(openapi_endpoint_after) > 0) {
        openapi_paths_after <- sapply(openapi_endpoint_after, function(e) e$path)
        expect_false("/v1/openapi.json" %in% openapi_paths_after)
      }

      # Mount again
      mount_openapi(pr, "http://localhost:8000/v1")
      openapi_endpoint_final <- pr$endpoints[["__no-preempt__"]]
      openapi_paths_final <- sapply(openapi_endpoint_final, function(e) e$path)
      expect_true("/v1/openapi.json" %in% openapi_paths_final)
    }
  )
})

test_that("apiPath works with environment variable", {
  skip_if_not_installed("withr")

  withr::with_envvar(
    list(PLUMBER_APIPATH = "/env/path"),
    {
      # Clear option to ensure env var is used
      withr::with_options(
        list(plumber.apiPath = NULL),
        {
          pr <- pr()
          pr$handle("GET", "/test", function() "test")
          mount_openapi(pr, "http://localhost:8000/env/path")

          openapi_endpoint <- pr$endpoints[["__no-preempt__"]]
          openapi_paths <- sapply(openapi_endpoint, function(e) e$path)
          expect_true("/env/path/openapi.json" %in% openapi_paths)
        }
      )
    }
  )
})

test_that("empty apiPath works correctly", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(plumber.apiPath = ""),
    {
      pr <- pr()
      pr$handle("GET", "/test", function() "test")
      mount_openapi(pr, "http://localhost:8000")

      openapi_endpoint <- pr$endpoints[["__no-preempt__"]]
      openapi_paths <- sapply(openapi_endpoint, function(e) e$path)
      expect_true("/openapi.json" %in% openapi_paths)
    }
  )
})

test_that("apiPath with trailing slash is handled correctly", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(plumber.apiPath = "/api/v1/"),
    {
      pr <- pr()
      pr$handle("GET", "/test", function() "test")
      mount_openapi(pr, "http://localhost:8000/api/v1/")

      # The implementation concatenates path + "/openapi.json"
      # so "/api/v1/" + "/openapi.json" = "/api/v1//openapi.json"
      # This test documents current behavior - may need adjustment if this is a bug
      openapi_endpoint <- pr$endpoints[["__no-preempt__"]]
      openapi_paths <- sapply(openapi_endpoint, function(e) e$path)

      # Check what path was actually created
      expect_true(any(grepl("openapi\\.json$", openapi_paths)))
    }
  )
})

test_that("overwriting message appears when route already exists with apiPath", {
  skip_if_not_installed("withr")

  withr::with_options(
    list(plumber.apiPath = "/api"),
    {
      pr <- pr()
      # Pre-register a route at the same path
      pr$handle("GET", "/api/openapi.json", function() "existing")

      # Now mount_openapi should show overwriting message
      expect_message(
        mount_openapi(pr, "http://localhost:8000/api"),
        "Overwritting existing `/openapi.json` route"
      )
    }
  )
})

#* @get /
#* @preempt __first__
function(){
  "first"
}

#* @get /abc
function(){
  "abc get"
}

#* @post /abc
function(){
  "abc post"
}

#* @filter filt1
function(req, res){
  forward()
}

#* @filter filt2
function(req, res){
  forward()
}

#* @use /dog
#* @preempt filt2
function(){
  "dog use"
}

#* @get /dog
#* @preempt filt1
function(){
  "dog get"
}

#* @get /error
function(){
  stop("ERROR")
}

#* @get /response123
function(res){
  res$body <- "overridden"
  res$status <- 123
  res
}

#* @get /response200
function(res){
  res$body <- "overridden"
  res$status <- 200
  res
}

#* @get /path1
#* @get /path2
function(){
  "dual path"
}

#* Save a file with a particular file name. Ex: `time.txt`
#* @serializer text
#* @get /name
function() {
  as_attachment(Sys.time(), "time.txt")
}



#* Save a file as the route. Ex: `no_name`
#* @serializer text
#* @get /no_name
function() {
  as_attachment(Sys.time())
}

#* Display within browser. Possible as the mime type is `text/plain`
#* @serializer text
#* @get /inline
function() {
  Sys.time()
}

test_serializer <- function() {
  endpoint_serializer(
    serializer = function(value, req, res, errorHandler) {
      list(status=201L, headers=list(), body=as.character(value))
    },
    preexec_hook = function(req, res, data) {
      cat("preexec\n")
    },
    postexec_hook = function(value, req, res, data) {
      cat("postexec\n")
      value
    },
    aroundexec_hook = function(..., .next) {
      cat("around pre\n")
      value <- .next(...)
      cat("around post\n")
      value
    }
  )
}

#' @plumber
function(pr) {
  register_serializer("test_serializer", test_serializer)
}

#' @serializer test_serializer
#' @get /test_serializer
function() {
  cat("exec\n")
  4
}

#' Does not produce a plot
#' @serializer png
#' @get /no_plot
function(n = 100) {
  dt <- runif(n)
  # hist(dt)
  dt
}

#' Deprecated functions
#'
#' @describeIn deprecated See [register_serializer()]
#' @export
#' @keywords internal
addSerializer <- function(name, serializer, verbose = TRUE) {
  lifecycle::deprecate_warn("1.0.0", "addSerializer()", "register_serializer()")

  register_serializer(name = name, serializer = serializer, verbose = verbose)
}

#' @describeIn deprecated See [get_character_set()]
#' @export
getCharacterSet <- function(contentType = NULL) {
  lifecycle::deprecate_warn("1.0.0", "getCharacterSet()", "get_character_set()")
  get_character_set(content_type = contentType)
}


#' @describeIn deprecated See [session_cookie()]
#' @export
sessionCookie <- function(
  key,
  name = "plumber",
  expiration = FALSE,
  http = TRUE,
  secure = FALSE,
  sameSite = FALSE
) {
  lifecycle::deprecate_warn("1.0.0", "sessionCookie()", "session_cookie()")
  session_cookie(
    key = key,
    name = name,
    expiration = expiration,
    http = http,
    secure = secure,
    same_site = sameSite
  )
}

#' Register a Serializer
#'
#' A serializer is responsible for translating a generated R value into output
#' that a remote user can understand. For instance, the \code{serializer_json}
#' serializes R objects into JSON before returning them to the user. The list of
#' available serializers in plumber is global.
#'
#' There are three main building-block serializers:
#' * `serializer_headers`: the base building-block serializer that is required to have [as_attachment()] work
#' * `serializer_content_type()`: for setting the content type. (Calls `serializer_headers()`)
#' * `serializer_device()`: add endpoint hooks to turn a graphics device on and off in addition to setting the content type. (Uses `serializer_content_type()`)
#'
#' @param name The name of the serializer (character string)
#' @param serializer The serializer function to be added.
#' This function should accept arguments that can be supplied when [plumb()]ing a file.
#' This function should return a function that accepts four arguments: `value`, `req`, `res`, and `errorHandler`.
#' See `print(serializer_json)` for an example.
#'
#' @param verbose Logical value which determines if a message should be printed when overwriting serializers
#' @describeIn register_serializer Register a serializer with a name
#' @export
#' @examples
#' # `serializer_json()` calls `serializer_content_type()` and supplies a serialization function
#' print(serializer_json)
#' # serializer_content_type() calls `serializer_headers()` and supplies a serialization function
#' print(serializer_content_type)
register_serializer <- function(name, serializer, verbose = TRUE) {
  if (name %in% registered_serializers()) {
    if (isTRUE(verbose)) {
      message("Overwriting serializer: ", name)
    }
  }
  .globals$serializers[[name]] <- serializer
}
#' @describeIn register_serializer Return a list of all registered serializers
#' @export
registered_serializers <- function() {
  sort(names(.globals$serializers))
}

get_registered_serializer <- function(name) {
  serializer <- .globals$serializers[[name]]
  if (is.null(serializer)) {
    stop("'", name, "' is not a registered serializer. See `?registered_serializers`")
  }

  serializer
}



# internal function to use directly within this file only. (performance purposes)
# Other files should use `serializer_identity()` to avoid confusion
serializer_identity_ <- function(val, req, res, errorHandler) {
  tryCatch({
    res$body <- val
    res$toResponse()
  }, error = function(err) {
    errorHandler(req, res, err)
  })
}
serializer_identity <- function(){
  serializer_identity_
}


#' Return an attachment response
#'
#' This will set the appropriate fields in the `Content-Disposition` header value.
#' To make sure the attachment is used, be sure your serializer eventually calls `serializer_headers`
#'
#' @param value Response value to be saved
#' @param filename File name to use when saving the attachment.
#'   If no `filename` is provided, the `value` will be treated as a regular attachment
#' @return Object with class `"plumber_attachment"`
#' @export
#' @examples
#' \dontrun{
#' # plumber.R
#'
#' #' @get /data
#' #' @serializer csv
#' function() {
#'   # will cause the file to be saved as `iris.csv`, not `data` or `data.csv`
#'   as_attachment(iris, "iris.csv")
#' }
#' }
as_attachment <- function(value, filename = NULL) {
  stopifnot(is.character(filename) || is.null(filename))
  if (is.character(filename)) {
    stopifnot(length(filename) == 1)
  }
  structure(
    list(
      value = value,
      filename = filename
    ),
    class = "plumber_attachment"
  )
}


#' Plumber Serializers
#'
#' Serializers are used in Plumber to transform the R object produced by a
#' filter/endpoint into an HTTP response that can be returned to the client. See
#' [here](https://www.rplumber.io/articles/rendering-output.html#serializers-1) for
#' more details on Plumber serializers and how to customize their behavior.
#' @describeIn serializers Add a static list of headers to each return value. Will add `Content-Disposition` header if a value is the result of `as_attachment()`.
#' @param ... extra arguments supplied to respective internal serialization function.
#' @param headers `list()` of headers to add to the response object
#' @param serialize_fn Function to serialize the data. The result object will be converted to a character string. Ex: [jsonlite::toJSON()].
#' @export
serializer_headers <- function(headers = list(), serialize_fn = identity) {
  stopifnot(is.function(serialize_fn))
  stopifnot(is.list(headers))

  function(val, req, res, errorHandler) {
    tryCatch({

      # handle `as_attachment()` before serializing
      if (inherits(val, "plumber_attachment")) {
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Disposition
        #> Used on the body itself, Content-Disposition has no effect.
        ## Therefore `Content-Type` MUST exist
        if (!is.null(headers[["Content-Type"]])) {
          # do not overwrite pre-existing header value
          if (is.null(headers[["Content-Disposition"]])) {
            headers[["Content-Disposition"]] <-
              if (is.null(val$filename)) {
                "attachment"
              } else {
                #> path information should be stripped
                filename <- basename(val$filename)
                if (stri_detect_fixed(filename, "\"") || stri_detect_fixed(filename, "'")) {
                  stop("`filename` may not contain quotes")
                }
                paste0("attachment; filename=\"", filename, "\"")
              }
          }
        }
        # make `val` the contained value and not the `as_attachment()` structure
        val <- val$value
      }

      # set headers after upgrading headers
      Map(names(headers), headers, f = function(header, header_val) {
        # `res` is an R6 object
        res$setHeader(header, header_val)
      })

      # serialize
      val <- serialize_fn(val)

      # return value
      serializer_identity_(val, req, res, errorHandler)
    }, error = function(err) {
      errorHandler(req, res, err)
    })
  }
}



#' @describeIn serializers Adds a `Content-Type` header to the response object
# name can not change!
#' @param type The value to provide for the `Content-Type` HTTP header.
#' @export
serializer_content_type <- function(type, serialize_fn = identity) {
  if (missing(type)){
    stop("You must provide the custom content type to the serializer_content_type")
  }

  stopifnot(length(type) == 1)
  stopifnot(is.character(type))
  stopifnot(nchar(type) > 0)

  serializer_headers(
    list("Content-Type" = type),
    serialize_fn
  )
}

#' @describeIn serializers Octet serializer. If content is received that does
#'   not have a `"raw"` type, then an error will be thrown.
#' @export
serializer_octet <- function(..., type = "application/octet-stream") {
  serializer_content_type(type, function(val) {
    if (!is.raw(val)) {
      stop("The Octet-Stream serializer received non-raw data")
    }
    val
  })
}


#' @describeIn serializers CSV serializer. See also: [readr::format_csv()]
#' @export
serializer_csv <- function(..., type = "text/csv; charset=UTF-8") {
  if (!requireNamespace("readr", quietly = TRUE)) {
    stop("`readr` must be installed for `serializer_csv` to work")
  }

  serializer_content_type(type, function(val) {
    readr::format_csv(val, ...)
  })
}

#' @describeIn serializers TSV serializer. See also: [readr::format_tsv()]
#' @export
serializer_tsv <- function(..., type = "text/tab-separated-values; charset=UTF-8") {
  if (!requireNamespace("readr", quietly = TRUE)) {
    stop("`readr` must be installed for `serializer_tsv` to work")
  }

  serializer_content_type(type, function(val) {
    readr::format_tsv(val, ...)
  })
}



#' @describeIn serializers HTML serializer
#' @export
serializer_html <- function(type = "text/html; charset=UTF-8") {
  serializer_content_type(type)
}


#' @describeIn serializers JSON serializer. See also: [jsonlite::toJSON()]
#' @export
#' @importFrom jsonlite toJSON
serializer_json <- function(..., type = "application/json") {
  serializer_content_type(type, function(val) {
    toJSON(val, ...)
  })
}

#' @describeIn serializers JSON serializer with `auto_unbox` defaulting to `TRUE`. See also: [jsonlite::toJSON()]
#' @inheritParams jsonlite::toJSON
#' @export
serializer_unboxed_json <- function(auto_unbox = TRUE, ..., type = "application/json") {
  serializer_json(auto_unbox = auto_unbox, ..., type = type)
}

#' @describeIn serializers GeoJSON serializer. See also [geojsonsf::sf_geojson()] and [[geojsonsf::sfc_geojson()]].
#' @export
serializer_geojson <- function(..., type = "application/geo+json") {
  if (!requireNamespace("geojsonsf", quietly = TRUE)) {
    stop("`geojsonsf` must be installed for `serializer_geojson` to work")
  }
  serializer_content_type(type, function(val) {
    if (inherits(val, "sfc")) return(geojsonsf::sfc_geojson(val, ...))
    if (inherits(val, "sf"))  return(geojsonsf::sf_geojson(val, ...))
    stop("Did not receive an `sf` or `sfc` object. ")
  })
}



#' @describeIn serializers RDS serializer. See also: [base::serialize()]
#' @inheritParams base::serialize
#' @export
serializer_rds <- function(version = "2", ascii = FALSE, ..., type = "application/rds") {
  if (identical(version, "3")) {
    if (package_version(R.version) < "3.5") {
      stop(
        "R versions before 3.5 do not know how to serialize with `version = \"3\"`",
        "\n Current R version: ", as.character(package_version(R.version))
      )
    }
  }
  serializer_content_type(type, function(val) {
    base::serialize(val, NULL, ascii = ascii, version = version, ...)
  })
}

#' @describeIn serializers feather serializer. See also: [arrow::write_feather()]
#' @export
serializer_feather <- function(type = "application/vnd.apache.arrow.file") {
  if (!requireNamespace("arrow", quietly = TRUE)) {
    stop("`arrow` must be installed for `serializer_feather` to work")
  }
  serializer_write_file(
    fileext = ".feather",
    type = type,
    write_fn = function(val, tmpfile) {
      arrow::write_feather(val, tmpfile)
    }
  )
}

#' @describeIn serializers parquet serializer. See also: [arrow::write_parquet()]
#' @export
serializer_parquet <- function(type = "application/vnd.apache.parquet") {
  if (!requireNamespace("arrow", quietly = TRUE)) {
    stop("`arrow` must be installed for `serializer_parquet` to work")
  }
  serializer_write_file(
    fileext = ".parquet",
    type = type,
    write_fn = function(val, tmpfile) {
      arrow::write_parquet(val, tmpfile)
    }
  )
}


#' @describeIn serializers YAML serializer. See also: [yaml::as.yaml()]
#' @export
serializer_yaml <- function(..., type = "text/x-yaml; charset=UTF-8") {
  if (!requireNamespace("yaml", quietly = TRUE)) {
    stop("yaml must be installed for the yaml serializer to work")
  }
  serializer_content_type(type, function(val) {
    yaml::as.yaml(val, ...)
  })
}

#' @describeIn serializers Text serializer. See also: [as.character()]
#' @export
serializer_text <- function(..., serialize_fn = as.character, type = "text/plain; charset=UTF-8") {
  serializer_content_type(type, function(val) {
    serialize_fn(val, ...)
  })
}



#' @describeIn serializers Text serializer. See also: [format()]
#' @export
serializer_format <- function(..., type = "text/plain; charset=UTF-8") {
  serializer_text(..., serialize_fn = format, type = type)
}

#' @describeIn serializers Text serializer. Captures the output of [print()]
#' @export
serializer_print <- function(..., type = "text/plain; charset=UTF-8") {
  serializer_text(
    type = type,
    serialize_fn = function(x) {
      paste0(
        collapse = "\n",
        utils::capture.output({
          print(x, ...)
        })
      )
    }
  )
}
#' @describeIn serializers Text serializer. Captures the output of [cat()]
#' @export
serializer_cat <- function(..., type = "text/plain; charset=UTF-8") {
  serializer_text(
    type = type,
    serialize_fn = function(x) {
      paste0(
        collapse = "\n",
        utils::capture.output({
          cat(x, ...)
        })
      )
    }
  )
}

#' @describeIn serializers Write output to a temp file whose contents are read back as a serialized response. `serializer_write_file()` creates (and cleans up) a temp file, calls the serializer (which should write to the temp file), and then reads the contents back as the serialized value.  If the content `type` starts with `"text"`, the return result will be read into a character string, otherwise the result will be returned as a raw vector.
#' @param write_fn Function that should write serialized content to the temp file provided. `write_fn` should have the function signature of `function(value, tmp_file){}`.
#' @param fileext A non-empty character vector giving the file extension. This value will try to be inferred from the content type provided.
#' @export
serializer_write_file <- function(
  type,
  write_fn,
  fileext = NULL
) {

  # try to be nice and get the file extension from the
  fileext <- fileext %||% get_fileext(type) %||% ""

  serializer_content_type(type, function(val) {
    tmpfile <- tempfile(fileext = fileext)
    on.exit({
      if (file.exists(tmpfile)) {
        unlink(tmpfile)
      }
    }, add = TRUE)

    # write to disk
    write_fn(val, tmpfile)

    # read back results
    if (grepl("^text", type)) {
      paste(readLines(tmpfile), collapse = "\n")
    } else {
      readBin(tmpfile, what = "raw", n = file.info(tmpfile)$size)
    }
  })
}




#' @describeIn serializers htmlwidget serializer. See also: [htmlwidgets::saveWidget()]
#' @export
serializer_htmlwidget <- function(..., type = "text/html; charset=UTF-8") {
  if (!requireNamespace("htmlwidgets", quietly = TRUE)) {
    stop("The htmlwidgets package is not available but is required in order to use the htmlwidgets serializer",
          call. = FALSE)
  }

  serializer_write_file(
    # Write out a temp file. htmlwidgets (or pandoc?) seems to require that this
    # file end in .html or the selfcontained=TRUE argument has no effect.
    fileext = ".html",
    type = type,
    write_fn = function(val, tmpfile) {
      # Write the widget out to a file (doesn't currently support in-memory connections - pandoc)
      # Must write a self-contained file. We're not serving a directory of assets
      # in response to this request, just one HTML file.
      htmlwidgets::saveWidget(val, tmpfile, selfcontained = TRUE, ...)
    }
  )
}


serializer_xml <- function() {
  #if (!requireNamespace("XML", quietly = TRUE)) {
  #  stop("The XML package is not available but is required in order to use the XML serializer.",
  #       call. = FALSE)
  #}
  stop("XML serialization not yet implemented. Please see the discussion at https://github.com/rstudio/plumber/issues/65")
}



#' Endpoint Serializer with Hooks
#'
#' This method allows serializers to return `preexec`, `postexec`, and `aroundexec` (\lifecycle{experimental}) hooks in addition to a serializer.
#' This is useful for graphics device serializers which need a `preexec` and `postexec` hook to capture the graphics output.
#'
#' `preexec` and `postexec` hooks happened directly before and after a route is executed.
#' These hooks are specific to a single [PlumberEndpoint]'s route calculation.
#'
#' @param serializer Serializer method to be used.  This method should already have its initialization arguments applied.
#' @param preexec_hook Function to be run directly before a [PlumberEndpoint] calls its route method.
#' @param postexec_hook Function to be run directly after a [PlumberEndpoint] calls its route method.
#' @param aroundexec_hook Function to be run around a [PlumberEndpoint] call. Must handle a `.next` argument to continue execution. \lifecycle{experimental}
#'
#' @export
#' @examples
#' # The definition of `serializer_device` returns
#' # * a `serializer_content_type` serializer
#' # * `aroundexec` hook
#' print(serializer_device)
endpoint_serializer <- function(
  serializer,
  preexec_hook = NULL,
  postexec_hook = NULL,
  aroundexec_hook = NULL
) {

  stopifnot(is.function(serializer))
  structure(
    list(
      serializer = serializer,
      preexec_hook = preexec_hook,
      postexec_hook = postexec_hook,
      aroundexec_hook = aroundexec_hook
    ),
    class = "plumber_endpoint_serializer"
  )
}

self_set_serializer <- function(self, serializer) {
  if (inherits(serializer, "plumber_endpoint_serializer")) {
    self$serializer <- serializer$serializer
    if (!is.null(serializer$preexec_hook)) {
      self$registerHook("preexec", serializer$preexec_hook)
    }
    if (!is.null(serializer$postexec_hook)) {
      self$registerHook("postexec", serializer$postexec_hook)
    }
    if (!is.null(serializer$aroundexec_hook)) {
      self$registerHook("aroundexec", serializer$aroundexec_hook)
    }
  } else {
    self$serializer <- serializer
  }

  self
}


#' @describeIn serializers Helper method to create graphics device serializers, such as [serializer_png()]. See also: [endpoint_serializer()]
#' @param dev_on Function to turn on a graphics device.
#' The graphics device `dev_on` function will receive any arguments supplied to the serializer in addition to `filename`.
#' `filename` points to the temporary file name that should be used when saving content.
#' @param dev_off Function to turn off the graphics device. Defaults to [grDevices::dev.off()]
#' @export
serializer_device <- function(type, dev_on, dev_off = grDevices::dev.off) {

  stopifnot(!missing(type))

  stopifnot(!missing(dev_on))
  stopifnot(is.function(dev_on))
  stopifnot(length(formals(dev_on)) > 0)
  if (!any(c("filename", "...") %in% names(formals(dev_on)))) {
    stop("`dev_on` must contain an arugment called `filename` or have `...`")
  }

  stopifnot(is.function(dev_off))

  endpoint_serializer(
    serializer = serializer_content_type(type),
    aroundexec_hook = function(..., .next) {
      tmpfile <- tempfile()

      dev_on(filename = tmpfile)
      device_id <- dev.cur()
      dev_off_once <- once(function() dev_off(device_id))

      success <- function(value) {
        dev_off_once()
        if (!file.exists(tmpfile)) {
          stop("The device output file is missing. Did you produce an image?", call. = FALSE)
        }
        con <- file(tmpfile, "rb")
        on.exit({close(con)}, add = TRUE)
        img <- readBin(con, "raw", file.info(tmpfile)$size)
        img
      }

      cleanup <- function() {
        dev_off_once()
        on.exit({
          # works even if the file does not exist
          unlink(tmpfile)
        }, add = TRUE)
      }

      # This is just a flag to ensure we don't cleanup() if the .next(...) is
      # async.
      async <- FALSE

      on.exit({
        if (!async) {
          cleanup()
        }
      }, add = TRUE)

      result <- promises::with_promise_domain(createGraphicsDevicePromiseDomain(device_id), {
        .next(...)
      })
      if (is.promising(result)) {
        async <- TRUE
        result %>% then(success) %>% finally(cleanup)
      } else {
        success(result)
      }
    }
  )
}

#' @describeIn serializers JPEG image serializer. See also: [grDevices::jpeg()]
#' @export
serializer_jpeg <- function(..., type = "image/jpeg") {
  serializer_device(
    type = type,
    dev_on = function(filename) {
      grDevices::jpeg(filename, ...)
    }
  )
}
#' @describeIn serializers PNG image serializer. See also: [grDevices::png()]
#' @export
serializer_png <- function(..., type = "image/png") {
  serializer_device(
    type = type,
    dev_on = function(filename) {
      grDevices::png(filename, ...)
    }
  )
}
#' @describeIn serializers SVG image serializer. See also: [grDevices::svg()]
#' @export
serializer_svg <- function(..., type = "image/svg+xml") {
  serializer_device(
    type = type,
    dev_on = function(filename) {
      grDevices::svg(filename, ...)
    }
  )
}
#' @describeIn serializers BMP image serializer. See also: [grDevices::bmp()]
#' @export
serializer_bmp <- function(..., type = "image/bmp") {
  serializer_device(
    type = type,
    dev_on = function(filename) {
      grDevices::bmp(filename, ...)
    }
  )
}
#' @describeIn serializers TIFF image serializer. See also: [grDevices::tiff()]
#' @export
serializer_tiff <- function(..., type = "image/tiff") {
  serializer_device(
    type = type,
    dev_on = function(filename) {
      grDevices::tiff(filename, ...)
    }
  )
}
#' @describeIn serializers PDF image serializer. See also: [grDevices::pdf()]
#' @export
serializer_pdf <- function(..., type = "application/pdf") {
  serializer_device(
    type = type,
    dev_on = function(filename) {
      grDevices::pdf(filename, ...)
    }
  )
}






add_serializers_onLoad <- function() {
  register_serializer("null",        serializer_identity)
  register_serializer("contentType", serializer_content_type)

  # octet-stream
  register_serializer("octet", serializer_octet)

  # html
  register_serializer("html", serializer_html)

  # objects
  register_serializer("json",        serializer_json)
  register_serializer("unboxedJSON", serializer_unboxed_json)
  register_serializer("rds",         serializer_rds)
  register_serializer("csv",         serializer_csv)
  register_serializer("tsv",         serializer_tsv)
  register_serializer("feather",     serializer_feather)
  register_serializer("parquet",     serializer_parquet)
  register_serializer("yaml",        serializer_yaml)
  register_serializer("geojson",     serializer_geojson)

  # text
  register_serializer("text",   serializer_text)
  register_serializer("format", serializer_format)
  register_serializer("print",  serializer_print)
  register_serializer("cat",    serializer_cat)

  # htmlwidget
  register_serializer("htmlwidget", serializer_htmlwidget)

  # devices
  register_serializer("device", serializer_device)
  register_serializer("jpeg",   serializer_jpeg)
  register_serializer("png",    serializer_png)
  register_serializer("svg",    serializer_svg)
  register_serializer("bmp",    serializer_bmp)
  register_serializer("tiff",   serializer_tiff)
  register_serializer("pdf",    serializer_pdf)


  ## Do not register until implemented
  # register_serializer("xml", serializer_xml)
}

# From https://github.com/rstudio/shiny/blob/ee13087d575d378fba2fae34664725dc7452df2d/R/imageutils.R
#' @importFrom grDevices dev.set dev.cur
# if the graphics device was not maintained for the promises, two promises could break how graphics are recorded
## Bad
## * Open p1 device
## * Open p2 device
## * Draw p1 in p2 device
## * Draw p2 in p2 device
## * Close cur device (p2)
## * Close cur device (p1) (which is empty)
##
## Good (and implemented using the function below)
## * Open p1 device in p1
## * Open p2 device in p2
## * Draw p1 in p1 device in p1
## * Draw p2 in p2 device in p2
## * Close p1 device in p1
## * Close p2 device in p2
createGraphicsDevicePromiseDomain <- function(which = dev.cur()) {
  force(which)

  if (which < 2) {
    stop(
      "`createGraphicsDevicePromiseDomain()` was called without opening a device first.",
      " Open a new graphics device before calling."
    )
  }

  promises::new_promise_domain(
    wrapOnFulfilled = function(onFulfilled) {
      force(onFulfilled)
      function(...) {
        old <- dev.cur()
        dev_set(which)
        on.exit(dev_set(old))

        onFulfilled(...)
      }
    },
    wrapOnRejected = function(onRejected) {
      force(onRejected)
      function(...) {
        old <- dev.cur()
        dev_set(which)
        on.exit(dev_set(old))

        onRejected(...)
      }
    },
    wrapSync = function(expr) {
      old <- dev.cur()
      dev_set(which)
      on.exit(dev_set(old))

      force(expr)
    }
  )
}

dev_set <- function(i) {
  # make sure to not open a new device when calling `dev.set(1)`
  if (i > 1) {
    dev.set(i)
  } else {
    warning("Can not set `.Device` to the `null device`. Was `dev.off()` manually called?")
  }
}

context("Serializer")

test_that("Responses returned directly aren't serialized", {
  res <- PlumberResponse$new("")

  r <- pr(test_path("files/router.R"))
  val <- r$serve(make_req("GET", "/response123"), res)
  expect_equal(val$body, NULL)
  expect_equal(val$status, 123)
  val <- r$serve(make_req("GET", "/response200"), res)
  expect_equal(val$body, "overridden")
  expect_equal(val$status, 200)
})

test_that("JSON is the default serializer", {
  res <- PlumberResponse$new()

  r <- pr(test_path("files/router.R"))
  expect_equal(r$serve(make_req("GET", "/"), res)$headers$`Content-Type`, "application/json")
})

test_that("Overridden serializers apply on filters and endpoints", {

  with_tmp_serializers({
    customSer <- function(){
      function(val, req, res, errorHandler){
        list(status=201L, headers=list(), body="CUSTOM")
      }
    }
    register_serializer("custom", customSer)

    custom2Ser <- function(){
      function(val, req, res, errorHandler){
        list(status=201L, headers=list(), body="CUSTOM2")
      }
    }
    register_serializer("custom2", custom2Ser)

    register_serializer("customOneArg", function(single){
      function(val, req, res, errorHandler){
        list(status=200L, headers=list(), body=list(val=val, arg=single))
      }
    })

    register_serializer("customMultiArg", function(first, second, third){
      function(val, req, res, errorHandler){
        list(status=200L, headers=list(),
            body=list(val=val, args=list(first=first, second=second, third=third)))
      }
    })

    r <- pr(test_path("files/serializer.R"))
    res <- PlumberResponse$new("json")
    expect_equal(r$serve(make_req("GET", "/"), res)$body, "CUSTOM")
    expect_equal(res$serializer, customSer())

    res <- PlumberResponse$new("json")
    expect_equal(r$serve(make_req("GET", "/filter-catch"), res)$body, "CUSTOM2")
    expect_equal(res$serializer, custom2Ser())

    req <- make_req("GET", "/something")
    res <- PlumberResponse$new(customSer())
    expect_equal(r$serve(req, res)$body, "CUSTOM")

    # have the filter set the serializer
    req <- make_req("GET", "/something", qs = "type=json")
    res <- PlumberResponse$new(NULL) # default to something that doesn't exist!
    expect_equal(r$serve(req, res)$body, jsonlite::toJSON(4))

    res <- PlumberResponse$new("json")
    expect_equal(r$serve(make_req("GET", "/another"), res)$body, "CUSTOM3")

    res <- PlumberResponse$new()
    body <- r$serve(make_req("GET", "/single-arg-ser"), res)$body
    expect_equal(body$val, "COA")
    expect_equal(body$arg, "hi there")

    res <- PlumberResponse$new()
    body <- r$serve(make_req("GET", "/multi-arg-ser"), res)$body
    expect_equal(body$val, "MAS")
    expect_equal(body$args$first, "A")
    expect_equal(body$args$second, 8)
    expect_equal(body$args$third, 4.3)
  })

  # due to covr changing some code, the return answer is very strange
  # the tests below should be skipped on covr
  testthat::skip_on_covr()

  res <- PlumberResponse$new()
  expect_equal(r$serve(make_req("GET", "/short-json"), res)$body, jsonlite::toJSON("JSON"))
  expect_equal_functions(res$serializer, serializer_json())

  res <- PlumberResponse$new()
  expect_equal(r$serve(make_req("GET", "/short-html"), res)$body, "HTML")
  expect_equal_functions(res$serializer, serializer_html())
})

# test_that("Overridding the attached serializer in code works.", {
#
# })

test_that("Redundant serializers fail", {

  with_tmp_serializers({
    register_serializer("inc", function(val, req, res, errorHandler){
      list(status=201L, headers=list(), body="CUSTOM2")
    })
    expect_error(pr(test_path("files/serializer-redundant.R")), regexp="Multiple @serializers")
  })

})

test_that("Empty serializers fail", {
  expect_error(pr(test_path("files/serializer-empty.R")), regexp="No @serializer specified")
})

test_that("Non-existant serializers fail", {
  expect_error(pr(test_path("files/serializer-nonexistent.R")), regexp="No such @serializer")
})


test_that("serializer_identity serializes properly", {
  v <- "<html><h1>Hi!</h1></html>"
  val <- serializer_identity()(v, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$body, v)
})

test_that("serializer_identity errors call error handler", {
  errors <- 0
  errHandler <- function(req, res, err){
    errors <<- errors + 1
  }

  expect_equal(errors, 0)
  serializer_identity()(parse(stop("I crash")), list(), PlumberResponse$new("json"), errorHandler = errHandler)
  expect_equal(errors, 1)
})

test_that("Error handler is passed to serializer", {

  with_tmp_serializers({
    res <- PlumberResponse$new()

    register_serializer("failingSer", function() {
      function(val, req, res, errorHandler){
        errorHandler(req, res, simpleError("A serializer error"))
      }
    })

    r <- pr(test_path("files/serializer-error.R"))

    r$setErrorHandler(function(req, res, err) {
      msg <- paste("Handled:", conditionMessage(err))
      stop(msg)
    })

    expect_error(r$serve(make_req("GET", "/fail"), res),
                regexp = "Handled: A serializer error")
  })
})

context("JSON serializer")

test_that("JSON serializes properly", {
  l <- list(a=1, b=2, c="hi")
  val <- serializer_json()(l, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "application/json")
  expect_equal(val$body, jsonlite::toJSON(l))

  l <- list(a=1, b=2, c="hi", na=NA)
  val <- serializer_json()(l, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "application/json")
  expect_equal(val$body, jsonlite::toJSON(l, na = 'null'))

  l <- list(a=1, b=2, c="hi", na=NA)
  val <- serializer_json(na = 'string')(l, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "application/json")
  expect_equal(val$body, jsonlite::toJSON(l, na = 'string'))
})

test_that("Errors call error handler", {
  errors <- 0
  errHandler <- function(req, res, err){
    errors <<- errors + 1
  }

  expect_equal(errors, 0)
  serializer_json()(parse(text="hi"), list(), PlumberResponse$new("json"), errorHandler = errHandler)
  expect_equal(errors, 1)
})

context("Unboxed JSON serializer")

test_that("Unboxed JSON serializes properly", {
  l <- list(a=1, b=2, c="hi")
  val <- serializer_unboxed_json()(l, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "application/json")
  expect_equal(val$body, jsonlite::toJSON(l, auto_unbox = TRUE))


  l <- list(a=1, b=2, c="hi", na=NA)
  val <- serializer_unboxed_json()(l, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "application/json")
  expect_equal(val$body, jsonlite::toJSON(l, auto_unbox = TRUE, na = 'null'))

  l <- list(a=1, b=2, c="hi", na=NA)
  val <- serializer_unboxed_json(na = 'string')(l, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "application/json")
  expect_equal(val$body, jsonlite::toJSON(l,  auto_unbox = TRUE, na = 'string'))
})

test_that("Unboxed JSON errors call error handler", {
  errors <- 0
  errHandler <- function(req, res, err){
    errors <<- errors + 1
  }

  expect_equal(errors, 0)
  serializer_unboxed_json()(parse(text="hi"), list(), PlumberResponse$new("json"), errorHandler = errHandler)
  expect_equal(errors, 1)
})

context("block parsing")

test_that("trimws works", {
  expect_equal(trimws("    hi there \t  "), "hi there")
  expect_equal(trimws("hi there\t"), "hi there")
  expect_equal(trimws("hi "), "hi")
})

test_that("plumbBlock works", {
  lines <- c(
    "#* Plumber comment not reached",
    "NULL",
    "#* Plumber comments",
    "#* Plumber description",
    "#* second line",
    "",
    "  ",
    "# Normal comments",
    "#' @get /",
    "#' @post /",
    "#' @filter test",
    "#' @serializer json")
  b <- plumbBlock(length(lines), lines)
  expect_length(b$paths, 2)
  # Paths order follow original code
  expect_equal(b$paths[[1]], list(verb="GET", path="/"))
  expect_equal(b$paths[[2]], list(verb="POST", path="/"))
  expect_equal(b$filter, "test")
  expect_equal(b$comments, "Plumber comments")
  expect_equal(b$description, "Plumber description\nsecond line")

  # due to covr changing some code, the return answer is very strange
  # the tests below should be skipped on covr
  testthat::skip_on_covr()

  expect_equal_functions(b$serializer, serializer_json())
})

test_that("plumbBlock images", {
  lines <- c("#'@png")
  expect_warning({
    b <- plumbBlock(length(lines), lines)
  })
  expect_equal(b$serializer, serializer_png())

  lines <- c("#'@jpeg")
  expect_warning({
    b <- plumbBlock(length(lines), lines)
  })
  expect_equal(b$serializer, serializer_jpeg())
  lines <- c("#'@png")
  expect_warning({
    b <- plumbBlock(length(lines), lines)
  })
  expect_equal(b$serializer, serializer_png())

  # Whitespace is fine
  lines <- c("#' @jpeg    \t ")
  expect_warning({
    b <- plumbBlock(length(lines), lines)
  })
  expect_equal(b$serializer, serializer_jpeg())

  # No whitespace is fine
  lines <- c("#' @jpeg(w=1)")
  expect_warning({
    b <- plumbBlock(length(lines), lines)
  })
  expect_equal(b$serializer, serializer_jpeg(w=1), check.environment=FALSE)

  # Additional chars after name don't count as image tags
  lines <- c("#' @jpegs")
  expect_error(
    expect_warning({plumbBlock(length(lines), lines)}),
    "Supplemental arguments to the serializer"
  )

  # Properly formatted arguments work
  lines <- c("#'@jpeg (width=100)")
  expect_warning({
    b <- plumbBlock(length(lines), lines)
  })
  expect_equal(b$serializer, serializer_jpeg(width = 100), check.environment=FALSE)

  # Ill-formatted arguments return a meaningful error
  lines <- c("#'@jpeg width=100")
  expect_error(
    expect_warning({plumbBlock(length(lines), lines)}),
    "Supplemental arguments to the serializer"
  )
})

test_that("Block can't be multiple mutually exclusive things", {

  srcref <- c(3,4)
  addE <- function(){ fail() }
  addF <- function(){ fail() }
  addA <- function(){ fail() }
  expect_error({
    evaluateBlock(srcref, c("#' @get /", "#' @assets /", "function(){}"),
                  function(){}, addE, addF, addA)
  }, "A single function can only be")

})

test_that("Block can't contain duplicate tags", {
  lines <- c("#* @tag test",
            "#* @tag test")
  expect_error(plumbBlock(length(lines), lines), "Duplicate tag specified.")
})

test_that("@json parameters work", {

  # due to covr changing some code, the return answer is very strange
  testthat::skip_on_covr()

  plumb_block_check <- function(lines) {
    if (grepl("@json", lines, fixed = TRUE)) {
      expect_warning(
        plumbBlock(length(lines), lines)
      )
    } else {
      plumbBlock(length(lines), lines)
    }
  }
  expect_block_fn <- function(lines, fn) {
    expect_equal_functions(plumb_block_check(lines)$serializer, fn)
  }
  expect_block_error <- function(lines, ...) {
    expect_error({
      plumb_block_check(lines)
    }, ...)
  }

  expect_block_fn("#' @serializer json", serializer_json())
  expect_block_fn("#' @json", serializer_json())
  expect_block_fn("#' @json()", serializer_json())
  expect_block_fn("#' @serializer unboxedJSON", serializer_unboxed_json())

  expect_block_fn("#' @serializer json list(na = 'string')", serializer_json(na = 'string'))
  expect_block_fn("#' @json(na = 'string')", serializer_json(na = 'string'))

  expect_block_fn("#* @serializer unboxedJSON list(na = \"string\")", serializer_unboxed_json(na = 'string'))
  expect_block_fn("#' @json(auto_unbox = TRUE, na = 'string')", serializer_json(auto_unbox = TRUE, na = 'string'))


  expect_block_fn("#' @json (    auto_unbox = TRUE, na = 'string'    )", serializer_json(auto_unbox = TRUE, na = 'string'))
  expect_block_fn("#' @json (auto_unbox          =       TRUE    ,      na      =      'string'   )             ", serializer_json(auto_unbox = TRUE, na = 'string'))
  expect_block_fn("#' @serializer json list   (      auto_unbox          =       TRUE    ,      na      =      'string'   )             ", serializer_json(auto_unbox = TRUE, na = 'string'))


  expect_block_error("#' @serializer json list(na = 'string'", "unexpected end of input")
  expect_block_error("#' @json(na = 'string'", "must be surrounded by parentheses")
  expect_block_error("#' @json (na = 'string'", "must be surrounded by parentheses")
  expect_block_error("#' @json ( na = 'string'", "must be surrounded by parentheses")
  expect_block_error("#' @json na = 'string')", "must be surrounded by parentheses")
  expect_block_error("#' @json list(na = 'string')", "must be surrounded by parentheses")

})


test_that("@html parameters produce an error", {
  # due to covr changing some code, the return answer is very strange
  testthat::skip_on_covr()

  plumb_block_check <- function(lines) {
    if (grepl("@html", lines, fixed = TRUE)) {
      expect_warning(
        plumbBlock(length(lines), lines)
      )
    } else {
      plumbBlock(length(lines), lines)
    }
  }
  expect_block_fn <- function(lines, fn) {
    expect_equal_functions(plumb_block_check(lines)$serializer, fn)
  }
  expect_block_error <- function(lines, ...) {
    expect_error({
      plumb_block_check(lines)
    }, ...)
  }

  expect_block_fn("#' @serializer html", serializer_html())

  expect_block_fn("#' @serializer html list()", serializer_html())
  expect_block_fn("#' @serializer html list(         )", serializer_html())
  expect_block_fn("#' @serializer html list     (         )     ", serializer_html())

  expect_block_fn("#' @html", serializer_html())
  expect_block_fn("#' @html()", serializer_html())
  expect_block_fn("#' @html ()", serializer_html())
  expect_block_fn("#' @html ( )", serializer_html())
  expect_block_fn("#' @html ( ) ", serializer_html())
  expect_block_fn("#' @html         (       )       ", serializer_html())

  expect_block_error("#' @serializer html list(key = \"val\")", "unused argument")
  expect_block_error("#' @html(key = \"val\")", "unused argument")
  expect_block_error("#' @html (key = \"val\")", "unused argument")

  expect_block_error("#' @html (key = \"val\")", "unused argument")
})

test_that("@parser parameters produce an error or not", {
  # due to covr changing some code, the return answer is very strange
  testthat::skip_on_covr()

  expect_block_parser <- function(lines, fn) {
    b <- plumbBlock(length(lines), lines)
    expect_equal(b$parsers, fn)
  }
  expect_block_error <- function(lines, ...) {
    expect_error({
      plumbBlock(length(lines), lines)
    }, ...)
  }


  expected <- list(octet = list())
  expect_block_parser("#' @parser octet",  expected)

  expect_block_parser("#' @parser octet list()", expected)
  expect_block_parser("#' @parser octet list(         )", expected)
  expect_block_parser("#' @parser octet list     (         )     ", expected)

  expect_error({
    evaluateBlock(
      srcref = 3, # which evaluates to line 2
      file = c("#' @get /test", "#' @parser octet list(key = \"val\")"),
      expr = as.expression(substitute(identity)),
      envir = new.env(),
      addEndpoint = function(a, b, ...) { stop("should not reach here")},
      addFilter = as.null,
      pr = pr()
    )
  }, "unused argument (key = \"val\")", fixed = TRUE)
})
test_that("Plumbing block use the right environment", {
  expect_silent(plumb(test_path("files/plumb-envir.R")))
})


test_that("device serializers produce a structure", {
  # due to covr changing some code, the return answer is very strange
  testthat::skip_on_covr()

  expect_s3_block <- function(lines, serializer_fn) {
    block <- plumbBlock(length(lines), lines)
    expect_s3_class(block$serializer, "plumber_endpoint_serializer")
    serializer_info <- serializer_fn()
    expect_equal(block$serializer$serializer, serializer_info$serializer)
    expect_true(is.function(block$serializer$serializer))
    expect_equal(block$serializer$preexec_hook, serializer_info$preexec_hook)
    expect_equal(block$serializer$postexec_hook, serializer_info$postexec_hook)
  }

  expect_s3_block("#' @serializer jpeg", serializer_jpeg)
  expect_s3_block("#' @serializer png", serializer_png)
  expect_s3_block("#' @serializer svg", serializer_svg)
  expect_s3_block("#' @serializer bmp", serializer_bmp)
  expect_s3_block("#' @serializer tiff", serializer_tiff)
  expect_s3_block("#' @serializer pdf", serializer_pdf)
})

test_that("Tags can contains space", {
  lines <- c("#* @tag 'test space'",
             "#* @tag \"test space2\"")
  expect_equal(plumbBlock(length(lines), lines)$tags, c("test space", "test space2"))
})

test_that("single character tag and response", {
  lines <- c(
    "#' @tag a",
    "#' @response 2 b",
    "#' @response 4 b c")
  b <- plumbBlock(length(lines), lines)
  expect_equal(b$tags, "a")
  expect_equal(b$responses, list(`2` = list(description = "b"), `4` = list(description = "b c")))
})

test_that("block respect original order of lines for comments, tags and responses", {
  lines <- c(
    "#' @tag aaa",
    "#' @tag bbb",
    "#' comments first line",
    "#' comments second line",
    "#' comments third line",
    "#' @response 200 ok",
    "#' @response 404 not ok")
  b <- plumbBlock(length(lines), lines)
  expect_equal(b$comments, "comments first line")
  expect_equal(b$description, "comments second line\ncomments third line")
  expect_equal(b$tags, c("aaa", "bbb"))
  expect_equal(b$responses, list(`200`=list(description="ok"), `404` = list(description="not ok")))
})

test_that("srcref values are set while plumbing from a file", {

  root <- plumb_api("plumber", "01-append")
  endpt <- root$endpoints[[1]][[1]]
  expect_s3_class(endpt$srcref, "srcref")

  root_with_no_srcref <- pr() %>% pr_get("/", force)
  endpt_with_no_srcref <- root_with_no_srcref$endpoints[[1]][[1]]
  expect_equal(endpt_with_no_srcref$srcref, NULL)
})


# TODO: more testing around filter, assets, endpoint, etc.

test_that("as_attachment shape", {
  val <- as_attachment(iris, "name.csv")
  expect_s3_class(val, "plumber_attachment")
  expect_equal(val$value, iris)
  expect_equal(val$filename, "name.csv")
})

test_that("disposition is not added unless content type exists", {
  val <- as_attachment("<html><h1>Hi!</h1></html>")
  res <- serializer_headers(list(), identity)(val, list(), PlumberResponse$new(), stop)
  expect_null(res$headers$`Content-Disposition`)
})

test_that("disposition is not added unless content type exists", {
  val <- as_attachment("<html><h1>Hi!</h1></html>")
  res <- serializer_content_type("text/html; charset=UTF-8", identity)(val, list(), PlumberResponse$new(), stop)
  expect_equal(res$headers$`Content-Type`, "text/html; charset=UTF-8")
  expect_equal(res$headers$`Content-Disposition`, "attachment")
  expect_equal(res$body, val$value)
  expect_equal(res$status, 200L)
})

test_that("disposition file can not contain quotes", {
  val <- as_attachment("<html><h1>Hi!</h1></html>", filename = "quote\".csv")
  expect_error({
    serializer_content_type("text/html; charset=UTF-8", identity)(val, list(), PlumberResponse$new(), stop)
  })
  val <- as_attachment("<html><h1>Hi!</h1></html>", filename = "quote'.csv")
  expect_error({
    serializer_content_type("text/html; charset=UTF-8", identity)(val, list(), PlumberResponse$new(), stop)
  })
})

test_that("disposition is added", {
  val <- as_attachment("<html><h1>Hi!</h1></html>", filename = "dispo.html")
  res <- serializer_content_type("text/html; charset=UTF-8", identity)(val, list(), PlumberResponse$new(), stop)
  expect_equal(res$headers$`Content-Type`, "text/html; charset=UTF-8")
  expect_equal(res$headers$`Content-Disposition`,"attachment; filename=\"dispo.html\"")
  expect_equal(res$body, val$value)
  expect_equal(res$status, 200L)
})

test_that("disposition is not overwritten", {
  val <- as_attachment("<html><h1>Hi!</h1></html>", filename = "dispo.html")
  res <- serializer_headers(
    list(
      "Content-Type" = "text/html; charset=UTF-8",
      "Content-Disposition" = "attachment; filename=\"original.html\""
    ),
    identity
  )(val, list(), PlumberResponse$new(), stop)

  expect_equal(res$headers$`Content-Type`, "text/html; charset=UTF-8")
  expect_equal(res$headers$`Content-Disposition`,"attachment; filename=\"original.html\"")
  expect_equal(res$body, val$value)
  expect_equal(res$status, 200L)
})

test_that("octet serializes raw objects properly", {

  content <- charToRaw("Beware of bugs in the above code; I have only proved it correct, not tried it.")

  val <- serializer_octet()(content, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "application/octet-stream")
  expect_equal(val$body, content)
  expect_equal(rawToChar(val$body), rawToChar(content))
})

test_that("octet throws on non-raw objects", {

  content <- "Beware of bugs in the above code; I have only proved it correct, not tried it."

  expect_error(
    serializer_octet()(
      content,
      list(),
      PlumberResponse$new(),
      function(req, res, err) stop(err)
    ),
    "received non-raw data",
  )
})

context("rds serializer")

test_that("rds serializes properly", {
  v <- iris[0,]
  attr(v, "origin") <- iris
  val <- serializer_rds()(v, list(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "application/rds")
  expect_equal(val$body, serialize(v, NULL, ascii = FALSE, version = 2L))
  expect_equal(unserialize(val$body), v)
})

#' @keywords internal
#' @title Hookable
Hookable <- R6Class(
  "Hookable",
  public=list(
    #' @description Register a hook on a router
    #' @param stage a character string.
    #' @param handler a hook function.
    registerHook = function(stage, handler) {
      stopifnot(is.function(handler))
      private$hooks[[stage]] <- c(private$hooks[[stage]], handler)
    },
    #' @description Register hooks on a router
    #' @param handlers a named list of hook functions.
    registerHooks = function(handlers){
      for (i in 1:length(handlers)){
        stage <- names(handlers)[i]
        h <- handlers[[i]]

        self$registerHook(stage, h)
      }
    }
  ), private=list(
    hooks = list( ),

    # Because we're passing in a `value` argument here, `runHooks` will return either the
    # unmodified `value` argument back, or will allow one or more hooks to modify the value,
    # in which case the modified value will be returned. Hooks declare that they intend to
    # modify the value by accepting a parameter named `value`, in which case their returned
    # value will be used as the updated value.
    runHooks = function(stage, args) {
      if (missing(args)) {
        args <- list()
      }

      stageHooks <- private$hooks[[stage]]
      if (length(stageHooks) == 0) {
        # if there is nothing to execute, return early
        return(args$value)
      }

      runSteps(
        NULL,
        errorHandlerStep = stop,
        append(
          unlist(lapply(stageHooks, function(stageHook) {
            stageHookArgs <- list()
            list(
              function(...) {
                stageHookArgs <<- getRelevantArgs(args, func = stageHook)
              },
              function(...) {
                do.call(stageHook, stageHookArgs) #TODO: envir=private$envir?
              },
              # `do.call` could return a promise. Wait for it's return value
              # if "value" exists in the original args, overwrite it for futher execution
              function(value, ...) {
                if ("value" %in% names(stageHookArgs)) {
                  # Special case, retain the returned value from the hook
                  # and pass it in as the value for the next handler.
                  # Ultimately, return value from this function
                  args$value <<- value
                }
                NULL
              }
            )
          })),
          list(
            function(...) {
              # Return the value as passed in or as explcitly modified by one or more hooks.
              return(args$value)
            }
          )
        )
      )
    },
    # Some stages (aroundexec) use a continuation passing style instead of callback style.
    # https://en.wikipedia.org/wiki/Continuation-passing_style
    # https://expressjs.com/en/guide/using-middleware.html
    runHooksAround = function(stage, args = list(), .next) {
      stageHooks <- private$hooks[[stage]]

      # Execute the specified (i) hook. If i == 0, execute the .next continuation.
      execHook <- function(i, hookArgs) {
        if (i == 0) {
          do.call(.next, getRelevantArgs(hookArgs, func = .next))
        } else {
          # Need to pass continuation to the hook
          hookArgs <- c(hookArgs, .next = nextHook(i - 1))
          stageHook <- stageHooks[[i]]
          do.call(stageHook, getRelevantArgs(hookArgs, func = stageHook))
        }
      }

      nextHook <- function(i) {
        function(...) {
          execHook(i, list(...))
        }
      }

      execHook(i = length(stageHooks), args)
    }
  )
)

context("Hookable")

test_that("simple extension works", {
  simpleHook <- R6Class(
    "simplehook",
    inherit = Hookable,
    public = list(
      exercise = function(hookName, args){
        private$runHooks(hookName, args)
      }
    )
  )

  events <- NULL
  s <- simpleHook$new()
  s$registerHook("abcd", function(arg1){
    events <<- c(events, arg1)
  })

  s$registerHook("defg", function(arg2){
    events <<- c(events, arg2)
  })

  expect_null(events)

  s$exercise("abcd", list(arg1="arg1here", unused="test"))
  expect_equal(events, "arg1here")

  s$exercise("defg", list(arg2="arg2here"))
  expect_equal(events, c("arg1here", "arg2here"))
})

test_that("registerHooks works", {
  simpleHook <- R6Class(
    "simplehook",
    inherit = Hookable,
    public = list(
      exercise = function(hookName, args){
        private$runHooks(hookName, args)
      }
    )
  )

  events <- NULL
  s <- simpleHook$new()
  s$registerHooks(list(
    defg = function(arg2){
      events <<- c(events, arg2)
    }, abcd = function(arg1){
      events <<- c(events, arg1)
    }))

  expect_null(events)

  s$exercise("abcd", list(arg1="arg1here", unused="test"))
  expect_equal(events, "arg1here")

  s$exercise("defg", list(arg2="arg2here"))
  expect_equal(events, c("arg1here", "arg2here"))
})

test_that("overloading extension works", {
  simpleHook <- R6Class(
    "simplehook",
    inherit = Hookable,
    public = list(
      registerHook = function(hook=c("hook1", "hook2"), fun){
        hook <- match.arg(hook)
        super$registerHook(hook, fun)
      },
      exercise = function(hookName, args){
        private$runHooks(hookName, args)
      }
    )
  )

  s <- simpleHook$new()
  expect_error(s$registerHook("abcd", function(arg1){
      events <<- c(events, arg1)
    })
  )

  events <- NULL
  s$registerHook("hook2", function(){
    events <<- c(events, "hook2!")
  })

  expect_null(events)

  # Works with missing args
  s$exercise("hook2")
  expect_equal(events, "hook2!")
})

test_that("value forwarding works across stacked hooks", {
  simpleHook <- R6Class(
    "simplehook",
    inherit = Hookable,
    public = list(
      exercise = function(hookName, args){
        private$runHooks(hookName, args)
      }
    )
  )

  increment <- function(value){
    value + 1
  }

  s <- simpleHook$new()
  s$registerHook("valForward", increment)

  # Register the same hook twice. Should see the value increment by two each call since the
  # values are getting forwarded from the first hook into the second.
  s$registerHook("valForward", increment)

  s$registerHook("noVal", function(){
    # Doesn't take a value parameter, so shouldn't be treated specially for value handling.
    return(3)
  })

  v <- s$exercise("valForward", list(value=0))
  expect_equal(v, 2)
  v <- s$exercise("noVal", list(value=0))
  expect_equal(v, 0)
})

# Shiny has been tried and tested on different encodings
# adoption from https://github.com/rstudio/shiny/blob/6ede0194c6e52d18be3ec13514defd8e0f22a3c1/R/utils.R#L1459-L1532

isWindows <- function() .Platform$OS.type == 'windows'

# assume file is encoded in UTF-8, but warn against BOM
checkEncoding <- function(file) {
  # skip *nix because its locale is normally UTF-8 based (e.g. en_US.UTF-8), and
  # *nix users have to make a conscious effort to save a file with an encoding
  # that is not UTF-8; if they choose to do so, we cannot do much about it
  # except sitting back and seeing them punished after they choose to escape a
  # world of consistency (falling back to getOption('encoding') will not help
  # because native.enc is also normally UTF-8 based on *nix)
  if (!isWindows()) return('UTF-8')
  size <- file.info(file)[, 'size']
  if (is.na(size)) stop('Cannot access the file ', file)
  # BOM is 3 bytes, so if the file contains BOM, it must be at least 3 bytes
  if (size < 3L) return('UTF-8')

  # check if there is a BOM character: this is also skipped on *nix, because R
  # on *nix simply ignores this meaningless character if present, but it hurts
  # on Windows
  if (identical(charToRaw(readChar(file, 3L, TRUE)), charToRaw('\UFEFF'))) {
    warning('You should not include the Byte Order Mark (BOM) in ', file, '. ',
            'Please re-save it in UTF-8 without BOM. See ',
            'http://shiny.rstudio.com/articles/unicode.html for more info.')
    return('UTF-8-BOM')
  }
  x <- readChar(file, size, useBytes = TRUE)
  if (is.na(iconv(x, 'UTF-8', 'UTF-8'))) {
    warning('The input file ', file, ' does not seem to be encoded in UTF8')
  }
  'UTF-8'
}

# read a file using UTF-8 and (on Windows) convert to native encoding if possible
readUTF8 <- function(file) {
  enc <- checkEncoding(file)
  file <- base::file(file, encoding = enc)
  on.exit(close(file), add = TRUE)
  x <- enc2utf8(readLines(file, warn = FALSE))
  tryNativeEncoding(x)
}

# if the UTF-8 string can be represented in the native encoding, use native encoding
tryNativeEncoding <- function(string) {
  if (!isWindows()) return(string)
  string2 <- enc2native(string)
  if (identical(enc2utf8(string2), string)) string2 else string
}

# similarly, try to source() a file with UTF-8
sourceUTF8 <- function(file, envir = globalenv()) {
  exprs <- parseUTF8(file)

  eval(exprs, envir)
}


parseUTF8 <- function(file) {
  lines <- readUTF8(file)
  enc <- if (any(Encoding(lines) == 'UTF-8')) 'UTF-8' else 'unknown'
  src <- srcfilecopy(file, lines, isFile = TRUE)  # source reference info

  # oddly, parse(file) does not work when file contains multibyte chars that
  # **can** be encoded natively on Windows (might be a bug in base R); we
  # rewrite the source code in a natively encoded temp file and parse it in this
  # case (the source reference is still pointed to the original file, though)
  if (isWindows() && enc == 'unknown') {
    file <- tempfile(); on.exit(unlink(file), add = TRUE)
    writeLines(lines, file)
  }
  
  # keep the source locations within the file while parsing
  exprs <- try(parse(file, keep.source = TRUE, srcfile = src, encoding = enc))
  if (inherits(exprs, "try-error")) {
    stop("Error sourcing ", file)
  }

  exprs
}

PlumberResponse <- R6Class(
  "PlumberResponse",
  public = list(
    initialize = function(serializer=serializer_json()){
      self$serializer <- serializer
    },
    status = 200L,
    body = NULL,
    headers = stats::setNames(list(), character()),
    serializer = NULL,
    setHeader = function(name, value){
      he <- list()
      he[[name]] <- value
      self$headers <- c(self$headers, he)
    },
    toResponse = function(){
      h <- self$headers

      body <- self$body

      if (is.character(body)) {
        charset <- get_character_set(h$HTTP_CONTENT_TYPE)
        Encoding(body) <- charset
      }

      if (self$status %in% c(100:199, 204, 304)) {
        body <- NULL
      }

      list(
        status = self$status,
        headers = h,
        body = body
      )
    },
    # TODO if name and value are a vector of same length, call set cookie many times
    setCookie = function(name, value, path = NULL, expiration = FALSE, http = FALSE, secure = FALSE, same_site = FALSE) {
      self$setHeader("Set-Cookie", cookieToStr(
        name = name,
        value = value,
        path = path,
        expiration = expiration,
        http = http,
        secure = secure,
        same_site = same_site
      ))
    },
    removeCookie = function(name, path = NULL, http = FALSE, secure = FALSE, same_site = FALSE, ...) {
      self$setHeader("Set-Cookie", removeCookieStr(
        name = name,
        path = path,
        http = http,
        secure = secure,
        same_site = same_site
      ))
    }
  )
)

removeCookieStr <- function(name, path = NULL, http = FALSE, secure = FALSE, same_site = FALSE) {
  str <- paste0(name, "=; ")
  if (!missing(path) && is.character(path) && nchar(path) > 0) {
    str <- paste0(str, "Path=", path, "; ")
  }
  if (!missing(http) && http){
    str <- paste0(str, "HttpOnly; ")
  }
  if (!missing(secure) && secure){
    str <- paste0(str, "Secure; ")
  }

  if (!missing(same_site) && is.character(same_site)) {
    str <- paste0(str, "SameSite=", same_site, "; ")
  }

  str <- paste0(str, "Expires=Thu, 01 Jan 1970 00:00:00 GMT")
  str
}

#' @noRd
cookieToStr <- function(
  name,
  value,
  path = NULL,
  expiration = FALSE,
  http = FALSE,
  secure = FALSE,
  same_site = FALSE,
  now = Sys.time() # used for testing. Should not be used in regular code.
){
  val <- httpuv::encodeURIComponent(as.character(value))
  str <- paste0(name, "=", val, "; ")

  if (!missing(path) && is.character(path) && nchar(path) > 0) {
    str <- paste0(str, "Path=", path, "; ")
  }

  if (!missing(http) && http){
    str <- paste0(str, "HttpOnly; ")
  }

  if (!missing(secure) && secure){
    str <- paste0(str, "Secure; ")
  }

  if (!missing(same_site) && is.character(same_site)) {
    str <- paste0(str, "SameSite=", same_site, "; ")
  }

  if (!missing(expiration)){
    if (is.numeric(expiration)){
      # Number of seconds in the future
      expy <- now + expiration
      expyStr <- format(expy, format="%a, %e %b %Y %T", tz="GMT", usetz=TRUE)

      str <- paste0(str, "Expires= ", expyStr, "; ")
      str <- paste0(str, "Max-Age= ", expiration, "; ")
    } else if (inherits(expiration, "POSIXt")){
      seconds <- difftime(expiration, now, units="secs")
      # TODO: DRY
      expyStr <- format(expiration, format="%a, %e %b %Y %T", tz="GMT", usetz=TRUE)
      str <- paste0(str, "Expires= ", expyStr, "; ")
      str <- paste0(str, "Max-Age= ", as.integer(seconds), "; ")
    } # interpret all other values as session cookies.
  }

  # Trim last '; '
  ret <- substr(str, 0, nchar(str)-2)

  # double check size limit isn't reached
  cookieByteSize <- length(charToRaw(ret))
  # http://browsercookielimits.squawky.net/#limits
  #  typical browsers support 4096.  A couple safari based browsers max out at 4093.
  if (cookieByteSize > 4093) {
    warning(
      "Cookie being saved is too large",
      " (> 4093 bytes; found ", cookieByteSize, " bytes).",
      " Browsers may not support such a large value.\n",
      "Consider using a database and only storing minimal information.")
  }

  ret
}

context("Responses")

test_that("response properly sets basic cookies", {
  res <- PlumberResponse$new()
  res$setCookie("abc", "two words")
  head <- res$toResponse()$headers
  expect_equal(head[["Set-Cookie"]], "abc=two%20words")
})

test_that("response sets non-char cookies", {
  res <- PlumberResponse$new()
  res$setCookie("abc", 123)
  head <- res$toResponse()$headers
  expect_equal(head[["Set-Cookie"]], "abc=123")
})

test_that("can set multiple same-named headers", {
  res <- PlumberResponse$new()
  res$setHeader("head", "test")
  res$setHeader("head", "another")

  test <- FALSE
  another <- FALSE

  pres <- res$toResponse()
  for (i in 1:length(pres$headers)){
    n <- names(pres$headers)[i]
    if (n == "head"){
      if (pres$headers[[i]] == "test"){
        test <- TRUE
      } else if (pres$headers[[i]] == "another"){
        another <- TRUE
      }
    }
  }

  expect_true(test)
  expect_true(another)
})

test_that("response properly sets cookies with multiple options", {
  res <- PlumberResponse$new()
  res$setCookie("abc", "two words", http=TRUE, secure=TRUE, same_site="None")
  head <- res$toResponse()$headers
  expect_equal(head[["Set-Cookie"]], "abc=two%20words; HttpOnly; Secure; SameSite=None")
})

test_that("Body on HTTP responses which forbid it", {
  res <- PlumberResponse$new()
  res$body <- "Hello"
  res$status <- 204
  expect_null(res$toResponse()$body)
})

#* @filter testFun
function(){ }

#* @filter testFun2
function(){ }

#* @filter testFun3
function(){ }

#* @preempt testFun
#* @get /
function(){

}

#* @preempt test
#* Excluded
function(){

}

#* @preempt testFun2
#* @get /
#*

function(){

}

#*@preempt testFun3
#*@post /
function(){

}

runStepsIfForwarding <- function(initialValue, errorHandlerStep, steps) {
  runStepsUntil(
    initialValue = initialValue,
    errorHandlerStep = errorHandlerStep,
    conditionFn = function(value) {
      !hasForwarded()
    },
    steps = steps
  )
}


withWarn1 <- function(expr) {
  oldWarn <- options("warn")[[1]]
  on.exit({
    options(warn = oldWarn)
  })
  options(warn = 1)

  force(expr)
}


runSteps <- function(initialValue, errorHandlerStep, steps) {
  runStepsUntil(
    initialValue,
    errorHandlerStep,
    function(x) {
      # always continue
      FALSE
    },
    steps
  )
}

#' @import promises
runStepsUntil <- function(initialValue, errorHandlerStep, conditionFn, steps) {
  x <- initialValue

  stepCount <- length(steps)
  nextStepPos <- 1L

  # Goals of `runStep`:
  ## - loop through functions that produce results synchronously (avoid recursion where possible)
  ## - If a result is a promise like object, wait for it's result, then recursively call `runStep`
  # Current ways to exit `runStep`:
  ## 1. no more functions to execute
  ## 2. nextStep returns a promise, which causes recursive (async) calculations to be returned
  ## 3. w/ synchronous result, `conditionFn` returns TRUE. Return result
  # Error Handling
  ## A tryCatch is provided around the initial runStep to capture sync errors
  ## A %...!% (promise catch) is added to the end of the async recursive call.
  ##  This captures both async errors and future sync errors.
  runStep <- function() {

    while (TRUE) {

      if (nextStepPos > stepCount) {
        return(x)
      }

      nextStep <- steps[[nextStepPos]]
      nextStepPos <<- nextStepPos + 1L

      # if NULL is provided, skip it
      if (is.null(nextStep)) {
        next
      }
      if (!is.function(nextStep)) {
        stop("runStepsUntil only knows how to handle functions or NULL values. Received something of classes: ", paste0(class(nextStep), collapse = ", "))
      }

      result <- nextStep(x)
      if (is.promising(result)) {
        # async
        resultOfNextStep <-
          result %...>%
          (function(value) {
            # message("WAITED!")
            x <<- value
            if (conditionFn(x)) {
              return(x) # all done, return
            } else {
              return(runStep()) # must recurse
            }
          }) %...!% errorHandlerStep

        return(resultOfNextStep)
      }

      # sync
      x <<- result
      if (conditionFn(x)) {
        return(x) # all done, return
      }
      # else, loop through like normal
      next
    } # end while (TRUE)
    # never reached
  }

  # catch sync error and return it
  tryCatch(runStep(), error = errorHandlerStep)
}





# steps_one <- list(
#   function(x) {
#     print("Got here 1")
#     1 # not a promise
#   },
#   function(x) {
#     stopifnot(x == 1)
#     print("Got here 2")
#     promise_resolve(2)
#   },
#   function(x) {
#     stopifnot(x == 2)
#     print("Got here 3")
#     3
#   },
#   function(x) {
#     stopifnot(x == 3)
#     print("Got here 4")
#     4
#   }
# )
#
# steps_two <- list(
#   function(x) {
#     runStep(steps_one, x)
#   },
#   function(x) {
#     print("Got here 1a")
#     1 # not a promise
#   },
#   function(x) {
#     print("Got here 2a")
#     2 # not a promise
#   }
# )
#
# runStep(steps_two); print("END")


currentExecName <- "currentExec"
getCurrentExec <- function() {
  .globals[[currentExecName]]
}
withCurrentExecDomain <- function(req, res, expr) {
  # Create a new environment for this particular request/response
  execEnv <- new.env(parent = emptyenv())
  execEnv$req <- req
  execEnv$res <- res

  domain <- createVarPromiseDomain(.globals, currentExecName, execEnv)
  promises::with_promise_domain(domain, expr)
}



# From Shiny.
# Creates a promise domain that always ensures `env[[name]] == value` when
# any code is being run in this domain.
createVarPromiseDomain <- function(env, name, value) {
  force(env)
  force(name)
  force(value)

  promises::new_promise_domain(
    wrapOnFulfilled = function(onFulfilled) {
      function(...) {
        orig <- env[[name]]
        env[[name]] <- value
        on.exit(env[[name]] <- orig)

        onFulfilled(...)
      }
    },
    wrapOnRejected = function(onRejected) {
      function(...) {
        orig <- env[[name]]
        env[[name]] <- value
        on.exit(env[[name]] <- orig)

        onRejected(...)
      }
    },
    wrapSync = function(expr) {
      orig <- env[[name]]
      env[[name]] <- value
      on.exit(env[[name]] <- orig)

      force(expr)
    }
  )
}

forward_class <- "plumber_forward"
#' Forward Request to The Next Handler
#'
#' This function is used when a filter is done processing a request and wishes
#' to pass control off to the next handler in the chain. If this is not called
#' by a filter, the assumption is that the filter fully handled the request
#' itself and no other filters or endpoints should be evaluated for this
#' request.
#' @export
forward <- function() {
  exec <- getCurrentExec()
  exec$forward <- TRUE
}
hasForwarded <- function() {
  getCurrentExec()$forward
}
resetForward <- function() {
  exec <- getCurrentExec()
  exec$forward <- FALSE
}

#' plumber step R6 class
#' @description an object representing a step in the lifecycle of the treatment
#' of a request by a plumber router.
PlumberStep <- R6Class(
  "PlumberStep",
  inherit=Hookable,
  public = list(
    #' @field srcref from step block
    srcref = NULL,
    #' @field lines lines from step block
    lines = NA,
    #' @field serializer step serializer function
    serializer = NULL,
    #' @description Create a new [PlumberStep()] object
    #' @param expr step expr
    #' @param envir step environment
    #' @param lines step block
    #' @param serializer step serializer
    #' @param srcref `srcref` attribute from block
    #' @return A new `PlumberStep` object
    initialize = function(expr, envir, lines, serializer, srcref){
      private$expr <- expr
      if (is.expression(expr)) {
        private$func <- eval(expr, envir)
      } else {
        private$func <- expr
      }
      throw_if_func_is_not_a_function(private$func)
      private$envir <- envir

      if (!missing(srcref)) {
        self$srcref <- srcref
      }
      if (!missing(lines)){
        self$lines <- lines
      }

      if (!missing(serializer)){
        self$serializer <- serializer
      }
    },
    #' @description step execution function
    #' @param req,res Request and response objects created by a Plumber request
    exec = function(req, res) {
      hookEnv <- new.env(parent = emptyenv())

      # use a function as items could possibly be added to `req$args` in each step
      args_for_formal_matching <- function() {
        args <- c(
          # add in `req`, `res` as they have been removed from `req$args`
          list(req = req, res = res),
          req$args
        )
      }

      preexecStep <- function(...) {
        private$runHooks("preexec", c(list(data = hookEnv), args_for_formal_matching()))
      }
      execStep <- function(...) {
        private$runHooksAround("aroundexec", args_for_formal_matching(), .next = function(...) {
          relevant_args <- getRelevantArgs(list(...), func = private$func)
          do.call(private$func, relevant_args, envir = private$envir)
        })
      }
      postexecStep <- function(value, ...) {
        private$runHooks("postexec", c(list(data = hookEnv, value = value), args_for_formal_matching()))
      }
      runSteps(
        NULL,
        function(error) {
          # rethrow error
          stop(error)
        },
        list(
          preexecStep,
          execStep,
          postexecStep
        )
      )
    },
    #' @description step hook registration method
    #' @param stage a character string.
    #' @param handler a step handler function.
    registerHook = function(stage=c("preexec", "postexec", "aroundexec"), handler){
      stage <- match.arg(stage)
      super$registerHook(stage, handler)
    }
  ),
  private = list(
    envir = NA,
    expr = NA,
    func = NA
  )
)

# @param positional list with names where they were provided.
getRelevantArgs <- function(args, func) {
  # Extract the names of the arguments this function supports.
  fargs <- names(formals(func))

  if (length(fargs) == 0) {
    # no matches
    return(list())
  }

  # fast return
  # also works with unnamed arguments
  if (identical(fargs, "...")) {
    return(args)
  }

  # If only req and res are found in function definition...
  # Only call using the first matches of req and res.
  if (all(fargs %in% c("req", "res"))) {
    ret <- list()
    # using `$` will retrieve the 1st occurance of req,res
    # args$req <- req is used within `Plumber$route()`
    if ("req" %in% fargs) {
      ret$req <- args$req
    }
    if ("res" %in% fargs) {
      ret$res <- args$res
    }
    return(ret)
  }

  # The remaining code MUST work with unnamed arguments
  # If there is no `...`, then the unnamed args will not be in `fargs` and will be removed
  # If there is `...`, then the unnamed args will not be in `fargs` and will be passed through

  if (!("..." %in% fargs)) {
    # Use the named arguments that match, drop the rest.
    args <- args[names(args) %in% fargs]
  }

  # dedupe matched formals
  arg_names <- names(args)
  is_farg <- arg_names %in% fargs
  # keep only the first matched formal argument (and all other non-`farg` params)
  args <- args[(is_farg & !duplicated(arg_names)) | (!is_farg)]

  args
}

#' Plumber Endpoint
#'
#' Defines a terminal handler in a Plumber router.
#'
#' @export
PlumberEndpoint <- R6Class(
  "PlumberEndpoint",
  inherit = PlumberStep,
  public = list(
    #' @field verbs a character vector. http methods. For historical reasons we have
    #' to accept multiple verbs for a single path. Now it's simpler to just parse
    #' each separate verb/path into its own endpoint, so we just do that.
    verbs = NA,
    #' @field path a character string. endpoint path
    path = NA,
    #' @field comments endpoint comments
    comments = NA,
    #' @field description endpoint description
    description = NA,
    #' @field responses endpoint responses
    responses = NA,
    #' @description retrieve endpoint typed parameters
    getTypedParams = function() {
      data.frame(
        name = private$regex$names,
        type = private$regex$types,
        isArray = private$regex$areArrays,
        stringsAsFactors = FALSE
      )
    },
    #' @field params endpoint parameters
    params = NA,
    #' @field tags endpoint tags
    tags = NA,
    #' @field parsers step allowed parsers
    parsers = NULL,
    #' @description ability to serve request
    #' @param req a request object
    #' @return a logical. `TRUE` when endpoint can serve request.
    canServe = function(req) {
      req$REQUEST_METHOD %in% self$verbs && self$matchesPath(req$PATH_INFO)
    },
    #' @description determines if route matches requested path
    #' @param path a url path
    #' @return a logical. `TRUE` when endpoint matches the requested path.
    matchesPath = function(path) {
      !is.na(stri_match_first_regex(path, private$regex$regex)[1,1])
    },
    #' @description Create a new `PlumberEndpoint` object
    #' @param verbs Endpoint verb Ex: `"GET"`, `"POST"`
    #' @param path Endpoint path. Ex: `"/index.html"`, `"/foo/bar/baz"`
    #' @param expr Endpoint function or expression that evaluates to a function.
    #' @param envir Endpoint environment
    #' @param serializer Endpoint serializer. Ex: [serializer_json()]
    #' @template pr_setParsers__parsers
    #' @param srcref `srcref` attribute from block
    #' @param lines Endpoint block
    #' @param params Endpoint params
    #' @param comments,description,responses,tags Values to be used within the OpenAPI Spec
    #' @details Parameters values are obtained from parsing blocks of lines in a plumber file.
    #' They can also be provided manually for historical reasons.
    #' @return A new `PlumberEndpoint` object
    initialize = function(verbs, path, expr, envir, serializer, parsers, lines, params, comments, description, responses, tags, srcref) {

      self$verbs <- verbs

      private$expr <- expr
      if (is.expression(expr)){
        private$func <- eval(expr, envir)
      } else {
        private$func <- expr
      }
      throw_if_func_is_not_a_function(private$func)
      private$envir <- envir

      self$setPath(path)

      if (!missing(serializer) && !is.null(serializer)){
        self_set_serializer(self, serializer)
      }

      if (!missing(parsers) && !is.null(parsers)) {
        self$parsers <- make_parser(parsers)
      }
      if (!missing(srcref)) {
        self$srcref <- srcref
      }
      if (!missing(lines)){
        self$lines <- lines
      }
      if (!missing(params)){
        self$params <- params
      }
      if (!missing(comments)){
        self$comments <- comments
      }
      if (!missing(description)){
        self$description <- description
      }
      if (!missing(responses)){
        self$responses <- responses
      }
      if(!missing(tags) && !is.null(tags)){
        # make sure we box tags in json using I()
        # single tags should be converted to json as:
        # tags: ["tagName"] and not tags: "tagName"
        self$tags <- I(tags)
      }
    },
    #' @description retrieve endpoint path parameters
    #' @param path endpoint path
    getPathParams = function(path){
      extractPathParams(private$regex, path)
    },
    #' @description retrieve endpoint function
    getFunc = function() {
      private$func
    },
    #' @description retrieve endpoint expression parameters
    getFuncParams = function() {
      getArgsMetadata(private$func)
    },
    #' @description retrieve endpoint defined parameters
    getEndpointParams = function() {
      if (length(self$params) == 0) {
        return(list())
      }
      if (isTRUE(is.na(self$params))) {
        return(list())
      }
      self$params
    },
    # It would not make sense to have `$getPath()` and deprecate `$path`
    #' @description Updates `$path` with a sanitized `path` and updates the internal path meta-data
    #' @param path Path to set `$path`. If missing a beginning slash, one will be added.
    setPath = function(path) {
      if (substr(path, 1,1) != "/") {
        path <- paste0("/", path)
      }
      # private$func is not updated after initialization
      self$path <- path
      private$regex <- createPathRegex(path, self$getFuncParams())
      path
    }
  ),
  private = list(
    regex = NULL
  )
)

PlumberFilter <- R6Class(
  "PlumberFilter",
  inherit = PlumberStep,
  public = list(
    name = NA,
    initialize = function(name, expr, envir, serializer, lines, srcref){
      self$name <- name
      private$expr <- expr
      if (is.expression(expr)){
        private$func <- eval(expr, envir)
      } else {
        private$func <- expr
      }
      throw_if_func_is_not_a_function(private$func)
      private$envir <- envir

      if (!missing(serializer)){
        self_set_serializer(self, serializer)
      }
      if (!missing(srcref)) {
        self$srcref <- srcref
      }
      if (!missing(lines)){
        self$lines <- lines
      }
    }
  )
)


throw_if_func_is_not_a_function <- function(func) {
  if(!is.function(func)) {
    stop("`expr` did not evaluate to a function")
  }
}

context("Endpoints")

test_that("Endpoints execute in their environment", {
  env <- new.env()
  assign("a", 5, envir=env)

  foo <- parse(text="foo <- function(){ a }")

  r <- PlumberEndpoint$new('verb', 'path', foo, env, 1:2)
  expect_equal(r$exec(req = list(), res = 2), 5)
})

test_that("Missing lines are ok", {
  expect_silent({
    PlumberEndpoint$new('verb', 'path', { function() { 1 }}, new.env(parent = globalenv()))
  })
})

test_that("Endpoints are exec'able with named arguments.", {
  foo <- parse(text="foo <- function(x){ x + 1 }")
  r <- PlumberEndpoint$new('verb', 'path', foo, new.env(parent = globalenv()))
  expect_equal(r$exec(req = list(args = list(x = 3)), res = 20), 4)
})

test_that("Unnamed arguments do not throw an error", {
  foo <- parse(text="foo <- function(){ -1 }")
  r <- PlumberEndpoint$new('verb', 'path', foo, new.env(parent = globalenv()))
  expect_equal(r$exec(req = list(args = list(3)), res = 100), -1)

  foo <- parse(text="foo <- function(req, res, x, ...){ x + sum(1, ...) }")
  r <- PlumberEndpoint$new('verb', 'path', foo, new.env(parent = globalenv()))
  expect_equal(r$exec(req = list(args = list(x = 3, 1)), res = 100), 5)
})

test_that("Ellipses allow any named args through", {
  foo <- parse(text="function(req, res, ...){ sum(unlist(list(...))) }")
  r <- PlumberEndpoint$new('verb', 'path', foo, new.env(parent = globalenv()))
  expect_equal(r$exec(req = list(args = list(a=1, b=2, c=3)), res = 20), 6)

  for ( txt in c(
      "", # with no req or res formals
      "req, res, "
  )) {
    foo <- parse(text=paste0("function(", txt, "...){ ret <- list(...); ret[!names(ret) %in% c('req', 'res')] }"))
    r <- PlumberEndpoint$new('verb', 'path', foo, new.env(parent = globalenv()))
    expect_equal(r$exec(req = list(args = list(a="aa", b="ba")), res = 2), list(a="aa", b="ba"))
    expect_equal(r$exec(req = list(args = list(a="aa1", a="aa2", b = "ba")), res = 2), list(a="aa1", a="aa2", b = "ba"))

    foo <- parse(text=paste0("function(", txt, "a, ...){ ret <- list(a = a, ...); ret[!names(ret) %in% c('req', 'res')] }"))
    r <- PlumberEndpoint$new('verb', 'path', foo, new.env(parent = globalenv()))
    expect_equal(r$exec(req = list(args = list(a="aa1", a="aa2", b = "ba")), res = 2), list(a = "aa1", b = "ba"))
  }
})

test_that("If only req and res are defined, duplicated arguments do not throw an error", {
  full_req <- list(args = list(req = 1, req = 2, res = 3, res = 4))
  full_res <- list(args = list(res = 1, res = 2, res = 3, res = 4))
  foo <- parse(text="function(req){ req }")
  r <- PlumberEndpoint$new('verb', 'path', foo, new.env(parent = globalenv()))
  expect_equal(r$exec(req = full_req, res = full_res), full_req)

  foo <- parse(text="function(res){ res }")
  r <- PlumberEndpoint$new('verb', 'path', foo, new.env(parent = globalenv()))
  expect_equal(r$exec(req = full_req, res = full_res), full_res)

  foo <- parse(text="function(req, res){ list(req, res) }")
  r <- PlumberEndpoint$new('verb', 'path', foo, new.env(parent = globalenv()))
  expect_equal(r$exec(req = full_req, res = full_res), list(full_req, full_res))

  foo <- parse(text="function(req, res, ...){ list(req, res, ...) }")
  r <- PlumberEndpoint$new('verb', 'path', foo, new.env(parent = globalenv()))
  expect_equal(r$exec(req = full_req, res = full_res), list(full_req, full_res))
})

test_that("Programmatic endpoints work", {
  r <- Plumber$new()

  serializer <- "ser"
  expr <- expression(function(req, res){res$setHeader("expr", TRUE)})

  r$handle("GET", "/", expr, "queryString", serializer)
  expect_equal(length(r$endpoints), 1)

  end <- r$endpoints[[1]][[1]]
  expect_equal(end$verbs, "GET")
  expect_equal(end$path, "/")
  expect_equal(names(r$endpoints)[1], "queryString")
  expect_equal(end$serializer, serializer)

  res <- PlumberResponse$new()
  req <- list()
  end$exec(req=req, res=res)

  h <- res$headers
  expect_true(h$expr)
})

test_that("Programmatic endpoints with functions work", {
  r <- Plumber$new()

  expr <- function(req, res){res$setHeader("expr", TRUE)}

  r$handle("GET", "/", expr)
  expect_equal(length(r$endpoints), 1)

  end <- r$endpoints[[1]][[1]]
  expect_equal(end$verbs, "GET")
  expect_equal(end$path, "/")

  res <- PlumberResponse$new()
  req <- list()
  end$exec(req=req, res=res)

  h <- res$headers
  expect_true(h$expr)
})

test_that("Path rewrites correctly", {
  foo_pr <- pr() %>%
    pr_get("/foo", function() "foo")

  foo <- foo_pr$endpoints[[1]][[1]]

  expect_true("foo" %in% names(foo_pr$routes))
  expect_true(foo$matchesPath("/foo"))
  expect_false(foo$matchesPath("/bar"))

  foo$setPath("bar")

  expect_false("foo" %in% names(foo_pr$routes))
  expect_false(foo$matchesPath("/foo"))
  expect_true(foo$matchesPath("/bar"))
})

# TODO: delete once we require R 3.3.0
trimws <- function(string){
  string <- gsub("^\\s+", "", string)
  gsub("\\s+$", "", string)
}

stopOnLine <- function(lineNum, line, msg){
  stop("Error on line #", lineNum, ": '", line, "' - ", msg)
}

#' @param lineNum The line number just above the function we're documenting
#' @param file A character vector representing all the lines in the file
#' @param envir An environment where to evaluate parsed expressions
#' @importFrom utils tail
#' @noRd
plumbBlock <- function(lineNum, file, envir = parent.frame()){
  paths <- NULL
  preempt <- NULL
  filter <- NULL
  serializer <- NULL
  parsers <- NULL
  assets <- NULL
  params <- NULL
  comments <- NULL
  responses <- NULL
  tags <- NULL
  routerModifier <- NULL
  while (lineNum > 0 && (stri_detect_regex(file[lineNum], pattern="^#['\\*]?|^\\s*$") || stri_trim_both(file[lineNum]) == "")){

    line <- file[lineNum]

    # If the line does not start with a plumber tag `#*` or `#'`, continue to next line
    if (!stri_detect_regex(line, pattern="^#['\\*]")) {
      lineNum <- lineNum - 1
      next
    }

    epMat <- stri_match(line, regex="^#['\\*]\\s*@(get|put|post|use|delete|head|options|patch)(\\s+(.*)$)?")
    if (!is.na(epMat[1,2])){
      p <- stri_trim_both(epMat[1,4])

      if (is.na(p) || p == ""){
        stopOnLine(lineNum, line, "No path specified.")
      }

      if (is.null(paths)){
        paths <- list()
      }

      paths[[length(paths)+1]] <- list(verb = enumerateVerbs(epMat[1,2]), path = p)
    }

    filterMat <- stri_match(line, regex="^#['\\*]\\s*@filter(\\s+(.*)$)?")
    if (!is.na(filterMat[1,1])){
      f <- stri_trim_both(filterMat[1,3])

      if (is.na(f) || f == ""){
        stopOnLine(lineNum, line, "No @filter name specified.")
      }

      if (!is.null(filter)){
        # Must have already assigned.
        stopOnLine(lineNum, line, "Multiple @filters specified for one function.")
      }

      filter <- f
    }

    preemptMat <- stri_match(line, regex="^#['\\*]\\s*@preempt(\\s+(.*)\\s*$)?")
    if (!is.na(preemptMat[1,1])){
      p <- stri_trim_both(preemptMat[1,3])
      if (is.na(p) || p == ""){
        stopOnLine(lineNum, line, "No @preempt specified")
      }
      if (!is.null(preempt)){
        # Must have already assigned.
        stopOnLine(lineNum, line, "Multiple @preempts specified for one function.")
      }
      preempt <- p
    }

    assetsMat <- stri_match(line, regex="^#['\\*]\\s*@assets(\\s+(\\S*)(\\s+(\\S+))?\\s*)?$")
    if (!is.na(assetsMat[1,1])){
      dir <- stri_trim_both(assetsMat[1,3])
      if (is.na(dir) || dir == ""){
        stopOnLine(lineNum, line, "No directory specified for @assets")
      }
      prefixPath <- stri_trim_both(assetsMat[1,5])
      if (is.na(prefixPath) || prefixPath == ""){
        prefixPath <- "/public"
      }
      if (!is.null(assets)){
        # Must have already assigned.
        stopOnLine(lineNum, line, "Multiple @assets specified for one entity.")
      }
      assets <- list(dir=dir, path=prefixPath)
    }

    serMat <- stri_match(line, regex="^#['\\*]\\s*@serializer(\\s+([^\\s]+)\\s*(.*)\\s*$)?")
    if (!is.na(serMat[1,1])){
      s <- stri_trim_both(serMat[1,3])
      if (is.na(s) || s == ""){
        stopOnLine(lineNum, line, "No @serializer specified")
      }
      if (!is.null(serializer)){
        # Must have already assigned.
        stopOnLine(lineNum, line, "Multiple @serializers specified for one function.")
      }

      if (!(s %in% registered_serializers())){
        stopOnLine(lineNum, line, paste0("No such @serializer registered: ", s))
      }

      ser <- get_registered_serializer(s)

      if (!is.na(serMat[1, 4]) && serMat[1,4] != ""){
        # We have an arg to pass in to the serializer
        argList <- tryCatch({
          eval(parse(text=serMat[1,4]), envir)
        }, error = function(e) {
          stopOnLine(lineNum, line, e)
        })
      } else {
        argList <- list()
      }
      tryCatch({
        serializer <- do.call(ser, argList)
      }, error = function(e) {
        stopOnLine(lineNum, line, paste0("Error creating serializer: ", s, "\n", e))
      })

    }

    shortSerMat <- stri_match(line, regex="^#['\\*]\\s*@(json|html|jpeg|png|svg)(.*)$")
    if (!is.na(shortSerMat[1,2])) {
      s <- stri_trim_both(shortSerMat[1,2])
      .Deprecated(msg = paste0(
        "Plumber tag `#* @", s, "` is deprecated.\n",
        "Use `#* @serializer ", s, "` instead."
      ))
      if (!is.null(serializer)){
        # Must have already assigned.
        stopOnLine(lineNum, line, "Multiple @serializers specified for one function (shorthand serializers like @json count, too).")
      }

      if (!is.na(s) && !(s %in% registered_serializers())){
        stopOnLine(lineNum, line, paste0("No such @serializer registered: ", s))
      }
      shortSerAttr <- trimws(shortSerMat[1,3])
      if(!identical(shortSerAttr, "") && !grepl("^\\(.*\\)$", shortSerAttr)){
        stopOnLine(lineNum, line, paste0("Supplemental arguments to the serializer must be surrounded by parentheses, as in `#' @", s, "(na='null')`"))
      }

      if (shortSerAttr != "") {
        # We have an arg to pass in to the serializer
        argList <- tryCatch({
          eval(parse(text=paste0("list", shortSerAttr)), envir)
        }, error = function(e) {
          stopOnLine(lineNum, line, e)
        })
      } else {
        argList <- list()
      }
      tryCatch({
        serializer <- do.call(get_registered_serializer(s), argList)
      }, error = function(e) {
        stopOnLine(lineNum, line, paste0("Error creating serializer: ", s, "\n", e))
      })
    }

    parsersMat <- stri_match(line, regex="^#['\\*]\\s*@parser(\\s+([^\\s]+)\\s*(.*)\\s*$)?")
    if (!is.na(parsersMat[1,1])){
      parser_alias <- stri_trim_both(parsersMat[1,3])
      if (is.na(parser_alias) || parser_alias == ""){
        stopOnLine(lineNum, line, "No @parser specified")
      }

      if (!parser_alias %in% registered_parsers()){
        stopOnLine(lineNum, line, paste0("No such @parser registered: ", parser_alias))
      }

      if (!is.na(parsersMat[1, 4]) && parsersMat[1,4] != ""){
        # We have an arg to pass in to the parser
        arg_list <- tryCatch({
          eval(parse(text=parsersMat[1,4]), envir)
        }, error = function(e) {
          stopOnLine(lineNum, line, e)
        })
      } else {
        arg_list <- list()
      }
      if (is.null(parsers)) {
        parsers <- list()
      }
      parsers[[parser_alias]] <- arg_list
    }

    responseMat <- stri_match(line, regex="^#['\\*]\\s*@response\\s+(\\w+)\\s+(\\S.*)\\s*$")
    if (!is.na(responseMat[1,1])){
      resp <- list()
      resp[[responseMat[1,2]]] <- list(description=responseMat[1,3])
      responses <- c(responses, resp)
    }

    paramMat <- stri_match(line, regex="^#['\\*]\\s*@param(\\s+([^\\s:]+):?([^\\s*]+)?(\\*)?(?:\\s+(.*))?\\s*$)?")
    if (!is.na(paramMat[1,2])){
      name <- paramMat[1,3]
      if (is.na(name)){
        stopOnLine(lineNum, line, "No parameter specified.")
      }
      plumberType <- stri_replace_all(paramMat[1,4], "$1", regex = "^\\[([^\\]]*)\\]$")
      apiType <- plumberToApiType(plumberType)
      isArray <- stri_detect_regex(paramMat[1,4], "^\\[[^\\]]*\\]$")
      isArray[is.na(isArray)] <- defaultIsArray
      required <- identical(paramMat[1,5], "*")

      params[[name]] <- list(desc=paramMat[1,6], type=apiType, required=required, isArray=isArray)
    }

    tagMat <- stri_match(line, regex="^#['\\*]\\s*@tag\\s+(\"[^\"]+\"|'[^']+'|\\S+)\\s*")
    if (!is.na(tagMat[1,1])){
      t <- stri_trim_both(tagMat[1,2], pattern = "[[\\P{Wspace}]-[\"']]")
      if (is.na(t) || t == ""){
        stopOnLine(lineNum, line, "No tag specified.")
      }
      if (t %in% tags){
        stopOnLine(lineNum, line, "Duplicate tag specified.")
      }
      tags <- c(tags, t)
    }

    commentMat <- stri_match(line, regex="^#['\\*]\\s*([^@\\s].*$)")
    if (!is.na(commentMat[1,2])){
      comments <- c(comments, trimws(commentMat[1,2]))
    }

    routerModifierMat <- stri_match(line, regex="^#['\\*]\\s*@plumber")
    if (!is.na(routerModifierMat[1,1])) {
      routerModifier <- TRUE
    }

    lineNum <- lineNum - 1
  }

  list(
    paths = rev(paths),
    preempt = preempt,
    filter = filter,
    serializer = serializer,
    parsers = rev(parsers),
    assets = assets,
    params = rev(params),
    comments = tail(comments, 1),
    description = paste0(rev(comments)[-1], collapse = "\n"),
    responses = rev(responses),
    tags = rev(tags),
    routerModifier = routerModifier
  )
}

#' Evaluate and activate a "block" of code found in a plumber API file.
#' @noRd
evaluateBlock <- function(srcref, file, expr, envir, addEndpoint, addFilter, pr) {
  lines <- srcref[c(1,3)]
  lineNum <- lines[1] - 1

  block <- plumbBlock(lineNum, file, envir)

  if (sum(!is.null(block$filter), !is.null(block$paths), !is.null(block$assets), !is.null(block$routerModifier)) > 1){
    stopOnLine(lineNum, file[lineNum], "A single function can only be a filter, an API endpoint, an asset or a Plumber object modifier (@filter AND @get, @post, @assets, @plumber, etc.)")
  }

  # ALL if statements possibilities must eventually call eval(expr, envir)
  if (!is.null(block$paths)){
    lapply(block$paths, function(p) {
      ep <- PlumberEndpoint$new(
        verbs = p$verb,
        path = p$path,
        expr = expr,
        envir = envir,
        serializer = block$serializer,
        parsers = block$parsers,
        lines = lines,
        srcref = srcref,
        params = block$params,
        comments = block$comments,
        description = block$description,
        responses = block$responses,
        tags = block$tags
      )

      addEndpoint(ep, block$preempt)
    })
  } else if (!is.null(block$filter)){
    filter <- PlumberFilter$new(block$filter, expr, envir, block$serializer,
      lines = lines, srcref = srcref)
    addFilter(filter)

  } else if (!is.null(block$assets)){
    path <- block$assets$path

    # Leading slash
    if (substr(path, 1,1) != "/"){
      path <- paste0("/", path)
    }

    stat <- PlumberStatic$new(block$assets$dir, expr)
    pr$mount(path, stat)

  } else if (!is.null(block$routerModifier)) {
    if (is.expression(expr)) {
      func <- tryCatch({
        eval(expr, envir)
      }, error = function(e) {
        stopOnLine(lineNum, file[lineNum], e)
      })

      if (!is.function(func)) {
        stopOnLine(lineNum, file[lineNum], "Invalid expression for @plumber tag, please use the form `function(pr) { }`.")
      }

      # Use time as an ID
      # Creating a new pr() takes at least a millisec
      pr_id <- as.numeric(Sys.time())
      pr$flags$id <- pr_id
      on.exit({
        pr$flags$id <- NULL
      }, add = TRUE)

      # process func
      func_ret <- func(pr)

      if (inherits(func_ret, "Plumber")) {
        func_ret_id <- func_ret$flags$id
        if (!identical(pr_id, func_ret_id)) {
          stopOnLine(lineNum, file[lineNum], "Plumber object returned is not the same as the one provided.")
        }
      }
    }
  } else {
    tryCatch({
      eval(expr, envir)
    }, error = function(e) {
      stopOnLine(lineNum, file[lineNum], e)
    })
  }

  # Show that we are returning nothing
  # Only modify pr in place
  return()
}

# calculate all OpenAPI Type information at once and use created information throughout package
apiTypesInfo <- list()
plumberToApiTypeMap <- list()
defaultApiType <- structure("string", default = TRUE)
defaultIsArray <- structure(FALSE, default = TRUE)

add_api_info_onLoad <- function() {
  addApiInfo <- function(apiType, plumberTypes,
                         regex = NULL, converter = NULL,
                         format = NULL,
                         location = NULL,
                         realType = NULL) {
    apiTypesInfo[[apiType]] <<-
      list(
        regex = regex,
        converter = converter,
        format = format,
        location = location,
        realType = realType,
        # Q: Do we need to safe guard against special characters, such as `,`?
        # https://github.com/rstudio/plumber/pull/532#discussion_r439584727
        # A: https://swagger.io/docs/specification/serialization/
        # > Additionally, the allowReserved keyword specifies whether the reserved
        # > characters :/?#[]@!$&'()*+,;= in parameter values are allowed to be sent as they are,
        # > or should be percent-encoded. By default, allowReserved is false, and reserved characters
        # > are percent-encoded. For example, / is encoded as %2F (or %2f), so that the parameter
        # > value quotes/h2g2.txt will be sent as quotes%2Fh2g2.txt
        regexArray = paste0("(?:(?:", regex, "),?)+"),
        converterArray = function(x) {converter(stri_split_fixed(x, ",")[[1]])}
      )

    for (plumberType in plumberTypes) {
      plumberToApiTypeMap[[plumberType]] <<- apiType
    }
    # make sure it could be called again
    plumberToApiTypeMap[[apiType]] <<- apiType

    invisible(TRUE)
  }

  addApiInfo(
    "boolean",
    c("bool", "boolean", "logical"),
    "[01tfTF]|true|false|TRUE|FALSE",
    as.logical,
    location = c("query", "path")
  )
  addApiInfo(
    "number",
    c("dbl", "double", "float", "number", "numeric"),
    "-?\\\\d*\\\\.?\\\\d+",
    as.numeric,
    format = "double",
    location = c("query", "path")
  )
  addApiInfo(
    "integer",
    c("int", "integer"),
    "-?\\\\d+",
    as.integer,
    format = "int64",
    location = c("query", "path")
  )
  addApiInfo(
    "string",
    c("chr", "str", "character", "string"),
    "[^/]+",
    as.character,
    location = c("query", "path")
  )
  addApiInfo(
    "object",
    c("list", "data.frame", "df", "object"),
    location = "requestBody"
  )
  addApiInfo(
    "file",
    c("file", "binary"),
    location = "requestBody",
    format = "binary",
    realType = "string"
  )
}


#' Parse the given plumber type and return the typecast value
#' @noRd
plumberToApiType <- function(type, inPath = FALSE) {
  if (length(type) > 1) {
    return(vapply(type, plumberToApiType, character(1), inPath, USE.NAMES = FALSE))
  }
  # default type is "string" type
  if (is.na(type)) {
    return(defaultApiType)
  }

  apiType <- plumberToApiTypeMap[[as.character(type)]]
  if (is.null(apiType)) {
    warning(
      "Unrecognized type: ", type, ". Using type: ", defaultApiType,
      call. = FALSE
    )
    apiType <- defaultApiType
  }
  if (inPath && !"path" %in% apiTypesInfo[[apiType]]$location) {
    warning(
      "Unsupported path parameter type: ", type, ". Using type: ", defaultApiType,
      call. = FALSE
    )
    apiType <- defaultApiType
  }

  return(apiType)
}

#' Filter OpenAPI Types
#' @noRd
filterApiTypes <- function(matches, property) {
  names(Filter(function(x) {any(matches %in% x[[property]])}, apiTypesInfo))
}

# Visit /async
# While /async is loading, visit /sync many times

library(promises)

add_async_sleep <- function(p) {
  for (i in 1:10) {
    p <- then(p, function(value) {
      value
    })
  }
  p
}

# use name_ as a placeholder for name when there are extra args
time <- function(name_, name = name_) {
  list(
    name = name,
    time = as.numeric(Sys.time())
  )
}

new_promise <- function() {
  promise_resolve(TRUE)
}

#' @get /async
function() {
  new_promise() %>%
    add_async_sleep() %...>%
    time("async")
}

#' @get /sync
function() {
  time("sync")
}

#' @get /async-bad
function() {
  new_promise() %>%
    add_async_sleep() %...>%
    stop("async-bad - expected error here")
}

#' @get /sync-bad
function() {
  stop("sync-bad - expected error here")
}


#' @get /number-one
function() {
  1
}


#' @serializer png
#' @get /promise_plot1
function(n = 100) {
  promises::promise_resolve(runif(n)) %...>%
  {
    dt <- .
    hist(dt)
  }
}

#' @serializer png
#' @get /promise_plot2
function(n = 100) {
  promises::promise_resolve(runif(n)) %...>%
  {
    dt <- .
    plot(dt)
  }
}

#* @serializer custom
#* @get /
function(){
  1
}

#* @filter foo
#* @serializer custom2
function(req, res){
  if (req$PATH_INFO != '/filter-catch'){
    forward()
  }
}

#* @filter foo2
function(req, res, type="") {
  if (type == "json"){
    res$serializer <- plumber:::serializer_json()
  }
  forward()
}

#* @get /something
function(){
  4
}

#* @get /another
function(req, res){
  res$serializer <-  function(val, req, res, errorHandler){
    list(status=201L, headers=list(), body="CUSTOM3")
  }
  5
}

#* @get /short-json
#* @serializer json
function(){
  "JSON"
}

#* @get /short-html
#* @serializer html
function(){
  "HTML"
}

#* @get /single-arg-ser
#* @serializer customOneArg list(single="hi there")
function(){
  "COA"
}

#* @get /multi-arg-ser
#* @serializer   customMultiArg list(first=LETTERS[1], second=4+4, third=4.3)
function(){
  "MAS"
}

library(stringi)

#* @preempt auth
#* @use /
function(a=1){
  list(msg=paste0("Welcome to the root URL! a = ", a))
}

#* @filter auth
function(req, res){
  if (!stri_startswith_fixed(req$QUERY_STRING, "?user=")){
    # Don't continue
    res$status <- 401
    return(list(err="Not authorized"))
  }

  user <- substr(req$QUERY_STRING, 7, nchar(req$QUERY_STRING))
  req$username <- user

  forward()
}

#* @get /me
function(req, res){
  list(name=req$username)
}

#* @get /error
#* @preempt auth
function(req, res){
  stop("I throw an error!")
}

#* @get /set
#* @preempt auth
function(req){
  req$testVal <- 1
  req$testVal
}

#* @get /get
#* @preempt auth
function(req){
  req$testVal
}

#* > This is an HTML file that will demonstrate the HTTPUV bug in which req's that
#* > share a TCP channel also share an environment. This is why we force connections
#* > to close for now.
#* The above bug does not exist anymore given httpuv >= 1.4.0
#* The test path below has been updated to include expected and unexpected output
#* @get /test
#* @preempt auth
#* @serializer html
function(){
  '<html>
    <head>
    </head>
    <body>
      <h3>Expected answer (good): `{}`<h3>
      <h3>Shared TCP Connections answer (bad): `[1]`.</h3>
      <br/>
      <h3>Answer:</h3>
      <span id="answer"></span>
      <script src="https://code.jquery.com/jquery-2.1.4.min.js"></script>
      <script>
        $.get("/get").done(function(){
          $.get("/set").done(function(){
            $.get("/get").done(function(a){
              window.document.getElementById("answer").append(JSON.stringify(a, null, "  "))
            });
          });
        });
      </script>
    </body>
  </html>'
}

make_req <- function(verb = "GET", path = "/", qs="", body="", args = c(), pr = NULL, ...){
  req <- as.environment(list(...))
  req$REQUEST_METHOD <- toupper(verb)
  req$PATH_INFO <- path
  req$QUERY_STRING <- qs

  if (is.character(body)) {
    body <- charToRaw(body)
  }
  stopifnot(is.raw(body))
  req$rook.input <- list(read_lines = function(){ rawToChar(body) },
                         read = function(){ body },
                         rewind = function(){ length(body) })
  req$bodyRaw <- body
  req$pr <- pr
  req
}

bodyFilter <- function(req){
  handled <- req$.internal$bodyHandled
  if (is.null(handled) || handled != TRUE) {

    rook_input <- req$rook.input

    # This will return raw bytes
    # store raw body into req$bodyRaw
    req$bodyRaw <- rook_input$read()

    delayedAssign(
      "postBody",
      {
        rook_input$rewind()
        paste0(rook_input$read_lines(), collapse = "\n")
      },
      assign.env = req
    )

    req$.internal$bodyHandled <- TRUE
  }
  forward()
}

req_body_parser <- function(req, parsers = NULL) {
  if (length(parsers) == 0) {
    return(NULL)
  }
  type <- req$HTTP_CONTENT_TYPE
  bodyRaw <- req$bodyRaw
  if (is.null(bodyRaw)) {
    return(NULL)
  }
  body <- parse_body(bodyRaw, type, parsers)
  # store parsed body into req$body
  body
}
req_body_args <- function(req) {

  body <- req$body
  if (length(body) == 0) {
    return(list())
  }

  # Copy name over so that it is clearer as to the goal of the code below
  # The value returned from this function is set to `ret$argsBody`
  args_body <- body

  if (inherits(args_body, "plumber_multipart")) {
    args_body <- combine_keys(args_body, type = "multi")

  } else if (!is.null(args_body)) {
    # if it's a vector, then we should maybe bundle it as a list
    # this will allow for req$args to have the single piece of information
    # but it will deter from trying to formal name match against MANY false positive values
    if (!is.list(args_body)) {
      args_body_names <- names(args_body)
      # if there are no names at all, wrap it in a unnamed list to pass it through
      if (is.null(args_body_names) || all(args_body_names == "")) {
        args_body <- list(args_body)
      }
    }
  }
  args_body
}

parse_body <- function(body, content_type = NULL, parsers = NULL) {
  if (!is.raw(body)) {body <- charToRaw(body)}
  toparse <- list(value = body, content_type = content_type, parsers = parsers)
  parse_raw(toparse)
}

parse_raw <- function(toparse) {
  if (length(toparse$value) == 0L) return(list())
  parser <- parser_picker(
    toparse$content_type,
    toparse$value[1],
    toparse$filename,
    toparse$parsers
  )
  if (is.null(parser)) {
    message("No suitable parser found to handle request body type ", toparse$content_type, ".")
    return(list())
  }
  do.call(parser, toparse)
}

looks_like_json <- local({
  square_brace <- as.raw(91L)
  curly_brace <- as.raw(123L)
  function(first_byte) {
    first_byte == square_brace || first_byte == curly_brace
  }
})
parser_picker <- function(content_type, first_byte, filename = NULL, parsers = NULL) {

  content_type <- cleanup_content_type(content_type)

  # parse as json or a form
  if (length(content_type) == 0) {
    # fast default to json when first byte is 7b (ascii {) or 5b (ascii [)
    if (looks_like_json(first_byte)) {
      return(parsers$alias$json)
    }

    return(parsers$alias$form)
  }

  parser <- parsers$fixed[[content_type]]

  # return known parser (exact match)
  if (!is.null(parser)) {
    return(parser)
  }

  fpm <- stri_detect_regex(
    content_type,
    names(parsers$regex),
    max_count = 1
  )

  # return known parser (first regex pattern match)
  if (any(fpm)) {
    return(parsers$regex[[which(fpm)[1]]])
  }

  # parse as a form submission
  if (is.null(filename)) {
    return(parsers$alias$form)
  }

  # octet
  parsers$alias$octet
}


#' Manage parsers
#'
#' A parser is responsible for decoding the raw body content of a request into
#' a list of arguments that can be mapped to endpoint function arguments.
#' For instance, [parser_json()] parse content-type `application/json`.
#'
#' @param alias An alias to map parser from the `@parser` plumber tag to the global parsers list.
#' @param parser The parser function to be added. This build the parser function. See Details for more information.
#' @param fixed A character vector of fixed string to be matched against a request `content-type` to use `parser`.
#' @param regex A character vector of [regex] string to be matched against a request `content-type` to use `parser`.
#' @param verbose Logical value which determines if a warning should be
#' displayed when alias in map are overwritten.
#'
#' @details
#' When `parser` is evaluated, it should return a parser function.
#' Parser matching is done first by `content-type` header matching with `fixed` then by using
#' regular expressions with `regex`. Note that plumber strips `; charset*` from `content-type` header before matching.
#'
#' Plumber will try to use [parser_json()] (if available) when no `content-type` header is found and
#' the request body starts with `{` or `[`.
#'
#' Functions signature should include `value`, `...` and
#' possibly `content_type`, `filename`. Other parameters may be provided
#' if you want to use the headers from [webutils::parse_multipart()].
#'
#' Parser function structure is something like below.
#' ```r
#' function(parser_arguments_here) {
#'   # return a function to parse a raw value
#'   function(value, ...) {
#'     # do something with raw value
#'   }
#' }
#' ```
#'
#' @examples
#' # `content-type` header is mostly used to look up charset and adjust encoding
#' parser_dcf <- function(...) {
#'   function(value, content_type = "text/x-dcf", ...) {
#'     charset <- get_character_set(content_type)
#'     value <- rawToChar(value)
#'     Encoding(value) <- charset
#'     read.dcf(value, ...)
#'   }
#' }
#'
#' # Could also leverage existing parsers
#' parser_dcf <- function(...) {
#'   parser_read_file(function(tmpfile) {
#'     read.dcf(tmpfile, ...)
#'   })
#' }
#'
#' # Register the newly created parser
#' \dontrun{register_parser("dcf", parser_dcf, fixed = "text/x-dcf")}
#' @export
register_parser <- function(
  alias,
  parser,
  fixed = NULL,
  regex = NULL,
  verbose = TRUE
) {

  if (!is.null(.globals$parsers[[alias]])) {
    if (isTRUE(verbose)) {
      warning("Overwriting parser: ", alias)
    }
  }

  stopifnot(is.function(parser))

  if (length(c(fixed, regex)) == 0) {
    stop("At least one value of `fixed` and `regex` is required to register a parser")
  }

  # Init the parser function with outside arguments
  init_parser_func <- function(...) {
    parser_function <- do.call(parser, list(...))

    parser_formals <- formals(parser_function)
    if (!("..." %in% names(parser_formals))) {
      stop("For parser '", alias, "', please add a `...` argument to the returned function for possible future parameter expansion")
    }

    create_list <- function(names) {
      setNames(
        replicate(
          length(names),
          parser_function),
        names
      )
    }

    parser_info <- list(
      alias = create_list(alias)
    )
    if (length(fixed) > 0) {
      parser_info$fixed <- create_list(fixed)
    }
    if (length(regex) > 0) {
      parser_info$regex <- create_list(regex)
    }

    return(parser_info)
  }

  .globals$parsers[[alias]] <- init_parser_func
  invisible(.globals$parsers)
}

#' @describeIn register_parser Return all registered parsers
#' @export
registered_parsers <- function() {
  sort(names(.globals$parsers))
}

# ' @describeIn register_parser Select from global parsers and create a combined parser list for programmatic use.
# ' @param aliases Can be one of:
# '   * A character vector of `alias` names.
# '   * A named `list()` whose keys are `alias` names and values are arguments to be applied with [do.call()]
# '   * A `TRUE` value, which will default to combining all parsers. This is great for seeing what is possible, but not great for security purposes.
# '   * Already combined parsers. (Will be returned immediately.)
# '
# ' If `"all"` is found in any `alias` character value or list name, all remaining parsers will be added.  When using a list, aliases already defined will maintain their existing argument values.  All other parser aliases will use their default arguments.
# ' @examples
# ' # provide a character string
# ' make_parser("json")
# '
# ' # provide a named list with no arguments
# ' make_parser(list(json = list()))
# '
# ' # provide a named list with arguments; include `rds`
# ' make_parser(list(json = list(simplifyVector = FALSE), rds = list()))
# '
# ' # default plumber parsers
# ' make_parser(c("json", "form", "text", "octet", "multi"))
make_parser <- function(aliases) {
  if (inherits(aliases, "plumber_parsed_parsers")) {
    return(aliases)
  }
  if (isTRUE(aliases)) {
    # use all parsers
    aliases <- "all"
  }
  if (is.character(aliases)) {
    if (any(is.na(aliases))) {
      stop("aliases can not be `NA` values")
    }
    if ("all" %in% aliases) {
      # use all available parsers expect `all` and `none`
      aliases <- setdiff(registered_parsers(), c("all", "none"))
    }
    # turn aliases into a named list with empty values
    aliases <- setNames(
      replicate(length(aliases), {list()}),
      aliases
    )
  }

  stopifnot(is.list(aliases))
  if (is.null(names(aliases))) {
    stop("aliases must be able to convert to a named list")
  }

  local({
    aliases_not_found <- !(names(aliases) %in% registered_parsers())
    if (any(aliases_not_found)) {
      missing_aliases <- names(aliases)[aliases_not_found]
      stop("Aliases not available: ", paste0(missing_aliases, collapse = ", "), ". See: registered_parsers()")
    }
  })

  # if "all" is found, remove "all" and add all remaining registered parsers (except 'none') to the `aliases` list
  if ("all" %in% names(aliases)) {
    all_parser_names <- setdiff(registered_parsers(), c("all", "none"))
    # remove to avoid infinite recursion
    aliases$all <- NULL
    names_to_add <- setdiff(all_parser_names, names(aliases))
    if (length(names_to_add)) {
      aliases[names_to_add] <- replicate(length(names_to_add), list())
    }
  }

  # convert parser functions into initialized information
  parser_infos <-
    lapply(
      names(aliases),
      function(alias) {
        # get init function
        init_parser_func <- .globals$parsers[[alias]]
        # call outer parser function to init the params for inner function
        do.call(init_parser_func, aliases[[alias]])
      }
    )

  # combine information into a single list
  combined_parser_info <-
    Reduce(
      function(cur_parser_info, parser_info) {
        utils::modifyList(cur_parser_info, parser_info)
      },
      parser_infos,
      init = list()
    )

  class(combined_parser_info) <- c("plumber_parsed_parsers", class(combined_parser_info))
  combined_parser_info
}

#' Plumber Parsers
#'
#' Parsers are used in Plumber to transform request body received
#' by the API. Extra parameters may be provided to parser
#' functions when enabling them on router. This will allow for
#' non-default behavior.
#'
#' Parsers are optional. When unspecified, only default endpoint parsers are enabled.
#' You can use `@parser NAME` tag to enable parser on endpoint.
#' Multiple parsers can be enabled on the same endpoint using multiple `@parser NAME` tags.
#'
#' User should be aware that `rds` parsing should only be done from a
#' trusted source. Do not accept `rds` files blindly.
#'
#' See [registered_parsers()] for a list of registered parsers names.
#'
#' @param ... parameters supplied to the appropriate internal function
#' @describeIn parsers Form query string parser
#' @examples
#' \dontrun{
#' # Overwrite `text/json` parsing behavior to not allow JSON vectors to be simplified
#' #* @parser json list(simplifyVector = FALSE)
#' # Activate `rds` parser in a multipart request
#' #* @parser multi
#' #* @parser rds
#' pr <- Plumber$new()
#' pr$handle("GET", "/upload", function(rds) {rds}, parsers = c("multi", "rds"))
#' }
#' @export
parser_form <- function() {
  parser_text(parseQS)
}


#' @describeIn parsers JSON parser. See [jsonlite::parse_json()] for more details. (Defaults to using `simplifyVectors = TRUE`)
#' @export
parser_json <- function(...) {
  parser_text(function(txt_value) {
    safeFromJSON(txt_value, ...)
  })
}

#' @describeIn parsers GeoJSON parser. See [geojsonsf::geojson_sf()] for more details.
#' @export
parser_geojson <- function(...) {
  if (!requireNamespace("geojsonsf", quietly = TRUE)) {
    stop("`geojsonsf` must be installed for `parser_geojson` to work")
  }
  parser_text(function(val) {
    geojsonsf::geojson_sf(val, ...)
  })
}

#' @describeIn parsers Helper parser to parse plain text
#' @param parse_fn function to further decode a text string into an object
#' @export
parser_text <- function(parse_fn = identity) {
  stopifnot(is.function(parse_fn))
  function(value, content_type = NULL, ...) {
    charset <- get_character_set(content_type)
    txt_value <- rawToChar(value)
    Encoding(txt_value) <- charset
    parse_fn(txt_value)
  }
}


#' @describeIn parsers YAML parser. See [yaml::yaml.load()] for more details.
#' @export
parser_yaml <- function(...) {
  parser_text(function(val) {
    if (!requireNamespace("yaml", quietly = TRUE)) {
      stop("yaml must be installed for the yaml parser to work")
    }
    yaml::yaml.load(val, ..., eval.expr = FALSE)
  })
}

#' @describeIn parsers CSV parser. See [readr::read_csv()] for more details.
#' @export
parser_csv <- function(...) {
  parse_fn <- function(raw_val) {
    if (!requireNamespace("readr", quietly = TRUE)) {
      stop("`readr` must be installed for `parser_csv` to work")
    }
    readr::read_csv(raw_val, ...)
  }
  function(value, ...) {
    parse_fn(value)
  }
}


#' @describeIn parsers TSV parser. See [readr::read_tsv()] for more details.
#' @export
parser_tsv <- function(...) {
  parse_fn <- function(raw_val) {
    if (!requireNamespace("readr", quietly = TRUE)) {
      stop("`readr` must be installed for `parser_tsv` to work")
    }
    readr::read_tsv(raw_val, ...)
  }
  function(value, ...) {
    parse_fn(value)
  }
}


#' @describeIn parsers Helper parser that writes the binary body to a file and reads it back again using `read_fn`.
#'   This parser should be used when reading from a file is required.
#' @param read_fn function used to read a the content of a file. Ex: [readRDS()]
#' @export
parser_read_file <- function(read_fn = readLines) {
  stopifnot(is.function(read_fn))
  function(value, filename = "", ...) {
    tmp <- tempfile("plumb", fileext = paste0("_", basename(filename)))
    on.exit({
      if (file.exists(tmp)) {
        file.remove(tmp)
      }
    }, add = TRUE)
    writeBin(value, tmp)
    read_fn(tmp)
  }
}


#' @describeIn parsers RDS parser. See [readRDS()] for more details.
#' @export
parser_rds <- function(...) {
  parser_read_file(function(tmpfile) {
    # `readRDS()` does not work with `rawConnection()`
    readRDS(tmpfile, ...)
  })
}

#' @describeIn parsers feather parser. See [arrow::read_feather()] for more details.
#' @export
parser_feather <- function(...) {
  parser_read_file(function(tmpfile) {
    if (!requireNamespace("arrow", quietly = TRUE)) {
      stop("`arrow` must be installed for `parser_feather` to work")
    }
    arrow::read_feather(tmpfile, ...)
  })
}

#' @describeIn parsers parquet parser. See [arrow::read_parquet()] for more details.
#' @export
parser_parquet <- function(...) {
  parser_read_file(function(tmpfile) {
    if (!requireNamespace("arrow", quietly = TRUE)) {
      stop("`arrow` must be installed for `parser_parquet` to work")
    }
    arrow::read_parquet(tmpfile, ...)
  })
}

#' @describeIn parsers Octet stream parser. Returns the raw content.
#' @export
parser_octet <- function() {
  function(value, ...) {
    return(value)
  }
}


#' @describeIn parsers Multi part parser. This parser will then parse each individual body with its respective parser.  When this parser is used, `req$body` will contain the updated output from [webutils::parse_multipart()] by adding the `parsed` output to each part.  Each part may contain detailed information, such as `name` (required), `content_type`, `content_disposition`, `filename`, (raw, original) `value`, and `parsed` (parsed `value`).  When performing Plumber route argument matching, each multipart part will match its `name` to the `parsed` content.
#' @export
#' @importFrom webutils parse_multipart
parser_multi <- function() {
  function(value, content_type, parsers, ...) {
    if (!stri_detect_fixed(content_type, "boundary=", case_insensitive = TRUE))
      stop("No boundary found in multipart content-type header: ", content_type)
    boundary <- stri_match_first_regex(content_type, "boundary=([^; ]{2,})", case_insensitive = TRUE)[,2]
    toparse <- parse_multipart(value, boundary)

    # set the names of the items as the `name` of each item
    toparse_names <- vapply(toparse, function(x) {
      name <- x$name
      # null or character(0)
      if (length(name) == 0) {
        return("")
      }
      name
    }, character(1))
    names(toparse) <- toparse_names

    # content-type detection
    ret <- lapply(toparse, function(x) {
      if (
        is.null(x$content_type) ||
        # allows for files to be shipped as octect, but parsed using the matching value in `knownContentTypes`
        # (Ex: `.rds` files -> `application/rds` which has a proper RDS parser)
        isTRUE(stri_detect_fixed(x$content_type, "application/octet-stream"))
      ) {
        if (!is.null(x$filename)) {
          # Guess content-type from file extension
          x$content_type <- getContentType(tools::file_ext(x$filename))
        }
      }
      # copy over to allow to return the updated `x` without `parsers`
      item <- x
      # add `parsers` to allow `parse_raw` to work
      item$parsers <- parsers
      # store the parsed information into `x`
      x$parsed <- parse_raw(item)
      # return the updated `webutils::parse_multipart()` output
      x
    })

    # set a class so `req$argsBody` can be reduced to a named list of values
    class(ret) <- "plumber_multipart"
    ret
  }
}

#' @describeIn parsers No parser. Will not process the postBody.
#' @export
parser_none <- function() {
  function(value, ...) {
    value
  }
}

register_parsers_onLoad <- function() {
  # parser alias names for plumbing
  register_parser("csv",     parser_csv,     fixed = c("application/csv", "application/x-csv", "text/csv", "text/x-csv"))
  register_parser("json",    parser_json,    fixed = c("application/json", "text/json"))
  register_parser("multi",   parser_multi,   fixed = "multipart/form-data", regex = "^multipart/")
  register_parser("octet",   parser_octet,   fixed = "application/octet-stream")
  register_parser("form",    parser_form,   fixed = "application/x-www-form-urlencoded")
  register_parser("rds",     parser_rds,     fixed = "application/rds")
  register_parser("feather", parser_feather, fixed = c("application/vnd.apache.arrow.file", "application/feather"))
  register_parser("parquet", parser_parquet, fixed = "application/vnd.apache.parquet")
  register_parser("text",    parser_text,    fixed = "text/plain", regex = "^text/")
  register_parser("tsv",     parser_tsv,     fixed = c("application/tab-separated-values", "text/tab-separated-values"))
  # yaml types: https://stackoverflow.com/a/38000954/591574
  register_parser("yaml",    parser_yaml,    fixed = c("text/vnd.yaml", "application/yaml", "application/x-yaml", "text/yaml", "text/x-yaml"))
  register_parser("none",    parser_none,    regex = "*")
  register_parser("geojson", parser_geojson, fixed = c("application/geo+json", "application/vdn.geo+json"))

  parser_all <- function() {
    stop("This function should never be called. It should be handled by `make_parser('all')`")
  }
  register_parser("all", parser_all, regex = "*")
}

context("body parsing")

test_that("JSON is consumed on POST", {
  expect_equal(parse_body('{"a":"1"}', content_type = NULL, parsers = make_parser("json")), list(a = "1"))
  expect_equal(parse_body('[1,2,3]', content_type = NULL, parsers = make_parser("json")), 1:3)
})

test_that("ending in `==` does not produce a unexpected key", {
  # See https://github.com/rstudio/plumber/issues/463
  expect_equal(parse_body("randomcharshere==", content_type = NULL, parsers = make_parser("form")), list())
})

test_that("Form query strings on post are handled correctly", {
  expect_equivalent(parse_body("a=", parsers = make_parser("form")), list()) # It's technically a named list()
  expect_equal(parse_body("a=1&b=&c&d=1", content_type = NULL, make_parser("form")), list(a="1", d="1"))
})

test_that("Able to handle UTF-8", {
  expect_equal(parse_body('{"text":"Ã©lise"}', content_type = "application/json", parsers = make_parser("json"))$text, "Ã©lise")
})

#charset moved to part parsing
test_that("filter passes on content-type", {
  req <- make_req(
    body = "this is a body",
    HTTP_CONTENT_TYPE = "text/html; charset=testset",
  )
  with_mock(
    parse_body = function(body, content_type = "unknown", parsers = NULL) {
      print(content_type)
      body
    },
    {
      expect_output(req_body_parser(req, make_parser("text")), "text/html; charset=testset")
    },
    .env = "plumber"
  )
})

# parsers
test_that("Test text parser", {
  expect_equal(parse_body("Ceci est un texte.", "text/html", make_parser("text")), "Ceci est un texte.")
})

test_that("Test yaml parser", {
  skip_if_not_installed("yaml")

  r_object <- list(a=1,b=list(c=2,d=list(e=3,f=4:6)))
  expect_equal(parse_body(charToRaw(yaml::as.yaml(r_object)), "application/x-yaml", make_parser("yaml")), r_object)
})

test_that("Test csv parser", {
  skip_if_not_installed("readr")

  tmp <- tempfile()
  on.exit({
    file.remove(tmp)
  }, add = TRUE)

  r_object <- cars
  write.csv(r_object, tmp, row.names = FALSE)
  val <- readBin(tmp, "raw", 1000)

  parsed <- parse_body(val, "application/csv", make_parser("csv"))
  # convert from readr tibble to data.frame
  parsed <- as.data.frame(parsed, stringsAsFactors = FALSE)
  attr(parsed, "spec") <- NULL

  expect_equal(parsed, r_object)
})

test_that("Test tsv parser", {
  skip_if_not_installed("readr")

  tmp <- tempfile()
  on.exit({
    file.remove(tmp)
  }, add = TRUE)

  r_object <- cars
  write.table(r_object, tmp, sep = "\t", row.names = FALSE)
  val <- readBin(tmp, "raw", 1000)

  parsed <- parse_body(val, "application/tab-separated-values", make_parser("tsv"))
  # convert from readr tibble to data.frame
  parsed <- as.data.frame(parsed, stringsAsFactors = FALSE)
  attr(parsed, "spec") <- NULL

  expect_equal(parsed, r_object)
})

test_that("Test feather parser", {
  skip_if_not_installed("arrow")

  tmp <- tempfile()
  on.exit({
    file.remove(tmp)
  }, add = TRUE)

  r_object <- iris
  arrow::write_feather(r_object, tmp)
  val <- readBin(tmp, "raw", 10000)

  parsed <- parse_body(val, "application/vnd.apache.arrow.file", make_parser("feather"))
  # convert from feather tibble to data.frame
  parsed <- as.data.frame(parsed, stringsAsFactors = FALSE)
  attr(parsed, "spec") <- NULL

  expect_equal(parsed, r_object)
})

test_that("Test parquet parser", {
  skip_if_not_installed("arrow")

  tmp <- tempfile()
  on.exit({
    file.remove(tmp)
  }, add = TRUE)

  r_object <- iris
  arrow::write_parquet(r_object, tmp)
  val <- readBin(tmp, "raw", 10000)

  parsed <- parse_body(val, "application/vnd.apache.parquet", make_parser("parquet"))
  # convert from parquet tibble to data.frame
  parsed <- as.data.frame(parsed, stringsAsFactors = FALSE)
  attr(parsed, "spec") <- NULL

  expect_equal(parsed, r_object)
})

test_that("Test geojson parser", {
  skip_if_not_installed("geojsonsf")
  skip_if_not_installed("sf")

  # Test sf object w/ fields
  geojson <- '{"type":"FeatureCollection","features":[{"type":"Feature","properties":{"a":3},"geometry":{"type":"Point","coordinates":[1,2]}},{"type":"Feature","properties":{"a":4},"geometry":{"type":"Point","coordinates":[3,4]}}]}'
  parsed <- parse_body(geojson, "application/geo+json", make_parser("geojson"))
  expect_equal(parsed, geojsonsf::geojson_sf(geojson))

  # Test sfc
  geojson <- '[
  { "type":"Point","coordinates":[0,0]},
  {"type":"LineString","coordinates":[[0,0],[1,1]]}
  ]'
  parsed <- parse_body(geojson, "application/geo+json", make_parser("geojson"))
  expect_equal(parsed, geojsonsf::geojson_sf(geojson))

  # Test simple sf object
  geojson <- '{ "type" : "Point", "coordinates" : [0, 0] }'
  parsed <- parse_body(geojson, "application/geo+json", make_parser("geojson"))
  expect_equal(parsed, geojsonsf::geojson_sf(geojson))

  # Test geojson file
  tmp <- tempfile()
  on.exit({
    file.remove(tmp)
  }, add = TRUE)

  writeLines(geojson, tmp)
  val <- readBin(tmp, "raw", 1000)
  parsed <- parse_body(val, "application/geo+json", make_parser("geojson"))
  expect_equal(parsed, geojsonsf::geojson_sf(geojson))

})

test_that("Test multipart output is reduced for argument matching", {
  bin_file <- test_path("files/multipart-file-names.bin")
  body <- readBin(bin_file, what = "raw", n = file.info(bin_file)$size)
  req <- make_req(
    body = body,
    HTTP_CONTENT_TYPE = "multipart/form-data; boundary=---------------------------286326291134907228894146459692"
  )

  req$body <- req_body_parser(req, make_parser(c("multi", "octet", "json")))
  body_args <- req_body_args(req)

  expect_s3_class(req$body, "plumber_multipart")
  expect_equal(names(req$body), c("files", "files", "files", "files", "dt", "namedval", "namedval", "namedval", "namedval"))
  for(part in req$body) {

    expect_equal(part$content_disposition, "form-data")
    expect_true(is.character(part$name))
    expect_true(is.raw(part$value))

    if (part$name == "dt") {
      expect_true(is.null(part$content_type))
    } else {
      expect_true(!is.null(part$content_type))
    }

    if (part$name == "dt" || identical(part$filename, "has_name3.json")) {
      expect_equal(part$parsed, jsonlite::parse_json("{}"))
    } else {
      expect_true(is.raw(part$parsed))
    }
  }

  expect_true(!inherits(body_args, "plumber_multipart"))
  expect_equal(names(body_args), c("files", "dt", "namedval"))

  expect_equal(length(body_args$files), 4)
  expect_equal(names(body_args$files), c("avatar2-small.png", "text1.bin", "text2.bin", "text3.bin"))
  for (parsed in body_args$files) {
    expect_true(is.raw(parsed))
  }

  expect_equal(body_args$dt, jsonlite::parse_json("{}"))

  expect_equal(length(body_args$namedval), 4)
  expect_equal(names(body_args$namedval), c("has_name.bin", "", "has_name2.bin", "has_name3.json"))
  for (parsed in body_args$namedval[-4]) {
    expect_true(is.raw(parsed))
  }
  expect_equal(body_args$namedval$`has_name3.json`, jsonlite::parse_json("{}"))
})


test_that("Test multipart parser", {
  # also tests rds and the octet -> content type conversion

  bin_file <- test_path("files/multipart-form.bin")
  body <- readBin(bin_file, what = "raw", n = file.info(bin_file)$size)
  req <- make_req(
    body = body,
    HTTP_CONTENT_TYPE = "multipart/form-data; boundary=----WebKitFormBoundaryMYdShB9nBc32BUhQ"
  )
  req$body <- req_body_parser(req, make_parser(c("multi", "json", "rds", "octet")))
  body_args <- req_body_args(req)

  expect_s3_class(req$body, "plumber_multipart")
  expect_equal(names(req$body), c("json", "img1", "img2", "rds"))
  for(part in req$body) {
    expect_equal(part$content_disposition, "form-data")
    expect_true(is.character(part$name))
    expect_true(is.raw(part$value))

    if (part$name == "json") {
      expect_true(is.null(part$content_type))
    } else {
      expect_true(!is.null(part$content_type))
    }

    switch(part$name,
      "json" = expect_equal(part$parsed, list(a=2,b=4,c=list(w=3,t=5))),
      "rds" = expect_equal(part$parsed, women),
      {
        if (part$name == "img1") expect_equal(part$filename, "avatar2-small.png")
        if (part$name == "img2") expect_equal(part$filename, "ragnarok_small.png")
        expect_true(is.raw(part$parsed))
        expect_gt(length(part$parsed), 100)
      }
    )
  }

  expect_true(!inherits(body_args, "plumber_multipart"))
  expect_equal(names(body_args), c("json", "img1", "img2", "rds"))
  expect_equal(body_args[["rds"]], list("women.rds" = women))
  expect_true(is.raw(body_args[["img1"]][["avatar2-small.png"]]))
  expect_gt(length(body_args[["img1"]][["avatar2-small.png"]]), 100)
  expect_true(is.raw(body_args[["img2"]][["ragnarok_small.png"]]))
  expect_gt(length(body_args[["img2"]][["ragnarok_small.png"]]), 100)
  expect_equal(body_args[["json"]], list(a=2,b=4,c=list(w=3,t=5)))
})


test_that("Test multipart respect content-type", {
  skip_if_not_installed("readr")

  bin_file <- test_path("files/multipart-ctype.bin")
  body <- readBin(bin_file, what = "raw", n = file.info(bin_file)$size)
  req <- make_req(
    body = body,
    HTTP_CONTENT_TYPE = "multipart/form-data; boundary=---------------------------90908882332870323642673870272"
  )
  req$body <- req_body_parser(req, make_parser(c("multi", "tsv")))
  body_args <- req_body_args(req)

  expect_s3_class(req$body, "plumber_multipart")
  expect_equal(length(req$body), 1)
  expect_equal(names(req$body), "sample_name")

  expect_equal(req$body$sample_name$content_disposition, "form-data")
  expect_true(is.character(req$body$sample_name$name))
  expect_true(is.raw(req$body$sample_name$value))
  expect_equal(req$body$sample_name$content_type, "text/tab-separated-values")

  expect_s3_class(req$body$sample_name$parsed, "data.frame")
  expect_equal(colnames(req$body$sample_name$parsed), c("x", "y", "z"))
  expect_equal(nrow(req$body$sample_name$parsed), 11)

  expect_true(!inherits(body_args, "plumber_multipart"))
  expect_s3_class(body_args[["sample_name"]][["sample.tsv"]], "data.frame")
  expect_equal(colnames(body_args[["sample_name"]][["sample.tsv"]]), c("x", "y", "z"))
  expect_equal(nrow(body_args[["sample_name"]][["sample.tsv"]]), 11)
})

test_that("Test an array of files upload", {
  bin_file <- test_path("files/multipart-files-array.bin")
  body <- readBin(bin_file, what = "raw", n = file.info(bin_file)$size)
  body <- readBin(bin_file, what = "raw", n = file.info(bin_file)$size)
  req <- make_req(
    body = body,
    HTTP_CONTENT_TYPE = "multipart/form-data; boundary=---------------------------286326291134907228894146459692"
  )
  req$body <- req_body_parser(req, make_parser(c("multi", "octet", "json")))
  body_args <- req_body_args(req)

  expect_s3_class(req$body, "plumber_multipart")
  expect_equal(names(req$body), c("files", "files", "files", "files", "dt"))

  for(i in seq_along(req$body)) {
    part <- req$body[[i]]
    expect_equal(part$content_disposition, "form-data")
    expect_true(is.character(part$name))
    expect_true(is.raw(part$value))

    if (i == 1) {
      expect_equal(part$name, "files")
      expect_equal(part$filename, "avatar2-small.png")
      expect_equal(part$content_type, "image/png")
    } else if (i == 5) {
      expect_equal(part$name, "dt")
      expect_equal(part$content_type, NULL)
      expect_equal(part$parsed, jsonlite::parse_json("{}"))
    } else {
      expect_equal(part$name, "files")
      expect_equal(part$filename, paste0("text", i - 1, ".bin"))
      expect_equal(part$content_type, "application/octet-stream")
      expect_equal(rawToChar(part$parsed), letters[i - 1])
    }
  }

  expect_true(!inherits(body_args, "plumber_multipart"))
  expect_equal(names(body_args), c("files", "dt"))
  expect_equal(names(body_args$files), c("avatar2-small.png", "text1.bin", "text2.bin", "text3.bin"))
  expect_equal(rawToChar(body_args$files[[2]]), "a")
  expect_equal(rawToChar(body_args$files[[3]]), "b")
  expect_equal(rawToChar(body_args$files[[4]]), "c")
  expect_equal(body_args[["dt"]], jsonlite::parse_json("{}"))
})

context("feather serializer")

test_that("feather serializes properly", {
  skip_if_not_installed("arrow")

  d <- data.frame(a=1, b=2, c="hi")
  val <- serializer_feather()(d, data.frame(), PlumberResponse$new(), stop)
  expect_equal(val$status, 200L)
  expect_equal(val$headers$`Content-Type`, "application/vnd.apache.arrow.file")

  # can test  by doing a full round trip if we believe the parser works via `test-parse-body.R`
  parsed <- parse_body(val$body, "application/vnd.apache.arrow.file", make_parser("feather"))
  # convert from feather tibble to data.frame
  parsed <- as.data.frame(parsed, stringsAsFactors = FALSE)
  attr(parsed, "spec") <- NULL

  expect_equal(parsed, d)
})

test_that("Errors call error handler", {
  skip_if_not_installed("arrow")

  errors <- 0
  errHandler <- function(req, res, err){
    errors <<- errors + 1
  }

  expect_equal(errors, 0)
  serializer_feather()(parse(text="hi"), data.frame(), PlumberResponse$new("csv"), errorHandler = errHandler)
  expect_equal(errors, 1)
})

test_that("Errors are rendered correctly with debug TRUE", {
  skip_if_not_installed("arrow")

  pr <- pr() %>% pr_get("/", function() stop("myerror"), serializer = serializer_feather()) %>% pr_set_debug(TRUE)
  capture.output(res <- pr$serve(make_req(pr = pr), PlumberResponse$new("csv")))

  expect_match(res$body, "Error in (function () : myerror", fixed = TRUE)
})

context("Parsers tag")

test_that("parsers can be combined", {

  expect_parsers <- function(names, target_names, sort_items = TRUE) {
    aliases <- names(make_parser(names)$alias)
    if (sort_items) {
      aliases <- sort(aliases)
      target_names <- sort(target_names)
    }
    expect_equal(aliases, target_names)
  }

  expect_parsers("json", "json")

  expect_parsers(c("form", "json"), c("form", "json"), sort_items = FALSE)

  expect_parsers("all", setdiff(registered_parsers(), c("all", "none")))
  expect_parsers(list(all = list()), setdiff(registered_parsers(), c("all", "none")))
  expect_parsers(TRUE, setdiff(registered_parsers(), c("all", "none")))



  # make sure parameters are not overwritten even when including all
  parsers_plain <- make_parser(list(all = list(), json = list(simplifyVector = FALSE)))
  json_input <- parsers_plain$alias$json(charToRaw(jsonlite::toJSON(1:3)))
  expect_equal(json_input, list(1,2,3))

  parsers_guess <- make_parser(list(all = list(), json = list(simplifyVector = TRUE)))
  json_input <- parsers_guess$alias$json(charToRaw(jsonlite::toJSON(1:3)))
  expect_equal(json_input, c(1,2,3))

  # check that parsers return already combined parsers
  expect_equal(make_parser(parsers_plain), parsers_plain)
})

test_that("parsers work", {
  r <- pr(test_path("files/parsers.R"))
  res <- PlumberResponse$new()

  expect_identical(r$route(make_req("POST", "/default", body='{"a":1}'), res), structure(list(1L), names = "a"))
  expect_identical(r$route(make_req("POST", "/none", body='{"a":1}'), res), structure(list(), names = character()))
  expect_identical(r$route(make_req("POST", "/all", body='{"a":1}'), res), structure(list(1L), names = "a"))
  bin_file <- test_path("files/multipart-ctype.bin")
  bin_body <- readBin(bin_file, "raw", file.info(bin_file)$size)
  expect_identical(r$route(make_req("POST", "/none", body=rawToChar(bin_body)), res), structure(list(), names = character()))
  expect_message(r$route(make_req("POST", "/json", body=rawToChar(bin_body)), res), "No suitable parser found")

  bin_file <- test_path("files/multipart-form.bin")
  bin_body <- readBin(bin_file, "raw", file.info(bin_file)$size)


  req <- new.env()
  req$REQUEST_METHOD <- "POST"
  req$PATH_INFO <- "/all"
  req$QUERY_STRING <- ""
  req$HTTP_CONTENT_TYPE <- "multipart/form-data; boundary=----WebKitFormBoundaryMYdShB9nBc32BUhQ"
  req$rook.input <- list(read_lines = function(){ stop("should not be executed") },
                         read = function(){ bin_body },
                         rewind = function(){ length(bin_body) })

  parsed_body <- r$route(req, PlumberResponse$new())
  expect_equal(names(parsed_body), c("json", "img1", "img2", "rds"))
  expect_equal(parsed_body[["rds"]], list("women.rds" = women))
  expect_true(is.raw(parsed_body[["img1"]][["avatar2-small.png"]]))
  expect_equal(parsed_body[["json"]], list(a=2,b=4,c=list(w=3,t=5)))


  # expect parsers match
  expect_equal(r$routes$none$parsers, make_parser("none"))
  expect_equal(r$routes$all$parsers, make_parser("all"))
  expect_equal(r$routes$default$parsers, NULL)
  expect_equal(r$routes$json$parsers, make_parser("json"))
  # Paths order follow original code
  expect_equal(r$routes$mixed$parsers, make_parser(c("form", "json")))
  expect_equal(r$routes$repeated$parsers, make_parser("json"))
})

#' Plumber options
#'
#' There are a number of global options that affect Plumber's behavior. These can
#' be set globally with [options()] or with [options_plumber()]. Options set using
#' [options_plumber()] should not include the `plumber.` prefix.
#'
#' \describe{
#' \item{`plumber.port`}{Port Plumber will attempt to use to start http server.
#' If the port is already in use, server will not be able to start. Defaults to `NULL`.}
#' \item{`plumber.docs`}{Name of the visual documentation interface to use. Defaults to `TRUE`, which will use `"swagger"`.}
#' \item{`plumber.docs.callback`}{A function. Called with
#' a single parameter corresponding to the visual documentation url after Plumber server is ready. This can be used
#' by RStudio to open the docs when then API is ran from the editor. Defaults to option `NULL`.}
#' \item{`plumber.trailingSlash`}{Logical value which allows the router to redirect any request
#' that has a matching route with a trailing slash. For example, if set to `TRUE` and the
#' GET route `/test/` existed, then a GET request of `/test?a=1` would redirect to
#' `/test/?a=1`. Defaults to `FALSE`. This option will default to `TRUE` in a future release.}
#' \item{`plumber.methodNotAllowed`}{`r lifecycle::badge("experimental")`
#' Logical value which allows the router to notify that an
#' unavailable method was requested, but a different request method is allowed. For example,
#' if set to `TRUE` and the GET route `/test` existed, then a POST request of `/test` would
#' receive a 405 status and the allowed methods. Defaults to `TRUE`.}
#' \item{`plumber.apiURL`}{Server urls for OpenAPI Specification respecting
#' pattern `scheme://host:port/path`. Other `api*` options will be ignored when set.}
#' \item{`plumber.apiScheme`}{Scheme used to build OpenAPI url and server url for
#' OpenAPI Specification. Defaults to `http`, or an empty string
#' when used outside a running router.}
#' \item{`plumber.apiHost`}{Host used to build docs url and server url for
#' OpenAPI Specification. Defaults to `host` defined by `run` method, or an empty string
#' when used outside a running router.}
#' \item{`plumber.apiPort`}{Port used to build OpenAPI url and server url for
#' OpenAPI Specification. Defaults to `port` defined by `run` method, or an empty string
#' when used outside a running router.}
#' \item{`plumber.apiPath`}{Path used to build OpenAPI url and server url for
#' OpenAPI Specification. Defaults to an empty string.}
#' \item{`plumber.maxRequestSize`}{Maximum length in bytes of request body. Body larger
#' than maximum are rejected with http error 413. `0` means unlimited size. Defaults to `0`.}
#' \item{`plumber.sharedSecret`}{Shared secret used to filter incoming request.
#' When `NULL`, secret is not validated. Otherwise, Plumber compares secret with http header
#' `PLUMBER_SHARED_SECRET`. Failure to match results in http error 400. Defaults to `NULL`.}
#' \item{`plumber.legacyRedirects`}{Plumber will redirect legacy route `/__swagger__/` and
#' `/__swagger__/index.html` to `../__docs__/` and `../__docs__/index.html`. You can disable this
#' by settings this option to `FALSE`. Defaults to `TRUE`}
#' }
#'
#' @param ... Ignored. Should be empty
#' @param port,docs,docs.callback,trailingSlash,methodNotAllowed,apiScheme,apiHost,apiPort,apiPath,apiURL,maxRequestSize,sharedSecret,legacyRedirects See details
#' @return
#' The complete, prior set of [options()] values.
#' If a particular parameter is not supplied, it will return the current value.
#' If no parameters are supplied, all returned values will be the current [options()] values.
#' @export
options_plumber <- function(
  ...,
  port                 = getOption("plumber.port"),
  docs                 = getOption("plumber.docs"),
  docs.callback        = getOption("plumber.docs.callback"),
  trailingSlash        = getOption("plumber.trailingSlash"),
  methodNotAllowed     = getOption("plumber.methodNotAllowed"),
  apiURL               = getOption("plumber.apiURL"),
  apiScheme            = getOption("plumber.apiScheme"),
  apiHost              = getOption("plumber.apiHost"),
  apiPort              = getOption("plumber.apiPort"),
  apiPath              = getOption("plumber.apiPath"),
  maxRequestSize       = getOption("plumber.maxRequestSize"),
  sharedSecret         = getOption("plumber.sharedSecret"),
  legacyRedirects      = getOption("plumber.legacyRedirects")
) {
  ellipsis::check_dots_empty()

  # Make sure all fallback options are disabled
  if (!missing(docs.callback) && is.null(docs.callback)) {
    options("plumber.swagger.url" = NULL)
  }

  options(
    plumber.port                 =   port,
    plumber.docs                 =   docs,
    plumber.docs.callback        =   docs.callback,
    plumber.trailingSlash        =   trailingSlash,
    plumber.methodNotAllowed     =   methodNotAllowed,
    plumber.apiURL               =   apiURL,
    plumber.apiScheme            =   apiScheme,
    plumber.apiHost              =   apiHost,
    plumber.apiPort              =   apiPort,
    plumber.apiPath              =   apiPath,
    plumber.maxRequestSize       =   maxRequestSize,
    plumber.sharedSecret         =   sharedSecret,
    plumber.legacyRedirects      =   legacyRedirects
  )
}

#' Convert the endpoints as they exist on the Plumber object to a list which can
#' be converted into a OpenAPI Specification for these endpoints
#' @noRd
endpointSpecification <- function(routerEndpointEntry, path = routerEndpointEntry$path) {
  ret <- list()

  # We are sensitive to trailing slashes. Should we be?
  # Yes - 12/2018
  cleanedPath <- gsub("<([^:>]+)(:[^>]+)?>", "{\\1}", path)
  ret[[cleanedPath]] <- list()

  # Get the params from the path
  pathParams <- routerEndpointEntry$getTypedParams()
  # Get the params from endpoint expression
  funcParams <- routerEndpointEntry$getFuncParams()
  # Get the plumber decoration defined endpoint params
  endpointParams <- routerEndpointEntry$getEndpointParams()
  for (verb in routerEndpointEntry$verbs) {
    params <- parametersSpecification(endpointParams, pathParams, funcParams)

    # If we haven't already documented a path param, we should add it here.
    # FIXME: warning("Undocumented path parameters: ", paste0())

    resps <- responsesSpecification(routerEndpointEntry)

    endptSpec <- list(
      summary = routerEndpointEntry$comments,
      description = routerEndpointEntry$description,
      responses = resps,
      parameters = params$parameters,
      requestBody = params$requestBody,
      tags = routerEndpointEntry$tags
    )

    ret[[cleanedPath]][[tolower(verb)]] <- endptSpec
  }

  ret
}

defaultResponse <- list(
  "200" = list(
    description = "OK"
  ),
  "500" = list(
    description = "Internal Server Error",
    content = list("application/json" = list(schema = list(type = "string")))
  ),
  "default" = list(
    description = "Default response."
  )
)
responsesSpecification <- function(endpts){
  if (!inherits(endpts, "PlumberEndpoint")) {
    return(defaultResponse)
  }
  resps <- defaultResponse
  if (is.list(endpts$responses)){
    resps <- utils::modifyList(defaultResponse, endpts$responses)
  }
  for (resp in names(resps)) {
    if (!length(resps[[resp]]$content)) {
      ctype <- NULL
      if (is.function(endpts$serializer)) {
        # Must safe-guard against partial name matching
        # since we are reaching into the function env
        ctype <- environment(endpts$serializer)[["headers"]][["Content-Type"]]
      }
      if (isTRUE(nchar(ctype) > 0)) {
        ctype <- stri_split_regex(ctype, "[ ;]")[[1]][1]
        schema <- list(type = ifelse(grepl("^text", ctype), "string", "object"))
        resps[[resp]]$content <- setNames(list(list(schema = schema)), ctype)
      }
    }
  }
  resps
}

#' Extract the OpenAPI parameters Specification from the endpoint
#' paramters.
#' @noRd
parametersSpecification <- function(endpointParams, pathParams, funcParams = NULL){

  params <- list(
    parameters = list(),
    requestBody = NULL
  )
  inBody <- filterApiTypes("requestBody", "location")
  inRaw <- filterApiTypes("binary", "format")
  for (p in unique(c(names(endpointParams), pathParams$name, names(funcParams)))) {

    # Dealing with priorities endpointParams > pathParams > funcParams
    # For each p, find out which source to trust for :
    #   `type`, `isArray`, `required`
    # - `description` comes from endpointParams
    # - `isArray` defines both `style` and `explode`
    # - `default` and `example` comes from funcParams
    # - `location` change to "path" when p is in pathParams and
    #   unused when `type` is "object" or "file"
    # - When type is `object`, create a requestBody with content
    #   default to "application/json"
    # - When type is `file`, change requestBody content to
    #   multipart/form-data

    if (p %in% pathParams$name) {
      location <- "path"
      required <- TRUE
      style <- "simple"
      explode <- FALSE
      type <- priorizeProperty(defaultApiType,
                               pathParams[pathParams$name == p,]$type,
                               endpointParams[[p]]$type,
                               funcParams[[p]]$type)
      type <- plumberToApiType(type, inPath = TRUE)
      isArray <- priorizeProperty(defaultIsArray,
                                  pathParams[pathParams$name == p,]$isArray,
                                  endpointParams[[p]]$isArray,
                                  funcParams[[p]]$isArray)
    } else {
      location <- "query"
      style <- "form"
      explode <- TRUE
      type <- priorizeProperty(defaultApiType,
                               endpointParams[[p]]$type,
                               funcParams[[p]]$type)
      type <- plumberToApiType(type)
      isArray <- priorizeProperty(defaultIsArray,
                                  endpointParams[[p]]$isArray,
                                  funcParams[[p]]$isArray)
      required <- priorizeProperty(funcParams[[p]]$required,
                                   endpointParams[[p]]$required)
    }

    # Building OpenAPI Specification
    if (type %in% inBody) {
      if (length(params$requestBody) == 0L) {
        params$requestBody$content$`application/json`[["schema"]] <-
          list(type = "object", properties = list())
      }
      property <- list(
        type = type,
        format = apiTypesInfo[[type]]$format,
        example = funcParams[[p]]$example,
        description = endpointParams[[p]]$desc
      )
      if (type %in% inRaw) {
        names(params$requestBody$content) <- "multipart/form-data"
        property$type <- apiTypesInfo[[type]]$realType
        property$example <- NULL
      }
      if (isArray) {
        property$items <- list(
          type = property$type,
          format = property$format,
          example = property$example
        )
        property$type <- "array"
        property$format <- NULL
        property$example <- NULL
      }
      params$requestBody[[1]][[1]][[1]]$properties[[p]] <- property
      if (required) { params$requestBody[[1]][[1]][[1]]$required <-
        c(p, params$requestBody[[1]][[1]][[1]]$required)}
    } else {
      paramList <- list(
        name = p,
        description = endpointParams[[p]]$desc,
        `in` = location,
        required = required,
        schema = list(
          type = type,
          format = apiTypesInfo[[type]]$format,
          default = funcParams[[p]]$default
        )
      )
      if (isArray) {
        paramList$schema <- list(
          type = "array",
          items = list(
            type = type,
            format = apiTypesInfo[[type]]$format
          ),
          default = funcParams[[p]]$default
        )
        paramList$style <- style
        paramList$explode <- explode
      }
      params$parameters[[length(params$parameters) + 1]] <- paramList
    }

  }
  params
}

#' Check na
#' @noRd
isNa <- function(x) {
  if (is.list(x)) {
    return(FALSE)
  }
  is.na(x)
}

#' Check na or null
#' @noRd
isNaOrNull <- function(x) {
  any(isNa(x)) || is.null(x)
}

#' Remove na or null
#' @noRd
removeNaOrNulls <- function(x) {
  # preemptively stop
  if (!is.list(x)) {
    return(x)
  }
  if (length(x) == 0) {
    return(x)
  }

  # remove any `NA` or `NULL` elements
  toRemove <- vapply(x, isNaOrNull, logical(1))
  if (any(toRemove)) {
    x[toRemove] <- NULL
  }

  # Recurse through list
  # Store to `x[]` to not overwrite any classes or attributes
  x[] <-
    Map(
      rlang::names2(x), # Ask again. It has prolly changed
      x,
      f = function(key, value) {
        switch(key,
          "example" = {
            # Don't do anything
            value
          },
          "examples" = {
            # Remove all NA or NULL values from fields other than `value`
            # https://spec.openapis.org/oas/v3.1.0.html#example-object
            ret <- value # Copy val for less confusion in code
            notValuePos <- rlang::names2(ret) != "value"
            ret[notValuePos] <- removeNaOrNulls(ret[notValuePos])
            ret
          },
          {
            if (
              # Ignore extensions
              grepl("^x-", key)
            ) {
              # Return value as is
              value
            } else {
              # Recurse through list
              removeNaOrNulls(value)
            }
          }
        )
      }
    )

  x
}

#' For OpenAPI Specification
#' @noRd
priorizeProperty <- function(...) {
  l <- list(...)
  if (length(l) > 0L) {
    isnullordefault <- vapply(l, function(x) {isNaOrNull(x) || isTRUE(attributes(x)$default)}, logical(1))
    # return the position of the first FALSE value or position 1 if all values are TRUE
    return(l[[which.min(isnullordefault)]])
  }
  NULL # do not return any value
}

#' Check if x is JSON serializable
#' @noRd
#' @importFrom jsonlite toJSON
isJSONserializable <- function(x) {
  tryCatch(
    {
      toJSON(x)
      TRUE
    },
    error = function(cond) {
      # Do we need to test for specific errors?
      FALSE
    }
  )
}

#' Extract metadata on args of endpoint_func
#' @noRd
getArgsMetadata <- function(endpoint_func) {
  #return same format as getTypedParams or params?
  args <- formals(endpoint_func)
  lapply(args[! (names(args) %in% c("req", "res", "..."))], function(arg) {
    required <- identical(arg, formals(function(x){})$x)
    if (is.call(arg) || is.name(arg)) {
      arg <- tryCatch(
        eval(arg, envir = environment(endpoint_func)),
        error = function(cond) {NA})
    }
    # Check that it is possible to transform arg value into
    # an example for the OpenAPI spec. Valid transform are
    # either a logical, a numeric, a character or a list that
    # is json serializable. Otherwise set to NA.
    if (!is.logical(arg) && !is.numeric(arg) && !is.character(arg)
        && !(is.list(arg) && isJSONserializable(arg))) {
      message("Argument of class ", class(arg), " cannot be used to set default value in OpenAPI Specifications.")
      arg <- NA
    }
    type <- if (isNaOrNull(arg)) {NA} else {typeof(arg)}
    type <- plumberToApiType(type)
    isArray <- if (isTRUE(length(arg) > 1L)) {TRUE} else {defaultIsArray}
    list(
      default = arg,
      example = arg,
      required = required,
      isArray = isArray,
      type = type
    )
  })
}

cookieFilter <- function(req){
  cookie <- req$HTTP_COOKIE
  req$cookies <- parseCookies(cookie)
  forward()
}

#' @noRd
parseCookies <- function(cookie) {
  if (is.null(cookie) || nchar(cookie) == 0) {
    return(list())
  }
  cookie <- strsplit(cookie, ";", fixed=TRUE)[[1]]
  cookie <- sub("\\s*([\\S*])\\s*", "\\1", cookie, perl=TRUE)

  cookieList <- stri_split_fixed(str = cookie, pattern = "=", n = 2)

  # Handle any non-existent cookie values.
  for (i in seq_along(cookieList)) {
    if(length(cookieList[[i]]) == 1) {
      cookieList[[i]][[2]] <- ""
    }
  }

  cookies <- vapply(cookieList, "[[", character(1), 2)
  decodedCookies <- as.list(httpuv::decodeURIComponent(cookies))
  cookieNames <- vapply(cookieList, "[[", character(1), 1)
  names(decodedCookies) <- cookieNames
  decodedCookies
}

requireRmd <- function(fun_name){
  if (!requireNamespace("rmarkdown", quietly = TRUE)) {
    stop("The rmarkdown package is not available but is required in order to use ", fun_name,
         call. = FALSE)
  }
}

#' Send File Contents as Response
#'
#' Returns the file at the given path as the response. If you want an endpoint to return a file as an attachment for user to download see [as_attachment()].
#'
#' \code{include_html} will merely return the file with the proper
#' \code{content_type} for HTML. \code{include_md} and \code{include_rmd} will
#' process the given markdown file through \code{rmarkdown::render} and return
#' the resultant HTML as a response.
#'
#' @param file The path to the file to return
#' @param res The response object into which we'll write
#' @param content_type If provided, the given value will be sent as the
#'  `Content-Type` header in the response. Defaults to the contentType of the file extension.
#' To disable the `Content-Type` header, set `content_type = NULL`.
#' @export
include_file <- function(file, res, content_type = getContentType(tools::file_ext(file))){
  # TODO stream this directly to the request w/o loading in memory
  lines <- paste(readLines(file), collapse="\n")
  res$serializer <- "null"
  res$body <- c(res$body, lines)

  if (!is.null(content_type)) {
    res$setHeader("Content-Type", content_type)
  }

  res
}

#' @rdname include_file
#' @export
include_html <- function(file, res){
  include_file(file, res, content_type="text/html; charset=UTF-8")
}

#' @rdname include_file
#' @param format Passed as the \code{output_format} to \code{rmarkdown::render}
#' @export
include_md <- function(file, res, format = NULL){
  requireRmd("include_md")

  f <- rmarkdown::render(file, format, quiet=TRUE)
  include_html(f, res)
}

#' @rdname include_file
#' @export
include_rmd <- function(file, res, format = NULL){
  requireRmd("include_rmd")

  f <- rmarkdown::render(file, format, quiet=TRUE)
  include_html(f, res)
}

#' @noRd
sharedSecretFilter <- function(req, res){
  secret <- getOption("plumber.sharedSecret", NULL)
  if (!is.null(secret)){
    supplied <- req$HTTP_PLUMBER_SHARED_SECRET
    if (!identical(supplied, secret)){
      res$status <- 400
      # Force the route to return as unboxed json
      res$serializer <- serializer_unboxed_json()
      # Using output similar to `defaultErrorHandler()`
      li <- list(error = "400 - Bad request")
      
      # Don't overly leak data unless they opt-in
      if (is.function(req$pr$getDebug) && isTRUE(req$pr$getDebug())) {
        li$message <- "Shared secret mismatch"
      }
      return(li)
    }
  }

  forward()
}

#' Determine if Plumber object
#'
#' @param pr Hopefully a [`Plumber`] object
#' @return Logical value if `pr` inherits from [`Plumber`]
#' @export
#' @examples
#' is_plumber(Plumber$new()) # TRUE
#' is_plumber(list()) # FALSE
is_plumber <- function(pr) {
  inherits(pr, "Plumber")
}

validate_pr <- function(pr) {
  if (!is_plumber(pr)) {
    stop("`pr` must be an object of class `Plumber`.")
  }
  invisible(TRUE)
}



#' Create a new Plumber router
#'
#' @param filters A list of Plumber filters
#' @param file Path to file to plumb
#' @param envir An environment to be used as the enclosure for the routers execution
#'
#' @return A new [`Plumber`] router
#'
#' @examples
#' \dontrun{
#' pr() %>%
#'   pr_run()
#' }
#'
#' @export
pr <- function(file = NULL,
               filters = defaultPlumberFilters,
               envir = new.env(parent = .GlobalEnv)) {
  Plumber$new(file = file, filters = filters, envir = envir)
}

#' Add handler to Plumber router
#'
#' This collection of functions creates handlers for a Plumber router.
#'
#' The generic [pr_handle()] creates a handle for the given method(s). Specific
#' functions are implemented for the following HTTP methods:
#' * `GET`
#' * `POST`
#' * `PUT`
#' * `DELETE`
#' * `HEAD`
#' Each function mutates the Plumber router in place and returns
#' the updated router.
#'
#' @template param_pr
#' @param methods Character vector of HTTP methods
#' @param path The endpoint path
#' @param handler A handler function
#' @param preempt A preempt function
#' @param serializer A Plumber serializer
#' @param endpoint A `PlumberEndpoint` object
#' @param ... Additional arguments for `PlumberEndpoint`
#'
#' @return A Plumber router with the handler added
#'
#' @examples
#' \dontrun{
#' pr() %>%
#'   pr_handle("GET", "/hi", function() "Hello World") %>%
#'   pr_run()
#'
#' pr() %>%
#'   pr_handle(c("GET", "POST"), "/hi", function() "Hello World") %>%
#'   pr_run()
#'
#' pr() %>%
#'   pr_get("/hi", function() "Hello World") %>%
#'   pr_post("/echo", function(req, res) {
#'     if (is.null(req$body)) return("No input")
#'     list(
#'       input = req$body
#'     )
#'   }) %>%
#'   pr_run()
#' }
#'
#' @export
pr_handle <- function(pr,
                      methods,
                      path,
                      handler,
                      preempt,
                      serializer,
                      endpoint,
                      ...) {
  validate_pr(pr)
  pr$handle(methods = methods,
            path = path,
            handler = handler,
            preempt = preempt,
            serializer = serializer,
            endpoint = endpoint,
            ...)
  pr
}

#' @rdname pr_handle
#' @export
pr_get <- function(pr,
                   path,
                   handler,
                   preempt,
                   serializer,
                   endpoint,
                   ...) {
  pr_handle(pr,
            "GET",
            path = path,
            handler = handler,
            preempt = preempt,
            serializer = serializer,
            endpoint = endpoint,
            ...)
}

#' @rdname pr_handle
#' @export
pr_post <- function(pr,
                    path,
                    handler,
                    preempt,
                    serializer,
                    endpoint,
                    ...) {
  pr_handle(pr,
            "POST",
            path = path,
            handler = handler,
            preempt = preempt,
            serializer = serializer,
            endpoint = endpoint,
            ...)
}

#' @rdname pr_handle
#' @export
pr_put <- function(pr,
                   path,
                   handler,
                   preempt,
                   serializer,
                   endpoint,
                   ...) {
  pr_handle(pr,
            "PUT",
            path = path,
            handler = handler,
            preempt = preempt,
            serializer = serializer,
            endpoint = endpoint,
            ...)
}

#' @rdname pr_handle
#' @export
pr_delete <- function(pr,
                      path,
                      handler,
                      preempt,
                      serializer,
                      endpoint,
                      ...) {
  pr_handle(pr,
            "DELETE",
            path = path,
            handler = handler,
            preempt = preempt,
            serializer = serializer,
            endpoint = endpoint,
            ...)
}

#' @rdname pr_handle
#' @export
pr_head <- function(pr,
                    path,
                    handler,
                    preempt,
                    serializer,
                    endpoint,
                    ...) {
  pr_handle(pr,
            "HEAD",
            path = path,
            handler = handler,
            preempt = preempt,
            serializer = serializer,
            endpoint = endpoint,
            ...)
}

#' Mount a Plumber router
#'
#' Plumber routers can be “nested” by mounting one into another
#' using the `mount()` method. This allows you to compartmentalize your API
#' by paths which is a great technique for decomposing large APIs into smaller
#' files. This function mutates the Plumber router ([pr()]) in place and
#' returns the updated router.
#'
#' @param pr The host Plumber router.
#' @param path A character string. Where to mount router.
#' @param router A Plumber router. Router to be mounted.
#'
#' @return A Plumber router with the supplied router mounted
#'
#' @examples
#' \dontrun{
#' pr1 <- pr() %>%
#'   pr_get("/hello", function() "Hello")
#'
#' pr() %>%
#'   pr_get("/goodbye", function() "Goodbye") %>%
#'   pr_mount("/hi", pr1) %>%
#'   pr_run()
#' }
#'
#' @export
pr_mount <- function(pr,
                     path,
                     router) {
  validate_pr(pr)
  pr$mount(path = path, router = router)
  pr
}

#' Register a hook
#'
#' Plumber routers support the notion of "hooks" that can be registered
#' to execute some code at a particular point in the lifecycle of a request.
#' Plumber routers currently support four hooks:
#'  1. `preroute(data, req, res)`
#'  2. `postroute(data, req, res, value)`
#'  3. `preserialize(data, req, res, value)`
#'  4. `postserialize(data, req, res, value)`
#' In all of the above you have access to a disposable environment in the `data`
#' parameter that is created as a temporary data store for each request. Hooks
#' can store temporary data in these hooks that can be reused by other hooks
#' processing this same request.
#'
#' One feature when defining hooks in Plumber routers is the ability to modify
#' the returned value. The convention for such hooks is: any function that accepts
#' a parameter named `value` is expected to return the new value. This could
#' be an unmodified version of the value that was passed in, or it could be a
#' mutated value. But in either case, if your hook accepts a parameter
#' named `value`, whatever your hook returns will be used as the new value
#' for the response.
#'
#' You can add hooks using the `pr_hook`, or you can add multiple
#' hooks at once using `pr_hooks`, which takes a named list in
#' which the names are the names of the hooks, and the values are the
#' handlers themselves.
#'
#' @template param_pr
#' @param stage A character string. Point in the lifecycle of a request.
#' @param handler A hook function.
#' @param handlers A named list of hook handlers
#'
#' @return A Plumber router with the defined hook(s) added
#'
#' @examples
#' \dontrun{
#' pr() %>%
#'   pr_hook("preroute", function(req){
#'     cat("Routing a request for", req$PATH_INFO, "...\n")
#'   }) %>%
#'   pr_hooks(list(
#'     preserialize = function(req, value){
#'       print("About to serialize this value:")
#'       print(value)
#'
#'       # Must return the value since we took one in. Here we're not choosing
#'       # to mutate it, but we could.
#'       value
#'     },
#'     postserialize = function(res){
#'       print("We serialized the value as:")
#'       print(res$body)
#'     }
#'   )) %>%
#'   pr_handle("GET", "/", function(){ 123 }) %>%
#'   pr_run()
#' }
#'
#' @export
pr_hook <- function(pr,
                    stage,
                    handler) {
  validate_pr(pr)
  pr$registerHook(stage = stage, handler = handler)
  pr
}

#' @rdname pr_hook
#' @export
pr_hooks <- function(pr,
                     handlers) {
  validate_pr(pr)
  pr$registerHooks(handlers)
  pr
}

#' Store session data in encrypted cookies.
#'
#' \code{plumber} uses the crypto R package \code{sodium}, to encrypt/decrypt
#' \code{req$session} information for each server request.
#'
#' The cookie's secret encryption \code{key} value must be consistent to maintain
#' \code{req$session} information between server restarts.
#'
#' @section Storing secure keys:
#' While it is very quick to get started with user session cookies using
#' \code{plumber}, please exercise precaution when storing secure key information.
#' If a malicious person were to gain access to the secret \code{key}, they would
#' be able to eavesdrop on all \code{req$session} information and/or tamper with
#' \code{req$session} information being processed.
#'
#' Please: \itemize{
#' \item Do NOT store keys in source control.
#' \item Do NOT store keys on disk with permissions that allow it to be accessed by everyone.
#' \item Do NOT store keys in databases which can be queried by everyone.
#' }
#'
#' Instead, please: \itemize{
#' \item Use a key management system, such as
#' \href{https://github.com/r-lib/keyring}{'keyring'} (preferred)
#' \item Store the secret in a file on disk with appropriately secure permissions,
#'   such as "user read only" (\code{Sys.chmod("myfile.txt", mode = "0600")}),
#'   to prevent others from reading it.
#' } Examples of both of these solutions are done in the Examples section.
#'
#' @template param_pr
#' @param key The secret key to use. This must be consistent across all R sessions
#'   where you want to save/restore encrypted cookies. It should be produced using
#'   \code{\link{random_cookie_key}}. Please see the "Storing secure keys" section for more details
#'   complex character string to bolster security.
#' @param name The name of the cookie in the user's browser.
#' @param path The URI path that the cookie will be available in future requests.
#'    Defaults to the request URI. Set to \code{"/"} to make cookie available to
#'    all requests at the host.
#' @param expiration A number representing the number of seconds into the future
#'   before the cookie expires or a \code{POSIXt} date object of when the cookie expires.
#'   Defaults to the end of the user's browser session.
#' @param http Boolean that adds the \code{HttpOnly} cookie flag that tells the browser
#'   to save the cookie and to NOT send it to client-side scripts. This mitigates \href{https://developer.mozilla.org/en-US/docs/Glossary/Cross-site_scripting}{cross-site scripting}.
#'   Defaults to \code{TRUE}.
#' @param secure Boolean that adds the \code{Secure} cookie flag.  This should be set
#'   when the route is eventually delivered over \href{https://en.wikipedia.org/wiki/HTTPS}{HTTPS}.
#' @param same_site A character specifying the SameSite policy to attach to the cookie.
#'   If specified, one of the following values should be given: "Strict", "Lax", or "None".
#'   If "None" is specified, then the \code{secure} flag MUST also be set for the modern browsers to
#'   accept the cookie. An error will be returned if \code{same_site = "None"} and \code{secure = FALSE}.
#'   If not specified or a non-character is given, no SameSite policy is attached to the cookie.
#' @seealso \itemize{
#' \item \href{https://github.com/r-lib/sodium}{'sodium'}: R bindings to 'libsodium'
#' \item \href{https://doc.libsodium.org/}{'libsodium'}: A Modern and Easy-to-Use Crypto Library
#' \item \href{https://github.com/r-lib/keyring}{'keyring'}: Access the system credential store from R
#' \item \href{https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie#Directives}{Set-Cookie flags}: Descriptions of different flags for \code{Set-Cookie}
#' \item \href{https://developer.mozilla.org/en-US/docs/Glossary/Cross-site_scripting}{Cross-site scripting}: A security exploit which allows an attacker to inject into a website malicious client-side code
#' }
#' @examples
#' \dontrun{
#'
#' ## Set secret key using `keyring` (preferred method)
#' keyring::key_set_with_value("plumber_api", password = plumber::random_cookie_key())
#'
#'
#' pr() %>%
#'   pr_cookie(
#'     keyring::key_get("plumber_api"),
#'     name = "counter"
#'   ) %>%
#'   pr_get("/sessionCounter", function(req) {
#'     count <- 0
#'     if (!is.null(req$session$counter)){
#'       count <- as.numeric(req$session$counter)
#'     }
#'     req$session$counter <- count + 1
#'     return(paste0("This is visit #", count))
#'   }) %>%
#'   pr_run()
#'
#'
#' #### -------------------------------- ###
#'
#'
#' ## Save key to a local file
#' pswd_file <- "normal_file.txt"
#' cat(plumber::random_cookie_key(), file = pswd_file)
#' # Make file read-only
#' Sys.chmod(pswd_file, mode = "0600")
#'
#' pr() %>%
#'   pr_cookie(
#'     readLines(pswd_file, warn = FALSE),
#'     name = "counter"
#'   ) %>%
#'   pr_get("/sessionCounter", function(req) {
#'     count <- 0
#'     if (!is.null(req$session$counter)){
#'       count <- as.numeric(req$session$counter)
#'     }
#'     req$session$counter <- count + 1
#'     return(paste0("This is visit #", count))
#'   }) %>%
#'   pr_run()
#' }
#' @export
pr_cookie <- function(pr,
                      key,
                      name = "plumber",
                      expiration = FALSE,
                      http = TRUE,
                      secure = FALSE,
                      same_site = FALSE,
                      path = NULL) {
  validate_pr(pr)
  pr$registerHooks(
    session_cookie(
      key = key,
      name = name,
      expiration = expiration,
      http = http,
      secure = secure,
      same_site = same_site,
      path = path
    )
  )
  pr
}



#' Add a filter to Plumber router
#'
#' Filters can be used to modify an incoming request, return an error, or return
#' a response prior to the request reaching an endpoint.
#'
#' @template param_pr
#' @param name A character string. Name of filter
#' @param expr An expr that resolve to a filter function or a filter function
#' @param serializer A serializer function
#'
#' @return The Plumber router with the defined filter added
#'
#' @examples
#' \dontrun{
#' pr() %>%
#'   pr_filter("foo", function(req, res) {
#'     print("This is filter foo")
#'     forward()
#'   }) %>%
#'   pr_get("/hi", function() "Hello") %>%
#'   pr_run()
#' }
#'
#' @export
pr_filter <- function(pr,
                      name,
                      expr,
                      serializer) {
  validate_pr(pr)
  pr$filter(name = name, expr = expr, serializer = serializer)
  pr
}

#' Start a server using `plumber` object
#'
#' `port` does not need to be explicitly assigned.
#'
#'
#' @template param_pr
#' @param host A string that is a valid IPv4 or IPv6 address that is owned by
#' this server, which the application will listen on. "0.0.0.0" represents
#' all IPv4 addresses and "::/0" represents all IPv6 addresses.
#' @param port A number or integer that indicates the server port that should
#' be listened on. Note that on most Unix-like systems including Linux and
#' Mac OS X, port numbers smaller than 1025 require root privileges.
#' @param ... Should be empty.
#' @param debug If `TRUE`, it will provide more insight into your API errors.
#'   Using this value will only last for the duration of the run.
#'   If [pr_set_debug()] has not been called, `debug` will default to `interactive()` at [pr_run()] time
#' @param docs Visual documentation value to use while running the API.
#'   This value will only be used while running the router.
#'   If missing, defaults to information previously set with [pr_set_docs()].
#'   For more customization, see [pr_set_docs()] for examples.
#' @param swaggerCallback An optional single-argument function that is called
#'   back with the URL to an OpenAPI user interface when one becomes ready. If
#'   missing, defaults to information set with [pr_set_docs_callback()].
#'   This value will only be used while running the router.
#' @param quiet If `TRUE`, don't print routine startup messages.
#'
#' @examples
#' \dontrun{
#' pr() %>%
#'   pr_run()
#'
#' pr() %>%
#'   pr_run(
#'     # manually set port
#'     port = 5762,
#'     # turn off visual documentation
#'     docs = FALSE,
#'     # do not display startup messages
#'     quiet = TRUE
#'   )
#' }
#'
#' @export
pr_run <- function(pr,
                   host = '127.0.0.1',
                   port = getOption('plumber.port', NULL),
                   ...,
                   debug = missing_arg(),
                   docs = missing_arg(),
                   swaggerCallback = missing_arg(),
                   quiet = FALSE
) {
  validate_pr(pr)
  ellipsis::check_dots_empty()
  pr$run(host = host,
         port = port,
         debug = debug,
         docs = docs,
         swaggerCallback = swaggerCallback,
         quiet = quiet)
}


#' Add a static route to the `plumber` object
#'
#' @template param_pr
#' @param path The mounted path location of the static folder
#' @param direc The local folder to be served statically
#'
#' @examples
#' \dontrun{
#' pr() %>%
#'   pr_static("/path", "./my_folder/location") %>%
#'   pr_run()
#' }
#'
#' @export
pr_static <- function(
  pr,
  path,
  direc
) {
  pr_mount(pr, path, PlumberStatic$new(direc))
}

context("Tidy Plumber API")

test_that("pr creates new router", {
  p <- pr()
  expect_equal(class(p), c("Plumber", "Hookable", "R6"))
})

test_that("pr_handle functions add routes", {
  p <- pr() %>%
    pr_handle(c("GET", "POST"),
              "/foo",
              function() "bar")
  ep <- p$endpoints[[1]][[1]]
  expect_equal(ep$verbs, c("GET", "POST"))
  expect_equal(ep$path, "/foo")

  p <- pr() %>%
    pr_get("/foo", function() "bar")
  ep <- p$endpoints[[1]][[1]]
  expect_equal(ep$verbs, c("GET"))
  expect_equal(ep$path, "/foo")

  p <- pr() %>%
    pr_post("/foo", function() "bar")
  ep <- p$endpoints[[1]][[1]]
  expect_equal(ep$verbs, c("POST"))
  expect_equal(ep$path, "/foo")

  p <- pr() %>%
    pr_put("/foo", function() "bar")
  ep <- p$endpoints[[1]][[1]]
  expect_equal(ep$verbs, c("PUT"))
  expect_equal(ep$path, "/foo")

  p <- pr() %>%
    pr_delete("/foo", function() "bar")
  ep <- p$endpoints[[1]][[1]]
  expect_equal(ep$verbs, c("DELETE"))
  expect_equal(ep$path, "/foo")

  p <- pr() %>%
    pr_head("/foo", function() "bar")
  ep <- p$endpoints[[1]][[1]]
  expect_equal(ep$verbs, c("HEAD"))
  expect_equal(ep$path, "/foo")
})

test_that("pr_mount mounts router", {
  p1 <- pr() %>%
    pr_get("/hello", function() "Hello")

  p2 <- pr() %>%
    pr_mount("/hi", p1)

  expect_equal(length(p2$mounts), 1)
})

test_that("pr_static mounts a static router", {
  p1 <- pr() %>%
    pr_static("/ex", system.file("plumber", package = "plumber"))

  expect_equal(length(p1$mounts), 1)
  expect_equal(names(p1$mounts), "/ex/")
  expect_s3_class(p1$mounts[[1]], "PlumberStatic")
})

test_that("pr_hooks registers hooks", {
  p <- pr() %>%
    pr_hook("preroute", function() print("Pre-route hook")) %>%
    pr_get("/hello", function() "Hello")

  req <- make_req("GET", "/hello")

  expect_output(p$call(req), "Pre-route hook")

  hooks <- list(
    preroute = function() print("Pre-route hook"),
    postroute = function() print("Post-route hook"),
    preserialize = function() print("Pre-serialize hook"),
    postserialize = function() print("Post-serialize hook")
  )

  p <- pr() %>%
    pr_hooks(hooks) %>%
    pr_get("/hello", function() "Hello")

  expect_output(p$call(req), "Pre-route hook")
  expect_output(p$call(req), "Pre-serialize hook")
  expect_output(p$call(req), "Post-route hook")
  expect_output(p$call(req), "Post-serialize hook")
})

test_that("pr_cookie adds cookie", {
  p <- pr() %>%
    pr_cookie(
      random_cookie_key(),
      name = "counter"
    ) %>%
    pr_get("/sessionCounter", function(req) {
      count <- 0
      if (!is.null(req$session$counter)){
        count <- as.numeric(req$session$counter)
      }
      req$session$counter <- count + 1
      return(paste0("This is visit #", count))
    })

  req <- make_req("GET", "/sessionCounter")
  expect_match(p$call(req)$headers$`Set-Cookie`, "^counter=")
})

test_that("pr_cookie adds path in cookie", {
  p <- pr() %>%
    pr_cookie(
      random_cookie_key(),
      name = "counter",
      path = "/test"
    ) %>%
    pr_get("/test/route", function(req) {
      count <- 0
      if (!is.null(req$session$counter)){
        count <- as.numeric(req$session$counter)
      }
      req$session$counter <- count + 1
      return(paste0("This is visit #", count))
    }) %>%
    pr_get("/testing", function(req) {
      return(paste0("Matching visits #", as.numeric(req$session$counter)))
    })

  req1 <- make_req("GET", "/test/route")
  cookie_header1 <- p$call(req1)$headers$`Set-Cookie`
  expect_match(cookie_header1, "^counter=")
  expect_match(cookie_header1, "Path=/test")

  req2 <- make_req("GET", "/testing")
  cookie_header2 <- p$call(req2)$headers$`Set-Cookie`
  # Path does not match cookie location, so no cookie is set
  expect_equal(cookie_header2, NULL)
})


test_that("pr default functions perform as expected", {
  # Serializer
  serialized <- function(...) {
    serializer_content_type("text/plain", function(val) {
      paste0("Serialized value: '", val, "'")
    })
  }

  p <- pr() %>%
    pr_set_serializer(serialized()) %>%
    pr_get("/hello", function() "Hello")

  req <- make_req("GET", "/hello")

  res <- p$call(req)

  expect_equal(res$body, "Serialized value: 'Hello'")

  # 404 Handler
  handler_404 <- function(req, res) {
    res$status <- 404
    res$body <- "Oops"
  }

  p <- pr() %>%
    pr_get("/hello", function() "Hello") %>%
    pr_set_404(handler_404)

  req <- make_req("GET", "/foo")

  res <- p$call(req)

  expect_equal(res$status, 404)
  expect_equal(res$body, jsonlite::toJSON("Oops"))

  # Error handler
  handler_error <- function(req, res, err){
    res$status <- 500
    list(error = "Custom Error Message")
  }

  p <- pr() %>%
    pr_get("/error", function() log("a")) %>%
    pr_set_error(handler_error)

  req <- make_req("GET", "/error")

  res <- p$call(req)

  expect_equal(res$status, 500)
  expect_equal(jsonlite::fromJSON(res$body)[[1]], "Custom Error Message")
})

test_that("pr_filter adds filters", {
  p <- pr()
  initial_filters <- length(p$filters)
  p <- p %>%
    pr_filter("foo", function() {
      print("Filter foo")
      forward()
    }) %>%
    pr_get("/hello", function() "Hello")

  expect_gt(length(p$filters), initial_filters)
})

return_inputs <- function(...) {
  ret <- list(...)
  ret$req <- NULL
  ret$res <- NULL
  ret
}


#* @post /default
return_inputs

#* @post /json
#* @parser json
return_inputs

#* @post /mixed
#* @parser form
#* @parser json
return_inputs

#* @post /repeated
#* @parser json
#* @parser json
return_inputs

#* @post /none
#* @parser none
return_inputs

#* @post /all
#* @parser all
return_inputs

# Taken from shiny
# @author shiny authors
resolve_path <- function(dir, relpath) {
  abs.path <- file.path(dir, relpath)
  if (!file.exists(abs.path))
    return(NULL)
  abs.path <- normalizePath(abs.path, winslash='/', mustWork=TRUE)
  dir <- normalize_dir_path(dir)
  if (nchar(abs.path) <= nchar(dir) + 1)
    return(NULL)
  if (substr(abs.path, 1, nchar(dir)) != dir ||
      substr(abs.path, nchar(dir)+1, nchar(dir)+1) != '/') {
    return(NULL)
  }
  return(abs.path)
}

normalize_dir_path <- function(dir) {
  dir <- normalizePath(dir, winslash = '/', mustWork = TRUE)
  # trim the possible trailing slash under Windows (#306)
  if (isWindows()) {
    dir <- sub('/$', '', dir)
  }
  dir
}

# We can't naively compare serializers using expect equal without losing codecov results.
# codecov modifies the source of the functions so they are no longer comparable when
# deparsed, which causes the tests to fail only for codecov.
# Here we'll make our own comparison function.
# covr adds lines to measure coverage but also adds brackets to capture expressions differently.
# So we use this rough heuristic to just take the word characters without whitespace and compare
# those. It's not perfect, but it would almost always fail if you were comparing to different
# functions.
expect_equal_functions <- function(object, expected){
  do <- deparse(object)
  de <- deparse(expected)

  do <- gsub(".*covr:::count.*", NA, do)
  do <- do[!is.na(do)]
  do <- paste(do, collapse="")
  do <- gsub("[^\\w]", "", do, perl=TRUE)

  de <- gsub(".*covr:::count.*", NA, de)
  de <- de[!is.na(de)]
  de <- paste(de, collapse="")
  de <- gsub("[^\\w]", "", de, perl=TRUE)

  expect_equal(do, de)
}

context("Preempts")

test_that("preempts work", {
  r <- pr(test_path("files/preempt.R"))
  expect_equal(length(r$endpoints), 3)

  expect_length(r$endpoints[["testFun"]], 1)
  expect_length(r$endpoints[["testFun2"]], 1)
  expect_length(r$endpoints[["testFun3"]], 1)
})

test_that("Redundant preempts fail", {
  expect_error(pr(test_path("files/preempt-redundant.R")), regexp="Multiple @preempts")
})

test_that("Empty preempts fail", {
  expect_error(pr(test_path("files/preempt-empty.R")), regexp="No @preempt specified")
})

test_that("Non-existant preempts fail", {
  expect_error(pr(test_path("files/preempt-nonexistent.R")), regexp="The given @preempt")
})

context("JSON")

test_that("safeFromJSON is safe", {
  # Valid JSON works
  a <- safeFromJSON('{"key": "value"}')
  expect_equal(a, list(key="value"))

  # File paths fail
  expect_error(safeFromJSON("/etc/passwd")) # error from jsonlite::parse_json()

  # Remote URLs fail
  expect_error(safeFromJSON("http://server.org/data.json")) # error from jsonlite::parse_json()
})

---
title: "test"
---

Overview
--------

This is an R Markdown document. Markdown is a simple formatting syntax
for authoring HTML, PDF, and MS Word documents. For more details on
using R Markdown see <http://rmarkdown.rstudio.com>.

R Output
--------

When you click the **Knit** button a document will be generated that
includes both content as well as the output of any embedded R code
chunks within the document. You can embed an R code chunk like this:

    summary(cars)

    ##      speed           dist       
    ##  Min.   : 4.0   Min.   :  2.00  
    ##  1st Qu.:12.0   1st Qu.: 26.00  
    ##  Median :15.0   Median : 36.00  
    ##  Mean   :15.4   Mean   : 42.98  
    ##  3rd Qu.:19.0   3rd Qu.: 56.00  
    ##  Max.   :25.0   Max.   :120.00

Plots
-----

You can also embed plots, for example:

Note that the `echo = FALSE` parameter was added to the code chunk to
prevent printing of the R code that generated the plot.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr.R
\name{is_plumber}
\alias{is_plumber}
\title{Determine if Plumber object}
\usage{
is_plumber(pr)
}
\arguments{
\item{pr}{Hopefully a \code{\link{Plumber}} object}
}
\value{
Logical value if \code{pr} inherits from \code{\link{Plumber}}
}
\description{
Determine if Plumber object
}
\examples{
is_plumber(Plumber$new()) # TRUE
is_plumber(list()) # FALSE
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr_set.R
\name{pr_set_debug}
\alias{pr_set_debug}
\title{Set debug value to include error messages of routes cause an error}
\usage{
pr_set_debug(pr, debug = interactive())
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{debug}{\code{TRUE} provides more insight into your API errors.}
}
\value{
The Plumber router with the new debug setting.
}
\description{
To hide any error messages in production, set the debug value to \code{FALSE}.
The \code{debug} value is enabled by default for \code{\link[=interactive]{interactive()}} sessions.
}
\examples{
\dontrun{
# Will contain the original error message
pr() \%>\%
  pr_set_debug(TRUE) \%>\%
  pr_get("/boom", function() stop("boom")) \%>\%
  pr_run()

# Will NOT contain an error message
pr() \%>\%
  pr_set_debug(FALSE) \%>\%
  pr_get("/boom", function() stop("boom")) \%>\%
  pr_run()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serializer.R
\name{serializer_headers}
\alias{serializer_headers}
\alias{serializer_content_type}
\alias{serializer_octet}
\alias{serializer_csv}
\alias{serializer_tsv}
\alias{serializer_html}
\alias{serializer_json}
\alias{serializer_unboxed_json}
\alias{serializer_geojson}
\alias{serializer_rds}
\alias{serializer_feather}
\alias{serializer_parquet}
\alias{serializer_yaml}
\alias{serializer_text}
\alias{serializer_format}
\alias{serializer_print}
\alias{serializer_cat}
\alias{serializer_write_file}
\alias{serializer_htmlwidget}
\alias{serializer_device}
\alias{serializer_jpeg}
\alias{serializer_png}
\alias{serializer_svg}
\alias{serializer_bmp}
\alias{serializer_tiff}
\alias{serializer_pdf}
\title{Plumber Serializers}
\usage{
serializer_headers(headers = list(), serialize_fn = identity)

serializer_content_type(type, serialize_fn = identity)

serializer_octet(..., type = "application/octet-stream")

serializer_csv(..., type = "text/csv; charset=UTF-8")

serializer_tsv(..., type = "text/tab-separated-values; charset=UTF-8")

serializer_html(type = "text/html; charset=UTF-8")

serializer_json(..., type = "application/json")

serializer_unboxed_json(auto_unbox = TRUE, ..., type = "application/json")

serializer_geojson(..., type = "application/geo+json")

serializer_rds(version = "2", ascii = FALSE, ..., type = "application/rds")

serializer_feather(type = "application/vnd.apache.arrow.file")

serializer_parquet(type = "application/vnd.apache.parquet")

serializer_yaml(..., type = "text/x-yaml; charset=UTF-8")

serializer_text(
  ...,
  serialize_fn = as.character,
  type = "text/plain; charset=UTF-8"
)

serializer_format(..., type = "text/plain; charset=UTF-8")

serializer_print(..., type = "text/plain; charset=UTF-8")

serializer_cat(..., type = "text/plain; charset=UTF-8")

serializer_write_file(type, write_fn, fileext = NULL)

serializer_htmlwidget(..., type = "text/html; charset=UTF-8")

serializer_device(type, dev_on, dev_off = grDevices::dev.off)

serializer_jpeg(..., type = "image/jpeg")

serializer_png(..., type = "image/png")

serializer_svg(..., type = "image/svg+xml")

serializer_bmp(..., type = "image/bmp")

serializer_tiff(..., type = "image/tiff")

serializer_pdf(..., type = "application/pdf")
}
\arguments{
\item{headers}{\code{list()} of headers to add to the response object}

\item{serialize_fn}{Function to serialize the data. The result object will be converted to a character string. Ex: \code{\link[jsonlite:fromJSON]{jsonlite::toJSON()}}.}

\item{type}{The value to provide for the \code{Content-Type} HTTP header.}

\item{...}{extra arguments supplied to respective internal serialization function.}

\item{auto_unbox}{automatically \code{\link[jsonlite:unbox]{unbox()}} all atomic vectors of length 1. It is usually safer to avoid this and instead use the \code{\link[jsonlite:unbox]{unbox()}} function to unbox individual elements.
An exception is that objects of class \code{AsIs} (i.e. wrapped in \code{\link[=I]{I()}}) are not automatically unboxed. This is a way to mark single values as length-1 arrays.}

\item{version}{the workspace format version to use.  \code{NULL}
    specifies the current default version (3). The only other supported
    value is 2, the default from \R 1.4.0 to \R 3.5.0.}

\item{ascii}{a logical.  If \code{TRUE} or \code{NA}, an ASCII
    representation is written; otherwise (default) a binary one.
    See also the comments in the help for \code{\link[base]{save}}.}

\item{write_fn}{Function that should write serialized content to the temp file provided. \code{write_fn} should have the function signature of \code{function(value, tmp_file){}}.}

\item{fileext}{A non-empty character vector giving the file extension. This value will try to be inferred from the content type provided.}

\item{dev_on}{Function to turn on a graphics device.
The graphics device \code{dev_on} function will receive any arguments supplied to the serializer in addition to \code{filename}.
\code{filename} points to the temporary file name that should be used when saving content.}

\item{dev_off}{Function to turn off the graphics device. Defaults to \code{\link[grDevices:dev]{grDevices::dev.off()}}}
}
\description{
Serializers are used in Plumber to transform the R object produced by a
filter/endpoint into an HTTP response that can be returned to the client. See
\href{https://www.rplumber.io/articles/rendering-output.html#serializers-1}{here} for
more details on Plumber serializers and how to customize their behavior.
}
\section{Functions}{
\itemize{
\item \code{serializer_headers()}: Add a static list of headers to each return value. Will add \code{Content-Disposition} header if a value is the result of \code{as_attachment()}.

\item \code{serializer_content_type()}: Adds a \code{Content-Type} header to the response object

\item \code{serializer_octet()}: Octet serializer. If content is received that does
not have a \code{"raw"} type, then an error will be thrown.

\item \code{serializer_csv()}: CSV serializer. See also: \code{\link[readr:format_delim]{readr::format_csv()}}

\item \code{serializer_tsv()}: TSV serializer. See also: \code{\link[readr:format_delim]{readr::format_tsv()}}

\item \code{serializer_html()}: HTML serializer

\item \code{serializer_json()}: JSON serializer. See also: \code{\link[jsonlite:fromJSON]{jsonlite::toJSON()}}

\item \code{serializer_unboxed_json()}: JSON serializer with \code{auto_unbox} defaulting to \code{TRUE}. See also: \code{\link[jsonlite:fromJSON]{jsonlite::toJSON()}}

\item \code{serializer_geojson()}: GeoJSON serializer. See also \code{\link[geojsonsf:sf_geojson]{geojsonsf::sf_geojson()}} and [\code{\link[geojsonsf:sfc_geojson]{geojsonsf::sfc_geojson()}}].

\item \code{serializer_rds()}: RDS serializer. See also: \code{\link[base:serialize]{base::serialize()}}

\item \code{serializer_feather()}: feather serializer. See also: \code{\link[arrow:write_feather]{arrow::write_feather()}}

\item \code{serializer_parquet()}: parquet serializer. See also: \code{\link[arrow:write_parquet]{arrow::write_parquet()}}

\item \code{serializer_yaml()}: YAML serializer. See also: \code{\link[yaml:as.yaml]{yaml::as.yaml()}}

\item \code{serializer_text()}: Text serializer. See also: \code{\link[=as.character]{as.character()}}

\item \code{serializer_format()}: Text serializer. See also: \code{\link[=format]{format()}}

\item \code{serializer_print()}: Text serializer. Captures the output of \code{\link[=print]{print()}}

\item \code{serializer_cat()}: Text serializer. Captures the output of \code{\link[=cat]{cat()}}

\item \code{serializer_write_file()}: Write output to a temp file whose contents are read back as a serialized response. \code{serializer_write_file()} creates (and cleans up) a temp file, calls the serializer (which should write to the temp file), and then reads the contents back as the serialized value.  If the content \code{type} starts with \code{"text"}, the return result will be read into a character string, otherwise the result will be returned as a raw vector.

\item \code{serializer_htmlwidget()}: htmlwidget serializer. See also: \code{\link[htmlwidgets:saveWidget]{htmlwidgets::saveWidget()}}

\item \code{serializer_device()}: Helper method to create graphics device serializers, such as \code{\link[=serializer_png]{serializer_png()}}. See also: \code{\link[=endpoint_serializer]{endpoint_serializer()}}

\item \code{serializer_jpeg()}: JPEG image serializer. See also: \code{\link[grDevices:png]{grDevices::jpeg()}}

\item \code{serializer_png()}: PNG image serializer. See also: \code{\link[grDevices:png]{grDevices::png()}}

\item \code{serializer_svg()}: SVG image serializer. See also: \code{\link[grDevices:cairo]{grDevices::svg()}}

\item \code{serializer_bmp()}: BMP image serializer. See also: \code{\link[grDevices:png]{grDevices::bmp()}}

\item \code{serializer_tiff()}: TIFF image serializer. See also: \code{\link[grDevices:png]{grDevices::tiff()}}

\item \code{serializer_pdf()}: PDF image serializer. See also: \code{\link[grDevices:pdf]{grDevices::pdf()}}

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr_set.R
\name{pr_set_api_spec}
\alias{pr_set_api_spec}
\title{Set the OpenAPI Specification}
\usage{
pr_set_api_spec(pr, api)
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{api}{This can be
\itemize{
\item an OpenAPI Specification formatted list object
\item a function that accepts the OpenAPI Specification autogenerated by \code{plumber} and returns a OpenAPI Specification formatted list object.
\item a path to an OpenAPI Specification
}

The value returned will not be validated for OAS compatibility.}
}
\value{
The Plumber router with the new OpenAPI Specification object or function.
}
\description{
Allows to modify OpenAPI Specification autogenerated by \code{plumber}.
}
\details{
Note, the returned value will be sent through \code{\link[=serializer_unboxed_json]{serializer_unboxed_json()}} which will turn all length 1 vectors into atomic values.
To force a vector to serialize to an array of size 1, be sure to call \code{\link[=as.list]{as.list()}} on your value. \code{list()} objects are always serialized to an array value.
}
\examples{
\dontrun{
# Set the API Spec to a function to use the auto-generated OAS object
pr() \%>\%
  pr_set_api_spec(function(spec) {
    spec$info$title <- Sys.time()
    spec
  }) \%>\%
  pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) \%>\%
  pr_run()

# Set the API Spec using an object
pr() \%>\%
  pr_set_api_spec(my_custom_object) \%>\%
  pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) \%>\%
  pr_run()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr_set.R
\name{pr_set_404}
\alias{pr_set_404}
\title{Set the handler that is called when the incoming request can't be served}
\usage{
pr_set_404(pr, fun)
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{fun}{A handler function}
}
\value{
The Plumber router with a modified 404 handler
}
\description{
This function allows a custom error message to be returned when a request
cannot be served by an existing endpoint or filter.
}
\examples{
\dontrun{
handler_404 <- function(req, res) {
  res$status <- 404
  res$body <- "Oops"
}

pr() \%>\%
  pr_get("/hi", function() "Hello") \%>\%
  pr_set_404(handler_404) \%>\%
  pr_run()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plumber-step.R
\name{PlumberStep}
\alias{PlumberStep}
\title{plumber step R6 class}
\description{
an object representing a step in the lifecycle of the treatment
of a request by a plumber router.
}
\section{Super class}{
\code{plumber::Hookable} -> \code{PlumberStep}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{srcref}}{from step block}

\item{\code{lines}}{lines from step block}

\item{\code{serializer}}{step serializer function}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-PlumberStep-new}{\code{PlumberStep$new()}}
\item \href{#method-PlumberStep-exec}{\code{PlumberStep$exec()}}
\item \href{#method-PlumberStep-registerHook}{\code{PlumberStep$registerHook()}}
\item \href{#method-PlumberStep-clone}{\code{PlumberStep$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Hookable" data-id="registerHooks"><a href='../../plumber/html/Hookable.html#method-Hookable-registerHooks'><code>plumber::Hookable$registerHooks()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberStep-new"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberStep-new}{}}}
\subsection{Method \code{new()}}{
Create a new \code{\link[=PlumberStep]{PlumberStep()}} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberStep$new(expr, envir, lines, serializer, srcref)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{expr}}{step expr}

\item{\code{envir}}{step environment}

\item{\code{lines}}{step block}

\item{\code{serializer}}{step serializer}

\item{\code{srcref}}{\code{srcref} attribute from block}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{PlumberStep} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberStep-exec"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberStep-exec}{}}}
\subsection{Method \code{exec()}}{
step execution function
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberStep$exec(req, res)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req, res}}{Request and response objects created by a Plumber request}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberStep-registerHook"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberStep-registerHook}{}}}
\subsection{Method \code{registerHook()}}{
step hook registration method
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberStep$registerHook(
  stage = c("preexec", "postexec", "aroundexec"),
  handler
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{stage}}{a character string.}

\item{\code{handler}}{a step handler function.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberStep-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberStep-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberStep$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr.R
\name{pr_handle}
\alias{pr_handle}
\alias{pr_get}
\alias{pr_post}
\alias{pr_put}
\alias{pr_delete}
\alias{pr_head}
\title{Add handler to Plumber router}
\usage{
pr_handle(pr, methods, path, handler, preempt, serializer, endpoint, ...)

pr_get(pr, path, handler, preempt, serializer, endpoint, ...)

pr_post(pr, path, handler, preempt, serializer, endpoint, ...)

pr_put(pr, path, handler, preempt, serializer, endpoint, ...)

pr_delete(pr, path, handler, preempt, serializer, endpoint, ...)

pr_head(pr, path, handler, preempt, serializer, endpoint, ...)
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{methods}{Character vector of HTTP methods}

\item{path}{The endpoint path}

\item{handler}{A handler function}

\item{preempt}{A preempt function}

\item{serializer}{A Plumber serializer}

\item{endpoint}{A \code{PlumberEndpoint} object}

\item{...}{Additional arguments for \code{PlumberEndpoint}}
}
\value{
A Plumber router with the handler added
}
\description{
This collection of functions creates handlers for a Plumber router.
}
\details{
The generic \code{\link[=pr_handle]{pr_handle()}} creates a handle for the given method(s). Specific
functions are implemented for the following HTTP methods:
\itemize{
\item \code{GET}
\item \code{POST}
\item \code{PUT}
\item \code{DELETE}
\item \code{HEAD}
Each function mutates the Plumber router in place and returns
the updated router.
}
}
\examples{
\dontrun{
pr() \%>\%
  pr_handle("GET", "/hi", function() "Hello World") \%>\%
  pr_run()

pr() \%>\%
  pr_handle(c("GET", "POST"), "/hi", function() "Hello World") \%>\%
  pr_run()

pr() \%>\%
  pr_get("/hi", function() "Hello World") \%>\%
  pr_post("/echo", function(req, res) {
    if (is.null(req$body)) return("No input")
    list(
      input = req$body
    )
  }) \%>\%
  pr_run()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-cookie.R
\name{session_cookie}
\alias{session_cookie}
\title{Store session data in encrypted cookies.}
\usage{
session_cookie(
  key,
  name = "plumber",
  expiration = FALSE,
  http = TRUE,
  secure = FALSE,
  same_site = FALSE,
  path = NULL
)
}
\arguments{
\item{key}{The secret key to use. This must be consistent across all R sessions
where you want to save/restore encrypted cookies. It should be produced using
\code{\link{random_cookie_key}}. Please see the "Storing secure keys" section for more details
complex character string to bolster security.}

\item{name}{The name of the cookie in the user's browser.}

\item{expiration}{A number representing the number of seconds into the future
before the cookie expires or a \code{POSIXt} date object of when the cookie expires.
Defaults to the end of the user's browser session.}

\item{http}{Boolean that adds the \code{HttpOnly} cookie flag that tells the browser
to save the cookie and to NOT send it to client-side scripts. This mitigates \href{https://developer.mozilla.org/en-US/docs/Glossary/Cross-site_scripting}{cross-site scripting}.
Defaults to \code{TRUE}.}

\item{secure}{Boolean that adds the \code{Secure} cookie flag.  This should be set
when the route is eventually delivered over \href{https://en.wikipedia.org/wiki/HTTPS}{HTTPS}.}

\item{same_site}{A character specifying the SameSite policy to attach to the cookie.
If specified, one of the following values should be given: "Strict", "Lax", or "None".
If "None" is specified, then the \code{secure} flag MUST also be set for the modern browsers to
accept the cookie. An error will be returned if \code{same_site = "None"} and \code{secure = FALSE}.
If not specified or a non-character is given, no SameSite policy is attached to the cookie.}

\item{path}{The URI path that the cookie will be available in future requests.
Defaults to the request URI. Set to \code{"/"} to make cookie available to
all requests at the host.}
}
\description{
\code{plumber} uses the crypto R package \code{sodium}, to encrypt/decrypt
\code{req$session} information for each server request.
}
\details{
The cookie's secret encryption \code{key} value must be consistent to maintain
\code{req$session} information between server restarts.
}
\section{Storing secure keys}{

While it is very quick to get started with user session cookies using
\code{plumber}, please exercise precaution when storing secure key information.
If a malicious person were to gain access to the secret \code{key}, they would
be able to eavesdrop on all \code{req$session} information and/or tamper with
\code{req$session} information being processed.

Please: \itemize{
\item Do NOT store keys in source control.
\item Do NOT store keys on disk with permissions that allow it to be accessed by everyone.
\item Do NOT store keys in databases which can be queried by everyone.
}

Instead, please: \itemize{
\item Use a key management system, such as
\href{https://github.com/r-lib/keyring}{'keyring'} (preferred)
\item Store the secret in a file on disk with appropriately secure permissions,
such as "user read only" (\code{Sys.chmod("myfile.txt", mode = "0600")}),
to prevent others from reading it.
} Examples of both of these solutions are done in the Examples section.
}

\examples{
\dontrun{

## Set secret key using `keyring` (preferred method)
keyring::key_set_with_value("plumber_api", plumber::random_cookie_key())


# Load a plumber API
plumb_api("plumber", "01-append") \%>\%
  # Add cookie support via `keyring`
  pr_cookie(
    keyring::key_get("plumber_api")
  ) \%>\%
  pr_run()


#### -------------------------------- ###


## Save key to a local file
pswd_file <- "normal_file.txt"
cat(plumber::random_cookie_key(), file = pswd_file)
# Make file read-only
Sys.chmod(pswd_file, mode = "0600")


# Load a plumber API
plumb_api("plumber", "01-append") \%>\%
  # Add cookie support and retrieve secret key from file
  pr_cookie(
    readLines(pswd_file, warn = FALSE)
  ) \%>\%
  pr_run()
}
}
\seealso{
\itemize{
\item \href{https://github.com/r-lib/sodium}{'sodium'}: R bindings to 'libsodium'
\item \href{https://doc.libsodium.org/}{'libsodium'}: A Modern and Easy-to-Use Crypto Library
\item \href{https://github.com/r-lib/keyring}{'keyring'}: Access the system credential store from R
\item \href{https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie#Directives}{Set-Cookie flags}: Descriptions of different flags for \code{Set-Cookie}
\item \href{https://developer.mozilla.org/en-US/docs/Glossary/Cross-site_scripting}{Cross-site scripting}: A security exploit which allows an attacker to inject into a website malicious client-side code
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plumber.R
\name{Plumber}
\alias{Plumber}
\title{Package Plumber Router}
\description{
Package Plumber Router

Package Plumber Router
}
\details{
Routers are the core request handler in \pkg{plumber}. A router is responsible for
taking an incoming request, submitting it through the appropriate filters and
eventually to a corresponding endpoint, if one is found.

See the \href{https://www.rplumber.io/articles/programmatic-usage.html}{Programmatic Usage} article for additional
details on the methods available on this object.
}
\examples{

## ------------------------------------------------
## Method `Plumber$mount`
## ------------------------------------------------

\dontrun{
root <- pr()

users <- Plumber$new("users.R")
root$mount("/users", users)

products <- Plumber$new("products.R")
root$mount("/products", products)
}

## ------------------------------------------------
## Method `Plumber$registerHook`
## ------------------------------------------------

\dontrun{
pr <- pr()
pr$registerHook("preroute", function(req){
  cat("Routing a request for", req$PATH_INFO, "...\n")
})
pr$registerHooks(list(
  preserialize=function(req, value){
    print("About to serialize this value:")
    print(value)

    # Must return the value since we took one in. Here we're not choosing
    # to mutate it, but we could.
    value
  },
  postserialize=function(res){
    print("We serialized the value as:")
    print(res$body)
  }
))

pr$handle("GET", "/", function(){ 123 })
}

## ------------------------------------------------
## Method `Plumber$handle`
## ------------------------------------------------

\dontrun{
pr <- pr()
pr$handle("GET", "/", function(){
  "<html><h1>Programmatic Plumber!</h1></html>"
}, serializer=plumber::serializer_html())
}

## ------------------------------------------------
## Method `Plumber$setSerializer`
## ------------------------------------------------

\dontrun{
pr <- pr()
pr$setSerializer(serializer_unboxed_json())
}

## ------------------------------------------------
## Method `Plumber$set404Handler`
## ------------------------------------------------

\dontrun{
pr <- pr()
pr$set404Handler(function(req, res) {cat(req$PATH_INFO)})
}

## ------------------------------------------------
## Method `Plumber$setErrorHandler`
## ------------------------------------------------

\dontrun{
pr <- pr()
pr$setErrorHandler(function(req, res, err) {
  message("Found error: ")
  str(err)
})
}
}
\seealso{
\code{\link[=pr]{pr()}},
\code{\link[=pr_run]{pr_run()}},
\code{\link[=pr_get]{pr_get()}}, \code{\link[=pr_post]{pr_post()}},
\code{\link[=pr_mount]{pr_mount()}},
\code{\link[=pr_hook]{pr_hook()}}, \code{\link[=pr_hooks]{pr_hooks()}}, \code{\link[=pr_cookie]{pr_cookie()}},
\code{\link[=pr_filter]{pr_filter()}},
\code{\link[=pr_set_api_spec]{pr_set_api_spec()}}, \code{\link[=pr_set_docs]{pr_set_docs()}},
\code{\link[=pr_set_serializer]{pr_set_serializer()}}, \code{\link[=pr_set_parsers]{pr_set_parsers()}},
\code{\link[=pr_set_404]{pr_set_404()}}, \code{\link[=pr_set_error]{pr_set_error()}},
\code{\link[=pr_set_debug]{pr_set_debug()}},
\code{\link[=pr_set_docs_callback]{pr_set_docs_callback()}}
}
\section{Super class}{
\code{plumber::Hookable} -> \code{Plumber}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{flags}}{For internal use only}
}
\if{html}{\out{</div>}}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{endpoints}}{Plumber router endpoints read-only}

\item{\code{filters}}{Plumber router filters read-only}

\item{\code{mounts}}{Plumber router mounts read-only}

\item{\code{environment}}{Plumber router environment read-only}

\item{\code{routes}}{Plumber router routes read-only}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Plumber-new}{\code{Plumber$new()}}
\item \href{#method-Plumber-run}{\code{Plumber$run()}}
\item \href{#method-Plumber-mount}{\code{Plumber$mount()}}
\item \href{#method-Plumber-unmount}{\code{Plumber$unmount()}}
\item \href{#method-Plumber-registerHook}{\code{Plumber$registerHook()}}
\item \href{#method-Plumber-handle}{\code{Plumber$handle()}}
\item \href{#method-Plumber-removeHandle}{\code{Plumber$removeHandle()}}
\item \href{#method-Plumber-print}{\code{Plumber$print()}}
\item \href{#method-Plumber-serve}{\code{Plumber$serve()}}
\item \href{#method-Plumber-route}{\code{Plumber$route()}}
\item \href{#method-Plumber-call}{\code{Plumber$call()}}
\item \href{#method-Plumber-onHeaders}{\code{Plumber$onHeaders()}}
\item \href{#method-Plumber-onWSOpen}{\code{Plumber$onWSOpen()}}
\item \href{#method-Plumber-setSerializer}{\code{Plumber$setSerializer()}}
\item \href{#method-Plumber-setParsers}{\code{Plumber$setParsers()}}
\item \href{#method-Plumber-set404Handler}{\code{Plumber$set404Handler()}}
\item \href{#method-Plumber-setErrorHandler}{\code{Plumber$setErrorHandler()}}
\item \href{#method-Plumber-setDocs}{\code{Plumber$setDocs()}}
\item \href{#method-Plumber-setDocsCallback}{\code{Plumber$setDocsCallback()}}
\item \href{#method-Plumber-setDebug}{\code{Plumber$setDebug()}}
\item \href{#method-Plumber-getDebug}{\code{Plumber$getDebug()}}
\item \href{#method-Plumber-filter}{\code{Plumber$filter()}}
\item \href{#method-Plumber-setApiSpec}{\code{Plumber$setApiSpec()}}
\item \href{#method-Plumber-getApiSpec}{\code{Plumber$getApiSpec()}}
\item \href{#method-Plumber-addEndpoint}{\code{Plumber$addEndpoint()}}
\item \href{#method-Plumber-addAssets}{\code{Plumber$addAssets()}}
\item \href{#method-Plumber-addFilter}{\code{Plumber$addFilter()}}
\item \href{#method-Plumber-addGlobalProcessor}{\code{Plumber$addGlobalProcessor()}}
\item \href{#method-Plumber-openAPIFile}{\code{Plumber$openAPIFile()}}
\item \href{#method-Plumber-swaggerFile}{\code{Plumber$swaggerFile()}}
\item \href{#method-Plumber-clone}{\code{Plumber$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Hookable" data-id="registerHooks"><a href='../../plumber/html/Hookable.html#method-Hookable-registerHooks'><code>plumber::Hookable$registerHooks()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-new"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-new}{}}}
\subsection{Method \code{new()}}{
Create a new \code{Plumber} router

See also \code{\link[=plumb]{plumb()}}, \code{\link[=pr]{pr()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$new(file = NULL, filters = defaultPlumberFilters, envir)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{file}}{path to file to plumb}

\item{\code{filters}}{a list of Plumber filters}

\item{\code{envir}}{an environment to be used as the enclosure for the routers execution}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{Plumber} router
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-run"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-run}{}}}
\subsection{Method \code{run()}}{
Start a server using \code{Plumber} object.

See also: \code{\link[=pr_run]{pr_run()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$run(
  host = "127.0.0.1",
  port = getOption("plumber.port", NULL),
  swagger = deprecated(),
  debug = missing_arg(),
  swaggerCallback = missing_arg(),
  ...,
  docs = missing_arg(),
  quiet = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{host}}{a string that is a valid IPv4 or IPv6 address that is owned by
this server, which the application will listen on. "0.0.0.0" represents
all IPv4 addresses and "::/0" represents all IPv6 addresses.}

\item{\code{port}}{a number or integer that indicates the server port that should
be listened on. Note that on most Unix-like systems including Linux and
Mac OS X, port numbers smaller than 1025 require root privileges.

This value does not need to be explicitly assigned. To explicitly set it, see \code{\link[=options_plumber]{options_plumber()}}.}

\item{\code{swagger}}{Deprecated. Please use \code{docs} instead. See \verb{$setDocs(docs)} or \verb{$setApiSpec()} for more customization.}

\item{\code{debug}}{If \code{TRUE}, it will provide more insight into your API errors. Using this value will only last for the duration of the run. If a \verb{$setDebug()} has not been called, \code{debug} will default to \code{interactive()} at \verb{$run()} time. See \verb{$setDebug()} for more details.}

\item{\code{swaggerCallback}}{An optional single-argument function that is
called back with the URL to an OpenAPI user interface when one becomes
ready. If missing, defaults to information previously set with \verb{$setDocsCallback()}.
This value will only be used while running the router.}

\item{\code{...}}{Should be empty.}

\item{\code{docs}}{Visual documentation value to use while running the API.
This value will only be used while running the router.
If missing, defaults to information previously set with \code{setDocs()}.
For more customization, see \verb{$setDocs()} or \code{\link[=pr_set_docs]{pr_set_docs()}} for examples.}

\item{\code{quiet}}{If \code{TRUE}, don't print routine startup messages.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-mount"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-mount}{}}}
\subsection{Method \code{mount()}}{
Mount a Plumber router

Plumber routers can be “nested” by mounting one into another
using the \code{mount()} method. This allows you to compartmentalize your API
by paths which is a great technique for decomposing large APIs into smaller files.

See also: \code{\link[=pr_mount]{pr_mount()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$mount(path, router)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{a character string. Where to mount router.}

\item{\code{router}}{a Plumber router. Router to be mounted.}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
root <- pr()

users <- Plumber$new("users.R")
root$mount("/users", users)

products <- Plumber$new("products.R")
root$mount("/products", products)
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-unmount"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-unmount}{}}}
\subsection{Method \code{unmount()}}{
Unmount a Plumber router
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$unmount(path)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{a character string. Where to unmount router.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-registerHook"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-registerHook}{}}}
\subsection{Method \code{registerHook()}}{
Register a hook

Plumber routers support the notion of "hooks" that can be registered
to execute some code at a particular point in the lifecycle of a request.
Plumber routers currently support four hooks:
\enumerate{
\item \code{preroute(data, req, res)}
\item \code{postroute(data, req, res, value)}
\item \code{preserialize(data, req, res, value)}
\item \code{postserialize(data, req, res, value)}
}
In all of the above you have access to a disposable environment in the \code{data}
parameter that is created as a temporary data store for each request. Hooks
can store temporary data in these hooks that can be reused by other hooks
processing this same request.

One feature when defining hooks in Plumber routers is the ability to modify
the returned value. The convention for such hooks is: any function that accepts
a parameter named \code{value} is expected to return the new value. This could
be an unmodified version of the value that was passed in, or it could be a
mutated value. But in either case, if your hook accepts a parameter
named \code{value}, whatever your hook returns will be used as the new value
for the response.

You can add hooks using the \code{registerHook} method, or you can add multiple
hooks at once using the \code{registerHooks} method which takes a name list in
which the names are the names of the hooks, and the values are the
handlers themselves.

See also: \code{\link[=pr_hook]{pr_hook()}}, \code{\link[=pr_hooks]{pr_hooks()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$registerHook(
  stage = c("preroute", "postroute", "preserialize", "postserialize", "exit"),
  handler
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{stage}}{a character string. Point in the lifecycle of a request.}

\item{\code{handler}}{a hook function.}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pr <- pr()
pr$registerHook("preroute", function(req){
  cat("Routing a request for", req$PATH_INFO, "...\n")
})
pr$registerHooks(list(
  preserialize=function(req, value){
    print("About to serialize this value:")
    print(value)

    # Must return the value since we took one in. Here we're not choosing
    # to mutate it, but we could.
    value
  },
  postserialize=function(res){
    print("We serialized the value as:")
    print(res$body)
  }
))

pr$handle("GET", "/", function(){ 123 })
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-handle"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-handle}{}}}
\subsection{Method \code{handle()}}{
Define endpoints

The “handler” functions that you define in these handle calls
are identical to the code you would have defined in your plumber.R file
if you were using annotations to define your API. The handle() method
takes additional arguments that allow you to control nuanced behavior
of the endpoint like which filter it might preempt or which serializer
it should use.

See also: \code{\link[=pr_handle]{pr_handle()}}, \code{\link[=pr_get]{pr_get()}}, \code{\link[=pr_post]{pr_post()}}, \code{\link[=pr_put]{pr_put()}}, \code{\link[=pr_delete]{pr_delete()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$handle(
  methods,
  path,
  handler,
  preempt,
  serializer,
  parsers,
  endpoint,
  ...
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{methods}}{a character string. http method.}

\item{\code{path}}{a character string. Api endpoints}

\item{\code{handler}}{a handler function.}

\item{\code{preempt}}{a preempt function.}

\item{\code{serializer}}{a serializer function.}

\item{\code{parsers}}{a named list of parsers.}

\item{\code{endpoint}}{a \code{PlumberEndpoint} object.}

\item{\code{...}}{additional arguments for \link{PlumberEndpoint} \code{new} method (namely \code{lines}, \code{params}, \code{comments}, \code{responses} and \code{tags}. Excludes \code{envir}).}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pr <- pr()
pr$handle("GET", "/", function(){
  "<html><h1>Programmatic Plumber!</h1></html>"
}, serializer=plumber::serializer_html())
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-removeHandle"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-removeHandle}{}}}
\subsection{Method \code{removeHandle()}}{
Remove endpoints
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$removeHandle(methods, path, preempt = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{methods}}{a character string. http method.}

\item{\code{path}}{a character string. Api endpoints}

\item{\code{preempt}}{a preempt function.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-print"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-print}{}}}
\subsection{Method \code{print()}}{
Print representation of plumber router.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$print(prefix = "", topLevel = TRUE, ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{prefix}}{a character string. Prefix to append to representation.}

\item{\code{topLevel}}{a logical value. When method executed on top level
router, set to \code{TRUE}.}

\item{\code{...}}{additional arguments for recursive calls}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A terminal friendly representation of a plumber router.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-serve"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-serve}{}}}
\subsection{Method \code{serve()}}{
Serve a request
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$serve(req, res)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{request object}

\item{\code{res}}{response object}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-route"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-route}{}}}
\subsection{Method \code{route()}}{
Route a request
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$route(req, res)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{request object}

\item{\code{res}}{response object}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-call"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-call}{}}}
\subsection{Method \code{call()}}{
\pkg{httpuv} interface call function. (Required for \pkg{httpuv})
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$call(req)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{request object}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-onHeaders"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-onHeaders}{}}}
\subsection{Method \code{onHeaders()}}{
httpuv interface onHeaders function. (Required for \pkg{httpuv})
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$onHeaders(req)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{request object}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-onWSOpen"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-onWSOpen}{}}}
\subsection{Method \code{onWSOpen()}}{
httpuv interface onWSOpen function. (Required for \pkg{httpuv})
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$onWSOpen(ws)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{ws}}{WebSocket object}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-setSerializer"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-setSerializer}{}}}
\subsection{Method \code{setSerializer()}}{
Sets the default serializer of the router.

See also: \code{\link[=pr_set_serializer]{pr_set_serializer()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$setSerializer(serializer)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{serializer}}{a serializer function}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pr <- pr()
pr$setSerializer(serializer_unboxed_json())
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-setParsers"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-setParsers}{}}}
\subsection{Method \code{setParsers()}}{
Sets the default parsers of the router. Initialized to \code{c("json", "form", "text", "octet", "multi")}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$setParsers(parsers)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{parsers}}{Can be one of:
\itemize{
\item A \code{NULL} value
\item A character vector of parser names
\item A named \code{list()} whose keys are parser names names and values are arguments to be applied with \code{\link[=do.call]{do.call()}}
\item A \code{TRUE} value, which will default to combining all parsers. This is great for seeing what is possible, but not great for security purposes
}

If the parser name \code{"all"} is found in any character value or list name, all remaining parsers will be added.
When using a list, parser information already defined will maintain their existing argument values.  All remaining parsers will use their default arguments.

Example:

\if{html}{\out{<div class="sourceCode">}}\preformatted{# provide a character string
parsers = "json"

# provide a named list with no arguments
parsers = list(json = list())

# provide a named list with arguments; include `rds`
parsers = list(json = list(simplifyVector = FALSE), rds = list())

# default plumber parsers
parsers = c("json", "form", "text", "octet", "multi")
}\if{html}{\out{</div>}}}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-set404Handler"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-set404Handler}{}}}
\subsection{Method \code{set404Handler()}}{
Sets the handler that gets called if an
incoming request can’t be served by any filter, endpoint, or sub-router.

See also: \code{\link[=pr_set_404]{pr_set_404()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$set404Handler(fun)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{fun}}{a handler function.}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pr <- pr()
pr$set404Handler(function(req, res) {cat(req$PATH_INFO)})
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-setErrorHandler"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-setErrorHandler}{}}}
\subsection{Method \code{setErrorHandler()}}{
Sets the error handler which gets invoked if any filter or
endpoint generates an error.

See also: \code{\link[=pr_set_404]{pr_set_404()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$setErrorHandler(fun)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{fun}}{a handler function.}
}
\if{html}{\out{</div>}}
}
\subsection{Examples}{
\if{html}{\out{<div class="r example copy">}}
\preformatted{\dontrun{
pr <- pr()
pr$setErrorHandler(function(req, res, err) {
  message("Found error: ")
  str(err)
})
}
}
\if{html}{\out{</div>}}

}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-setDocs"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-setDocs}{}}}
\subsection{Method \code{setDocs()}}{
Set visual documentation to use for API

See also: \code{\link[=pr_set_docs]{pr_set_docs()}}, \code{\link[=register_docs]{register_docs()}}, \code{\link[=registered_docs]{registered_docs()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$setDocs(docs = getOption("plumber.docs", TRUE), ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{docs}}{a character value or a logical value. See \code{\link[=pr_set_docs]{pr_set_docs()}} for examples.
If using \code{\link[=options_plumber]{options_plumber()}}, the value must be set before initializing your Plumber router.}

\item{\code{...}}{Arguments for the visual documentation. See each visual documentation package for further details.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-setDocsCallback"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-setDocsCallback}{}}}
\subsection{Method \code{setDocsCallback()}}{
Set a callback to notify where the API's visual documentation is located.

When set, it will be called with a character string corresponding
to the API docs url. This allows RStudio to locate visual documentation.

If using \code{\link[=options_plumber]{options_plumber()}}, the value must be set before initializing your Plumber router.

See also: \code{\link[=pr_set_docs_callback]{pr_set_docs_callback()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$setDocsCallback(callback = getOption("plumber.docs.callback", NULL))}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{callback}}{a callback function for taking action on the docs url. (Also accepts \code{NULL} values to disable the \code{callback}.)}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-setDebug"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-setDebug}{}}}
\subsection{Method \code{setDebug()}}{
Set debug value to include error messages.

See also: \verb{$getDebug()} and \code{\link[=pr_set_debug]{pr_set_debug()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$setDebug(debug = interactive())}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{debug}}{\code{TRUE} provides more insight into your API errors.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-getDebug"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-getDebug}{}}}
\subsection{Method \code{getDebug()}}{
Retrieve the \code{debug} value. If it has never been set, the result of \code{interactive()} will be used.

See also: \verb{$getDebug()} and \code{\link[=pr_set_debug]{pr_set_debug()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$getDebug()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-filter"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-filter}{}}}
\subsection{Method \code{filter()}}{
Add a filter to plumber router

See also: \code{\link[=pr_filter]{pr_filter()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$filter(name, expr, serializer)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{a character string. Name of filter}

\item{\code{expr}}{an expr that resolve to a filter function or a filter function}

\item{\code{serializer}}{a serializer function}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-setApiSpec"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-setApiSpec}{}}}
\subsection{Method \code{setApiSpec()}}{
Allows to modify router autogenerated OpenAPI Specification

Note, the returned value will be sent through \code{\link[=serializer_unboxed_json]{serializer_unboxed_json()}} which will turn all length 1 vectors into atomic values.
To force a vector to serialize to an array of size 1, be sure to call \code{\link[=as.list]{as.list()}} on your value. \code{list()} objects are always serialized to an array value.

See also: \code{\link[=pr_set_api_spec]{pr_set_api_spec()}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$setApiSpec(api = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{api}}{This can be
\itemize{
\item an OpenAPI Specification formatted list object
\item a function that accepts the OpenAPI Specification autogenerated by \code{plumber} and returns a OpenAPI Specification formatted list object.
\item a path to an OpenAPI Specification
}

The value returned will not be validated for OAS compatibility.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-getApiSpec"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-getApiSpec}{}}}
\subsection{Method \code{getApiSpec()}}{
Retrieve OpenAPI file
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$getApiSpec()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-addEndpoint"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-addEndpoint}{}}}
\subsection{Method \code{addEndpoint()}}{
addEndpoint has been deprecated in v0.4.0 and will be removed in a coming release. Please use \code{handle()} instead.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$addEndpoint(
  verbs,
  path,
  expr,
  serializer,
  processors,
  preempt = NULL,
  params = NULL,
  comments
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{verbs}}{verbs}

\item{\code{path}}{path}

\item{\code{expr}}{expr}

\item{\code{serializer}}{serializer}

\item{\code{processors}}{processors}

\item{\code{preempt}}{preempt}

\item{\code{params}}{params}

\item{\code{comments}}{comments}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-addAssets"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-addAssets}{}}}
\subsection{Method \code{addAssets()}}{
addAssets has been deprecated in v0.4.0 and will be removed in a coming release. Please use \code{mount} and \code{PlumberStatic$new()} instead.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$addAssets(dir, path = "/public", options = list())}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dir}}{dir}

\item{\code{path}}{path}

\item{\code{options}}{options}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-addFilter"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-addFilter}{}}}
\subsection{Method \code{addFilter()}}{
\verb{$addFilter()} has been deprecated in v0.4.0 and will be removed in a coming release. Please use \verb{$filter()} instead.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$addFilter(name, expr, serializer, processors)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{name}}{name}

\item{\code{expr}}{expr}

\item{\code{serializer}}{serializer}

\item{\code{processors}}{processors}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-addGlobalProcessor"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-addGlobalProcessor}{}}}
\subsection{Method \code{addGlobalProcessor()}}{
\verb{$addGlobalProcessor()} has been deprecated in v0.4.0 and will be removed in a coming release. Please use \verb{$registerHook}(s) instead.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$addGlobalProcessor(proc)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{proc}}{proc}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-openAPIFile"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-openAPIFile}{}}}
\subsection{Method \code{openAPIFile()}}{
Deprecated. Retrieve OpenAPI file
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$openAPIFile()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-swaggerFile"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-swaggerFile}{}}}
\subsection{Method \code{swaggerFile()}}{
Deprecated. Retrieve OpenAPI file
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$swaggerFile()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plumber-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Plumber-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Plumber$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plumb.R
\name{plumb_api}
\alias{plumb_api}
\alias{available_apis}
\title{Process a Package's Plumber API}
\usage{
plumb_api(package = NULL, name = NULL, edit = FALSE)

available_apis(package = NULL)
}
\arguments{
\item{package}{Package to inspect}

\item{name}{Name of the package folder to \code{\link[=plumb]{plumb()}}.}

\item{edit}{Whether or not to open the API source code for viewing / editing}
}
\value{
A \code{\link{Plumber}} object. If either \code{package} or \code{name} is null, the appropriate \code{\link[=available_apis]{available_apis()}} will be returned.
}
\description{
So that packages can ship multiple plumber routers, users should store their Plumber APIs
in the \code{inst} subfolder \code{plumber} (\code{./inst/plumber/API_1/plumber.R}).
}
\details{
To view all available Plumber APIs across all packages, please call \code{available_apis()}.
A \code{package} value may be provided to only display a particular package's Plumber APIs.
}
\section{Functions}{
\itemize{
\item \code{plumb_api()}: \code{\link[=plumb]{plumb()}}s a package's Plumber API. Returns a \code{\link{Plumber}} router object

\item \code{available_apis()}: Displays all available package Plumber APIs. Returns a \code{data.frame} of \code{package}, \code{name}, and \code{source_directory} information.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plumber-step.R
\name{PlumberEndpoint}
\alias{PlumberEndpoint}
\title{Plumber Endpoint}
\description{
Plumber Endpoint

Plumber Endpoint
}
\details{
Defines a terminal handler in a Plumber router.

Parameters values are obtained from parsing blocks of lines in a plumber file.
They can also be provided manually for historical reasons.
}
\section{Super classes}{
\code{plumber::Hookable} -> \code{plumber::PlumberStep} -> \code{PlumberEndpoint}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{verbs}}{a character vector. http methods. For historical reasons we have
to accept multiple verbs for a single path. Now it's simpler to just parse
each separate verb/path into its own endpoint, so we just do that.}

\item{\code{path}}{a character string. endpoint path}

\item{\code{comments}}{endpoint comments}

\item{\code{description}}{endpoint description}

\item{\code{responses}}{endpoint responses}

\item{\code{params}}{endpoint parameters}

\item{\code{tags}}{endpoint tags}

\item{\code{parsers}}{step allowed parsers}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-PlumberEndpoint-getTypedParams}{\code{PlumberEndpoint$getTypedParams()}}
\item \href{#method-PlumberEndpoint-canServe}{\code{PlumberEndpoint$canServe()}}
\item \href{#method-PlumberEndpoint-matchesPath}{\code{PlumberEndpoint$matchesPath()}}
\item \href{#method-PlumberEndpoint-new}{\code{PlumberEndpoint$new()}}
\item \href{#method-PlumberEndpoint-getPathParams}{\code{PlumberEndpoint$getPathParams()}}
\item \href{#method-PlumberEndpoint-getFunc}{\code{PlumberEndpoint$getFunc()}}
\item \href{#method-PlumberEndpoint-getFuncParams}{\code{PlumberEndpoint$getFuncParams()}}
\item \href{#method-PlumberEndpoint-getEndpointParams}{\code{PlumberEndpoint$getEndpointParams()}}
\item \href{#method-PlumberEndpoint-setPath}{\code{PlumberEndpoint$setPath()}}
\item \href{#method-PlumberEndpoint-clone}{\code{PlumberEndpoint$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Hookable" data-id="registerHooks"><a href='../../plumber/html/Hookable.html#method-Hookable-registerHooks'><code>plumber::Hookable$registerHooks()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="PlumberStep" data-id="exec"><a href='../../plumber/html/PlumberStep.html#method-PlumberStep-exec'><code>plumber::PlumberStep$exec()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="PlumberStep" data-id="registerHook"><a href='../../plumber/html/PlumberStep.html#method-PlumberStep-registerHook'><code>plumber::PlumberStep$registerHook()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberEndpoint-getTypedParams"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberEndpoint-getTypedParams}{}}}
\subsection{Method \code{getTypedParams()}}{
retrieve endpoint typed parameters
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberEndpoint$getTypedParams()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberEndpoint-canServe"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberEndpoint-canServe}{}}}
\subsection{Method \code{canServe()}}{
ability to serve request
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberEndpoint$canServe(req)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{req}}{a request object}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
a logical. \code{TRUE} when endpoint can serve request.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberEndpoint-matchesPath"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberEndpoint-matchesPath}{}}}
\subsection{Method \code{matchesPath()}}{
determines if route matches requested path
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberEndpoint$matchesPath(path)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{a url path}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
a logical. \code{TRUE} when endpoint matches the requested path.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberEndpoint-new"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberEndpoint-new}{}}}
\subsection{Method \code{new()}}{
Create a new \code{PlumberEndpoint} object
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberEndpoint$new(
  verbs,
  path,
  expr,
  envir,
  serializer,
  parsers,
  lines,
  params,
  comments,
  description,
  responses,
  tags,
  srcref
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{verbs}}{Endpoint verb Ex: \code{"GET"}, \code{"POST"}}

\item{\code{path}}{Endpoint path. Ex: \code{"/index.html"}, \code{"/foo/bar/baz"}}

\item{\code{expr}}{Endpoint function or expression that evaluates to a function.}

\item{\code{envir}}{Endpoint environment}

\item{\code{serializer}}{Endpoint serializer. Ex: \code{\link[=serializer_json]{serializer_json()}}}

\item{\code{parsers}}{Can be one of:
\itemize{
\item A \code{NULL} value
\item A character vector of parser names
\item A named \code{list()} whose keys are parser names names and values are arguments to be applied with \code{\link[=do.call]{do.call()}}
\item A \code{TRUE} value, which will default to combining all parsers. This is great for seeing what is possible, but not great for security purposes
}

If the parser name \code{"all"} is found in any character value or list name, all remaining parsers will be added.
When using a list, parser information already defined will maintain their existing argument values.  All remaining parsers will use their default arguments.

Example:

\if{html}{\out{<div class="sourceCode">}}\preformatted{# provide a character string
parsers = "json"

# provide a named list with no arguments
parsers = list(json = list())

# provide a named list with arguments; include `rds`
parsers = list(json = list(simplifyVector = FALSE), rds = list())

# default plumber parsers
parsers = c("json", "form", "text", "octet", "multi")
}\if{html}{\out{</div>}}}

\item{\code{lines}}{Endpoint block}

\item{\code{params}}{Endpoint params}

\item{\code{comments, description, responses, tags}}{Values to be used within the OpenAPI Spec}

\item{\code{srcref}}{\code{srcref} attribute from block}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new \code{PlumberEndpoint} object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberEndpoint-getPathParams"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberEndpoint-getPathParams}{}}}
\subsection{Method \code{getPathParams()}}{
retrieve endpoint path parameters
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberEndpoint$getPathParams(path)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{endpoint path}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberEndpoint-getFunc"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberEndpoint-getFunc}{}}}
\subsection{Method \code{getFunc()}}{
retrieve endpoint function
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberEndpoint$getFunc()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberEndpoint-getFuncParams"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberEndpoint-getFuncParams}{}}}
\subsection{Method \code{getFuncParams()}}{
retrieve endpoint expression parameters
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberEndpoint$getFuncParams()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberEndpoint-getEndpointParams"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberEndpoint-getEndpointParams}{}}}
\subsection{Method \code{getEndpointParams()}}{
retrieve endpoint defined parameters
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberEndpoint$getEndpointParams()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberEndpoint-setPath"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberEndpoint-setPath}{}}}
\subsection{Method \code{setPath()}}{
Updates \verb{$path} with a sanitized \code{path} and updates the internal path meta-data
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberEndpoint$setPath(path)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{path}}{Path to set \verb{$path}. If missing a beginning slash, one will be added.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-PlumberEndpoint-clone"></a>}}
\if{latex}{\out{\hypertarget{method-PlumberEndpoint-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{PlumberEndpoint$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr_set.R
\name{pr_set_error}
\alias{pr_set_error}
\title{Set the error handler that is invoked if any filter or endpoint generates an
error}
\usage{
pr_set_error(pr, fun)
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{fun}{An error handler function. This should accept \code{req}, \code{res}, and the error value}
}
\value{
The Plumber router with a modified error handler
}
\description{
Set the error handler that is invoked if any filter or endpoint generates an
error
}
\examples{
\dontrun{
handler_error <- function(req, res, err){
  res$status <- 500
  list(error = "Custom Error Message")
}

pr() \%>\%
  pr_get("/error", function() log("a")) \%>\%
  pr_set_error(handler_error) \%>\%
  pr_run()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr_set.R
\name{pr_set_docs_callback}
\alias{pr_set_docs_callback}
\title{Set the \code{callback} to tell where the API visual documentation is located}
\usage{
pr_set_docs_callback(pr, callback = getOption("plumber.docs.callback", NULL))
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{callback}{a callback function for taking action on the docs url.}
}
\value{
The Plumber router with the new docs callback setting.
}
\description{
When set, it will be called with a character string corresponding
to the API visual documentation url. This allows RStudio to locate visual documentation.
}
\details{
If using \code{\link[=options_plumber]{options_plumber()}}, the value must be set before initializing your Plumber router.
}
\examples{
\dontrun{
pr() \%>\%
  pr_set_docs_callback(function(url) { message("API location: ", url) }) \%>\%
  pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) \%>\%
  pr_run()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hookable.R
\name{Hookable}
\alias{Hookable}
\title{Hookable}
\description{
Hookable

Hookable
}
\keyword{internal}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Hookable-registerHook}{\code{Hookable$registerHook()}}
\item \href{#method-Hookable-registerHooks}{\code{Hookable$registerHooks()}}
\item \href{#method-Hookable-clone}{\code{Hookable$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Hookable-registerHook"></a>}}
\if{latex}{\out{\hypertarget{method-Hookable-registerHook}{}}}
\subsection{Method \code{registerHook()}}{
Register a hook on a router
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Hookable$registerHook(stage, handler)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{stage}}{a character string.}

\item{\code{handler}}{a hook function.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Hookable-registerHooks"></a>}}
\if{latex}{\out{\hypertarget{method-Hookable-registerHooks}{}}}
\subsection{Method \code{registerHooks()}}{
Register hooks on a router
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Hookable$registerHooks(handlers)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{handlers}}{a named list of hook functions.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Hookable-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Hookable-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Hookable$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serializer.R
\name{endpoint_serializer}
\alias{endpoint_serializer}
\title{Endpoint Serializer with Hooks}
\usage{
endpoint_serializer(
  serializer,
  preexec_hook = NULL,
  postexec_hook = NULL,
  aroundexec_hook = NULL
)
}
\arguments{
\item{serializer}{Serializer method to be used.  This method should already have its initialization arguments applied.}

\item{preexec_hook}{Function to be run directly before a \link{PlumberEndpoint} calls its route method.}

\item{postexec_hook}{Function to be run directly after a \link{PlumberEndpoint} calls its route method.}

\item{aroundexec_hook}{Function to be run around a \link{PlumberEndpoint} call. Must handle a \code{.next} argument to continue execution. \lifecycle{experimental}}
}
\description{
This method allows serializers to return \code{preexec}, \code{postexec}, and \code{aroundexec} (\lifecycle{experimental}) hooks in addition to a serializer.
This is useful for graphics device serializers which need a \code{preexec} and \code{postexec} hook to capture the graphics output.
}
\details{
\code{preexec} and \code{postexec} hooks happened directly before and after a route is executed.
These hooks are specific to a single \link{PlumberEndpoint}'s route calculation.
}
\examples{
# The definition of `serializer_device` returns
# * a `serializer_content_type` serializer
# * `aroundexec` hook
print(serializer_device)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr_set.R
\name{pr_set_docs}
\alias{pr_set_docs}
\title{Set the API visual documentation}
\usage{
pr_set_docs(pr, docs = getOption("plumber.docs", TRUE), ...)
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{docs}{a character value or a logical value.
If using \code{\link[=options_plumber]{options_plumber()}}, the value must be set before initializing your Plumber router.}

\item{...}{Arguments for the visual documentation. See each visual documentation package for further details.}
}
\value{
The Plumber router with the new docs settings.
}
\description{
\code{docs} should be either a logical or a character value matching a registered visual documentation.
Multiple handles will be added to \code{\link{Plumber}} object. OpenAPI json
file will be served on paths \verb{/openapi.json}. Documentation
will be served on paths \verb{/__docs__/index.html} and \verb{/__docs__/}.
}
\examples{
\dontrun{
## View API using Swagger UI
# Official Website: https://swagger.io/tools/swagger-ui/
# install.packages("swagger")
if (require(swagger)) {
  pr() \%>\%
    pr_set_docs("swagger") \%>\%
    pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) \%>\%
    pr_run()
}

## View API using Redoc
# Official Website: https://github.com/Redocly/redoc
if (require(redoc)) {
  pr() \%>\%
    pr_set_docs("redoc") \%>\%
    pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) \%>\%
    pr_run()
}

## View API using RapiDoc
# Official Website: https://github.com/mrin9/RapiDoc
if (require(rapidoc)) {
  pr() \%>\%
    pr_set_docs("rapidoc") \%>\%
    pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) \%>\%
    pr_run()
}

## Disable the OpenAPI Spec UI
pr() \%>\%
  pr_set_docs(FALSE) \%>\%
  pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) \%>\%
  pr_run()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plumb.R
\name{plumb}
\alias{plumb}
\title{Process a Plumber API}
\usage{
plumb(file = NULL, dir = ".")
}
\arguments{
\item{file}{The file to parse as the plumber router definition.}

\item{dir}{The directory containing the \code{plumber.R} file to parse as the
plumber router definition. Alternatively, if an \code{entrypoint.R} file is
found, it will take precedence and be responsible for returning a runnable
router.}
}
\description{
Process a Plumber API
}
\details{
API routers are the core request handler in plumber. A router is responsible for
taking an incoming request, submitting it through the appropriate filters and
eventually to a corresponding endpoint, if one is found.

See the \href{https://www.rplumber.io/articles/programmatic-usage.html}{Programmatic Usage} article for additional
details on the methods available on this object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr.R
\name{pr_mount}
\alias{pr_mount}
\title{Mount a Plumber router}
\usage{
pr_mount(pr, path, router)
}
\arguments{
\item{pr}{The host Plumber router.}

\item{path}{A character string. Where to mount router.}

\item{router}{A Plumber router. Router to be mounted.}
}
\value{
A Plumber router with the supplied router mounted
}
\description{
Plumber routers can be “nested” by mounting one into another
using the \code{mount()} method. This allows you to compartmentalize your API
by paths which is a great technique for decomposing large APIs into smaller
files. This function mutates the Plumber router (\code{\link[=pr]{pr()}}) in place and
returns the updated router.
}
\examples{
\dontrun{
pr1 <- pr() \%>\%
  pr_get("/hello", function() "Hello")

pr() \%>\%
  pr_get("/goodbye", function() "Goodbye") \%>\%
  pr_mount("/hi", pr1) \%>\%
  pr_run()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr.R
\name{pr_static}
\alias{pr_static}
\title{Add a static route to the \code{plumber} object}
\usage{
pr_static(pr, path, direc)
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{path}{The mounted path location of the static folder}

\item{direc}{The local folder to be served statically}
}
\description{
Add a static route to the \code{plumber} object
}
\examples{
\dontrun{
pr() \%>\%
  pr_static("/path", "./my_folder/location") \%>\%
  pr_run()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr.R
\name{pr_run}
\alias{pr_run}
\title{Start a server using \code{plumber} object}
\usage{
pr_run(
  pr,
  host = "127.0.0.1",
  port = getOption("plumber.port", NULL),
  ...,
  debug = missing_arg(),
  docs = missing_arg(),
  swaggerCallback = missing_arg(),
  quiet = FALSE
)
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{host}{A string that is a valid IPv4 or IPv6 address that is owned by
this server, which the application will listen on. "0.0.0.0" represents
all IPv4 addresses and "::/0" represents all IPv6 addresses.}

\item{port}{A number or integer that indicates the server port that should
be listened on. Note that on most Unix-like systems including Linux and
Mac OS X, port numbers smaller than 1025 require root privileges.}

\item{...}{Should be empty.}

\item{debug}{If \code{TRUE}, it will provide more insight into your API errors.
Using this value will only last for the duration of the run.
If \code{\link[=pr_set_debug]{pr_set_debug()}} has not been called, \code{debug} will default to \code{interactive()} at \code{\link[=pr_run]{pr_run()}} time}

\item{docs}{Visual documentation value to use while running the API.
This value will only be used while running the router.
If missing, defaults to information previously set with \code{\link[=pr_set_docs]{pr_set_docs()}}.
For more customization, see \code{\link[=pr_set_docs]{pr_set_docs()}} for examples.}

\item{swaggerCallback}{An optional single-argument function that is called
back with the URL to an OpenAPI user interface when one becomes ready. If
missing, defaults to information set with \code{\link[=pr_set_docs_callback]{pr_set_docs_callback()}}.
This value will only be used while running the router.}

\item{quiet}{If \code{TRUE}, don't print routine startup messages.}
}
\description{
\code{port} does not need to be explicitly assigned.
}
\examples{
\dontrun{
pr() \%>\%
  pr_run()

pr() \%>\%
  pr_run(
    # manually set port
    port = 5762,
    # turn off visual documentation
    docs = FALSE,
    # do not display startup messages
    quiet = TRUE
  )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui.R
\name{register_docs}
\alias{register_docs}
\alias{registered_docs}
\title{Add visual documentation for plumber to use}
\usage{
register_docs(name, index, static = NULL)

registered_docs()
}
\arguments{
\item{name}{Name of the visual documentation}

\item{index}{A function that returns the HTML content of the landing page of the documentation.
Parameters (besides \code{req} and \code{res}) will be supplied as if it is a regular \code{GET} route.
Default parameter values may be used when setting the documentation \code{index} function.
See the example below.}

\item{static}{A function that returns the path to the static assets (images, javascript, css, fonts) the Docs will use.}
}
\description{
\code{\link[=register_docs]{register_docs()}} is used by other packages like \code{swagger}, \code{rapidoc}, and \code{redoc}.
When you load these packages, it calls \code{\link[=register_docs]{register_docs()}} to provide a user
interface that can interpret your plumber OpenAPI Specifications.
}
\examples{
\dontrun{
# Example from the `swagger` R package
register_docs(
  name = "swagger",
  index = function(version = "3", ...) {
    swagger::swagger_spec(
      api_path = paste0(
        "window.location.origin + ",
        "window.location.pathname.replace(",
          "/\\\\(__docs__\\\\\\\\/|__docs__\\\\\\\\/index.html\\\\)$/, \"\"",
        ") + ",
        "\"openapi.json\""
      ),
      version = version
    )
  },
  static = function(version = "3", ...) {
    swagger::swagger_path(version)
  }
)

# When setting the docs, `index` and `static` function arguments can be supplied
# * via `pr_set_docs()`
# * or through URL query string variables
pr() \%>\%
  # Set default argument `version = 3` for the swagger `index` and `static` functions
  pr_set_docs("swagger", version = 3) \%>\%
  pr_get("/plus/<a:int>/<b:int>", function(a, b) { a + b }) \%>\%
  pr_run()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr.R
\name{pr_filter}
\alias{pr_filter}
\title{Add a filter to Plumber router}
\usage{
pr_filter(pr, name, expr, serializer)
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{name}{A character string. Name of filter}

\item{expr}{An expr that resolve to a filter function or a filter function}

\item{serializer}{A serializer function}
}
\value{
The Plumber router with the defined filter added
}
\description{
Filters can be used to modify an incoming request, return an error, or return
a response prior to the request reaching an endpoint.
}
\examples{
\dontrun{
pr() \%>\%
  pr_filter("foo", function(req, res) {
    print("This is filter foo")
    forward()
  }) \%>\%
  pr_get("/hi", function() "Hello") \%>\%
  pr_run()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr.R
\name{pr}
\alias{pr}
\title{Create a new Plumber router}
\usage{
pr(
  file = NULL,
  filters = defaultPlumberFilters,
  envir = new.env(parent = .GlobalEnv)
)
}
\arguments{
\item{file}{Path to file to plumb}

\item{filters}{A list of Plumber filters}

\item{envir}{An environment to be used as the enclosure for the routers execution}
}
\value{
A new \code{\link{Plumber}} router
}
\description{
Create a new Plumber router
}
\examples{
\dontrun{
pr() \%>\%
  pr_run()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/content-types.R
\name{get_character_set}
\alias{get_character_set}
\title{Request character set}
\usage{
get_character_set(content_type = NULL)
}
\arguments{
\item{content_type}{Request Content-Type header}
}
\value{
Default to \code{UTF-8}. Otherwise return \code{charset} defined in request header.
}
\description{
Request character set
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr_set.R
\name{pr_set_serializer}
\alias{pr_set_serializer}
\title{Set the default serializer of the router}
\usage{
pr_set_serializer(pr, serializer)
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{serializer}{A serializer function}
}
\value{
The Plumber router with the new default serializer
}
\description{
By default, Plumber serializes responses to JSON. This function updates the
default serializer to the function supplied via \code{serializer}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options_plumber.R
\name{options_plumber}
\alias{options_plumber}
\title{Plumber options}
\usage{
options_plumber(
  ...,
  port = getOption("plumber.port"),
  docs = getOption("plumber.docs"),
  docs.callback = getOption("plumber.docs.callback"),
  trailingSlash = getOption("plumber.trailingSlash"),
  methodNotAllowed = getOption("plumber.methodNotAllowed"),
  apiURL = getOption("plumber.apiURL"),
  apiScheme = getOption("plumber.apiScheme"),
  apiHost = getOption("plumber.apiHost"),
  apiPort = getOption("plumber.apiPort"),
  apiPath = getOption("plumber.apiPath"),
  maxRequestSize = getOption("plumber.maxRequestSize"),
  sharedSecret = getOption("plumber.sharedSecret"),
  legacyRedirects = getOption("plumber.legacyRedirects")
)
}
\arguments{
\item{...}{Ignored. Should be empty}

\item{port, docs, docs.callback, trailingSlash, methodNotAllowed, apiScheme, apiHost, apiPort, apiPath, apiURL, maxRequestSize, sharedSecret, legacyRedirects}{See details}
}
\value{
The complete, prior set of \code{\link[=options]{options()}} values.
If a particular parameter is not supplied, it will return the current value.
If no parameters are supplied, all returned values will be the current \code{\link[=options]{options()}} values.
}
\description{
There are a number of global options that affect Plumber's behavior. These can
be set globally with \code{\link[=options]{options()}} or with \code{\link[=options_plumber]{options_plumber()}}. Options set using
\code{\link[=options_plumber]{options_plumber()}} should not include the \code{plumber.} prefix.
}
\details{
\describe{
\item{\code{plumber.port}}{Port Plumber will attempt to use to start http server.
If the port is already in use, server will not be able to start. Defaults to \code{NULL}.}
\item{\code{plumber.docs}}{Name of the visual documentation interface to use. Defaults to \code{TRUE}, which will use \code{"swagger"}.}
\item{\code{plumber.docs.callback}}{A function. Called with
a single parameter corresponding to the visual documentation url after Plumber server is ready. This can be used
by RStudio to open the docs when then API is ran from the editor. Defaults to option \code{NULL}.}
\item{\code{plumber.trailingSlash}}{Logical value which allows the router to redirect any request
that has a matching route with a trailing slash. For example, if set to \code{TRUE} and the
GET route \verb{/test/} existed, then a GET request of \verb{/test?a=1} would redirect to
\verb{/test/?a=1}. Defaults to \code{FALSE}. This option will default to \code{TRUE} in a future release.}
\item{\code{plumber.methodNotAllowed}}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
Logical value which allows the router to notify that an
unavailable method was requested, but a different request method is allowed. For example,
if set to \code{TRUE} and the GET route \verb{/test} existed, then a POST request of \verb{/test} would
receive a 405 status and the allowed methods. Defaults to \code{TRUE}.}
\item{\code{plumber.apiURL}}{Server urls for OpenAPI Specification respecting
pattern \verb{scheme://host:port/path}. Other \verb{api*} options will be ignored when set.}
\item{\code{plumber.apiScheme}}{Scheme used to build OpenAPI url and server url for
OpenAPI Specification. Defaults to \code{http}, or an empty string
when used outside a running router.}
\item{\code{plumber.apiHost}}{Host used to build docs url and server url for
OpenAPI Specification. Defaults to \code{host} defined by \code{run} method, or an empty string
when used outside a running router.}
\item{\code{plumber.apiPort}}{Port used to build OpenAPI url and server url for
OpenAPI Specification. Defaults to \code{port} defined by \code{run} method, or an empty string
when used outside a running router.}
\item{\code{plumber.apiPath}}{Path used to build OpenAPI url and server url for
OpenAPI Specification. Defaults to an empty string.}
\item{\code{plumber.maxRequestSize}}{Maximum length in bytes of request body. Body larger
than maximum are rejected with http error 413. \code{0} means unlimited size. Defaults to \code{0}.}
\item{\code{plumber.sharedSecret}}{Shared secret used to filter incoming request.
When \code{NULL}, secret is not validated. Otherwise, Plumber compares secret with http header
\code{PLUMBER_SHARED_SECRET}. Failure to match results in http error 400. Defaults to \code{NULL}.}
\item{\code{plumber.legacyRedirects}}{Plumber will redirect legacy route \verb{/__swagger__/} and
\verb{/__swagger__/index.html} to \verb{../__docs__/} and \verb{../__docs__/index.html}. You can disable this
by settings this option to \code{FALSE}. Defaults to \code{TRUE}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plumber-step.R
\name{forward}
\alias{forward}
\title{Forward Request to The Next Handler}
\usage{
forward()
}
\description{
This function is used when a filter is done processing a request and wishes
to pass control off to the next handler in the chain. If this is not called
by a filter, the assumption is that the filter fully handled the request
itself and no other filters or endpoints should be evaluated for this
request.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/includes.R
\name{include_file}
\alias{include_file}
\alias{include_html}
\alias{include_md}
\alias{include_rmd}
\title{Send File Contents as Response}
\usage{
include_file(file, res, content_type = getContentType(tools::file_ext(file)))

include_html(file, res)

include_md(file, res, format = NULL)

include_rmd(file, res, format = NULL)
}
\arguments{
\item{file}{The path to the file to return}

\item{res}{The response object into which we'll write}

\item{content_type}{If provided, the given value will be sent as the
\code{Content-Type} header in the response. Defaults to the contentType of the file extension.
To disable the \code{Content-Type} header, set \code{content_type = NULL}.}

\item{format}{Passed as the \code{output_format} to \code{rmarkdown::render}}
}
\description{
Returns the file at the given path as the response. If you want an endpoint to return a file as an attachment for user to download see \code{\link[=as_attachment]{as_attachment()}}.
}
\details{
\code{include_html} will merely return the file with the proper
\code{content_type} for HTML. \code{include_md} and \code{include_rmd} will
process the given markdown file through \code{rmarkdown::render} and return
the resultant HTML as a response.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serializer.R
\name{as_attachment}
\alias{as_attachment}
\title{Return an attachment response}
\usage{
as_attachment(value, filename = NULL)
}
\arguments{
\item{value}{Response value to be saved}

\item{filename}{File name to use when saving the attachment.
If no \code{filename} is provided, the \code{value} will be treated as a regular attachment}
}
\value{
Object with class \code{"plumber_attachment"}
}
\description{
This will set the appropriate fields in the \code{Content-Disposition} header value.
To make sure the attachment is used, be sure your serializer eventually calls \code{serializer_headers}
}
\examples{
\dontrun{
# plumber.R

#' @get /data
#' @serializer csv
function() {
  # will cause the file to be saved as `iris.csv`, not `data` or `data.csv`
  as_attachment(iris, "iris.csv")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/session-cookie.R
\name{random_cookie_key}
\alias{random_cookie_key}
\title{Random cookie key generator}
\usage{
random_cookie_key()
}
\value{
A 64 digit hexadecimal string to be used as a key for cookie encryption.
}
\description{
Uses a cryptographically secure pseudorandom number generator from \code{\link[sodium:helpers]{sodium::helpers()}} to generate a 64 digit hexadecimal string.  \href{https://github.com/r-lib/sodium}{'sodium'} wraps around \href{https://doc.libsodium.org/}{'libsodium'}.
}
\details{
Please see \code{\link{session_cookie}} for more information on how to save the generated key.
}
\seealso{
\code{\link{session_cookie}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/digital-ocean.R
\name{do_provision}
\alias{do_provision}
\alias{do_configure_https}
\alias{do_deploy_api}
\alias{do_forward}
\alias{do_remove_api}
\alias{do_remove_forward}
\title{DigitalOcean Plumber server}
\usage{
do_provision(...)

do_configure_https(...)

do_deploy_api(...)

do_forward(...)

do_remove_api(...)

do_remove_forward(...)
}
\description{
These methods are now defunct.
Please use the \href{https://github.com/meztez/plumberDeploy}{\code{plumberDeploy}} R package.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr_set.R
\name{pr_set_parsers}
\alias{pr_set_parsers}
\title{Set the default endpoint parsers for the router}
\usage{
pr_set_parsers(pr, parsers)
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{parsers}{Can be one of:
\itemize{
\item A \code{NULL} value
\item A character vector of parser names
\item A named \code{list()} whose keys are parser names names and values are arguments to be applied with \code{\link[=do.call]{do.call()}}
\item A \code{TRUE} value, which will default to combining all parsers. This is great for seeing what is possible, but not great for security purposes
}

If the parser name \code{"all"} is found in any character value or list name, all remaining parsers will be added.
When using a list, parser information already defined will maintain their existing argument values.  All remaining parsers will use their default arguments.

Example:

\if{html}{\out{<div class="sourceCode">}}\preformatted{# provide a character string
parsers = "json"

# provide a named list with no arguments
parsers = list(json = list())

# provide a named list with arguments; include `rds`
parsers = list(json = list(simplifyVector = FALSE), rds = list())

# default plumber parsers
parsers = c("json", "form", "text", "octet", "multi")
}\if{html}{\out{</div>}}}
}
\value{
The Plumber router with the new default \link{PlumberEndpoint} parsers
}
\description{
By default, Plumber will parse JSON, text, query strings, octet streams, and multipart bodies. This function updates the
default parsers for any endpoint that does not define their own parsers.
}
\details{
Note: The default set of parsers will be completely replaced if any value is supplied. Be sure to include all of your parsers that you would like to include.
Use \code{registered_parsers()} to get a list of available parser names.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/serializer.R
\name{register_serializer}
\alias{register_serializer}
\alias{registered_serializers}
\title{Register a Serializer}
\usage{
register_serializer(name, serializer, verbose = TRUE)

registered_serializers()
}
\arguments{
\item{name}{The name of the serializer (character string)}

\item{serializer}{The serializer function to be added.
This function should accept arguments that can be supplied when \code{\link[=plumb]{plumb()}}ing a file.
This function should return a function that accepts four arguments: \code{value}, \code{req}, \code{res}, and \code{errorHandler}.
See \code{print(serializer_json)} for an example.}

\item{verbose}{Logical value which determines if a message should be printed when overwriting serializers}
}
\description{
A serializer is responsible for translating a generated R value into output
that a remote user can understand. For instance, the \code{serializer_json}
serializes R objects into JSON before returning them to the user. The list of
available serializers in plumber is global.
}
\details{
There are three main building-block serializers:
\itemize{
\item \code{serializer_headers}: the base building-block serializer that is required to have \code{\link[=as_attachment]{as_attachment()}} work
\item \code{serializer_content_type()}: for setting the content type. (Calls \code{serializer_headers()})
\item \code{serializer_device()}: add endpoint hooks to turn a graphics device on and off in addition to setting the content type. (Uses \code{serializer_content_type()})
}
}
\section{Functions}{
\itemize{
\item \code{register_serializer()}: Register a serializer with a name

\item \code{registered_serializers()}: Return a list of all registered serializers

}}
\examples{
# `serializer_json()` calls `serializer_content_type()` and supplies a serialization function
print(serializer_json)
# serializer_content_type() calls `serializer_headers()` and supplies a serialization function
print(serializer_content_type)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{addSerializer}
\alias{addSerializer}
\alias{getCharacterSet}
\alias{sessionCookie}
\title{Deprecated functions}
\usage{
addSerializer(name, serializer, verbose = TRUE)

getCharacterSet(contentType = NULL)

sessionCookie(
  key,
  name = "plumber",
  expiration = FALSE,
  http = TRUE,
  secure = FALSE,
  sameSite = FALSE
)
}
\description{
Deprecated functions
}
\section{Functions}{
\itemize{
\item \code{addSerializer()}: See \code{\link[=register_serializer]{register_serializer()}}

\item \code{getCharacterSet()}: See \code{\link[=get_character_set]{get_character_set()}}

\item \code{sessionCookie()}: See \code{\link[=session_cookie]{session_cookie()}}

}}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate_api_spec.R
\name{validate_api_spec}
\alias{validate_api_spec}
\title{Validate OpenAPI Spec}
\usage{
validate_api_spec(
  pr,
  ...,
  ruleset = c("minimal", "recommended", "recommended-strict"),
  verbose = TRUE
)
}
\arguments{
\item{pr}{A Plumber API}

\item{...}{Ignored}

\item{ruleset}{Character that determines the ruleset to use for validation. Can be one of "minimal", "recommended",
or "recommended-strict". Defaults to "minimal". See \href{https://redocly.com/docs/cli/commands/lint#options}{\verb{@redocly/cli} options} for more details.}

\item{verbose}{Logical that determines if a "is valid" statement is displayed. Defaults to \code{TRUE}}
}
\description{
Validate an OpenAPI Spec using \href{https://redocly.com/docs/cli/commands/lint}{\verb{@redocly/cli}}.
}
\details{
If any warning or error is presented, an error will be thrown.

This function is \ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}} and may be altered, changed, or removed in the future.
}
\examples{
\dontrun{
pr <- plumb_api("plumber", "01-append")
validate_api_spec(pr)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pr.R
\name{pr_hook}
\alias{pr_hook}
\alias{pr_hooks}
\title{Register a hook}
\usage{
pr_hook(pr, stage, handler)

pr_hooks(pr, handlers)
}
\arguments{
\item{pr}{A Plumber API. Note: The supplied Plumber API object will also be updated in place as well as returned by the function.}

\item{stage}{A character string. Point in the lifecycle of a request.}

\item{handler}{A hook function.}

\item{handlers}{A named list of hook handlers}
}
\value{
A Plumber router with the defined hook(s) added
}
\description{
Plumber routers support the notion of "hooks" that can be registered
to execute some code at a particular point in the lifecycle of a request.
Plumber routers currently support four hooks:
\enumerate{
\item \code{preroute(data, req, res)}
\item \code{postroute(data, req, res, value)}
\item \code{preserialize(data, req, res, value)}
\item \code{postserialize(data, req, res, value)}
In all of the above you have access to a disposable environment in the \code{data}
parameter that is created as a temporary data store for each request. Hooks
can store temporary data in these hooks that can be reused by other hooks
processing this same request.
}
}
\details{
One feature when defining hooks in Plumber routers is the ability to modify
the returned value. The convention for such hooks is: any function that accepts
a parameter named \code{value} is expected to return the new value. This could
be an unmodified version of the value that was passed in, or it could be a
mutated value. But in either case, if your hook accepts a parameter
named \code{value}, whatever your hook returns will be used as the new value
for the response.

You can add hooks using the \code{pr_hook}, or you can add multiple
hooks at once using \code{pr_hooks}, which takes a named list in
which the names are the names of the hooks, and the values are the
handlers themselves.
}
\examples{
\dontrun{
pr() \%>\%
  pr_hook("preroute", function(req){
    cat("Routing a request for", req$PATH_INFO, "...\n")
  }) \%>\%
  pr_hooks(list(
    preserialize = function(req, value){
      print("About to serialize this value:")
      print(value)

      # Must return the value since we took one in. Here we're not choosing
      # to mutate it, but we could.
      value
    },
    postserialize = function(res){
      print("We serialized the value as:")
      print(res$body)
    }
  )) \%>\%
  pr_handle("GET", "/", function(){ 123 }) \%>\%
  pr_run()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated-R6.R
\name{hookable}
\alias{hookable}
\alias{plumber}
\title{Deprecated R6 functions}
\description{
Deprecated R6 functions

Deprecated R6 functions

Deprecated R6 functions

Deprecated R6 functions
}
\section{Functions}{
\itemize{
\item \code{hookable}: See \code{\link[=Hookable]{Hookable()}}

\item \code{plumber}: See \code{\link[=Plumber]{Plumber()}}

}}
\keyword{internal}
\section{Super class}{
\code{plumber::Hookable} -> \code{hookable}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-hookable-new}{\code{hookable$new()}}
\item \href{#method-hookable-clone}{\code{hookable$clone()}}
}
}
\if{html}{\out{
<details open><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Hookable" data-id="registerHook"><a href='../../plumber/html/Hookable.html#method-Hookable-registerHook'><code>plumber::Hookable$registerHook()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Hookable" data-id="registerHooks"><a href='../../plumber/html/Hookable.html#method-Hookable-registerHooks'><code>plumber::Hookable$registerHooks()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-hookable-new"></a>}}
\if{latex}{\out{\hypertarget{method-hookable-new}{}}}
\subsection{Method \code{new()}}{
Initialize a new \code{hookable}. Throws deprecated warning prompting user to use \code{\link{Hookable}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{hookable$new()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-hookable-clone"></a>}}
\if{latex}{\out{\hypertarget{method-hookable-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{hookable$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}
\section{Super classes}{
\code{plumber::Hookable} -> \code{plumber::Plumber} -> \code{plumber}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-plumber-new}{\code{plumber$new()}}
\item \href{#method-plumber-clone}{\code{plumber$clone()}}
}
}
\if{html}{\out{
<details><summary>Inherited methods</summary>
<ul>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Hookable" data-id="registerHooks"><a href='../../plumber/html/Hookable.html#method-Hookable-registerHooks'><code>plumber::Hookable$registerHooks()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="addAssets"><a href='../../plumber/html/Plumber.html#method-Plumber-addAssets'><code>plumber::Plumber$addAssets()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="addEndpoint"><a href='../../plumber/html/Plumber.html#method-Plumber-addEndpoint'><code>plumber::Plumber$addEndpoint()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="addFilter"><a href='../../plumber/html/Plumber.html#method-Plumber-addFilter'><code>plumber::Plumber$addFilter()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="addGlobalProcessor"><a href='../../plumber/html/Plumber.html#method-Plumber-addGlobalProcessor'><code>plumber::Plumber$addGlobalProcessor()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="call"><a href='../../plumber/html/Plumber.html#method-Plumber-call'><code>plumber::Plumber$call()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="filter"><a href='../../plumber/html/Plumber.html#method-Plumber-filter'><code>plumber::Plumber$filter()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="getApiSpec"><a href='../../plumber/html/Plumber.html#method-Plumber-getApiSpec'><code>plumber::Plumber$getApiSpec()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="getDebug"><a href='../../plumber/html/Plumber.html#method-Plumber-getDebug'><code>plumber::Plumber$getDebug()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="handle"><a href='../../plumber/html/Plumber.html#method-Plumber-handle'><code>plumber::Plumber$handle()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="mount"><a href='../../plumber/html/Plumber.html#method-Plumber-mount'><code>plumber::Plumber$mount()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="onHeaders"><a href='../../plumber/html/Plumber.html#method-Plumber-onHeaders'><code>plumber::Plumber$onHeaders()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="onWSOpen"><a href='../../plumber/html/Plumber.html#method-Plumber-onWSOpen'><code>plumber::Plumber$onWSOpen()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="openAPIFile"><a href='../../plumber/html/Plumber.html#method-Plumber-openAPIFile'><code>plumber::Plumber$openAPIFile()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="print"><a href='../../plumber/html/Plumber.html#method-Plumber-print'><code>plumber::Plumber$print()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="registerHook"><a href='../../plumber/html/Plumber.html#method-Plumber-registerHook'><code>plumber::Plumber$registerHook()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="removeHandle"><a href='../../plumber/html/Plumber.html#method-Plumber-removeHandle'><code>plumber::Plumber$removeHandle()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="route"><a href='../../plumber/html/Plumber.html#method-Plumber-route'><code>plumber::Plumber$route()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="run"><a href='../../plumber/html/Plumber.html#method-Plumber-run'><code>plumber::Plumber$run()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="serve"><a href='../../plumber/html/Plumber.html#method-Plumber-serve'><code>plumber::Plumber$serve()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="set404Handler"><a href='../../plumber/html/Plumber.html#method-Plumber-set404Handler'><code>plumber::Plumber$set404Handler()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="setApiSpec"><a href='../../plumber/html/Plumber.html#method-Plumber-setApiSpec'><code>plumber::Plumber$setApiSpec()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="setDebug"><a href='../../plumber/html/Plumber.html#method-Plumber-setDebug'><code>plumber::Plumber$setDebug()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="setDocs"><a href='../../plumber/html/Plumber.html#method-Plumber-setDocs'><code>plumber::Plumber$setDocs()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="setDocsCallback"><a href='../../plumber/html/Plumber.html#method-Plumber-setDocsCallback'><code>plumber::Plumber$setDocsCallback()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="setErrorHandler"><a href='../../plumber/html/Plumber.html#method-Plumber-setErrorHandler'><code>plumber::Plumber$setErrorHandler()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="setParsers"><a href='../../plumber/html/Plumber.html#method-Plumber-setParsers'><code>plumber::Plumber$setParsers()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="setSerializer"><a href='../../plumber/html/Plumber.html#method-Plumber-setSerializer'><code>plumber::Plumber$setSerializer()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="swaggerFile"><a href='../../plumber/html/Plumber.html#method-Plumber-swaggerFile'><code>plumber::Plumber$swaggerFile()</code></a></span></li>
<li><span class="pkg-link" data-pkg="plumber" data-topic="Plumber" data-id="unmount"><a href='../../plumber/html/Plumber.html#method-Plumber-unmount'><code>plumber::Plumber$unmount()</code></a></span></li>
</ul>
</details>
}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-plumber-new"></a>}}
\if{latex}{\out{\hypertarget{method-plumber-new}{}}}
\subsection{Method \code{new()}}{
Initialize a new \code{plumber}. Throws deprecated warning prompting user to use \code{\link{Plumber}}
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{plumber$new(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{params passed in to \code{Plumber$new()}}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-plumber-clone"></a>}}
\if{latex}{\out{\hypertarget{method-plumber-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{plumber$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse-body.R
\name{register_parser}
\alias{register_parser}
\alias{registered_parsers}
\title{Manage parsers}
\usage{
register_parser(alias, parser, fixed = NULL, regex = NULL, verbose = TRUE)

registered_parsers()
}
\arguments{
\item{alias}{An alias to map parser from the \verb{@parser} plumber tag to the global parsers list.}

\item{parser}{The parser function to be added. This build the parser function. See Details for more information.}

\item{fixed}{A character vector of fixed string to be matched against a request \code{content-type} to use \code{parser}.}

\item{regex}{A character vector of \link{regex} string to be matched against a request \code{content-type} to use \code{parser}.}

\item{verbose}{Logical value which determines if a warning should be
displayed when alias in map are overwritten.}
}
\description{
A parser is responsible for decoding the raw body content of a request into
a list of arguments that can be mapped to endpoint function arguments.
For instance, \code{\link[=parser_json]{parser_json()}} parse content-type \code{application/json}.
}
\details{
When \code{parser} is evaluated, it should return a parser function.
Parser matching is done first by \code{content-type} header matching with \code{fixed} then by using
regular expressions with \code{regex}. Note that plumber strips \verb{; charset*} from \code{content-type} header before matching.

Plumber will try to use \code{\link[=parser_json]{parser_json()}} (if available) when no \code{content-type} header is found and
the request body starts with \verb{\{} or \code{[}.

Functions signature should include \code{value}, \code{...} and
possibly \code{content_type}, \code{filename}. Other parameters may be provided
if you want to use the headers from \code{\link[webutils:parse_multipart]{webutils::parse_multipart()}}.

Parser function structure is something like below.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{function(parser_arguments_here) \{
  # return a function to parse a raw value
  function(value, ...) \{
    # do something with raw value
  \}
\}
}\if{html}{\out{</div>}}
}
\section{Functions}{
\itemize{
\item \code{registered_parsers()}: Return all registered parsers

}}
\examples{
# `content-type` header is mostly used to look up charset and adjust encoding
parser_dcf <- function(...) {
  function(value, content_type = "text/x-dcf", ...) {
    charset <- get_character_set(content_type)
    value <- rawToChar(value)
    Encoding(value) <- charset
    read.dcf(value, ...)
  }
}

# Could also leverage existing parsers
parser_dcf <- function(...) {
  parser_read_file(function(tmpfile) {
    read.dcf(tmpfile, ...)
  })
}

# Register the newly created parser
\dontrun{register_parser("dcf", parser_dcf, fixed = "text/x-dcf")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse-body.R
\name{parser_form}
\alias{parser_form}
\alias{parser_json}
\alias{parser_geojson}
\alias{parser_text}
\alias{parser_yaml}
\alias{parser_csv}
\alias{parser_tsv}
\alias{parser_read_file}
\alias{parser_rds}
\alias{parser_feather}
\alias{parser_parquet}
\alias{parser_octet}
\alias{parser_multi}
\alias{parser_none}
\title{Plumber Parsers}
\usage{
parser_form()

parser_json(...)

parser_geojson(...)

parser_text(parse_fn = identity)

parser_yaml(...)

parser_csv(...)

parser_tsv(...)

parser_read_file(read_fn = readLines)

parser_rds(...)

parser_feather(...)

parser_parquet(...)

parser_octet()

parser_multi()

parser_none()
}
\arguments{
\item{...}{parameters supplied to the appropriate internal function}

\item{parse_fn}{function to further decode a text string into an object}

\item{read_fn}{function used to read a the content of a file. Ex: \code{\link[=readRDS]{readRDS()}}}
}
\description{
Parsers are used in Plumber to transform request body received
by the API. Extra parameters may be provided to parser
functions when enabling them on router. This will allow for
non-default behavior.
}
\details{
Parsers are optional. When unspecified, only default endpoint parsers are enabled.
You can use \verb{@parser NAME} tag to enable parser on endpoint.
Multiple parsers can be enabled on the same endpoint using multiple \verb{@parser NAME} tags.

User should be aware that \code{rds} parsing should only be done from a
trusted source. Do not accept \code{rds} files blindly.

See \code{\link[=registered_parsers]{registered_parsers()}} for a list of registered parsers names.
}
\section{Functions}{
\itemize{
\item \code{parser_form()}: Form query string parser

\item \code{parser_json()}: JSON parser. See \code{\link[jsonlite:read_json]{jsonlite::parse_json()}} for more details. (Defaults to using \code{simplifyVectors = TRUE})

\item \code{parser_geojson()}: GeoJSON parser. See \code{\link[geojsonsf:geojson_sf]{geojsonsf::geojson_sf()}} for more details.

\item \code{parser_text()}: Helper parser to parse plain text

\item \code{parser_yaml()}: YAML parser. See \code{\link[yaml:yaml.load]{yaml::yaml.load()}} for more details.

\item \code{parser_csv()}: CSV parser. See \code{\link[readr:read_delim]{readr::read_csv()}} for more details.

\item \code{parser_tsv()}: TSV parser. See \code{\link[readr:read_delim]{readr::read_tsv()}} for more details.

\item \code{parser_read_file()}: Helper parser that writes the binary body to a file and reads it back again using \code{read_fn}.
This parser should be used when reading from a file is required.

\item \code{parser_rds()}: RDS parser. See \code{\link[=readRDS]{readRDS()}} for more details.

\item \code{parser_feather()}: feather parser. See \code{\link[arrow:read_feather]{arrow::read_feather()}} for more details.

\item \code{parser_parquet()}: parquet parser. See \code{\link[arrow:read_parquet]{arrow::read_parquet()}} for more details.

\item \code{parser_octet()}: Octet stream parser. Returns the raw content.

\item \code{parser_multi()}: Multi part parser. This parser will then parse each individual body with its respective parser.  When this parser is used, \code{req$body} will contain the updated output from \code{\link[webutils:parse_multipart]{webutils::parse_multipart()}} by adding the \code{parsed} output to each part.  Each part may contain detailed information, such as \code{name} (required), \code{content_type}, \code{content_disposition}, \code{filename}, (raw, original) \code{value}, and \code{parsed} (parsed \code{value}).  When performing Plumber route argument matching, each multipart part will match its \code{name} to the \code{parsed} content.

\item \code{parser_none()}: No parser. Will not process the postBody.

}}
\examples{
\dontrun{
# Overwrite `text/json` parsing behavior to not allow JSON vectors to be simplified
#* @parser json list(simplifyVector = FALSE)
# Activate `rds` parser in a multipart request
#* @parser multi
#* @parser rds
pr <- Plumber$new()
pr$handle("GET", "/upload", function(rds) {rds}, parsers = c("multi", "rds"))
}
}
